#!/bin/bash
### Copyright 1999-2017. Parallels IP Holdings GmbH. All Rights Reserved.
#

#
# Plesk script
#


upgrade_transfer_pmm_logs_11_6_1()
{
	local old_dir="${PRODUCT_ROOT_D}/PMM/logs"
	local new_dir="${PRODUCT_LOGS_D}/PMM"

	if [ -d "$old_dir" -a ! -L "$old_dir" ] ; then
		local moved='false'
		if [ ! -e "$new_dir" ] ; then
			mv -f "$old_dir" "$new_dir"
			if [ "$?" = "0" ] ; then
				moved='true'
			else
				warn "move '$old_dir' to '$new_dir'"
			fi
		fi
		if [ "$moved" = 'false' ]; then
			new_dir="${new_dir}.$product_suff"
			mv -f "$old_dir" "$new_dir" || 				{ warn "move '$old_dir' to '$new_dir'"; return 1; }
		fi

		ln -s "$new_dir" "$old_dir" || 			warn "unable to create compatibility symlink '$old_dir' to '$new_dir'"
	fi
}

upgrade_dump_tmp_d_13_0_2()
{
	local psa_conf=/etc/psa/psa.conf
	if [ -f "$psa_conf" ]; then
		DUMP_TMP_D=`conf_getvar $psa_conf DUMP_TMP_D`
		if [ "/tmp" = "$DUMP_TMP_D" ]; then
			conf_setvar "$psa_conf" "DUMP_TMP_D" "/usr/local/psa/PMM/tmp" 
		fi
	fi
}

update_dump_d_mode_17_5_3()
{
	local psa_conf=/etc/psa/psa.conf
	if [ -f "$psa_conf" ]; then
		DUMP_D=`conf_getvar $psa_conf DUMP_D`
		if [ ! -z "$DUMP_D" ]; then
			chmod 750 "$DUMP_D"
		fi
	fi
}
# vim:syntax=sh

generate_encryption_key()
{
	local key_file="/etc/psa/private/secret_key"
	local key_dir="`dirname $key_file`"
	local rc=0

	[ -d "$key_dir" ] || mkdir -p "$key_dir"

	if [ ! -e "$key_file" ]; then
		dd if=/dev/urandom of="$key_file" bs=16 count=1 2>/dev/null
	else
		rc=1
	fi
	fix_key_permissions
	return $rc
}

fix_key_permissions()
{
	local key_file="/etc/psa/private/secret_key"
	local key_dir="`dirname $key_file`"

	if [ -e "$key_file" ]; then
		chown psaadm:0 "$key_file"
		chmod 0600 "$key_file"
	fi

	if [ -d "$key_dir" ]; then
		chown psaadm:0 "$key_dir"
		chmod 0700 "$key_dir"
	fi
}

#admin

set_admin_params()
{

	ADMIN_ROOT=$PRODUCT_ROOT_D/admin

	UTILS_DIR=$ADMIN_ROOT/sbin

	admin_user="${product}adm"
	admin_UID=8444
	admin_group="${product}adm"
	admin_GID=8444

	ADMIN_CERT="$ADMIN_ROOT/conf/httpsd.pem"

	define_start_script
}

lock_keyupdate()
{
	export PLESK_KEYUPDATE_DISABLED=installation
}

## @@constructor set_admin_params
## @@constructor lock_keyupdate

set_admin_perms()
{
   	#remove stale file
	rm -f "$ADMIN_ROOT/htdocs/index.php3"

#	if [ "X$machine" != "Xlinux" -o "X$linux_distr" = "Xdebian" ]; then
#		set_admin_additional_perms
#	fi

	[ -d $ADMIN_ROOT/man ] &&
	set_ac 0 $admin_group 755 "$ADMIN_ROOT/man/*" &&
	for i in `find $ADMIN_ROOT/man -type d -print`; do
		set_ac root $admin_group 755 "$i"
		set_ac root $admin_group 644 "$i/*"
	done

	[ -d ${DUMP_D} ] &&
		set_ac $admin_user $admin_group 750 ${DUMP_D}

	#set locales permissions
	for dir in htdocs plib; do
		chmod -R 644 "$ADMIN_ROOT/$dir/locales"
		find "$ADMIN_ROOT/$dir/locales" -type d -exec chmod 755 {} \;
	done

	if [ "X$DEMO_VERSION" = "Xyes" ]; then
		set_ac root $admin_group 775 "$ADMIN_ROOT/htdocs/images/custom_buttons"
	else
		set_ac root $admin_group 4110 "$ADMIN_ROOT/sbin/wrapper"
		set_ac root $admin_group 4110 "$ADMIN_ROOT/sbin/mod_wrapper"
	fi
}

install_admin()
{
	p_echo

	p_echo "===> Installing Admin Server"

	# set default permissions on $ADMIN_ROOT and his contents
	set_admin_perms

	if [ ! -s "$ADMIN_CERT" ] ;then
		inten="copy default SSL Certificate for admin server"
		echo_try $inten
		get_certificate "$ADMIN_CERT" || die $inten
		set_ac 0 0 0400 "$ADMIN_CERT"
		suc
	fi

	if [ "X$DEMO_VERSION" = "Xyes" ]; then
		chown 0:psaadm $ADMIN_CERT || die
		chmod 440 $ADMIN_CERT || die
	else
		chown 0:0 $ADMIN_CERT || die
		chmod 400 $ADMIN_CERT || die
	fi
}

get_certificate()
{
	local dst="$1"
	local src="${certificate_file}"
# Check or generate new default certificate
	[ -s "$src" ] || generate_default_certificate
 	set_ac 0 0 0400 "${src}"
	cp -fp $src $dst
}

generate_default_certificate()
{
	# Currently I have no will to accurately rewrite the stuff below
	# so I just add support for (optional) file parameter. Actually
	# one need to accurately handle temporary files, directory creation
	# etc. Probably next time...
	local cert_file

	cert_file="${certificate_file}"
	if [ -s "$cert_file" ]; then
		p_echo "default certificate already exists"
		return
	fi

	# This var also can safely be made local. It's not used outside
	local OPENSSL_CNF is_temp_cnf rc

	OPENSSL_CNF=${PRODUCT_ROOT_D}/admin/conf/openssl.cnf
	is_temp_cnf=0

	if [ ! -f "${OPENSSL_CNF}" ]; then
		# Well, base psa package isn't yet installed. 
		# let's form a temporary config file
		OPENSSL_CNF=$(mktemp /tmp/openssl.cnf.XXXXXX)
		[ $? != 0 ] && die "Unable to create temporary file"
		cat >>${OPENSSL_CNF} <<EOF
[ req ]
attributes=req_attributes
distinguished_name=req_distinguished_name
default_md = sha256

[ req_attributes ]
challengePassword               = A challenge password
unstructuredName                = An optional company name

[ req_distinguished_name ]
countryName             = Country Name
stateOrProvinceName     = State or Province Name
localityName            = Locality Name
organizationName        = Organisation Name
organizationalUnitName  = Organization Unit Name
commonName              = Common Name
emailAddress            = Email Address
EOF
		is_temp_cnf=1
	fi

	local cert_file_d=`dirname "${cert_file}"`
	[ -d "${cert_file_d}" ] || mkdir -p "${cert_file_d}"

	echo "CH
.
Schaffhausen
Plesk
.
Plesk
info@plesk.com


" | openssl req -x509 -nodes -days 365 -newkey rsa:2048 	-config ${OPENSSL_CNF} -set_serial `date +%s` -keyout "${cert_file}_" 	-out "${cert_file}_" >> $product_log 2>&1

	# preserve exit code
	rc=$?

	if [ $rc -eq 0 ]; then
		cat "${cert_file}_" | sed -e 's/^\(-----END.*\)/\1/' > $cert_file
		set_ac 0 0 0400 "${cert_file}"
	fi
	# cleanup temporary files
	rm -f "${cert_file}_" ||:
	if [ "x$is_temp_cnf" = "x1" ]; then
		rm -f ${OPENSSL_CNF} ||:
	fi
	return $rc
}


clean_panel_sessions()
{
# use subshell to catch die's inside db_do:
	(
	db_do "DELETE FROM sessions"
	db_do "DELETE FROM SessionContexts"
	)
}
set_apache_params()
{
	apache_user="www-data"
	apache_UID=80
	apache_group="www-data"
	apache_GID=80

	user_apxs="/usr/bin/apxs2"

	apache_pid_file="$APACHE_ROOT/logs/httpd.pid"
	apache_lock_file="$APACHE_ROOT/logs/httpd.lock"
	product_lock_file="$HTTPD_CONF_D/cnf.lock"
	apache_service_name="apache2"

	apache_modules_d="/usr/lib/apache2/modules"

	apache_service="$apache_service_name"

	apache_httpd_conf="$HTTPD_CONF_D/httpd.conf"
	apache_httpd_conf2="$HTTPD_CONF_D/apache2.conf"
	apache_httpd_conf_in="$HTTPD_CONF_D/httpd.conf.in"

	apache_httpd_include="$HTTPD_INCLUDE_D/zz010_psa_httpd.conf"

	APACHE_CERT="$HTTPD_CONF_D/httpd.pem"
	APACHE_ROOT="/usr"

	min_suexec_UID=10000
	max_suexec_UID=16000
	min_suexec_GID=$min_suexec_UID
	max_suexec_GID=$max_suexec_UID

	suexec_storage=/usr/lib/plesk-9.0/suexec
	suexec=/usr/lib/apache2/suexec
	suexec_dir=/usr/lib/apache2
	suexec_file=suexec

	rpm_httpd_bin=/usr/sbin/httpd
}

module_exists()
{
	local dir
	[ -n "$1" ] || return 1
	test -e "${HTTPD_CONF_D}/mods-available/$1.load" && return
	if [ -n "${additional_apache_modules_d}" ]; then
		for dir in $additional_apache_modules_d; do 
			if [ -f "${dir}/mod_$1.so"  ] ; then return ; fi
			if [ -n "$2" -a -f "${dir}/$2" ] ; then return ; fi
		done
	fi
	return 1
}

# Run before it
#        read_conf
#        # set_common_params
#        set_apache_params
add_apache_module()
{
	local module_name="$1"
	local module_soname="$2" # optional
	local config="$3" # optional

	if ! module_exists $module_name "$module_soname"; then
		# TODO : fix message error
		p_echo "Shared object mod_${module_name}.so not found in ${apache_modules_d}...skipped."
		return
	fi

		if [ -x /usr/sbin/a2enmod ]; then
			/usr/sbin/a2enmod $module_name >> $product_log
		else
			warn "a2enmod utilily not found! Module $module_name was NOT enabled."
		fi
}

true apache_status_linux_debian

apache_status_linux_debian()
{
	get_pid "/usr/sbin/apache2" false
	local pid=$common_var
	if test "$pid" -ne 1; then
# running
		return 0
	fi
	return 1
}

httpdmng_reconfigure()
{
	local httpdmng_tool="${PRODUCT_ROOT_D}/admin/sbin/httpdmng"
	local what="$1"
	shift

	case "$what" in
		all|server) : ;;
		*) simply_die "Unexpected argument '$what' to httpdmng_reconfigure()" ;;
	esac

	if [ ! -x "$httpdmng_tool" ]; then
		p_echo "Unable to do 'httpdmng --reconfigure-$what': the utility is not yet on its place. Maybe will attempt later."
		return 1
	fi

	! "$httpdmng_tool" "--reconfigure-$what" -service-node local "$@" >> $product_log 2>&1 || return 0

	# There are 2 well known possible reasons of webserver templates reconfiguration failure:
	# 1. domain/server/webmail templates are well, but arbitrary domain unable to be configured due to  the some reasons,
	#	for example broken vhost structure
	# 2. broken domain/server/webmail templates


	# STEP 1. try to rollback to the default templates
	! httpdmng_reconfigure_rollback_to_default_templates "${what}" "$@" || return 0

	if [ "all" = "${what}" ]; then
	# STEP 2. ignore broken domains
		pp_echo "Warning: web server configuration is broken. We will try to repair it. This operation can take a lot of time, please do not interrupt the process."
		p_echo "Unable to rebuild web server configuration, possible there are broken domains"
		local inten="reconfigure web-server configurations skipping broken domains"
		echo_try "${inten}"
		if "$httpdmng_tool" "--reconfigure-$what" "-skip-broken" -service-node local "$@" >> $product_log 2>&1; then
			suc
			return 0
		else
	# STEP 3. try to rollback to the default templates, broken domains are ignored (possibly problem are in the server/webmail configs)
			warn "${inten}"
			! httpdmng_reconfigure_rollback_to_default_templates "${what}" "-skip-broken" -service-node local "$@" || return 0
		fi
	fi

	return 1
}

# NOTE: do not use function directly, it just helper for httpdmng_reconfigure !!!
# here we try to reconfigure webserver configs using default temnplates
httpdmng_reconfigure_rollback_to_default_templates()
{
	local template_d="${PRODUCT_ROOT_D}/admin/conf/templates"
	local inten="rebuild web server configs with default templates after reconfiguration failure"
	local tmp_d=
	local what="$1"
	shift

	[ -d "${template_d}/custom" ] || return 1
	p_echo "Unable to rebuild web server configuration with currently active default templates"
	echo_try "${inten}"

	if ! tmp_d=`mktemp -d "${template_d}/broken_templates_XXXXXXXX" 2>>$product_log`; then
		warn "${inten} - unable to create temporary directory for custom templates"
		return 1
	fi
	if ! mv -f "${template_d}/custom/"* "${tmp_d}"; then
		warn "${inten} - unable to move custom templates into ${tmp_d}"
		return 1
	fi
	rm -rf "${template_d}/custom"

	if ! "$httpdmng_tool" "--reconfigure-$what" -service-node local "$@" >> $product_log 2>&1; then
		warn "${inten} - using default templates does not solve the problem, restore custom ones"
		mv -f "${tmp_d}" "${template_d}/custom"
		return 1
	fi

	suc
	pp_echo "Custom templates which are breaking web server reconfiguration were moved to ${tmp_d}"
	cat <<EOF >"${tmp_d}/README"
This directory contains broken custom webserver templates which are breaking web server reconfiguration.
Custom templates were moved here on `date` during product update or upgrade.
EOF
}

move_out_custom_vhost_templates()
{
	local move_to="$1"
	local template_d="$PRODUCT_ROOT_D/admin/conf/templates"
	local move_to_d=

	[ -d "$template_d/custom" ] || return 0

	if ! move_to_d="`mktemp -d "$template_d/${move_to}_XXXXXXXX" 2>>"$product_log"`"; then
		warn "move out custom web server templates - cannot create temporary directory under $template_d"
		return 1
	fi

	local inten="move out custom web server templates from $template_d/custom/ to $move_to_d/"
	echo_try "$inten"

	if ! mv -f "$template_d/custom"/* "$move_to_d/"; then
		warn "$inten - unable to move custom templates"
		return 1
	fi

	rm -rf "$template_d/custom"

	suc

	pp_echo "Custom web server templates that may break web server reconfiguration were moved from $template_d/custom/ to $move_to_d/"
	cat <<-EOF > "$move_to_d/README"
		This directory contains previously active custom web server templates that
		may break web server reconfiguration.
		Custom templates were moved here on `date`
		during product update or upgrade.
		EOF
}

php_upgrade_system_handlers_pre_12_1_18()
{
	local php_versions_json="/etc/psa/php_versions.json"
	cp -f "${php_versions_json}" "${php_versions_json}.pre_upg"
}

php_upgrade_system_handlers_post_12_1_18()
{
	local php_versions_json="/etc/psa/php_versions.json"
	if [ -e "${php_versions_json}.pre_upg" ]; then
		mv -f "${php_versions_json}.pre_upg" "${php_versions_json}"
	fi
}

#-*- vim:ft=sh

repeat_until_succeeds()
{
	local repeat_func="$1"
	local modify_func="$2"
	local max_times="$3"

	for i in `seq 1 "$max_times"`; do
		eval "$repeat_func" && return 0 || eval "$modify_func"
	done
	return 1
}

apsc_try_create_db()
{
	local mysql_db_name="mysql"
	echo "CREATE DATABASE $apsc_db_name" | mysql >> "$product_log" 2>&1
}

apsc_modify_db_name()
{
	apsc_db_name="${apsc_db_name_base}_`get_random_string 4 'a-z0-9'`"
}

apsc_try_create_db_user()
{
	local mysql_db_name="mysql"
	echo "CREATE USER '$apsc_db_user'@'localhost' IDENTIFIED BY '$apsc_db_passwd'" | mysql >> "$product_log" 2>&1 && 	echo "GRANT ALL PRIVILEGES ON $apsc_db_name.* TO '$apsc_db_user'@'localhost'" | mysql >> "$product_log" 2>&1
}

apsc_modify_db_user()
{
	apsc_db_user="${apsc_db_user_base}_`get_random_string 4 'a-z0-9'`"
}

apsc_init_db()
{
	local mysql_db_name="$apsc_db_name"
	local mysql_user="--user=$apsc_db_user"
	local mysql_passwd="$apsc_db_passwd"
	local sql_dump_file="$PRODUCT_BOOTSTRAPPER_DIR/db/apsc_mysql_db.sql"

	[ -r "$sql_dump_file" ] || return 1
	mysql < "$sql_dump_file" >> "$product_log" 2>&1
}

apsc_remove_db()
{
	[ -n "$apsc_db_name" ] || return 1
	local mysql_db_name="mysql"
	echo "DROP DATABASE $apsc_db_name" | mysql >> "$product_log" 2>&1
}

apsc_remove_db_user()
{
	[ -n "$apsc_db_user" ] || return 1
	local mysql_db_name="mysql"
	echo "DROP USER '$apsc_db_user'@'localhost'" | mysql >> "$product_log" 2>&1
}

set_apsc_params()
{
	odbc_config_bin="/usr/bin/odbc_config" # present at least on CentOS 5
	if [ -x "$odbc_config_bin" ]; then
		odbc_dsn_conf=`$odbc_config_bin --odbcini 2>> "$product_log"`
		odbc_drivers_conf=`$odbc_config_bin --odbcinstini 2>> "$product_log"`
	fi
	if [ -z "$odbc_dsn_conf" ]; then
		odbc_dsn_conf="/etc/odbc.ini"
	fi
	if [ -z "$odbc_drivers_conf" ]; then
		odbc_drivers_conf="/etc/odbcinst.ini"
	fi

	odbc_isql_bin="/usr/bin/isql"
	odbc_iusql_bin="/usr/bin/iusql"

	odbc_mysql_drivers="/usr/lib/x86_64-linux-gnu/odbc/libmyodbc5w-plesk.so"
	for d in $odbc_mysql_drivers; do
		[ -s $d ] || continue
		odbc_mysql_driver="$d"
		break
	done

	if [ -z "$odbc_mysql_driver" ]; then
		die "No ODBC MySQL mysql drivers found, trying $odbc_mysql_drivers"
	fi

	if [ -s "/usr/lib/x86_64-linux-gnu/sw/libmysqlserver.so.2.0" ]; then
		apsc_driver_library="/usr/lib/x86_64-linux-gnu/sw/libmysqlserver.so.2.0"
	elif [ -s "/usr/lib/x86_64-linux-gnu/libmysqlserver.so.2.0" ]; then
		apsc_driver_library="/usr/lib/x86_64-linux-gnu/libmysqlserver.so.2.0"
	else
		die "find MySQL mysql platform driver for APSC"
	fi
}

apsc_sanity_check_binaries()
{
	# To easily detect breakage on new OSes
	[ -f "$odbc_mysql_driver" ] || die "find MySQL ODBC driver"
	[ -f "$apsc_driver_library" ] || die "find MySQL platform driver library"
}

read_apsc_connection_param_db()
{
	local param="$1"
	db_select "SELECT val FROM misc WHERE param like 'aps_$param'"
	echo "$db_select_output"
}

read_apsc_connection_params_db()
{
	# Warning: password can be encrypted. Don't use it.
	apsc_db_name=`read_apsc_connection_param_db "database"`
	apsc_db_user=`read_apsc_connection_param_db "login"`
	apsc_db_passwd=`read_apsc_connection_param_db "password"`
	apsc_db_host=`read_apsc_connection_param_db "host"`
}

save_apsc_connection_params_db()
{
	local register_script="$PRODUCT_ROOT_D/admin/plib/scripts/register_apsc_database.php"
	[ -f "$register_script" ] || die 'can not find "register_apsc_database.php"'

	sw_engine_pleskrun "$register_script" --register 		-host 'localhost' 		-port 3306 		-database "$apsc_db_name" 		-login "$apsc_db_user" 		-password "$apsc_db_passwd" >> "$product_log" 2>&1
}

save_apsc_connection_params_conf()
{
	conf_setval "$prod_conf_t" APS_DB_DRIVER_LIBRARY "$apsc_driver_library"
}

save_apsc_connection_params()
{
	save_apsc_connection_params_conf || die "store driver library path for APS controller db connection"
	save_apsc_connection_params_db || die "store APS controller db connection parameters into product db"
}

check_ini_section_exists()
{
	local section="$1"
	local file="$2"
	grep -q "\[$section\]" "$file" >/dev/null 2>&1
}

check_odbc_driver_exists()
{
	# We could use odbcinst here, but it wouldn't save us much trouble
	local driver="$1"
	check_ini_section_exists "$driver" "$odbc_drivers_conf"
}

check_odbc_dsn_exists()
{
	local dsn="$1"
	check_ini_section_exists "$dsn" "$odbc_dsn_conf"
}

remove_ini_section()
{
	local section="$1"
	local file="$2"
	[ -r "$file" ] || return 0
	awk "/\[.*\]/ { del = 0 } /\[$section\]/ { del = 1 } ( del != 1 ) { print }" "$file" > "$file.new" && 	mv -f "$file.new" "$file"
}

remove_odbc_driver()
{
	local driver="$1"
	remove_ini_section "$driver" "$odbc_drivers_conf"
}

remove_odbc_dsn()
{
	local dsn="$1"
	remove_ini_section "$dsn" "$odbc_dsn_conf"
}

apsc_try_create_odbc_driver()
{
	check_odbc_driver_exists "$apsc_odbc_driver_name" && return 1

	local odbc_mysql_driver64=
	if echo "$odbc_mysql_driver" | grep -q lib64 2>/dev/null ; then
		odbc_mysql_driver64="$odbc_mysql_driver"
	fi

	cp -f "$odbc_drivers_conf" "$odbc_drivers_conf.new" && 	cat <<EOF >> "$odbc_drivers_conf.new" && mv -f "$odbc_drivers_conf.new" "$odbc_drivers_conf"
[$apsc_odbc_driver_name]
Description = MySQL driver for Plesk
Driver      = $odbc_mysql_driver
Setup       = 
FileUsage   = 1
Driver64    = $odbc_mysql_driver64
Setup64     = 
UsageCount  = 1

EOF
	local rc="$?"
	[ "$rc" -eq 0 ] || rm -f "$odbc_drivers_conf.new"
	return "$rc"
}

apsc_try_create_odbc_dsn()
{
	check_odbc_dsn_exists "$apsc_dsn_name" && return 1

	cp -f "$odbc_dsn_conf" "$odbc_dsn_conf.new" && 	cat <<EOF >> "$odbc_dsn_conf.new" && mv -f "$odbc_dsn_conf.new" "$odbc_dsn_conf"
[$apsc_dsn_name]
Description = MySQL APS controller database (temporary DSN)
Driver      = $apsc_odbc_driver_name
SERVER      = localhost
USER        = $apsc_db_user
PASSWORD    = $apsc_db_passwd
PORT        = 3306
DATABASE    = $apsc_db_name

EOF
	local rc="$?"
	[ "$rc" -eq 0 ] || rm -f "$odbc_dsn_conf.new"
	return "$rc"
}

apsc_modify_odbc_driver()
{
	# just remove the one that bothers us
	remove_odbc_driver "$apsc_odbc_driver_name"
}

apsc_modify_odbc_dsn()
{
	apsc_dsn_name="${apsc_dsn_name_base}_`get_random_string 4 'a-z0-9'`"
}

check_apsc_installed()
{
	local db_name=`read_apsc_connection_param_db "database"`

	[ -n "$db_name" ] || return 1
	db_test_database "$db_name"
}

odbc_test_connection_using_dsn()
{
	local dsn="$1"
	local isql_bin="$2"
	[ -n "$isql_bin" ] || return 1
	p_echo "Attempting to check APS db connection via ODBC, using $isql_bin (empty output is good)"
	if [ ! -x "$isql_bin" ]; then
		p_echo " Executable $isql_bin was not found, skipped."
		return 1
	fi

	echo "" | $isql_bin -b "$dsn" >> "$product_log" 2>&1
	local rc="$?"

	[ "$rc" -eq 0 ] || p_echo " Failed, rc = $rc."
	return $rc
}

odbc_test_connection()
{
	# This function should be reimplemented via sw-engine. Until then skip the check.
	p_echo "APS db accessibility check was skipped."
	return 0

	if [ ! -x "$odbc_isql_bin" -a ! -x "$odbc_iusql_bin" ]; then
		warn "isql utilities were not found. APS db accessibility check was skipped."
		return 0
	fi

	local apsc_dsn_name_base="apsc"
	local apsc_dsn_name="$apsc_dsn_name_base"

	[ -r "$odbc_dsn_conf" ] || die "find ODBC config"
	repeat_until_succeeds apsc_try_create_odbc_dsn apsc_modify_odbc_dsn 50 || die "create ODBC DSN"

	odbc_test_connection_using_dsn "$apsc_dsn_name" "$odbc_isql_bin" || 	odbc_test_connection_using_dsn "$apsc_dsn_name" "$odbc_iusql_bin"
	local rc="$?"

	remove_odbc_dsn "$apsc_dsn_name"

	return $rc
}

install_apsc()
{
	set_apsc_params

	if ! check_apsc_installed; then
		inten="set up APS controller database"
		echo_try "$inten"

		apsc_db_name_base="apsc"
		apsc_db_name="$apsc_db_name_base"
		apsc_db_user_base="apsc"
		apsc_db_user="$apsc_db_user_base"
		apsc_db_passwd=`get_random_string 12 'A-Za-z0-9_'`

		repeat_until_succeeds apsc_try_create_db      apsc_modify_db_name 50 && 		repeat_until_succeeds apsc_try_create_db_user apsc_modify_db_user 50 && 		apsc_init_db || die "create database for APS controller"

		apsc_odbc_driver_name="MySQL"

		[ -r "$odbc_drivers_conf" ] || touch "$odbc_drivers_conf"
		apsc_sanity_check_binaries
		repeat_until_succeeds apsc_try_create_odbc_driver apsc_modify_odbc_driver 50 || die "create ODBC driver configuration"

		if odbc_test_connection ; then
			save_apsc_current_db_version "$product_db_version"
			save_apsc_connection_params

			suc
		else
			p_echo "Failed to establish test connection. Cleaning up."

			remove_odbc_driver "$apsc_odbc_driver_name"
			apsc_remove_db_user || true
			apsc_remove_db || true

			die "establish test connection to APS database via ODBC"
		fi
	fi
}

# vim:ft=sh:

# For upgrade we use admin credentials instead of apsc since password for the 
# latter might be encrypted.

apsc_upgrader_repair_arg()
{
        if [ -n "$1" -o -n "$do_repair" ]; then
                echo "--repair"
        fi
}

apsc_upgrader_action_name()
{
        if [ -n "`apsc_upgrader_repair_arg $1`" ]; then
                echo "upgrade and repair"
        else
                echo "upgrade"
        fi
}

upgrade_apsc_prep_stage()
{
	local action_name="`apsc_upgrader_action_name $1`"
	local optional_repair_arg="`apsc_upgrader_repair_arg $1`"

	# if DB is not created (clean installation or upgrade from versions where there is no APS DB) just exit from PRE stage
	# DB will be created in POST stage
	check_apsc_installed || return 0


	read_apsc_connection_params_db
	pp_echo "===> Cumulative APS controller database ($apsc_db_name) ${action_name} has been started."

	local apsc_db_version=`read_apsc_current_db_version`
	if [ -z "$apsc_db_version" ] || ! expr "$apsc_db_version" + 0 >/dev/null 2>&1 ; then
		# db_version is not set or invalid. Create it.
		apsc_db_version="$prev_short_version"
		save_apsc_current_db_version "$apsc_db_version"
	fi


	cu_run_upgrader "apsc" apsc_upgrader 		--load-config-data apsc_cu_load_config_data --store-config-data apsc_cu_store_config_data 		--prepare apsc_upgrader_prepare --rollback apsc_upgrader_rollback 		--from-version "$apsc_db_version" --stage "prep" $optional_repair_arg

	if [ "$?" -eq 0 ]; then
		pp_echo "===> Cumulative ${action_name} of APS controller database has been completed."
	else
		pp_echo "===> APS controller database was not upgraded completely. See installation log for details."
	fi
}

upgrade_apsc_post_stage()
{
	read_apsc_connection_params_db

	local action_name="`apsc_upgrader_action_name $1`"
	local optional_repair_arg="`apsc_upgrader_repair_arg $1`"
	pp_echo "===> Cumulative APS controller $action_name (final stage) has been started."

	if ! check_apsc_installed; then
		install_apsc
	else 	
		local apsc_db_version=`read_apsc_current_db_version`
		cu_run_upgrader "apsc" apsc_upgrader 			--load-config-data apsc_cu_load_config_data --store-config-data apsc_cu_store_config_data 			--commit apsc_upgrader_commit 			--from-version "$apsc_db_version" --stage "post" $optional_repair_arg --update-version
	fi

	if [ "$?" -eq 0 ]; then
		pp_echo "===> Cumulative $action_name of APS controller (final stage) has been completed."
	else
		pp_echo "===> APS controler settings was not upgraded completely. See installation log for details."
	fi
}

apsc_run_repair()
{
	local ret=0
	/usr/lib/plesk-9.0/vps_optimized_aspects/mysql-innodb off >> $product_log 2>&1
	if ! check_apsc_installed; then
		install_apsc
	else
		upgrade_apsc_prep_stage repair

		if ! apsc_passwd_repair; then
			warn "fix credentials for APS controller database"
			ret=1
		fi

		upgrade_apsc_post_stage repair
	fi
	return $ret
}

apsc_passwd_repair()
{
	check_apsc_installed || return 0
	read_apsc_connection_params_db

	is_local_mysql_instance "$apsc_db_host" || return 0

	! sw_engine_pleskrun "$PRODUCT_ROOT_D/admin/plib/scripts/check_apsc_connection.php" >>"$product_log" 2>&1 || return 0

	local apsc_db_passwd
	apsc_db_passwd=`get_random_string 12 'A-Za-z0-9_'`
	reset_user_password "$apsc_db_user" "$apsc_db_passwd" "$apsc_db_host" && 	save_apsc_connection_params_db && 	sw_engine_pleskrun "$PRODUCT_ROOT_D/admin/plib/scripts/check_apsc_connection.php" >>"$product_log" 2>&1

}

create_apsc_meta_info_table()
{
	local mysql_db_name="$apsc_db_name"
	local query="
		CREATE TABLE IF NOT EXISTS meta_info (
		    param   VARCHAR(255) CHARACTER SET ascii COLLATE ascii_bin NOT NULL PRIMARY KEY
		,   value   VARCHAR(2000) CHARACTER SET binary NOT NULL
		) ENGINE=InnoDB DEFAULT CHARSET=utf8;"
	
	echo "$query" | mysql >> "$product_log" 2>&1
}

read_apsc_current_db_version()
{
	local mysql_db_name="$apsc_db_name"
	db_test_table "meta_info" || return 0
	apsc_read_meta_info_attr 'db_version'
}

save_apsc_current_db_version()
{
	local apsc_db_version="$1"
	local mysql_db_name="$apsc_db_name"
	db_test_table "meta_info" || create_apsc_meta_info_table || die "create APSC meta_info DB table"
	db_do "REPLACE INTO meta_info (param, value) VALUES ('db_version', '$apsc_db_version')"
}

add_apsc_db_changes()
{
	local apsc_db_version="$1"
	local inten="upgrade APSC SQL database from $apsc_db_version"
	local upgrade_db_sql="$PRODUCT_BOOTSTRAPPER_DIR/db/apsc_mysql_db_upgrade_$apsc_db_version.sql"
	local sed_script="s/@@APSC_DATABASE@@/$apsc_db_name/g"

	local mysql_db_name="$apsc_db_name"
	sed -e "$sed_script" "$upgrade_db_sql" | mysql >> "$product_log" 2>&1 || die "$inten"
}

apsc_innodb_engine()
{
	local mysql_db_name="$apsc_db_name"
	db_select "SELECT TABLE_NAME FROM information_schema.TABLES where TABLE_SCHEMA = '$apsc_db_name' and ENGINE = 'MyISAM'"
	for tableName in $db_select_output; do
		db_do "ALTER TABLE $tableName ENGINE=InnoDB"
	done
}

apsc_read_meta_info_attr()
{
	db_select "SELECT value FROM meta_info WHERE param = '$1'"
	echo "$db_select_output"
}

apsc_cu_load_config_data()
{
	local upgrader_name="$1"
	local mysql_db_name="$apsc_db_name"

	[ "$upgrader_name" = "apsc" ] || return 0
	db_test_table "meta_info" || return 0

	cu_db_current_timestamp="`  apsc_read_meta_info_attr 'cu_current_timestamp'`"
	cu_db_continue_from="`      apsc_read_meta_info_attr 'cu_continue_from'`"
	cu_db_failed_actions_list="`apsc_read_meta_info_attr 'cu_failed_actions_list'`"
	cu_db_latest_timestamp="`   apsc_read_meta_info_attr 'cu_latest_timestamp'`"
}

apsc_cu_store_config_data()
{
	local upgrader_name="$1"
	local mysql_db_name="$apsc_db_name"

	[ "$upgrader_name" = "apsc" ] || return 0
	db_test_table "meta_info" || create_apsc_meta_info_table || die "create APSC meta_info DB table"

	db_do "REPLACE INTO meta_info (param, value) VALUES 		('cu_current_timestamp', '$cu_db_current_timestamp'), 		('cu_continue_from', '$cu_db_continue_from'), 		('cu_failed_actions_list', '$cu_db_failed_actions_list'), 		('cu_latest_timestamp', '$cu_db_latest_timestamp')"
}

apsc_upgrader_prepare()
{
	local apsc_db_name=`read_apsc_connection_param_db "database"`
	db_backup "mysql.preupgrade.apsc.${prev_version}-${product_version}" "$apsc_db_name"
	APSC_UPGRADER_DB_BACKUP="$dfile"
}

apsc_upgrader_rollback()
{
	[ -n "$APSC_UPGRADER_DB_BACKUP" ] || return 1

	pp_echo "===> Restoring APS controller database from backup '$APSC_UPGRADER_DB_BACKUP'"
	zcat "$APSC_UPGRADER_DB_BACKUP" | mysql_anydb || {
		pp_echo "     [ERROR] Database restoration failed. Try to restore database from '$APSC_UPGRADER_DB_BACKUP' manually."
		return
	}

	APSC_UPGRADER_DB_BACKUP=
}

apsc_upgrader_commit()
{
	save_apsc_current_db_version "$product_db_version"
}

apsc_upgrader()
{
	local ver="cu_check_version"
	local db="cu_check_stage prep"
	local files="cu_check_stage post"
	local do="cu_do"

	cu_start_upgrade_from_version 1101
	$db    $ver $do apsc_innodb_engine
	$db    $ver $do add_apsc_db_changes 11.1.0

	cu_start_upgrade_from_version 011006003
	$files $ver $do update_apsc_driver_library_path

	cu_start_upgrade_from_version 012001023
	$files $ver $do update_apsc_odbc_driver


	# DO NOT MODIFY ANYTHING BELOW THIS LINE! (without approval of review from the entire Linux team)
	call_optional_function dev_apsc_upgrader
	cu_verify_latest_timestamp_lt '2015/08/14 12:00:00'
}



# ----- APSC upgrade routines start here -----

# For a function to be recognized as an APSC db upgrader, you should call it from apsc_upgrader()
# with appropriate "tags", e.g.:
#  $ts "2012/11/12 15:08:45" $do APSC_DB_UPGRADE_FUNCTION_NAME
# where "2012/11/12 15:08:45" is a timestamp that is produced by `date '+%Y/%m/%d %H:%M:%S'` 
# or `date '+%Y%m%d%H%M%S'` or any other sane format with the same order of format specifiers.
#
# Please note that to safely use '$ts' here, one should call 'cu_upgrader_conf_init "apsc" ...'
# somewhere during APSC DB installation.
# 
# Upgrade function names may be arbitrary, but let's prefix them with APSC_DB_UPGRADE_FUNCTION_.


update_apsc_driver_library_path()
{
	set_apsc_params
	save_apsc_connection_params_conf
}

update_apsc_odbc_driver()
{
	set_apsc_params
	apsc_odbc_driver_name="MySQL"
	apsc_modify_odbc_driver
	repeat_until_succeeds apsc_try_create_odbc_driver apsc_modify_odbc_driver 50 || :
}

#awstats

set_awstats_params()
{
	AWSTATS_ROOT_D="$PRODUCT_ROOT_D/awstats"
}

## @@constructor set_awstats_params


update_statistics_path_in_awstats_confs()
{
	[ -z "$*" ] && return
	sed -i '{ /\/system\/[^/]\+\/statistics\//b; s:\(DirData=\)\"\(.*\)/\([^/]\+\)/statistics/\(.*\)\":\1\"\2/system/\3/statistics/\4\": }' "$@"
}

upgrade_domains_awstats_conf_11_5_24()
{
	update_statistics_path_in_awstats_confs `ls $PRODUCT_ROOT_D/etc/awstats/*.conf 2>/dev/null`
}

# Backup manager tables will be managed by plesk

pbm_create_certificate()
{
	local inten="generate certificate for backup signing"

	[ -x "$PRODUCT_ROOT_D/admin/sbin/backup_sign" ] || return 0

	if db_select "select val from misc where param='bu_cert_id'" && [ -n "$db_select_output"  ]; then
#	Nothing to be done
		return
	fi

	echo_try $inten

	if ! $PRODUCT_ROOT_D/admin/sbin/backup_sign generate >> $product_log 2>&1; then
		warn "backup certificate generation"
		return 1
	fi

	suc
}

install_cron_backup()
{
	local backup_file_before_10_5_24="/etc/cron.d/plesk-backup-manager"
	local backup_file="/etc/cron.d/plesk-backup-manager-task"

	if [ -s "${backup_file_before_10_5_24}" ]; then
		mv -f "${backup_file_before_10_5_24}" "${backup_file}"
	else
		rm -f "${backup_file_before_10_5_24}"
	fi

	[ ! -s "${backup_file}" ] || return 0
	minutes=`get_random_number 15`

	cat >"${backup_file}" <<-EOF
$minutes,`expr $minutes + 15`,`expr $minutes + 30`,`expr $minutes + 45`	*	*	*	*	root [ -x $PRODUCT_ROOT_D/admin/sbin/backupmng ] && $PRODUCT_ROOT_D/admin/sbin/backupmng >/dev/null 2>&1
EOF
}
# vim:ft=sh:

chroot_warning_prgtype()
{
	local prg type
	prg="$1"
	type="$2"

	p_echo "$prg: $type"
	p_echo "probably it will not work in chrooted accounts"
	warn "register $prg in chrooted environment"
}

install_chroot_program()
{
	local inten prg cp_update filetype libdir libdir64
	inten="register $1 in chroot"
	prg="$1"
	libdir="$2"
	libdir64="$3"

	filetype=`file -ib "$prg"`

	case "$filetype" in
	application/x-executable*statically' 'linked*) 	;;
	application/x-shellscript*|text/x-shellscript*) ;;
	application/x-executable*|application/x-sharedlib*)
		# sometimes 'file' say the executable is a shared library
		# newer OSes also say just "executable" without distibguishind between static and dynamic
		ldd "$prg" | awk '/not found/ { print; exit 1; }
						/not a dynamic executable/ { exit 0; }
						NR=4 && $1 !~ /^\// && $2=="=>" && $4 ~ /\(0x[[:xdigit:]]*\)/ {
							  if ($3 ~ /\/lib64\//) {
								dir = DIR64;
							  } else {
								dir = DIR;
							  }
							  if (system("cp -v -p " CP_UPDATE " " $3 " " dir "/" $1)!=0)
								{ exit 1; }
							}
					' DIR="$libdir" DIR64="$libdir64" CP_UPDATE="$cp_update" >> "$product_log" 2>&1 || report_problem "$inten"
		;;
	application/octet-strea*)
		;;
	*)
		chroot_warning_prgtype "$prg" "$filetype"
		;;
	esac
}

install_chroot_base()
{
	local bindir libdir libdir64

	CHROOT_ROOT_D="$HTTPD_VHOSTS_D/chroot"
	CHROOT_PROGRAMS="bash cat chmod cp grep head id less ln ls mkdir more 			mv pwd rm rmdir sash sh tail touch du vi true false scp groups 			wget zip unzip tar gzip gunzip curl unrar"

	os=`uname -s`
	if [ "${os}" != "Linux" ]; then
		die "unknown system $os" ;
	fi

	cp_update=-u
	bindir="$CHROOT_ROOT_D/bin"
	libdir="$CHROOT_ROOT_D/lib"

	syslibdir64="`readlink -f -m /lib64`"
	libdir64="$CHROOT_ROOT_D/$syslibdir64"

	copy="cp -v -p $cp_update"
	echo_try "install chrooted environment"

	# NOTE: previously during chroot upgrade all chroot environments in vhosts were removed and then added again via chrootmng.
	# This behavior was removed to support remote SN. Not sure it is in fact needed.
	# See also http://plesk-process.parallels.com/TargetProcess2/Project/Planning/UserStory/View.aspx?UserStoryID=70752

	install -m 750 -d "$CHROOT_ROOT_D"
	chown 0:0 "$CHROOT_ROOT_D"
	mkdir -m 755 -p "$bindir"
	mkdir -m 755 -p "$libdir"
	[ -d "$syslibdir64" ] && mkdir -m 755 -p "$libdir64"
	mkdir -m 755 -p "$CHROOT_ROOT_D/usr"
	mkdir -m 755 -p "$CHROOT_ROOT_D/usr/lib"


	[ -d /usr/libexec ] && mkdir -m 755 -p "$CHROOT_ROOT_D/usr/libexec"
	[ -d /libexec ] && mkdir -m 755 -p "$CHROOT_ROOT_D/libexec"
	mkdir -m 755 -p "$CHROOT_ROOT_D/etc"
	mkdir -m 1777 -p "$CHROOT_ROOT_D/tmp"
	mkdir -m 755 -p "$CHROOT_ROOT_D/var"
	mkdir -m 1777 -p "$CHROOT_ROOT_D/var/tmp"

	mkdir -m 755 -p "$CHROOT_ROOT_D/dev"
	[ -e "$CHROOT_ROOT_D/dev/null" ] || mknod -m 666 "$CHROOT_ROOT_D/dev/null" c 1 3
	chown root:root "$CHROOT_ROOT_D/dev/null"

	[ -e "$CHROOT_ROOT_D/dev/random" ] || mknod -m 644 "$CHROOT_ROOT_D/dev/random" c 1 8
	[ -e "$CHROOT_ROOT_D/dev/urandom" ] || mknod -m 644 "$CHROOT_ROOT_D/dev/urandom" c 1 9
	chown root:root "$CHROOT_ROOT_D/dev/random" "$CHROOT_ROOT_D/dev/urandom"


	cd "$bindir"
	for f in $CHROOT_PROGRAMS; do
		path=`which "$f" 2>>"$product_log"`
		found=$?
		if test $found -eq 0 -a -x "$path"; then
			cp -v -p "$path" "$f" >> "$product_log" 2>&1
			install_chroot_program "$f" "$libdir" "$libdir64"
		fi
	done

	echo_try "install sftp-server binary into chroot environment"

	sftp_path="$(awk '/^Subsystem[[:space:]]sftp.*$/ {print $3}' < /etc/ssh/sshd_config)"
	sftp_dir="${sftp_path%/*}"

	if [ -z "$sftp_dir" ]; then
		echo "Warning: sftp-server binary not found" >> "$product_log"
		echo "+ sftp connections will not be available for chrooted accounts" >> "$product_log"
		warn "installation of sftp-server binary into chroot environment"
	else
		mkdir -p "$CHROOT_ROOT_D$sftp_dir" >> "$product_log" 2>&1
		cp -v -p "$sftp_path" "$CHROOT_ROOT_D$sftp_path" >> "$product_log" 2>&1
		suc
		install_chroot_program "$CHROOT_ROOT_D$sftp_path" "$libdir" "$libdir64"
	fi

	for l in lib lib64 lib/i386-linux-gnu lib/x86_64-linux-gnu; do
		[ -d "/$l" ] || continue
		mkdir -p "$CHROOT_ROOT_D/$l"
		$copy /$l/ld-linux* /$l/libnss_*.so.2 "$CHROOT_ROOT_D/$l"
	done >> "$product_log" 2>&1

	if test -r /etc/termcap; then
		cp -v -p /etc/termcap "$CHROOT_ROOT_D/etc/termcap" >> "$product_log" 2>&1
	fi

	touch "$CHROOT_ROOT_D/etc/passwd"
	touch "$CHROOT_ROOT_D/etc/group"

	# /etc/resolv.conf may be a symlink on some systems (e.g., on Debian if resolvconf package is installed)
	if [ -f "/etc/resolv.conf" ]; then
		local hardlink_symlink_tgt="ln -v -f -T"
		# ln -L is not supported on some old OSes, such as CentOS/RedHat 5, Ubuntu 10.04. Hope nobody uses symlinks there.
		ln -L 2>&1 | grep -q invalid || hardlink_symlink_tgt="$hardlink_symlink_tgt -L"
		$hardlink_symlink_tgt "/etc/resolv.conf" "$CHROOT_ROOT_D/etc/resolv.conf" >> "$product_log" 2>&1
	fi

	if test -d "$PRODUCT_ROOT_D/chroot"; then
		cd "$PRODUCT_ROOT_D/chroot"

		for f in *; do
			tar cpf - -C "$f" . | tar xpf - -C "$CHROOT_ROOT_D" || report_problem "copy $f to chroot parttern"
			find $d -type f -perm +111 | while read p; do
				install_chroot_program "$p" "$libdir" "$libdir64"
			done
		done
	fi

	suc

	selinux_relabel_dir -F "$CHROOT_ROOT_D"
	selinux_relabel_dir -F "/etc/resolv.conf"
}

# import certificate into database when installing
import_default_certificate()
{
	db_select "SELECT val FROM misc WHERE param = 'default_certificate_id'"

	local default_certificate_id="$db_select_output"

	if [ -z "$default_certificate_id" ]; then
# Update misc table if needed
		echo_try "import default certificate"
		import_certificate "$certificate_file"  "default certificate"

		db_do "REPLACE INTO misc set param='default_certificate_id', val='$last_certificate_id'"
		default_certificate_id="$last_certificate_id"
		suc
	fi

	db_select "SELECT val FROM misc WHERE param = 'cert_rep_id'"
	local admin_repository_id="$db_select_output"

	if [ -z "$admin_repository_id" ]; then
		echo_try "push default certificate into repository"
		db_select "REPLACE INTO Repository SET component_id='${default_certificate_id}';SELECT LAST_INSERT_ID();"
		admin_repository_id="$db_select_output"
		db_do "REPLACE INTO misc set param='cert_rep_id', val='${admin_repository_id}'"
	else
#	simple certificate insertion to repository
		db_do "REPLACE INTO Repository SET rep_id='${admin_repository_id}', component_id='${default_certificate_id}'"
	fi
}

# import certificate from a file to certificates table
# $1 - certoificate file path
# $2 - short description of certificate
# last_certificate_id variable will contain its identifier
import_certificate()
{
	local certificate_file certificate_comment inten

	certificate_file=$1
	certificate_comment=$2

	cert_pub_val=`cat ${certificate_file} |
		awk 'BEGIN { state=0 }
		{
			if (state == 0) {
				if (index($0, "-BEGIN CERTIFICATE-") != 0) state = 1;
			};
			if (state == 1) {
				print $0;
				if (index($0, "-END CERTIFICATE-") != 0) state = 2;
			}
		}' |
		perl -n -e 's/([^ A-Za-z0-9])/"%" . sprintf("%02X", unpack("C", $1))/ge; s/ /\+/g; print $_;'`

	cert_priv_val=`cat ${certificate_file} |
		awk 'BEGIN { state=0 }
		{
			if (state == 0) {
				if ((index($0, "-BEGIN ") != 0) &&
					(index($0, " PRIVATE KEY-") != 0)) state = 1;
			};

			if (state == 1) {
				print $0;
				if ((index($0, "-END ") != 0) &&
					(index($0, " PRIVATE KEY-") != 0)) state = 2;
			}
		}' |
		perl -n -e 's/([^ A-Za-z0-9])/"%" . sprintf("%02X", unpack("C", $1))/ge; s/ /\+/g; print $_;'`

	if [ ! -d "${PRODUCT_ROOT_D}"/var/certificates ]; then
		inten="creating directory for certificates"
		mkdir -p "${PRODUCT_ROOT_D}"/var/certificates || die ${inten}
		chown root:0 "${PRODUCT_ROOT_D}"/var/certificates || die ${inten}
		chmod 500 "${PRODUCT_ROOT_D}"/var/certificates || die ${inten}
	fi

	# copy certificate file to file repository
	newcertfile=`mktemp "${PRODUCT_ROOT_D}"/var/certificates/certXXXXXXX`

	cp "${certificate_file}" "${newcertfile}"  || die "place certificate to its directory"
	chown root:0 "${newcertfile}"
	chmod 400 "${newcertfile}"

	newcertfile=`basename ${newcertfile}`
	db_select "INSERT INTO certificates (pvt_key, cert, name, ca_cert, csr, cert_file, ca_file)
		VALUES ('${cert_priv_val}', '${cert_pub_val}', '${certificate_comment}', '', '',
			'${newcertfile}', '');
		SELECT LAST_INSERT_ID()"

	last_certificate_id="$db_select_output"
}


drop_upgrade_flag()
{
        rm -f "${PRODUCT_ROOT_D}/version.upg"
}

install_rbash()
{
	p_echo

	p_echo "===> Installing rbash"

	p_echo " Checking that $rbash_shell registered as login shell..."
	egrep "^$rbash_shell" /etc/shells >> $product_log 2>&1
	case $? in
		0)	p_echo " $rbash_shell is already registered as login shell"
			;;
		*)	
			if [ -f $bash_shell ]; then
				ln -sf $bash_shell $rbash_shell
				inten="add $rbash_shell to /etc/shells"
				echo_try $inten
				echo "$rbash_shell" >> /etc/shells && suc || die $inten
			fi
			;;
	esac
	p_echo
}
get_old_domain()
{
    if [ "X$domain" = "X" ]; then
		get_config_parameters
    fi
}

stop_before_upgrade()
{
	$START_SH stop 2>/dev/null
}

# vim:ft=sh:

#courier-imap

set_courier_imap_params()
{
	COURIERIMAP_CONFDIR="/etc/courier-imap"
	IMAPD_CERT="/usr/share/imapd.pem"
	POP3D_CERT="/usr/share/pop3d.pem"
	COURIER_DHPARAMS="/usr/share/dhparams.pem"

	# Certificate paths for Courier-IMAP <= 3.0.8
	OLD_IMAPD_CERT="/usr/share/courier-imap/imapd.pem"
	OLD_POP3D_CERT="/usr/share/courier-imap/pop3d.pem"

	COURIER_DELIVER_QUOTA="/usr/bin/deliverquota"

	courier_imapd_service="courier-imapd"
	courier_imaps_service="courier-imaps"
	courier_pop3d_service="courier-pop3d"
	courier_pop3s_service="courier-pop3s"
	courier_authdaemon_service="courier-authdaemon"

	# Service name for Courier-IMAP <= 3.0.8
	old_courier_service="courier-imap"
}

## @@constructor set_courier_imap_params

courier_imap_generate_certs()
{
	# Generates POP3 and IMAP certificates or upgrades their location (3.0.8 => 4.12.0 upgrade)
	local restart_needed=0

	set_courier_imap_params

	if [ ! -f "$IMAPD_CERT" ]; then
		if [ -f "$OLD_IMAPD_CERT" ]; then
			inten="move certificate for Courier-IMAP server for IMAP-SSL during upgrade"
			mv -f "$OLD_IMAPD_CERT" "$IMAPD_CERT" || die "$inten"
		else
			inten="generate Certificate for Courier-IMAP server for IMAP-SSL"
			echo_try $inten
			get_certificate "$IMAPD_CERT" || die $inten
			set_ac 0 0 0400 "$IMAPD_CERT"
			suc
		fi
		restart_needed=110
	fi

	if [ ! -f "$POP3D_CERT" ]; then
		if [ -f "$OLD_POP3D_CERT" ]; then
			inten="move certificate for Courier-IMAP server for POP3-SSL during upgrade"
			mv -f "$OLD_POP3D_CERT" "$POP3D_CERT" || die "$inten"
		else
			inten="copy default SSL Certificate for Courier-IMAP server for POP3-SSL"
			echo_try $inten
			cp -fp "$IMAPD_CERT" "$POP3D_CERT" || die $inten
			suc
		fi
		restart_needed=110
	fi

	courier_imap_generate_dhparams

	return $restart_needed
}

courier_imap_generate_dhparams()
{
	if [ ! -f "$COURIER_DHPARAMS" ]; then
		local inten="generate dhparams file $COURIER_DHPARAMS for Courier-IMAP"
		echo_try "$inten"

		(
			umask 0277
			openssl dhparam -out "$COURIER_DHPARAMS" 1024 >> $product_log 2>&1
		)

		if [ "$?" != "0" ]; then
			warn "${inten}. TLS extension for POP/IMAP may not work."
		else
			suc
		fi
	fi
}

is_courier_imap_active()
{
	# For now it's rather "is Courier-IMAP installed?". Check is the same as in backend.
	set_courier_imap_params

	[ -x "$COURIER_DELIVER_QUOTA" ]
}

courier_imap_services_action()
{
	local action="$1"
	local use_exit_code="$2"
	local ret=0
	pleskrc courier_imapd "$action" || ret=1
	pleskrc courier_imaps "$action" || ret=1
	pleskrc courier_pop3d "$action" || ret=1
	pleskrc courier_pop3s "$action" || ret=1
	[ ! -n "$use_exit_code" ] || return $ret
}

# vim:ft=sh:

set_poplock_params()
{
	set_mail_params

	PLESK_POSTFIX_DB_ROOT="/var/spool/postfix/plesk"
	PLESK_POSTFIX_POP_DB_ROOT="/var/spool/postfix/plesk-pop"
	PLESK_POSTFIX_POPLOCK_DB="hash:/var/spool/postfix/plesk-pop/poplock"
	POSTFIX_POPLOCK_DB_PATH="/var/spool/postfix/plesk-pop/poplock.db"
	POSTFIX_POPLOCK_DB_PERM="0460"
	POSTFIX_POPLOCK_DB_USER="postfix"
	POSTFIX_POPLOCK_DB_GROUP="$MAIL_USERGROUP"

	QMAIL_POPLOCK_DB_DIR="/var/lib/plesk/mail/poplock"
	QMAIL_POPLOCK_DB_PATH="/var/lib/plesk/mail/poplock/poplock.db"
	QMAIL_POPLOCK_DB_PERM="0660"
	QMAIL_POPLOCK_DB_USER="$MAIL_USERGROUP"
	QMAIL_POPLOCK_DB_GROUP="nofiles"
}

poplock_db_qmail_upgrade_to_courier_4_12()
{
	[ -d "$QMAIL_POPLOCK_DB_DIR" ] || return 0
	[ -f "$QMAIL_POPLOCK_DB_PATH" ] || return 0
	get_userID "$QMAIL_POPLOCK_DB_USER" && get_groupID "$QMAIL_POPLOCK_DB_GROUP" || return 0

	set_ac "$QMAIL_POPLOCK_DB_USER" "$QMAIL_POPLOCK_DB_GROUP" 0770 "$QMAIL_POPLOCK_DB_DIR"
	set_ac "$QMAIL_POPLOCK_DB_USER" "$QMAIL_POPLOCK_DB_GROUP" "$QMAIL_POPLOCK_DB_PERM" "$QMAIL_POPLOCK_DB_PATH"
}

poplock_db_postfix_upgrade_to_courier_4_12()
{
	[ -d "$PLESK_POSTFIX_DB_ROOT" -a -d "$PLESK_POSTFIX_POP_DB_ROOT" ] || return 0
	local old_poplock_db_path="$PLESK_POSTFIX_DB_ROOT/`basename $POSTFIX_POPLOCK_DB_PATH`"
	[ -f "$old_poplock_db_path" -a ! -L "$old_poplock_db_path" ] || return 0
	get_userID "$POSTFIX_POPLOCK_DB_USER" && get_groupID "$POSTFIX_POPLOCK_DB_GROUP" || return 0

	mv -f "$old_poplock_db_path" "$POSTFIX_POPLOCK_DB_PATH" 2>>"$product_log"
	set_ac "$POSTFIX_POPLOCK_DB_USER" "$POSTFIX_POPLOCK_DB_GROUP" "$POSTFIX_POPLOCK_DB_PERM" "$POSTFIX_POPLOCK_DB_PATH"
	ln -s "../`basename $PLESK_POSTFIX_POP_DB_ROOT`/`basename $POSTFIX_POPLOCK_DB_PATH`" "$old_poplock_db_path" 2>>"$product_log"

	# we should be able to skip reconfiguring Postfix thanks to compatibility symlink, but we'll do it nonetheless
	[ -s "/etc/postfix/main.cf" ] || return 0
	
	set_postfix_params

	local poplock_hash_basename="`basename $PLESK_POSTFIX_POPLOCK_DB`"
	local filtered_value="`read_postconf_value 'mynetworks' | sed -e 's|^.*/'$poplock_hash_basename'$|'$PLESK_POSTFIX_POPLOCK_DB'|'`"
	write_postconf_value 'mynetworks' "$filtered_value" || return 1
	pleskrc postfix reload
}

upgrade_courier_to_4_12_pop_before_smtp_11_5_23()
{
	set_poplock_params

	# poplock.db files for both QMail and Postfix may be present (e.g., if there was a mail provider switch before upgrade)
	# therefore we attempt to run both upgraders. Each of them just does nothing if not applicable, so that's OK.
	poplock_db_qmail_upgrade_to_courier_4_12 && poplock_db_postfix_upgrade_to_courier_4_12
}
# vim:ft=sh:

set_dovecot_params()
{
	DOVECOT_CONFDIR="/etc/dovecot"
	DOVECOT_INCLUDE_DIR="/etc/dovecot/conf.d"
	DOVECOT_DIST_CONFDIR="/usr/share/doc/plesk-dovecot/dist-config"
	DOVECOT_CERT_DIR="$DOVECOT_CONFDIR/private"
	DOVECOT_CERT="$DOVECOT_CERT_DIR/ssl-cert-and-key.pem"

	DOVECOT_INTERNAL_USERGROUP="dovecot"
	DOVECOT_LOGIN_USERGROUP="dovenull"

	DOVECOT_LDA="/usr/lib/dovecot/dovecot-lda"

	dovecot_service="dovecot"
}

is_dovecot_active()
{
	# For now it's rather "is Dovecot installed?". Check is the same as in backend.
	set_dovecot_params

	[ -x "$DOVECOT_LDA" ]
}

# vim:syntax=sh

install_cron_somely()
{
	install_cron_somely_noop
}

install_cron_somely_noop()
{
	:
}

remove_from_crontab()
{
	program=$1

	echo_try "remove $program from crontab"

	cron_file=`mktemp /tmp/cronXXXXXX`
	$crontab -l 2>/dev/null | egrep -v "^[^#].*$program" > $cron_file
	$crontab $cron_file

	rm -f $cron_file
	suc
}

setup_support_url()
{
	local support_url

	support_url="$1"
	db_do "REPLACE INTO misc (param,val) values('support_url', '$support_url')"
}


# add_db_changes [--fix] version
# Runs ${PRODNAME}_db_upgrade_${version}.sql
# (with --fix, runs ${PRODNAME}_db_upgrade_${version}_fix.sql)
add_db_changes()
{
	if [ "$db_fix_check_stage" = "yes" ]; then
		return
	fi

	local inten upgrade_db_sql fix
	fix=""
	eval `sh_get_args '--fix) fix="_fix";;'`
	inten="upgrade SQL database from $1"
	upgrade_db_sql="$PRODUCT_BOOTSTRAPPER_DIR/db/${PRODNAME}_db_upgrade_${1}${fix}.sql"
	mysql < $upgrade_db_sql >> "$product_log" 2>&1 || die "$inten"
}

# db_fix_add_column [--init def_value] table column type
# Adds a column of the specified type to the table, optionally
# initializing it with def_value, which is an SQL expression. If the
# column already exists, does nothing
db_fix_add_column()
{
	local def_value=""
	eval `sh_get_args '--init) def_value="$2"; shift;;'`
	local table="$1"
	local column="$2"
	local type="$3"

	if [ "$db_fix_check_stage" = "yes" ]; then
		if ! db_test "DESC $table" "\$1 == \"$column\""; then
			if [ -z "$db_added_columns" ]; then
				db_added_columns="$table.$column"
			else
				db_added_columns="$db_added_columns|$table.$column"
			fi
		fi
		return
	fi

	local query="ALTER TABLE \`$table\` ADD \`$column\` $type"
	if [ -n "$def_value" ]; then
		query="$query; UPDATE \`$table\` SET \`$column\`=$def_value"
	fi
	db_do_if_not "DESC $table" "\$1 == \"$column\"" "$query"
}

# db_fix_change_column [--init def_value] table column type
# Changes a column of the specified type to the table, optionally
# initializing it with def_value, which is an SQL expression.
# Changes a column from the table if it exists
db_fix_change_column()
{
	if [ "$db_fix_check_stage" = "yes" ]; then
		return
	fi

	local def_value=""
	eval `sh_get_args '--init) def_value="$2"; shift;;'`
	local table="$1"
	local old_column="$2"
	local new_column="$3"
	local type="$4"

	local query="ALTER TABLE \`$table\` CHANGE \`$old_column\` \`$new_column\` $type"
	if [ -n "$def_value" ]; then
		query="$query; UPDATE \`$table\` SET \`$new_column\`=$def_value"
	fi
	db_do_if "DESC \`$table\`" "\$1 == \"$old_column\"" "$query"
}

# db_fix_del_column table column
# Deletes a column from the table if it exists
db_fix_del_column()
{
	if [ "$db_fix_check_stage" = "yes" ]; then
		return
	fi

	local table="$1"
	local column="$2"

	db_do_if "DESC \`$table\`" "\$1 == \"$column\"" "ALTER TABLE \`$table\` DROP \`$column\`"
}

# subset set1 set2
# Return true if set1 is subset of set2, false otherwise
# set1 and set2 are comma separated lists of strings
subset()
{
	local first=$1
	shift
	local second=$1

	local f=`mktemp`
	echo $first | tr ',' '\n' | sort | uniq > $f

	local s=`mktemp`
	echo $second | tr ',' '\n' | sort | uniq > $s

	local rc=`comm -23 $f $s | head -1`
	rm -f $f $s
	[ -z $rc ] && return 0 || return 1
}

# db_fix_add_key [--name key_name] table column [column ...]
# Adds a non-unique [compound] key on the specified columns of the
# table. If the key already exists, or there is another key that
# substitutes this one, nothing is done. Existing keys that are
# substituted by the new one are dropped. Explicit key name
# can be specified by option --name
db_fix_add_key()
{
	if [ "$db_fix_check_stage" = "yes" ]; then
		return
	fi

	local name=""
	eval `sh_get_args '--name) name="$2"; shift;;'`
	local table="$1"
	local columns="$2"
	shift 2
	local i
	for i; do
		columns="$columns,\`$i\`"
	done
	local type iname icolumns
	local found=no
	db_get_keys "$table" | (
		while read type iname icolumns; do
			if subset "$icolumns" "$columns" && [ "$name" = "$iname" ]; then
				found=yes
				break
			elif [ "$name" = "$iname" ] || subset "$columns" "$icolumns"; then
				if [ "$type" = "KEY" ]; then
					db_do "ALTER TABLE \`$table\` DROP KEY \`$iname\`"
				fi
			fi
		done

		if [ "$found" = no ]; then
			if [ -z "$name" ]; then
				db_do "ALTER TABLE \`$table\` ADD KEY ($columns)"
			else
				db_do "ALTER TABLE \`$table\` ADD KEY \`$name\` ($columns)"
			fi
		fi
		exit 0
	)
	local status="$?"
	if [ "$status" -ne 0 ]; then
		exit "$status"
	fi
}

# db_fix_drop_key [--name key_name] table column [column ...]
# Drops a non-unique [compound] key on the specified columns of the
# table. If the key does not exist, nothing is done
# if key_name is present, it will be removed by name, columns arguments
# are ignored
db_fix_drop_key()
{
	if [ "$db_fix_check_stage" = "yes" ]; then
		return
	fi

	local name=""
	eval `sh_get_args '--name) name="$2"; shift;;'`

	local table="$1"
	shift

	local columns=`implode "," "$@"`
	local type iname icolumns
	local found=no
	db_get_keys "$table" | (
		while read type iname icolumns; do
			if [ -n "$name" -a "$name" = "$iname" ] || [ "$icolumns" = "$columns" ]; then
				if [ "$type" = "KEY" ]; then
					db_do "ALTER TABLE $table DROP KEY $iname"
					break
				fi
			fi
		done
		exit 0
	)
	local status="$?"
	if [ "$status" -ne 0 ]; then
		exit "$status"
	fi
}

# db_fix_add_unique_key [--check-where condition] [--name key_name] table column [column ...]
# Adds a unique [compound] key on the specified columns of the table.
# If the key already exists, or there is another key that substitutes
# this one, nothing is done. Existing keys that are substituted by the
# new one are dropped. If --check-where is specified, limit checking
# to rows that satisfy the given SQL condition. Explicit key name
# can be specified by option --name
db_fix_add_unique_key()
{
	local check_where="" name=""
	eval `sh_get_args '--check-where) check_where="--where \"$2\""; shift;; --name) name="$2"; shift;;'`
	if [ "$db_fix_check_stage" = "yes" ]; then
		eval db_check_unique $check_where "$@"
		return
	fi

	local table="$1"
	local columns="$2"
	local esc_columns="$2"
	shift 2
	local i
	for i; do
		columns="$columns,$i"
		esc_columns="$esc_columns,\`$i\`"
	done
	local type iname icolumns
	local found=no
	db_get_keys "$table" | (
		while read type iname icolumns; do
			if subset "$icolumns" "$columns" && [ "$name" = "$iname" ]; then
				if [ "$type" != "KEY" ]; then
					found=yes
					break
				fi
			elif [ "$name" = "$iname" ] || subset "$columns" "$icolumns"; then
				if [ "$type" != "PRIMARY" ]; then
					db_do "ALTER TABLE \`$table\` DROP KEY \`$iname\`"
				fi
			fi
		done
		if [ "$found" = no ]; then
			if [ -z "$name" ]; then
				db_do "ALTER TABLE \`$table\` ADD UNIQUE ($esc_columns)"
			else
				db_do "ALTER TABLE \`$table\` ADD UNIQUE \`$name\` ($esc_columns)"
			fi
		fi
		exit 0
	)
	local status="$?"
	if [ "$status" -ne 0 ]; then
		exit "$status"
	fi
}

# db_fix_drop_unique_key [--name key_name] table column [column ...]
# Drops an unique [compound] key on the specified columns of the
# table. If the key does not exist, nothing is done
# if key_name is present, it will be removed by name, columns arguments
# are ignored
db_fix_drop_unique_key()
{
	if [ "$db_fix_check_stage" = "yes" ]; then
		return
	fi

	local name=""
	eval `sh_get_args '--name) name="$2"; shift;;'`

	local table="$1"
	shift
	local columns=`implode "," "$@"`
	local type iname icolumns
	local found=no
	db_get_keys "$table" | (
		while read type iname icolumns; do
			if [ -n "$name" -a "$name" = "$iname" ] || [ "$icolumns" = "$columns" ]; then
				if [ "$type" = "UNIQUE" ]; then
					db_do "ALTER TABLE $table DROP KEY $iname"
					break
				fi
			fi
		done
		exit 0
	)
	local status="$?"
	if [ "$status" -ne 0 ]; then
		exit "$status"
	fi
}

# db_fix_drop_table table_name
db_fix_drop_table()
{
	local table_name
	table_name="$1"

	db_do "DROP TABLE IF EXISTS \`$table_name\`"
}

# db_test test_query awk_script
# Runs test_query and processes it with awk_script. If the output is
# not empty, return 0, otherwise return 1. Hint: supply '1' for
# awk_script to test just for the presence of any output.
db_test()
{
	local any_db=
	eval `sh_get_args '--any-db) any_db=yes;;'`

	local test_query="$1"
	local awk_script="$2"

	if [ -n "$any_db" ]; then
		local output="`mysql_raw_anydb -e \"$test_query\" 2>>\"$product_log\"`"
	else
		local output="`mysql_raw -e \"$test_query\" 2>>\"$product_log\"`"
	fi
	local status=$?
	if [ "$status" -ne 0 ]; then
		p_echo "$output"
		die "run the following SQL query: $1"
	fi

	echo -n "$output" | awk -F '\t' -- "$awk_script" | test `wc -l` -ne 0
}

# db_do [--inten <inten>] query
# Runs query. If it fails, die
# the inten string describes the query reason (to make finding the bug simpler)
db_do()
{
	local desc="execute SQL query"
	eval `sh_get_args '--inten) desc=" (to $2)"; shift;;'`
	if [ "$db_fix_check_stage" = "yes" ]; then
		return
	fi

	local query="$1"

	mysql -e "$query" >>"$product_log" 2>&1 || die "$desc, the query was: $query"
}

# db_select <query>
# runs <query> via mysql_raw
# writes output to db_select_output
# if query fails, output errors and return 1
db_select()
{
	local desc="execute SQL query"
	local query="$1"
	local output="`mysql_raw -e \"$query\" 2>>\"$product_log\"`"
	local status="$?"
	if [ "$status" -ne "0" ]; then
		p_echo "$output"
		die "run the following SQL query: $query"
	fi

	db_select_output="$output"
	return 0
}

# db_do_if test_query awk_script query
# If db_test test_query awk_script returns zero (true), run the query.
# If any query fails, die
db_do_if()
{
	if [ "$db_fix_check_stage" = "yes" ]; then
		return
	fi

	local test_query="$1"
	local awk_script="$2"
	local query="$3"

	if db_test "$test_query" "$awk_script"; then
		db_do "$query"
	fi
}

# db_do_if_not test_query awk_script query
# If db_test test_query awk_script returns nonzero (false), run the
# query. If any query fails, die
db_do_if_not()
{
	if [ "$db_fix_check_stage" = "yes" ]; then
		return
	fi

	local test_query="$1"
	local awk_script="$2"
	local query="$3"

	if ! db_test "$test_query" "$awk_script"; then
		db_do "$query"
	fi
}

# db_check_not test_query awk_script message
# If db_test test_query awk_script returns zero (true), die with
# message
db_check_not()
{
	local test_query="$1"
	local awk_script="$2"
	local message="$3"

	if db_test "$test_query" "$awk_script"; then
		p_echo "$message"
		die
	fi
}

# db_check_unique [--where condition] table column [column ...]
# If the table contains duplicate tuples of values in the specified
# columns, die. If --where is specified, limit checking to rows that
# satisfy the given SQL condition
db_check_unique()
{
	local where=""
	eval `sh_get_args '--where) where="WHERE $2"; shift;;'`
	local table="$1"
	shift
	local columns="$1"
	shift
	local i
	for i; do
		if [ -n "$db_added_columns" ]; then
			eval "case $table.$i in $db_added_columns) return ;; esac"
		fi
		columns="$columns, \`$i\`"
	done
	db_check_not "SELECT COUNT(*) FROM \`$table\` $where GROUP BY $columns HAVING COUNT(*) > 1 LIMIT 1" 1 		"Database integrity check failed: duplicate values of ($columns) exist in table $table"
}

# db_get_keys table
# Outputs the list of keys for a table, one on a row, in the following format:
# name type columns
# Where type is one of PRIMARY, UNIQUE, and KEY, name is always
# PRIMARY for primary keys, and columns is a comma-separated list of
# columns without embedded spaces. All backtick characters are removed
db_get_keys()
{
	local table="$1"
	local query="SHOW CREATE TABLE \`$table\`"
	local output="`mysql_raw -e "$query"`"
	local status=$?
	if [ "$status" -ne 0 ]; then
		p_echo "$output"
		die "run the following SQL query: $1"
	fi

	echo "$output" | awk '
		$1 == "PRIMARY"	{ print "PRIMARY", "PRIMARY", $3 }
		$1 == "UNIQUE"	{ print "UNIQUE", $3, $4 }
		$1 == "KEY"		{ print "KEY", $2, $3 }' | 		sed -e 's/[`()]//g' -e 's/,[[:space:]]*$//g'
}

# db_test_database database
# Returns 0 if the database exists
db_test_database()
{
	local database="$1"

	local mysql_db_name="mysql"
	db_test "SHOW DATABASES" "\$1 == \"$database\""
}

# db_test_table table
# Returns 0 if the table exists
db_test_table()
{
	local table="$1"

	db_test "SHOW TABLES LIKE '$table'" 1
}

# db_test_column table column
# Returns 0 if table.column exists
db_test_column()
{
	local table="$1"
	local column="$2"

	db_test "DESC \`$table\`" "\$1 == \"$column\""
}

# --- db_install helpers ---

dbi_try_create_db()
{
	local mysql_db_name="mysql"
	echo "CREATE DATABASE $DB_NAME DEFAULT CHARSET=utf8" | mysql >> "$product_log" 2>&1
}

dbi_modify_db_name()
{
	DB_NAME="${DB_BASENAME}_`get_random_string 4 'a-z0-9'`"
}

dbi_modify_db_user()
{
	DB_USER_NAME="${DB_USER_BASENAME}_`get_random_string 4 'a-z0-9'`"
}

dbi_try_initialize_db()
{
	local initializer_func="$1"

	local mysql_db_name="$DB_NAME"
	if [ "$DB_USER_NAME" != "admin" ]; then
		local mysql_user="--user=$DB_USER_NAME"
		local mysql_passwd="$DB_PASSWORD"
	fi

	$initializer_func
}

db_install()
{
	# Usage: db_install "<DB description for log messages>" options ... (see below)
	local opt_db_desc="$1"
	shift

	local opt_name= opt_basename= opt_user_name= opt_user_basename=
	local opt_user_creator= opt_skip_when_func= opt_initializer_func= opt_commit_func=

	local inten="install $opt_db_desc database"

	while [ "$#" -gt 0 ]; do
		case "$1" in
			--name)
				# New DB name. If --basename is also specified, then this is the initial name to try.
				opt_name="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--basename)
				# New DB name prefix. Several tries would be made to create database with name '<basename>_<random suffix>'.
				opt_basename="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--user-name)
				# New DB user name. If --user-basename is also specified, then this is the initial name to try.
				opt_user_name="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--user-basename)
				# New DB user prefix. Several tries would be made to create user with name '<user basename>_<random suffix>'.
				opt_user_basename="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--user-creator)
				# DB user creation callback. If not present, 'admin' account will be used for DB creation and initialization.
				opt_user_creator="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--skip-when)
				# Predicate function that returns 0 when DB creation should be skipped (e.g. if DB already exists).
				opt_skip_when_func="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--initializer)
				# DB initializer callback. Initialize your DB schema and populate DB here. Required argument.
				# MySQL connection will be properly set up for this callback - use db_*() and mysql*() functions.
				opt_initializer_func="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--commit)
				# Commit callback. Save DB connection properties here. Following variable will be available in it:
				# DB_HOST, DB_NAME, DB_USER_NAME, DB_PASSWORD (empty if DB_USER_NAME = "admin").
				opt_commit_func="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			*)
				die "$inten: unknown key '$1'"
				;;
		esac
	done

	# Sanity checks
	[ -n "$opt_db_desc"          ] || die "$inten: empty database description"
	[ -z "$opt_user_creator"     ] || is_function "$opt_user_creator"     || die "$inten: invalid value for --user-creator argument"
	[ -z "$opt_skip_when_func"   ] || is_function "$opt_skip_when_func"   || die "$inten: invalid value for --skip-when argument"
	[ -n "$opt_initializer_func" ] && is_function "$opt_initializer_func" || die "$inten: invalid value for --initializer argument"
	[ -z "$opt_commit_func"      ] || is_function "$opt_commit_func"      || die "$inten: invalid value for --commit argument"
	[ -n "$opt_name$opt_basename" ] || die "$inten: at least --name or --basename must be specified"
	[ -z "$opt_user_creator" -o -n "$opt_user_name$opt_user_basename" ] || die "$inten: at least --user-name or --user-basename must be specified if --user-creator is set"

	# Further code assumes that current MySQL access credentials are sufficient for DB creation

	if [ -n "$opt_skip_when_func" ] && $opt_skip_when_func ; then
		p_echo "Installation of $opt_db_desc database is skipped"
		return 0
	fi

	echo_try "$inten"

	local DB_HOST="localhost"
	local DB_NAME="$opt_name"
	local DB_BASENAME="$opt_basename"
	local DB_USER_NAME="$opt_user_name"
	local DB_USER_BASENAME="$opt_user_basename"
	local DB_PASSWORD="`get_random_string 12 'A-Za-z0-9_'`"

	! [ -z "$DB_NAME" -a -n "$DB_BASENAME" ] || dbi_modify_db_name
	! [ -z "$DB_USER_NAME" -a -n "$DB_USER_BASENAME" ] || dbi_modify_db_user

	# Create DB
	if [ -n "$DB_BASENAME" ]; then
		repeat_until_succeeds dbi_try_create_db dbi_modify_db_name 50 || return 1
	else
		dbi_try_create_db || return 1
	fi

	# Create user if required, otherwise use current user
	if [ -n "$opt_user_creator" ]; then
		if [ -n "$DB_USER_BASENAME" ]; then
			repeat_until_succeeds "$opt_user_creator" dbi_modify_db_user 50 || return 1
		else
			$opt_user_creator || return 1
		fi
	else
		DB_USER_NAME="admin"
		DB_PASSWORD=
	fi

	dbi_try_initialize_db "$opt_initializer_func" || return 1

	[ -z "$opt_commit_func" ] || $opt_commit_func

	p_echo "Installation of $opt_db_desc database finished"
}


db_upgrade_add_indexes_for_performance_11_1_11()
{
    db_do_if_not "SHOW INDEX FROM clients" "\$3 == \"type\"" "CREATE INDEX type ON clients (type)"
    db_do_if_not "SHOW INDEX FROM domains" "\$3 == \"external_id\"" "CREATE INDEX external_id ON domains (external_id)"
}

db_upgrade_change_sys_users_home_to_utf8()
{
	db_fix_change_column "sys_users" "home" "home" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_general_ci NOT NULL"
}


db_upgrade_service_node_properties_move_property_to_configuration_11_1_12()
{
    local sourceName="$1"
    local destinationName="$3"
    local destinationSection="$2"
    db_select "SELECT \`serviceNodeId\` FROM \`ServiceNodeProperties\` WHERE \`name\` = '$sourceName'"
    for serviceNodeId in $db_select_output; do
        db_do_if_not             "SELECT \`name\` FROM \`ServiceNodeConfiguration\` WHERE \`serviceNodeId\` = $serviceNodeId AND \`section\` = '$destinationName' AND \`name\` = '$destinationSection'" 1             "INSERT INTO \`ServiceNodeConfiguration\` (\`serviceNodeId\`, \`section\`, \`name\`, \`value\`) SELECT '$serviceNodeId', '$destinationSection', '$destinationName', \`value\` FROM \`ServiceNodeProperties\` WHERE \`serviceNodeId\` = '$serviceNodeId' AND \`name\` = '$sourceName'"
    done
}

db_upgrade_service_node_properties_move_misc_to_configuration_11_1_12()
{
    local sourceName="$1"
    local destinationName="$3"
    local destinationSection="$2"
    db_select "SELECT \`id\` FROM \`ServiceNodes\` WHERE \`ipAddress\` = 'local'"
    for serviceNodeId in $db_select_output; do
        db_do_if_not             "SELECT \`name\` FROM \`ServiceNodeConfiguration\` WHERE \`serviceNodeId\` = $serviceNodeId AND \`section\` = '$destinationSection' AND \`name\` = '$destinationName'" 1             "INSERT INTO \`ServiceNodeConfiguration\` (\`serviceNodeId\`, \`section\`, \`name\`, \`value\`) SELECT '$serviceNodeId', '$destinationSection', '$destinationName', \`val\` FROM misc WHERE \`param\` = '$sourceName'"
    done
    db_do "DELETE FROM \`misc\` WHERE \`param\`= '$sourceName'"
}

db_upgrade_service_node_properties_11_1_12()
{
    if db_test_table 'ServiceNodeProperties'; then
        db_upgrade_service_node_properties_move_property_to_configuration_11_1_12 'server.nginx.enabled'             'webProxy' 'enabled'
        db_upgrade_service_node_properties_move_property_to_configuration_11_1_12 'server.nginx.backend-port'       'webProxy' 'backendPort'
        db_upgrade_service_node_properties_move_property_to_configuration_11_1_12 'server.nginx.ssl-backend-port'   'webProxy' 'sslBackendPort'
        db_upgrade_service_node_properties_move_property_to_configuration_11_1_12 'server.nginx.frontend-port'      'webProxy' 'frontendPort'
        db_upgrade_service_node_properties_move_property_to_configuration_11_1_12 'server.nginx.ssl-frontend-port'  'webProxy' 'sslFrontendPort'

        db_upgrade_service_node_properties_move_property_to_configuration_11_1_12 'server.webserver.httpPort'       'webServer' 'httpPort'
        db_upgrade_service_node_properties_move_property_to_configuration_11_1_12 'server.webserver.httpsPort'      'webServer' 'httpsPort'

        db_upgrade_service_node_properties_move_misc_to_configuration_11_1_12 'http_port' 'webServer' 'httpPort'
        db_upgrade_service_node_properties_move_misc_to_configuration_11_1_12 'https_port' 'webServer' 'httpsPort'

        db_select "SELECT SUBSTR(\`name\`, 26) FROM \`ServiceNodeProperties\` WHERE \`name\` LIKE 'coldfusionservice.config.%' GROUP BY \`name\`"
        for cfName in $db_select_output; do
            db_upgrade_service_node_properties_move_property_to_configuration_11_1_12 'coldfusionservice.config.$cfName' 'coldfusion' '$cfName'
        done

        db_upgrade_service_node_properties_move_property_to_configuration_11_1_12 'server.configuration.crontabSecureShell' 'scheduler' 'crontabSecureShell'

        db_fix_drop_table ServiceNodeProperties
    fi

    if db_test_table 'MailServerProperties'; then
        db_do "REPLACE INTO \`ServiceNodeConfiguration\` (\`serviceNodeId\`, \`section\`, \`name\`, \`value\`) SELECT \`serviceNodeId\`, 'mailServer', \`name\`, \`value\` FROM \`MailServerProperties\`";

        db_fix_drop_table 'MailServerProperties'
    fi
}

db_upgrade_custom_buttons_11_1_12()
{
    echo_try "Converting custom buttons from domain to client level"
    db_do "UPDATE custom_buttons cb, (SELECT cb.id, 4 AS level, d.cl_id AS level_id FROM custom_buttons cb INNER JOIN domains d on cb.level_id = d.id WHERE level = 8) sq SET cb.level = sq.level, cb.level_id = sq.level_id WHERE cb.id = sq.id"
}


db_upgrade_service_node_configuration_11_1_13()
{
    db_select "SELECT DISTINCT serviceNodeId FROM ServiceNodeConfiguration WHERE serviceNodeId NOT IN (SELECT serviceNodeId FROM ServiceNodeConfiguration WHERE section = 'WebProxy' AND name = 'enabled')"
    for serviceNodeId in $db_select_output; do

        p_echo "Fix web proxy settings for service node $serviceNodeId"

        db_select "SELECT value FROM ServiceNodeConfiguration WHERE section = 'WebProxy' AND name = 'backendPort' AND serviceNodeId = '$serviceNodeId'"
        local backendPort="$db_select_output"
        if [ -z "$backendPort" ]; then
            backendPort="7080"
        fi

        db_select "SELECT value FROM ServiceNodeConfiguration WHERE section = 'WebProxy' AND name = 'sslBackendPort' AND serviceNodeId = '$serviceNodeId'"
        local sslBackendPort="$db_select_output"
        if [ -z "$backendPort" ]; then
            backendPort="7081"
        fi

        local proxyEnabled="false"
        if db_test "SELECT * FROM ServiceNodeConfiguration WHERE name = 'webServer' AND (section = 'httpPort' AND value = '$backendPort' OR section = 'httpsPort' AND value = '$sslBackendPort') AND serviceNodeId = '$serviceNodeId'" 1; then
            proxyEnabled="true"
        fi

        p_echo "Frontend port = $frontendPort, backend port = $backendPort, proxy enabled = $proxyEnabled"

        db_do "INSERT INTO ServiceNodeConfiguration (serviceNodeId, section, name, value) VALUES ('$serviceNodeId', 'WebProxy', 'enabled', '$proxyEnabled')"

    done


    db_do "UPDATE IGNORE ServiceNodeConfiguration SET name = section, section = 'webServer' WHERE name = 'webServer'"
    db_do "DELETE FROM ServiceNodeConfiguration WHERE name = 'webServer'"

    db_do "UPDATE IGNORE ServiceNodeConfiguration SET name = section, section = 'webProxy' WHERE name = 'webProxy'"
    db_do "DELETE FROM ServiceNodeConfiguration WHERE name = 'webProxy'"

    db_do "UPDATE ServiceNodeConfiguration SET section = 'webProxy' WHERE section = 'WebProxy'"
}


db_upgrade_dns_templates_11_1_14()
{
	echo_try "drop rows from psa.dns_recs_t with empty values of column 'type'"
	db_do "DELETE from dns_recs_t WHERE \`type\`=''"
	suc
}


db_upgrade_subscription_properties_11_1_15()
{
	echo_try "Create subscription records"
	sw_engine_pleskrun $PRODUCT_ROOT_D/admin/plib/scripts/upgrade_subscriptions_108.php && suc || warn "fail\nCannot update subscriptions information. Please run $PRODUCT_ROOT_D/admin/plib/scripts/upgrade_subscriptions_108.php manually."

	echo_try "Move limits and permissions from main domain of subscription to subscription properties"
	sw_engine_pleskrun $PRODUCT_ROOT_D/admin/plib/scripts/upgrade_subscription_properties_11.1.16.php && suc || warn "fail\nCannot update subscription properties. Please run $PRODUCT_ROOT_D/admin/plib/scripts/upgrade_subscription_properties_11.1.16.php manually."
}


db_upgrade_forwarding_11_1_17()
{
    db_fix_add_column "forwarding" "http_code" "INT UNSIGNED NOT NULL DEFAULT 301"
}

db_upgrade_seo_safe_domain_aliases_11_1_17()
{
    db_fix_add_column "domain_aliases" "seoRedirect" "ENUM('false', 'true') NOT NULL DEFAULT 'false'"
}

db_upgrade_db_user_unique_key_11_1_17()
{
    db_fix_drop_unique_key "db_users" "db_id" "login"
    db_fix_drop_unique_key "db_users" "db_server_id" "login"
    db_fix_add_unique_key "db_users" "db_server_id" "db_id" "login"
}

db_upgrade_php_handler_id_11_1_18()
{
    db_fix_add_column --init php_handler_type hosting php_handler_id "varchar(255) NOT NULL DEFAULT 'module'"
    db_do "REPLACE INTO \`TmplData\` (\`tmpl_id\`, \`element\`, \`value\`) SELECT \`tmpl_id\`, 'php_handler_id', \`value\` FROM \`TmplData\` WHERE \`element\` = 'php_handler_type'"
    db_do "DELETE from \`TmplData\` WHERE \`element\` = 'php_handler_type'"
}

db_upgrade_drop_php_version_11_5_19()
{
	if db_test_column 'hosting' 'php_handler_type' && db_test_column 'hosting' 'php_version'; then
	    db_do "update ServiceNodeEnvironment e, ServiceNodes n, hosting h, sys_users u	        set h.php_handler_id = concat(php_handler_type, '-', php_version)    	    where n.id = e.serviceNodeId and name='PLATFORM' and value='windows' and h.sys_user_id = u.id and u.serviceNodeId = n.id;"
	fi
    db_fix_del_column 'hosting' 'php_handler_type'
    db_fix_del_column 'hosting' 'php_version'
}

db_upgrade_webmails_11_5_19()
{
    db_fix_add_column Webmails serviceNodeId "int(10) unsigned NOT NULL default 1"
}

db_upgrade_smb_11_5_19()
{
# 51
	db_fix_add_column smb_apsPackages isNonStorefrontCommercial "INT DEFAULT 0"
}

db_upgrade_enable_dns_connector_11_5_21()
{
	local connector=

	db_select "select p.val from cl_param p INNER JOIN clients c ON c.id = p.cl_id WHERE c.type = 'admin' AND p.param = 'dns-connector-class-name' LIMIT 1"
	case "$db_select_output" in
		Service_Dns_PoaConnector)
			connector="poa"
		;;
		Service_Dns_PleskConnector)
			connector="plesk"
		;;
	esac

	if [ -n "$connector" ]; then
		db_select "SELECT \`id\` FROM \`ServiceNodes\` WHERE \`ipAddress\` = 'local'"
		local nodes="$db_select_output"
		for serviceNodeId in $nodes; do
			db_do_if_not 				"SELECT \`value\` FROM \`ServiceNodeConfiguration\` WHERE \`serviceNodeId\` = $serviceNodeId AND \`section\` = 'dnsConnector' AND \`name\` = '$connector'" 1 				"INSERT INTO \`ServiceNodeConfiguration\` (\`serviceNodeId\`, \`section\`, \`name\`, \`value\`) values ($serviceNodeId, 'dnsConnector', '$connector', 'true')"
		done
	fi
	db_do "DELETE FROM cl_param WHERE param='dns-connector-class-name'"
}

db_upgrade_set_force_update_applications_notification_11_5_21()
{
	db_select "SELECT DISTINCT d.cl_id AS clientId FROM Subscriptions AS s INNER JOIN apsContexts AS a ON (s.id = a.subscriptionId) INNER JOIN domains AS d ON (s.object_type = 'domain' AND s.object_id = d.id) LEFT JOIN cl_param AS clp ON (d.cl_id = clp.cl_id AND clp.param = 'show_force_update_applications_notification') WHERE clp.cl_id IS NULL"
	for clientId in $db_select_output; do
		db_do "INSERT INTO \`cl_param\` (\`cl_id\`, \`param\`, \`val\`) values ($clientId, 'show_force_update_applications_notification', 'true')"
	done
}

db_add_public_ip_11_5_21()
{
	db_fix_add_column "IP_Addresses" "public_ip_address" "VARCHAR(39) CHARACTER SET ascii COLLATE ascii_general_ci  DEFAULT NULL"
}

db_upgrade_drop_ip_address_id_11_5_22()
{
    db_fix_del_column "hosting" "ip_address_id"
    db_fix_del_column "forwarding" "ip_address_id"
}

db_upgrade_set_property_unpaid_website_status_for_plans_11_5_22()
{
    db_select "SELECT t.id AS planId FROM Templates AS t LEFT JOIN TmplData AS td ON (t.id = td.tmpl_id AND td.element = 'unpaid_website_status') WHERE t.type='domain' AND td.value IS NULL"
    for planId in $db_select_output; do
        db_do "INSERT INTO \`TmplData\` (\`tmpl_id\`, \`element\`, \`value\`) values ($planId, 'unpaid_website_status', 'disabled')"
    done
}

db_upgrade_set_property_unpaid_website_status_for_subscriptions_11_5_22()
{
    db_select "SELECT s.id AS subscriptionId FROM Subscriptions AS s LEFT JOIN SubscriptionProperties AS sp ON (s.id = sp.subscription_id AND sp.name = 'unpaid_website_status') WHERE s.object_type='domain' AND sp.value IS NULL"
    for subscriptionId in $db_select_output; do
        db_do "INSERT INTO \`SubscriptionProperties\` (\`subscription_id\`, \`name\`, \`value\`) values ($subscriptionId, 'unpaid_website_status', 'disabled')"
    done
}

db_upgrade_correcting_property_unpaid_website_status_for_plans_11_5_24()
{
	db_do "UPDATE \`TmplData\` SET \`element\` = 'unpaid_website_status' WHERE \`element\` = 'unpaidWebsiteStatus';"
}

db_upgrade_correcting_property_unpaid_website_status_for_subscriptions_11_5_24()
{
	db_do "UPDATE \`SubscriptionProperties\` SET \`name\` = 'unpaid_website_status' WHERE \`name\` = 'unpaidWebsiteStatus';"
}

db_upgrade_maintenance_mode_id_11_5_24()
{
	db_select "SELECT \`d\`.\`id\` AS \`domainId\`, \`d\`.\`status\` AS \`domainStatus\`, COALESCE(\`w\`.\`id\`, \`d\`.\`id\`) AS \`webspaceId\`, \`d\`.\`cl_id\` AS \`clientId\`, COALESCE(\`w\`.\`permissions_id\`, \`d\`.\`permissions_id\`) AS \`permissionsId\` FROM \`hosting\` AS \`h\` INNER JOIN \`domains\` AS \`d\` ON \`d\`.\`id\` = \`h\`.\`dom_id\` LEFT JOIN \`domains\` AS \`w\` ON \`w\`.\`id\` = \`d\`.\`webspace_id\` WHERE \`h\`.\`maintenance_mode\` = 'true' AND \`d\`.\`status\` = 0"
    test -z "$db_select_output" && return 0
	echo "$db_select_output" | while read row; do
		domainId=`echo $row | awk '{print $1}'`
		domainStatus=`echo $row | awk '{print $2}'`
		webspaceId=`echo $row | awk '{print $3}'`
		clientId=`echo $row | awk '{print $4}'`
		permissionsId=`echo $row | awk '{print $5}'`

		if db_test "SELECT \`p\`.\`value\` FROM \`Subscriptions\` AS  \`s\` LEFT JOIN \`SubscriptionProperties\` AS \`sp\` ON \`sp\`.\`subscription_id\` = \`s\`.\`id\` AND \`sp\`.\`name\` = 'permissionsId' INNER JOIN \`Permissions\` AS \`p\` ON \`p\`.\`id\` = COALESCE(\`sp\`.\`value\`, $permissionsId) AND \`p\`.\`permission\` = 'manage_website_maintenance' AND \`p\`.\`value\` = 'false' WHERE \`s\`.\`object_id\` = $webspaceId AND \`s\`.\`object_type\` = 'domain'" 1; then
			if db_test "SELECT \`p\`.\`type\` FROM \`clients\` as \`c\` INNER JOIN \`clients\` as \`p\` ON \`p\`.\`id\` = \`c\`.\`parent_id\` AND \`p\`.\`type\` = 'reseller' WHERE \`c\`.\`id\` = $clientId" 1; then
				domainStatus=$(( $domainStatus | 32 ))
			else
				domainStatus=$(( $domainStatus | 16 ))
			fi
		else
			domainStatus=$(( $domainStatus | 64 ))
		fi

		db_do "UPDATE \`domains\` SET \`status\` = $domainStatus WHERE \`id\` = $domainId"
		db_do_if_not 			"SELECT \`dom_id\` FROM \`dom_param\` WHERE \`param\` = 'turnOffAction' AND \`dom_id\` = $domainId" 1 			"INSERT INTO \`dom_param\` (\`dom_id\`, \`param\`, \`val\`) VALUES ($domainId, 'turnOffAction', 'suspend')"
		db_do_if_not 			"SELECT \`dom_id\` FROM \`dom_param\` WHERE \`param\` = 'selfTurnOffAction' AND \`dom_id\` = $domainId" 1 			"INSERT INTO \`dom_param\` (\`dom_id\`, \`param\`, \`val\`) VALUES ($domainId, 'selfTurnOffAction', 'suspend')"
	done
}

db_upgrade_service_node_configuration_11_5_24()
{
    db_fix_change_column "ServiceNodeEnvironment" "value" "value" "TEXT CHARACTER SET utf8 COLLATE utf8_general_ci"
    db_fix_change_column "ServiceNodeConfiguration" "value" "value" "TEXT CHARACTER SET utf8 COLLATE utf8_general_ci"
}

db_upgrade_secret_keys_11_5_24()
{
    db_fix_add_column "secret_keys" "lookup_id" "VARCHAR(255) CHARACTER SET ascii COLLATE ascii_general_ci DEFAULT NULL"
    db_do_if_not "SHOW INDEX FROM secret_keys" "\$3 == \"lookup_id\"" "ALTER TABLE secret_keys ADD INDEX (lookup_id)"
    sw_engine_pleskrun $PRODUCT_ROOT_D/admin/plib/scripts/upgrade_secret_keys_11.5.25.php && suc || warn "fail\nencryption of secret keys"
}

db_upgrade_correcting_aps_tokens_11_5_29()
{
    db_fix_change_column "ApsTokens" "packageId" "applicationId" "VARCHAR(2000) CHARACTER SET binary NOT NULL"
}

db_smb_upgrade_11_5_29()
{
	db_fix_add_column smb_users locale "VARCHAR(17)"
}

db_upgrade_external_id_11_5_99()
{
    db_fix_add_column "Subscriptions" "external_id" "VARCHAR(255) CHARACTER SET ascii COLLATE ascii_bin NOT NULL DEFAULT ''"
    db_fix_add_column "dns_zone" "external_id" "VARCHAR(255) CHARACTER SET ascii COLLATE ascii_bin NOT NULL DEFAULT ''"
    db_fix_add_column "dns_recs" "external_id" "VARCHAR(255) CHARACTER SET ascii COLLATE ascii_bin NOT NULL DEFAULT ''"
}

db_upgrade_move_configuration_11_5_99()
{
    db_upgrade_move_cl_param_to_servicenode_config_11_5_99 'poa-api-url' 'POA'
    db_upgrade_move_cl_param_to_servicenode_config_11_5_99 'poa-api-login' 'POA'
    db_upgrade_move_cl_param_to_servicenode_config_11_5_99 'poa-api-password' 'POA'
    db_upgrade_move_cl_param_to_servicenode_config_11_5_99 'poa-ui-dispatcher-url' 'POA'

    db_upgrade_move_cl_param_to_servicenode_config_11_5_99 'ppb-url' 'Billing'
    db_upgrade_move_cl_param_to_servicenode_config_11_5_99 'integration-api-version' 'Billing'
    db_upgrade_move_cl_param_to_servicenode_config_11_5_99 'ppb-remote' 'Billing'
}

db_upgrade_move_cl_param_to_servicenode_config_11_5_99()
{
    local sourceName="$1"
    local destinationName="$sourceName"
    local destinationSection="$2"

    db_select "SELECT \`id\` FROM \`ServiceNodes\` WHERE \`ipAddress\` = 'local'"
    local serviceNodeId="$db_select_output";
    if [ "X${serviceNodeId}" = "X" ]; then
        die "Management service node record not found"
    fi

    db_select "SELECT \`id\` FROM \`clients\` WHERE \`type\` = 'admin'"
    local adminClientId="$db_select_output";
    if [ "X${adminClientId}" = "X" ]; then
        die "Admin client record not found"
    fi

    db_do_if_not         "SELECT \`name\` FROM \`ServiceNodeConfiguration\` WHERE \`serviceNodeId\` = $serviceNodeId AND \`section\` = '$destinationSection' AND \`name\` = '$destinationName'" 1         "INSERT INTO \`ServiceNodeConfiguration\` (\`serviceNodeId\`, \`section\`, \`name\`, \`value\`) SELECT '$serviceNodeId', '$destinationSection', '$destinationName', \`val\` FROM cl_param WHERE \`cl_id\` = $adminClientId AND \`param\` = '$sourceName'"

    db_do "DELETE FROM \`cl_param\` WHERE \`cl_id\` = $adminClientId AND \`param\` = '$sourceName'"
}

db_upgrade_iis_app_pools_cpu_limit_11_5_99()
{
    db_fix_add_column "IisAppPools" "cpuLimitAction" "enum('NoAction','KillW3wp','Throttle','ThrottleUnderLoad') NOT NULL default 'NoAction'"
    db_fix_add_column "IisAppPools" "cpuLimitInterval" "int(10) unsigned NOT NULL default '5'"
}


db_upgrade_sdk_add_services_11_5_99()
{
    db_upgrade_sdk_add_service "PleskDns"
    db_upgrade_sdk_add_service "PleskFileSystem"
}
 
db_upgrade_sdk_add_service()
{
    local serviceName="$1"
 
    db_do_if_not       "SELECT id FROM ServiceInstallations WHERE name = '${serviceName}' AND serviceNodeId = 1" 1       "INSERT INTO ServiceInstallations (name, status, serviceNodeId) values ('${serviceName}', 1, 1)"
 
    db_select "SELECT id FROM ServiceInstallations WHERE name = '${serviceName}' LIMIT 1"
    local serviceInstallationId="${db_select_output}"
 
    db_select "SELECT id FROM Subscriptions WHERE object_type='domain'"
    for subscriptionId in ${db_select_output}; do
      db_do_if_not         "SELECT si.id FROM ServiceInstances AS si INNER JOIN Subscriptions AS s ON (si.servicePackageId = s.id) WHERE si.serviceInstallationId = ${serviceInstallationId} AND s.id = ${subscriptionId} LIMIT 1" 1         "INSERT INTO ServiceInstances (serviceInstallationId, servicePackageId, status, enabled) values (${serviceInstallationId}, ${subscriptionId}, 1, 1)"
    done
}

db_upgrade_web_server_configurations_11_6_2()
{
    db_fix_add_column "Configurations" "errorFile" "blob NOT NULL"
}

db_upgrade_mail_description_11_6_2()
{
    db_fix_add_column "mail" "description" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_general_ci"
}

db_upgrade_clients_description_11_6_2()
{
    db_fix_add_column "clients" "description" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_general_ci"
}

db_upgrade_domains_description_11_6_3()
{
    db_fix_add_column "domains" "adminDescription" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_general_ci"
    db_fix_add_column "domains" "resellerDescription" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_general_ci"
    db_fix_add_column "domains" "description" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_general_ci"
}
db_upgrade_heterogenous_1100()
{
    db_fix_add_column "hosting" "maintenance_mode" "ENUM('false','true') NOT NULL DEFAULT 'false'"

    db_fix_add_column "clients" "passwd" "VARCHAR(20) NOT NULL DEFAULT ''"
    db_fix_add_column "clients" "ownership" "ENUM('false','true') NOT NULL DEFAULT 'true'"

    db_fix_add_column "dom_level_usrs" "passwd" "VARCHAR(20) NOT NULL DEFAULT ''"
    db_fix_add_column "dom_level_usrs" "uid" "VARCHAR(255) DEFAULT NULL"
    db_fix_add_column "dom_level_usrs" "ownership" "ENUM('false','true') NOT NULL DEFAULT 'true'"

    db_fix_change_column "domains" "guid" "guid" "VARCHAR(255) CHARACTER SET ascii COLLATE ascii_bin NOT NULL DEFAULT '00000000-0000-0000-0000-000000000000'"

    db_fix_add_column "disk_usage" "mysql_dbases" "BIGINT(20) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_add_column "disk_usage" "mssql_dbases" "BIGINT(20) UNSIGNED NOT NULL DEFAULT '0'"

    db_fix_change_column "hosting" "fp_adm" "fp_adm" "VARCHAR(20) CHARACTER SET ascii COLLATE ascii_bin NOT NULL"
    db_fix_change_column "hosting" "webstat" "webstat" "VARCHAR(20) NOT NULL DEFAULT 'none'"
    db_fix_add_column "hosting" "fp_no_index" "ENUM('false','true') NOT NULL DEFAULT 'false'"
    db_fix_add_column "hosting" "ssi_html" "ENUM('false','true') NOT NULL DEFAULT 'false'"
    db_fix_add_column "hosting" "php_isapi" "ENUM('false','true') NOT NULL DEFAULT 'false'"
    db_fix_add_column "hosting" "php_version" "VARCHAR(255) NOT NULL DEFAULT '4'"
    db_fix_add_column "hosting" "managed_runtime_version" "VARCHAR(255) NOT NULL DEFAULT '1.1'"
    db_fix_add_column "hosting" "at_domains" "ENUM('false','true') NOT NULL DEFAULT 'false'"
    db_fix_add_column "hosting" "write_modify" "VARCHAR(255) NOT NULL DEFAULT ''"
    db_fix_add_column "hosting" "webdeploy" "ENUM('false','true') NOT NULL DEFAULT 'false'"

    db_fix_add_column "web_users" "ssi_html" "ENUM('false','true') NOT NULL DEFAULT 'false'"
    db_fix_add_column "web_users" "php_isapi" "ENUM('false','true') NOT NULL DEFAULT 'false'"
    db_fix_add_column "web_users" "managed_runtime_version" "VARCHAR(255) NOT NULL DEFAULT '1.1'"
    db_fix_add_column "web_users" "write_modify" "VARCHAR(255) DEFAULT NULL"

    db_fix_add_column "ftp_users" "permission" "INT UNSIGNED NOT NULL DEFAULT 3";

    db_fix_add_column "DatabaseServers" "parameters_id" "INT(10) UNSIGNED NOT NULL"
	db_fix_add_column "DatabaseServers" "type" "ENUM('mysql','postgresql','mssql') NOT NULL"
	db_fix_change_column "DatabaseServers" "type" "type" "ENUM('mysql','postgresql','mssql') NOT NULL"

    db_fix_add_column "db_users" "passwd" "VARCHAR(16) NOT NULL DEFAULT ''"
    db_fix_add_column "db_users" "status" "ENUM('normal','orphaned') NOT NULL DEFAULT 'normal'"

    db_fix_change_column "TmplData" "element" "element" "VARCHAR(255) CHARACTER SET ascii COLLATE ascii_general_ci NOT NULL"
    db_fix_change_column "longtaskparams" "param" "param" "VARCHAR(255) CHARACTER SET ascii COLLATE ascii_general_ci NOT NULL"

    db_fix_add_column "protected_dirs" "dom_type" "ENUM('domain','subdomain') NOT NULL default 'domain'"

    db_fix_change_column "hosting" "php_handler_type" "php_handler_type" "ENUM('cgi', 'fastcgi', 'module', 'isapi') NOT NULL default 'module'"
    db_fix_change_column "subdomains" "php_handler_type" "php_handler_type" "ENUM('cgi', 'fastcgi', 'module', 'isapi') NOT NULL default 'module'"

    db_fix_add_column "mail" "disk_usage" "BIGINT DEFAULT 0 NOT NULL"

	if ! db_test_table "domain_aliases"; then
        db_do "RENAME TABLE domainaliases TO domain_aliases"
    fi
}

db_upgrade_apsmail_1100()
{
    db_fix_add_column "mail" "apsServiceAccountInstanceId" "INT UNSIGNED DEFAULT 0 NOT NULL"
    db_fix_add_column "mail" "apsAntivirusServiceInstanceId" "INT UNSIGNED DEFAULT 0 NOT NULL"
    db_fix_add_column "mail_aliases" "apsServiceInstanceId"  "INT UNSIGNED DEFAULT 0 NOT NULL"
    db_fix_add_column "mail_resp" "apsServiceInstanceId"  "INT UNSIGNED DEFAULT 0 NOT NULL"
    db_fix_add_column "mail_redir" "apsServiceInstanceId" "INT UNSIGNED DEFAULT 0 NOT NULL"
    db_fix_add_column "mail" "apsAntispamServiceInstanceId" "INT UNSIGNED DEFAULT 0 NOT NULL"
}

db_upgrade_service_nodes_1100()
{
    db_fix_drop_unique_key "ServiceNodes" "name"
    db_fix_change_column "ServiceNodes" "name" "description" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_general_ci NOT NULL DEFAULT ''"
    db_fix_change_column "ServiceNodes" "remoteIpAddress" "ipAddress" "VARCHAR(255) CHARACTER SET ascii COLLATE ascii_general_ci NOT NULL DEFAULT 'local'"
    db_fix_add_unique_key "ServiceNodes" "ipAddress"
    db_fix_add_column "ServiceNodes" "type" "VARCHAR(255) CHARACTER SET ascii COLLATE ascii_general_ci NOT NULL DEFAULT 'plesk'"
    db_do "UPDATE ServiceNodes SET ipAddress='local' WHERE ISNULL(ipAddress) OR ipAddress=''"
}

db_upgrade_webusers_miva_1100()
{
	db_fix_add_column "web_users" "miva" "ENUM('false','true') NOT NULL DEFAULT 'false'"
}

db_upgrade_mail_server_settings_1100()
{
    db_select "SELECT \`param\` FROM \`misc\` WHERE \`param\` IN (         'spf_enabled', 'spf_behavior', 'spf_guess', 'spf_rules', 'spf_exp',         'disable_pop_auth', 'disable_smtp_auth',         'message_submission', 'message_submission_ip',         'poplock_time', 'rbl', 'rbl_server', 'spf_rule', 'relay',         'domain_keys_sign', 'domain_keys_verify', 'ml_size',         'spamfilter_enabled', 'spamfilter_use_mailuser_prefs', 'spamfilter_max_children', 'spamfilter_max_addr_list_length',         'virusfilter'     )";
    for propertyName in $db_select_output; do
            db_do "DELETE FROM \`MailServerProperties\` WHERE \`name\`='$propertyName' AND \`serviceNodeId\`=1";
            db_do "INSERT INTO \`MailServerProperties\` (\`serviceNodeId\`, \`name\`, \`value\`) SELECT 1, \`param\`, \`val\` FROM \`misc\` WHERE \`param\` = '$propertyName'";
            db_do "DELETE FROM \`misc\` WHERE \`param\`='$propertyName'";
    done;
	db_fix_add_column "badmailfrom" "serviceNodeId" "INT UNSIGNED NOT NULL DEFAULT 1"
	db_fix_add_column "smtp_poplocks" "serviceNodeId" "INT UNSIGNED NOT NULL DEFAULT 1"

    db_fix_add_column "spamfilter" "serviceNodeId" "INT UNSIGNED NOT NULL DEFAULT 1"
	db_do_if 		"SHOW KEYS FROM spamfilter WHERE Key_name = 'username' AND Column_name = 'username'" 1 		"ALTER TABLE spamfilter DROP KEY \`username\`"
	db_do_if_not 		"SHOW KEYS FROM spamfilter WHERE Key_name = 'spamfilter_username_serviceNodeId' AND Column_name = 'serviceNodeId'" 1 		"ALTER TABLE spamfilter ADD UNIQUE KEY spamfilter_username_serviceNodeId (\`username\`, serviceNodeId)"

    db_fix_add_column "GL_settings" "serviceNodeId" "INT UNSIGNED NOT NULL DEFAULT 1"
	db_do_if 		"SHOW KEYS FROM GL_settings WHERE Key_name = 'param' AND Column_name = 'param'" 1 		"ALTER TABLE GL_settings DROP KEY \`param\`"
	db_do_if 		"SHOW KEYS FROM GL_settings WHERE Key_name = 'PRIMARY' AND Column_name = 'param'" 1 		"ALTER TABLE GL_settings DROP PRIMARY KEY"
	db_do_if_not 		"SHOW KEYS FROM GL_settings WHERE Key_name = 'PRIMARY' AND Column_name = 'serviceNodeId'" 1 		"ALTER TABLE GL_settings ADD PRIMARY KEY (serviceNodeId, \`param\`)"

    db_fix_add_column "GL_remote_domains" "serviceNodeId" "INT UNSIGNED NOT NULL DEFAULT 1"
	db_do_if 		"SHOW KEYS FROM GL_remote_domains WHERE Key_name = 'domain' AND Column_name = 'domain'" 1 		"ALTER TABLE GL_remote_domains DROP KEY \`domain\`"
	db_do_if_not 		"SHOW KEYS FROM GL_remote_domains WHERE Key_name = 'PRIMARY' AND Column_name = 'serviceNodeId'" 1 		"ALTER TABLE GL_remote_domains ADD PRIMARY KEY (serviceNodeId, \`domain\`)"
}

db_smb_upgrade_1100()
{
# 48
    db_do_if_not "SHOW INDEX FROM smb_users" "\$3 == \"ownerId\"" "CREATE INDEX ownerId_idx ON smb_users (ownerId)"
# 49
    db_do_if_not "SHOW INDEX FROM smb_roleGeneralPermissions" "\$3 == \"roleId\"" "CREATE INDEX roleId_idx ON smb_roleGeneralPermissions (roleId)"
# 50
    db_do_if_not "SHOW INDEX FROM smb_users" "\$3 == \"isBuiltIn\"" "CREATE INDEX isBuiltIn_idx ON smb_users (isBuiltIn)"
}

db_upgrade_longtasks_1101()
{
    db_fix_add_column "longtasks" "context" "VARCHAR(255) CHARACTER SET ascii COLLATE ascii_general_ci DEFAULT NULL"
    db_fix_change_column "longtasks" "status" "status" "VARCHAR(255) CHARACTER SET ascii COLLATE ascii_general_ci NOT NULL DEFAULT 'not_started'"
    db_fix_add_column "longtasks" "statusProcessed" "INT UNSIGNED NOT NULL DEFAULT 0"

    db_fix_change_column "longtasks" "progress_text" "progressStatus" "VARCHAR(255) CHARACTER SET ascii COLLATE ascii_general_ci NOT NULL DEFAULT 'queue'"
    db_fix_add_column "longtasks" "progressParams" "BLOB DEFAULT NULL"
    db_fix_change_column "longtasks" "progress_max" "progressMax" "INT UNSIGNED NOT NULL DEFAULT 1"
    db_fix_change_column "longtasks" "progress_value" "progressValue" "INT UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "longtasks" "start_time" "startTime" "TIMESTAMP"
    db_fix_change_column "longtasks" "finish_time" "finishTime" "TIMESTAMP"

    db_fix_del_column "longtasks" "title"
    db_fix_del_column "longtasks" "owner"
    db_fix_del_column "longtasks" "owner_type"

    db_fix_change_column "longtaskparams" "val" "val" "BLOB DEFAULT NULL"

    db_do "DELETE FROM \`longtasks\`";
    db_do "DELETE FROM \`longtaskparams\`";
}

db_upgrade_db_users_1101()
{
    db_fix_add_column "db_users" "dom_id" "INT UNSIGNED NOT NULL DEFAULT 0"
    db_fix_add_column "db_users" "db_server_id" "INT UNSIGNED NOT NULL DEFAULT 0"
    db_do "UPDATE db_users, data_bases SET db_users.dom_id = data_bases.dom_id, db_users.db_server_id = data_bases.db_server_id WHERE data_bases.id = db_users.db_id"
}
db_upgrade_enable_sendmail_default_values_12_0_10()
{
    db_plan_properties_default_values_12_0_9 'outgoing_messages_enable_sendmail'
    db_subscription_properties_default_values_12_0_9 'outgoing_messages_enable_sendmail'
}
db_upgrade_default_php_handler_12_0_12()
{
    db_fix_change_column "hosting" "php_handler_id" "php_handler_id" "varchar(255) NOT NULL DEFAULT 'fastcgi'"
}

db_upgrade_wordpress_instances_12_0_17()
{
    db_fix_drop_unique_key "WordpressInstances" "path"
    db_fix_add_unique_key "WordpressInstances" "subscriptionId" "path"
}

upgrade_cgi_bin_mode_defaults_12_0_5()
{
    local script="$PRODUCT_ROOT_D/admin/plib/scripts/upgrade_cgi_bin_mode.php"
    if [ -f "$script" ]; then
        sw_engine_pleskrun "$script" >> "$product_log" 2>&1
    fi
}

db_upgrade_service_plan_domain_12_0_6()
{
    db_select "SELECT t.id AS planId FROM Templates AS t LEFT JOIN TmplData AS td ON (t.id = td.tmpl_id AND td.element = 'mailservice') WHERE t.type='domain' AND td.value IS NULL"
    for planId in $db_select_output; do
        db_do "INSERT INTO \`TmplData\` (\`tmpl_id\`, \`element\`, \`value\`) values ($planId, 'mailservice', 'true')"
    done
}

db_upgrade_iis_app_pools_12_0_6()
{
    db_fix_add_column "IisAppPools" "maxProcesses" "int(10) unsigned NOT NULL default 1"
    db_fix_add_column "IisAppPools" "idleTimeout" "int(10) unsigned NOT NULL default 5"
    db_fix_add_column "IisAppPools" "idleTimeoutAction" "enum('Terminate','Suspend') NOT NULL default 'Terminate'"
    db_fix_add_column "IisAppPools" "recyclingByTime" "int(11) NOT NULL default 1740"
    db_fix_add_column "IisAppPools" "recyclingByRequests" "int(11) NOT NULL default -1"
    db_fix_add_column "IisAppPools" "recyclingByVirtualMemory" "bigint(11) NOT NULL default -1"
    db_fix_add_column "IisAppPools" "recyclingByPrivateMemory" "bigint(11) NOT NULL default -1"
}

db_upgrade_subscription_outgoing_messages_overlimit_percent_to_default_12_0_6()
{
    db_select "SELECT s.id FROM Subscriptions s LEFT JOIN SubscriptionProperties sp ON (s.id = sp.subscription_id AND sp.name = 'outgoing_messages_overlimit_percent')  WHERE sp.value IS NULL AND s.object_type = 'domain'"

    for subscription_id in $db_select_output; do
        db_do "INSERT INTO SubscriptionProperties (subscription_id, name, value) VALUES ('$subscription_id', 'outgoing_messages_overlimit_percent', '0')"
    done
}

db_upgrade_plan_outgoing_messages_overlimit_percent_to_default_12_0_6()
{
    db_select "SELECT t.id FROM Templates t LEFT JOIN TmplData td ON (t.id = td.tmpl_id AND td.element = 'outgoing_messages_overlimit_percent')  WHERE td.value IS NULL AND t.type = 'domain'"

    for tmpl_id in $db_select_output; do
        db_do "INSERT INTO TmplData (tmpl_id, element, value) VALUES ('$tmpl_id', 'outgoing_messages_overlimit_percent', '0')"
    done
}

db_upgrade_sessions_12_0_7()
{
    db_do "delete from sessions"
    db_fix_del_column "sessions" "click_time"
    db_fix_change_column "sessions" "login_time" "login_time" "INT UNSIGNED"
}

db_subscription_properties_default_values_12_0_8()
{
        local subscription_ids
        local property_name="$1"

        db_select "SELECT s.id FROM Subscriptions s LEFT JOIN SubscriptionProperties sp ON (s.id = sp.subscription_id AND sp.name = '$property_name')  WHERE sp.value IS NULL AND s.object_type = 'domain'"
        subscription_ids="$db_select_output"

        for subscription_id in $subscription_ids; do
            db_do "INSERT INTO SubscriptionProperties (subscription_id, name, value) VALUES ('$subscription_id', '$property_name', 'default')"
        done
}

db_plan_properties_default_values_12_0_8()
{
        local tmpl_ids
        local property_name="$1"

        db_select "SELECT t.id FROM Templates t LEFT JOIN TmplData td ON (t.id = td.tmpl_id AND td.element = '$property_name')  WHERE td.value IS NULL AND t.type = 'domain'"
        tmpl_ids="$db_select_output"

        for tmpl_id in $tmpl_ids; do
            db_do "INSERT INTO TmplData (tmpl_id, element, value) VALUES ('$tmpl_id', '$property_name', 'default')"
        done
}

db_upgrade_subscription_properties_default_values_12_0_8()
{
	db_subscription_properties_default_values_12_0_8 'outgoing_messages_enable_sendmail'
}

db_upgrade_plan_properties_default_values_12_0_8()
{
	db_plan_properties_default_values_12_0_8 'outgoing_messages_enable_sendmail'
}

db_subscription_properties_default_values_12_0_9()
{
        local subscription_ids
        local property_name="$1"

        db_select "SELECT s.id FROM Subscriptions s LEFT JOIN SubscriptionProperties sp ON (s.id = sp.subscription_id AND sp.name = '$property_name')  WHERE sp.value IS NULL AND s.object_type = 'domain'"
        subscription_ids="$db_select_output"

        for subscription_id in $subscription_ids; do
            db_do "INSERT INTO SubscriptionProperties (subscription_id, name, value) VALUES ('$subscription_id', '$property_name', 'default')"
        done
}

db_plan_properties_default_values_12_0_9()
{
        local tmpl_ids
        local property_name="$1"

        db_select "SELECT t.id FROM Templates t LEFT JOIN TmplData td ON (t.id = td.tmpl_id AND td.element = '$property_name')  WHERE td.value IS NULL AND t.type = 'domain'"
        tmpl_ids="$db_select_output"

        for tmpl_id in $tmpl_ids; do
            db_do "INSERT INTO TmplData (tmpl_id, element, value) VALUES ('$tmpl_id', '$property_name', 'default')"
        done
}

db_upgrade_mbox_outgoing_messages_to_default_12_0_9()
{
    db_select "SELECT m.id FROM mail AS m LEFT JOIN mn_param AS mnp ON (m.id = mnp.mn_id AND mnp.param = 'outgoing_messages_limit') WHERE mnp.val IS NULL"

    for mail_id in $db_select_output; do
        db_do "INSERT INTO mn_param (mn_id, param, val) VALUES ('$mail_id', 'outgoing_messages_limit', 'default')"
    done
}

db_upgrade_domain_outgoing_messages_to_default_12_0_9()
{
    db_select "SELECT d.id FROM domains AS d LEFT JOIN Parameters AS p ON (d.id = p.id AND p.parameter = 'outgoing_messages_limit') WHERE p.value IS NULL"

    for domain_id in $db_select_output; do
        db_do "INSERT INTO Parameters (id, parameter, value) VALUES ('$domain_id', 'outgoing_messages_limit', 'default')"
    done
}

db_upgrade_subscription_properties_default_values_12_0_9()
{
    db_subscription_properties_default_values_12_0_9 'outgoing_messages_domain_limit'
    db_subscription_properties_default_values_12_0_9 'outgoing_messages_mbox_limit'
    db_subscription_properties_default_values_12_0_9 'outgoing_messages_subscription_limit'
}

db_upgrade_plan_properties_default_values_12_0_9()
{
    db_plan_properties_default_values_12_0_9 'outgoing_messages_domain_limit'
    db_plan_properties_default_values_12_0_9 'outgoing_messages_mbox_limit'
    db_plan_properties_default_values_12_0_9 'outgoing_messages_subscription_limit'
}

db_upgrade_wordpress_instances_12_0_9()
{
    db_fix_add_column "WordpressInstances" "apsInstanceId" "INT UNSIGNED DEFAULT NULL"
    db_fix_del_column 'WordpressInstances' 'isAttached'
    db_do_if_not "SHOW INDEX FROM WordpressInstances" "\$3 == \"subscriptionId\"" "CREATE INDEX subscriptionId ON WordpressInstances (subscriptionId)"
}

db_upgrade_sessions_12_0_9()
{
    db_fix_drop_table "WhpSessions"
    db_fix_del_column "sessions" "data"
}

db_upgrade_iis_app_pools_12_1_4()
{
    db_fix_add_column "IisAppPools" "managedPipelineMode" "enum('Classic','Integrated') NOT NULL default 'Integrated'"
}

db_upgrade_features_new_12_1_4()
{
    db_do "DELETE FROM cl_param WHERE param LIKE 'feature_%'"
}

db_12_1_4_fix_default_null()
{
	db_do "UPDATE \`$1\` SET \`$2\` = 0 WHERE \`$2\` IS NULL"
}

db_upgrade_unified_db_12_1_4()
{
	# Set engine
	if db_test_table "LimitsReservation"; then
		db_do "ALTER TABLE LimitsReservation ENGINE=InnoDB"
	fi

	# Set collation to utf8
	#for table in accounts actions apsContexts apsContextsApplications APSLicenses APSLicenseTypes apsResources 	#	BackendCache Cards certificates cl_param clients Configurations cp_access custom_buttons data_bases 	#	DatabaseUserRemoteAccessRules db_users dns_recs dns_recs_t dns_refs dns_zone dom_param domain_aliases 	#	domains event_handlers exp_event GL_remote_domains GL_settings hosting IP_Addresses key_history 	#	key_history_params Limits locales log_actions log_components log_rotation Logos longtaskparams longtasks 	#	mail mail_aliases mail_resp MailLists misc mn_param Modules Notifications Parameters password_secrets 	#	pd_users Permissions PersistentCache PlanItems protected_dirs report_section resp_freq SBConfig SBResellers 	#	secret_keys ServiceNodeCache ServiceNodeCertificates ServiceNodeConfiguration ServiceNodeEnvironment 	#	ServiceNodes SessionContexts sessions Skins smtp_poplocks spamfilter spamfilter_preferences subdomains 	#	Subscriptions sys_users Templates TmplData upgrade_history WebApps Webmails 	#	LimitsReservation; do
	#		db_set_utf8 $table
	#done

    # add indexes
    db_fix_add_key --name "license_type_id" "APSLicenses" "license_type_id"
    db_fix_add_key --name "date_type" "avstatistics" "type" "date"
    db_fix_add_key --name "uid" "dom_level_usrs" "uid"
    db_fix_add_key --name "displayName" "domain_aliases" "displayName"
    db_fix_add_key --name "dns_zone_id" "domains" "dns_zone_id"
    db_fix_add_key --name "source" "exp_event" "source" "event_type"
    db_fix_add_key --name "source_2" "exp_event" "source" "event_type" "obj_class"
    db_fix_add_key --name "source_3" "exp_event" "source" "event_time" "event_type" "obj_class"
    db_fix_add_key --name "flushed" "exp_event" "flushed"
    db_fix_add_key --name "action_id" "log_components" "action_id"

    # rename indexes
    db_fix_drop_key --name "collectionRef" "LimitsReservation"
    db_fix_add_key --name "collection" "LimitsReservation" "collectionRef" "limitName"
    db_fix_add_key --name "subscriptionId_idx" "WordpressInstances" "subscriptionId"

    # rename unique keys
    db_fix_add_unique_key --name "client_app_item" "APSClientApplicationItems" "client_id" "app_item_id"
    db_fix_add_unique_key --name "id" "BackupsSettings" "id" "type" "param"
    db_fix_add_unique_key --name "name_and_db_server_id" "data_bases" "name" "db_server_id"
    db_fix_add_unique_key --name "mm_cl" "mass_mail_clients" "mm_id" "cl_id"
    db_fix_add_unique_key --name "mm_dom" "mass_mail_domains" "mm_id" "dom_id"
    db_fix_add_unique_key --name "reference" "ServiceInstances" "serviceInstallationId" "servicePackageId" "reference"
    db_fix_add_unique_key --name "object_id_and_type" "Subscriptions" "object_id" "object_type"
    db_fix_add_unique_key --name "login" "sys_users" "serviceNodeId" "login"
    db_fix_add_unique_key --name "path" "WordpressInstances" "subscriptionId" "path"

    # add default 0 to be same as Windows
    db_fix_change_column "anon_ftp" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "APSApplicationItems" "license_type_id" "license_type_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "apscategories" "parent_id" "parent_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "APSLicenses" "license_type_id" "license_type_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "BackupsScheduled" "obj_id" "obj_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "BackupsScheduled" "period" "period" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "cl_param" "cl_id" "cl_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "clients" "limits_id" "limits_id" "INT(10) UNSIGNED DEFAULT '0'"
    db_12_1_4_fix_default_null "clients" "params_id"
    db_fix_change_column "clients" "params_id" "params_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_12_1_4_fix_default_null "clients" "perm_id"
    db_fix_change_column "clients" "perm_id" "perm_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_12_1_4_fix_default_null "clients" "pool_id"
    db_fix_change_column "clients" "pool_id" "pool_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_12_1_4_fix_default_null "clients" "logo_id"
    db_fix_change_column "clients" "logo_id" "logo_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_12_1_4_fix_default_null "clients" "tmpl_id"
    db_fix_change_column "clients" "tmpl_id" "tmpl_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_12_1_4_fix_default_null "clients" "sapp_pool_id"
    db_fix_change_column "clients" "sapp_pool_id" "sapp_pool_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_12_1_4_fix_default_null "clients" "parent_id"
    db_fix_change_column "clients" "parent_id" "parent_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "ClientsTraffic" "cl_id" "cl_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "data_bases" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "db_users" "db_id" "db_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "db_users" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "db_users" "db_server_id" "db_server_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "disk_usage" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "dom_level_usrs" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "dom_level_usrs" "account_id" "account_id" "INT(10) UNSIGNED"
    db_fix_change_column "dom_param" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "domain_aliases" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "domains" "cl_id" "cl_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "domains" "cert_rep_id" "cert_rep_id" "INT(10) UNSIGNED DEFAULT '0'"
    db_12_1_4_fix_default_null "domains" "limits_id"
    db_fix_change_column "domains" "limits_id" "limits_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_12_1_4_fix_default_null "domains" "params_id"
    db_fix_change_column "domains" "params_id" "params_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_12_1_4_fix_default_null "domains" "vendor_id"
    db_fix_change_column "domains" "vendor_id" "vendor_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "DomainServices" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_12_1_4_fix_default_null "DomainServices" "parameters_id"
    db_fix_change_column "DomainServices" "parameters_id" "parameters_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_12_1_4_fix_default_null "DomainServices" "ipCollectionId"
    db_fix_change_column "DomainServices" "ipCollectionId" "ipCollectionId" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "DomainsTraffic" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "event_handlers" "action_id" "action_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_12_1_4_fix_default_null "externalWebmails" "enabled"
    db_fix_change_column "externalWebmails" "enabled" "enabled" "INT(10) NOT NULL DEFAULT '0'"
    db_fix_change_column "forwarding" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "ftp_users" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "ftp_users" "sys_user_id" "sys_user_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "hosting" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "hosting" "sys_user_id" "sys_user_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "IP_Addresses" "ssl_certificate_id" "ssl_certificate_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_12_1_4_fix_default_null "IP_Addresses" "default_domain_id"
    db_fix_change_column "IP_Addresses" "default_domain_id" "default_domain_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "key_history_params" "key_id" "key_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "log_actions" "action_id" "action_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "log_actions" "object_id" "object_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "log_components" "action_id" "action_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "mail" "account_id" "account_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "mail" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "mail" "perm_id" "perm_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "mail_aliases" "mn_id" "mn_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "mail_redir" "mn_id" "mn_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "mail_resp" "mn_id" "mn_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "MailLists" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "mn_param" "mn_id" "mn_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "ModuleSettings" "module_id" "module_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "pd_users" "account_id" "account_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "pd_users" "pd_id" "pd_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "PhpSettingsParameters" "id" "id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "PlanItemProperties" "plan_item_id" "plan_item_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "protected_dirs" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "report" "type" "type" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "report" "user_type" "user_type" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "report_auto" "report_id" "report_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "report_auto" "client" "client" "INT(10) UNSIGNED DEFAULT '0'"
    db_fix_change_column "report_auto" "domain" "domain" "INT(10) UNSIGNED DEFAULT '0'"
    db_fix_change_column "report_section" "report_id" "report_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "Repository" "component_id" "component_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "resp_attach" "rn_id" "rn_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "resp_forward" "rn_id" "rn_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "resp_freq" "rn_id" "rn_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "sessions" "type" "type" "SMALLINT(5) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "subdomains" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "subdomains" "sys_user_id" "sys_user_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "SubscriptionProperties" "subscription_id" "subscription_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "sys_users" "account_id" "account_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "TmplData" "tmpl_id" "tmpl_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "web_users" "dom_id" "dom_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "web_users" "sys_user_id" "sys_user_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "WebApps" "domain_service_id" "domain_service_id" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
    db_fix_change_column "ftp_users" "permission" "permission" "BIGINT(20) UNSIGNED NOT NULL DEFAULT '0'"

    # add other defaults to be same as Windows
    db_fix_change_column "APSCatalogUpdates" "update_type" "update_type" "ENUM('patch','upgrade') NOT NULL DEFAULT 'patch'"
    db_fix_change_column "DatabaseServers" "type" "type" "ENUM('mysql','postgresql','mssql') NOT NULL DEFAULT 'mysql'"
    db_fix_change_column "domains" "htype" "htype" "ENUM('none','vrt_hst','std_fwd','frm_fwd') NOT NULL DEFAULT 'none'"
    db_fix_change_column "exp_event" "event_type" "event_type" "ENUM('started','stopped','created','updated','deleted','status_changed','terminated','flushed','installed','uninstalled','siteapp_added','siteapp_removed','expired','exceeded','guid_changed') NOT NULL DEFAULT 'started'"
    db_fix_change_column "exp_event" "obj_class" "obj_class" "varchar(255) NOT NULL DEFAULT 'license'"
    db_do 'UPDATE `ip_pool` SET `type` = "shared" WHERE `type` IS NULL'
    db_fix_change_column "ip_pool" "type" "type" "ENUM('shared','exclusive') NOT NULL DEFAULT 'shared'"

    db_fix_change_column "APSCatalogUpdates" "received" "received" "DATETIME DEFAULT '0000-00-00 00:00:00'"
    db_fix_change_column "upgrade_history" "upgrade_date" "upgrade_date" "DATETIME NOT NULL DEFAULT '0000-00-00 00:00:00'"
    db_fix_change_column "ClientsTraffic" "date" "date" "DATE NOT NULL DEFAULT '0000-00-00'"
    db_fix_change_column "DomainsTraffic" "date" "date" "DATE NOT NULL DEFAULT '0000-00-00'"
    db_fix_change_column "MailMessagesStat" "date" "date" "DATE NOT NULL DEFAULT '0000-00-00'"
    db_fix_change_column "PleskPagesStat" "date" "date" "DATE NOT NULL DEFAULT '0000-00-00'"
    db_fix_change_column "SitePagesStat" "date" "date" "DATE NOT NULL DEFAULT '0000-00-00'"

    db_fix_change_column "exp_event" "event_time" "event_time" "TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP"

    # add notNull attribute where default is available
    db_fix_change_column "apsContexts" "subscriptionId" "subscriptionId" "INT(10) UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "BackupsScheduled" "content_type" "content_type" "ENUM('backup_content_all_at_domain','backup_content_vhost_only','backup_content_mail_only') NOT NULL DEFAULT 'backup_content_all_at_domain'"
    db_fix_change_column "PhpSettings" "noteId" "noteId" "INT(10) UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "PlanItems" "isVisible" "isVisible" "INT(10) UNSIGNED NOT NULL DEFAULT 1"
    db_fix_change_column "PlanItems" "applicableToSubscription" "applicableToSubscription" "INT(10) UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "PlanItems" "applicableToEmail" "applicableToEmail" "INT(10) UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "PlanItems" "applicableToSite" "applicableToSite" "INT(10) UNSIGNED NOT NULL DEFAULT 0"

    # drop NOT NULL DEFAULT ''
    db_fix_change_column "apsResourcesParameters" "value" "value" "VARCHAR(255)"
    db_fix_change_column "clients" "external_id" "external_id" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "DatabaseUserRemoteAccessRules" "firewallRuleId" "firewallRuleId" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "dns_recs" "external_id" "external_id" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "dns_zone" "external_id" "external_id" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "domains" "external_id" "external_id" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "hosting" "write_modify" "write_modify" "VARCHAR(255)"
    db_fix_change_column "ServiceNodeCertificates" "certFile" "certFile" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "ServiceNodeCertificates" "caFile" "caFile" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "ServiceNodes" "externalId" "externalId" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "Subscriptions" "external_id" "external_id" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "Templates" "external_id" "external_id" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin"

    # fix wrong type attributes
    db_fix_change_column "accounts" "password" "password" "TEXT CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "ai_vendor_sources" "url" "url" "VARCHAR(255) NOT NULL DEFAULT ''"
    db_fix_change_column "ai_vendor_sources" "login" "login" "VARCHAR(255) DEFAULT ''"
    db_fix_change_column "ai_vendor_sources" "password" "password" "VARCHAR(255) DEFAULT ''"
    db_fix_change_column "APSLicenseTypes" "application_name" "application_name" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin NOT NULL"
    db_fix_change_column "APSLicenseTypes" "application_versions" "application_versions" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin NOT NULL"
    db_fix_change_column "ApsTokens" "applicationId" "applicationId" "VARCHAR(2000) NOT NULL"
    db_fix_change_column "BackupsScheduled" "backup_time" "backup_time" "VARCHAR(8) NOT NULL default '00:00:00'"
    db_fix_change_column "BackendCache" "data" "data" "TEXT NOT NULL"
    db_fix_change_column "cl_param" "val" "val" "VARCHAR(255)"
    db_fix_change_column "data_bases" "name" "name" "VARCHAR(64) CHARACTER SET utf8 COLLATE utf8_bin NOT NULL"
    db_fix_change_column "DatabaseServers" "server_version" "server_version" "CHAR(255) NOT NULL"
    db_fix_change_column "DatabaseServers" "admin_login" "admin_login" "CHAR(255) NOT NULL"
    db_fix_change_column "DatabaseServers" "admin_password" "admin_password" "CHAR(255) NOT NULL"
    db_fix_change_column "dom_level_usrs" "passwd" "passwd" "VARCHAR(255) NOT NULL DEFAULT ''"
    db_fix_change_column "dom_param" "val" "val" "VARCHAR(255)"
    db_fix_change_column "domain_aliases" "status" "status" "BIGINT(20) UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "DomainOutgoingMessagesPeaks" "rejected" "rejected" "BIGINT(20) UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "DomainOutgoingMessagesStats" "limitValue" "limitValue" "BIGINT(20) NOT NULL DEFAULT -1"
    db_fix_change_column "DomainOutgoingMessagesStats" "passed" "passed" "BIGINT(20) UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "DomainOutgoingMessagesStats" "rejected" "rejected" "BIGINT(20) UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "DomainServices" "status" "status" "BIGINT(20) NOT NULL DEFAULT 0"
    db_fix_change_column "hosting" "www_root" "www_root" "TEXT CHARACTER SET utf8 COLLATE utf8_bin NOT NULL"
    db_fix_change_column "key_history_params" "val" "val" "VARCHAR(255) NOT NULL"
    db_fix_change_column "Limits" "limit_name" "limit_name" "VARCHAR(255) NOT NULL"
    db_fix_change_column "Limits" "value" "value" "VARCHAR(255)"
    db_fix_change_column "LimitsReservation" "value" "value" "BIGINT(20)"
    db_fix_change_column "log_components" "component" "component" "VARCHAR(255) NOT NULL"
    db_fix_change_column "log_components" "old_value" "old_value" "VARCHAR(255) NOT NULL"
    db_fix_change_column "log_components" "new_value" "new_value" "VARCHAR(255) NOT NULL"
    db_fix_change_column "longtaskparams" "val" "val" "MEDIUMTEXT"
    db_fix_change_column "mail" "disk_usage" "disk_usage" "BIGINT UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "MailOutgoingMessagesPeaks" "rejected" "rejected" "BIGINT(20) UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "MailOutgoingMessagesStats" "limitValue" "limitValue" "BIGINT(20) NOT NULL DEFAULT -1"
    db_fix_change_column "MailOutgoingMessagesStats" "passed" "passed" "BIGINT(20) UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "MailOutgoingMessagesStats" "rejected" "rejected" "BIGINT(20) UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "misc" "val" "val" "VARCHAR(2000)"
    db_fix_change_column "mn_param" "val" "val" "VARCHAR(255)"
    db_fix_change_column "Parameters" "parameter" "parameter" "VARCHAR(255) NOT NULL"
    db_fix_change_column "Parameters" "value" "value" "VARCHAR(255)"
    db_fix_change_column "Permissions" "permission" "permission" "VARCHAR(255) NOT NULL"
    db_fix_change_column "Permissions" "value" "value" "VARCHAR(255)"
    db_fix_change_column "SubscriptionOutgoingMessagesPeaks" "rejected" "rejected" "BIGINT(20) UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "SubscriptionOutgoingMessagesStats" "limitValue" "limitValue" "BIGINT(20) NOT NULL DEFAULT -1"
    db_fix_change_column "SubscriptionOutgoingMessagesStats" "passed" "passed" "BIGINT(20) UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "SubscriptionOutgoingMessagesStats" "rejected" "rejected" "BIGINT(20) UNSIGNED NOT NULL DEFAULT 0"
    db_fix_change_column "ServiceNodeCache" "path" "path" "TEXT CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "ServiceNodeCache" "cache" "cache" "TEXT CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "ServiceNodes" "accessKey" "accessKey" "TEXT CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "SessionContexts" "data" "data" "MEDIUMTEXT"
    db_fix_change_column "subdomains" "www_root" "www_root" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin NOT NULL DEFAULT 'httpdocs'"
    db_fix_change_column "suspend_handler_history" "session_id" "session_id" "VARCHAR(255)"
    db_fix_change_column "TmplData" "value" "value" "VARCHAR(255) NOT NULL"
    db_fix_change_column "WordpressInstanceProperties" "value" "value" "MEDIUMTEXT NOT NULL"

    # drop default=0 for foreign key columns
    db_fix_change_column "admin_aliases" "account_id" "account_id" "INT(10) UNSIGNED NOT NULL"
    db_fix_change_column "cf_dsn" "domain_id" "domain_id" "INT(10) UNSIGNED NOT NULL"
    db_fix_change_column "cf_dsn_params" "cf_dsn_id" "cf_dsn_id" "INT(10) UNSIGNED NOT NULL"
    db_fix_change_column "clients" "account_id" "account_id" "INT(10) UNSIGNED NOT NULL"
    db_fix_change_column "DatabaseServers" "parameters_id" "parameters_id" "INT(10) UNSIGNED"
    db_fix_change_column "dom_level_usrs" "account_id" "account_id" "INT(10) UNSIGNED"
    db_fix_change_column "IisAppPoolDomains" "domainId" "domainId" "INT(10) UNSIGNED NOT NULL"
    db_fix_change_column "IisAppPoolDomains" "poolId" "poolId" "INT(10) UNSIGNED NOT NULL"
    db_fix_change_column "IisAppPools" "ownerId" "ownerId" "INT(10) UNSIGNED NOT NULL"
    db_fix_change_column "mail" "userId" "userId" "INT(10) UNSIGNED"
    db_fix_change_column "shared_ssl_links" "sharedSslId" "sharedSslId" "INT(10) UNSIGNED NOT NULL"
    db_fix_change_column "SharedSslDomains" "clientId" "clientId" "INT(10) UNSIGNED NOT NULL"
    db_fix_change_column "SharedSslDomains" "domainId" "domainId" "INT(10) UNSIGNED NOT NULL"

    # drop NOT NULL where this constraint cannot be added on Windows
    db_fix_change_column "APSLicenses" "source" "source" "TEXT"
    db_fix_change_column "custom_buttons" "text" "text" "VARCHAR(255)"
    db_fix_change_column "DatabaseServers" "server_version" "server_version" "CHAR(255)"
    db_fix_change_column "domain_aliases" "dns_zone_id" "dns_zone_id" "INT(10) UNSIGNED"
    db_fix_change_column "domains" "dns_zone_id" "dns_zone_id" "INT(10) UNSIGNED"
    db_fix_change_column "event_handlers" "command" "command" "TEXT CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "key_history_params" "val" "val" "VARCHAR(255)"
    db_fix_change_column "log_actions" "date" "date" "DATETIME"
    db_fix_change_column "mail" "userId" "userId" "INT(10) UNSIGNED"
    db_fix_change_column "Notifications" "send2reseller" "send2reseller" "VARCHAR(1) CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "report" "name" "name" "VARCHAR(255)"
    db_fix_change_column "sessions" "login" "login" "VARCHAR(255)"
    db_fix_change_column "sessions" "ip_address" "ip_address" "VARCHAR(39)"
    db_fix_change_column "upgrade_history" "version_info" "version_info" "VARCHAR(50) CHARACTER SET utf8 COLLATE utf8_bin"
    db_fix_change_column "upgrade_history" "db_version" "db_version" "VARCHAR(50) CHARACTER SET utf8 COLLATE utf8_bin"
}

db_upgrade_unified_db_12_1_5()
{
    db_fix_change_column "certificates" "csr" "csr" "BLOB"
    db_fix_change_column "certificates" "pvt_key" "pvt_key" "BLOB"
    db_fix_change_column "certificates" "cert" "cert" "BLOB"
    db_fix_change_column "certificates" "ca_cert" "ca_cert" "BLOB"
    db_fix_change_column "certificates" "cert_file" "cert_file" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin NOT NULL DEFAULT ''"
    db_fix_change_column "certificates" "ca_file" "ca_file" "VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_bin NOT NULL DEFAULT ''"
    db_fix_change_column "certificates" "name" "name" "VARCHAR(255) NOT NULL DEFAULT ''"
    db_fix_change_column "Configurations" "file" "file" "BLOB"
    db_fix_change_column "Configurations" "errorFile" "errorFile" "BLOB"
    db_fix_change_column "db_users" "login" "login" "VARCHAR(128) NOT NULL"
    db_fix_change_column "GL_remote_domains" "serviceNodeId" "serviceNodeId" "INT(11) NOT NULL DEFAULT 1"
    db_fix_change_column "GL_settings" "serviceNodeId" "serviceNodeId" "INT(11) NOT NULL DEFAULT 1"
    db_fix_change_column "secret_keys" "client_id" "client_id" "INT(10) UNSIGNED"
    db_fix_change_column "Limits" "value" "value" "BIGINT(20)"
    db_fix_del_column "longtasks" "can_cancel"
    db_fix_del_column "ServiceNodes" "transport"
    db_fix_drop_unique_key --name "name" "dsn"
    db_fix_add_unique_key --name "name" "dsn" "name" "serviceNodeId"
    db_fix_add_key --name "parentDomainId" "domains" "parentDomainId"
}

db_upgrade_subscriptions_12_1_5()
{
    db_fix_change_column "Subscriptions" "uuid" "uuid" "VARCHAR(36)"
}

db_upgrade_generate_service_instance_references_12_1_5()
{
    sw_engine_pleskrun $PRODUCT_ROOT_D/admin/plib/scripts/upgrade_service_instance_references_12.1.5.php && suc || warn "fail\nCannot generate references for ServiceInstances"
}

db_drop_pmm_tables_12_1_5()
{
	db_fix_drop_table "PMM"
	db_fix_drop_table "PMMDefault"
}

db_upgrade_smb_unified_12_1_6()
{
	db_fix_add_key --name "subscriptionDomainId_idx" "smb_users" "subscriptionDomainId"

	db_fix_add_key --name "filterId_idx" "smb_apsBundleFilterItems" "filterId"
	db_fix_add_key --name "roleId_idx" "smb_roleGeneralPermissions" "roleId"
	db_fix_add_key --name "ownerId_idx" "smb_roles" "ownerId"
	db_fix_add_key --name "ownerId_idx" "smb_users" "ownerId"
	db_fix_add_key --name "isBuiltIn_idx" "smb_users" "isBuiltIn"
	db_fix_drop_unique_key --name "cacheId_2" "smb_apsPackages"

	db_fix_change_column "smb_apsBundleFilterItems" "filterId" "filterId" "INT(10) UNSIGNED NOT NULL DEFAULT '0'"
	db_fix_change_column "smb_users" "isDomainAdmin" "isDomainAdmin" "INT(11) DEFAULT '0'"
	db_fix_change_column "smb_users" "isLegacyUser" "isLegacyUser" "INT(11) DEFAULT '0'"
	db_fix_change_column "smb_apsPackages" "installed" "installed" "VARCHAR(26)"
	db_fix_change_column "smb_users" "subscriptionDomainId" "subscriptionDomainId" "INT(11) DEFAULT 0"
}

db_remove_tables_12_1_6()
{
	db_fix_drop_table "smb_apsImportedResources"
	db_fix_drop_table "smb_apsImportedSettings"
}

db_drop_storefront_12_1_7()
{
	db_fix_del_column "smb_apsPackages" "isNonStorefrontCommercial"
}

refresh_trusted_ips_12_1_8()
{
	sw_engine_pleskrun $PRODUCT_ROOT_D/admin/plib/scripts/refresh-trusted-ips.php && suc || warn "fail\nrefresh trusted ips for site preview"
}

db_upgrade_backups_scheduled_12_1_8()
{
	db_fix_change_column "BackupsScheduled" "split_size" "split_size" "BIGINT(20) UNSIGNED NOT NULL DEFAULT '0'"
}
set_drweb_params()
{
	drweb_user="drweb"
	drweb_group="drweb"

	drweb_service="drwebd"
}

true drweb_status
drweb_status()
{
	local pidfile="/var/drweb/run/drwebd.pid"
	if [ ! -r "$pidfile" ]; then
		p_echo "drweb is stopped (no pidfile found)"
		return 1
	fi

	local pid=$(head -1 "$pidfile" 2>/dev/null)
	if  [ -z "$pid" ]; then
		p_echo "drweb is stopped (wrong pidfile)"
		return 1
	fi

	if kill -0 "$pid" 2>/dev/null || ps -p "$pid" >/dev/null 2>&1 ; then
		p_echo "drwebd (pid $pid) is running..."
		return 0
	fi
	p_echo "drwebd is stopped"
	return 1
}

#proftpd

set_ftpd_params()
{
	anonftp_user="${product}ftp"
	anonftp_group="${product}ftp"

	PROFTPD_ROOT="$PRODUCT_ROOT_D/ftpd"
	PROFTPD_ETC_D="$PROFTPD_ROOT/etc"
}
upgrade_proftpd_107()
{
	echo_try "upgrade anon_ftp quota to new format"
	local domain;
	for domain in $HTTPD_VHOSTS_D/*; do
		[ -e $domain/anon_ftp/incoming/quotadir ] || continue
		rm -f $domain/anon_ftp/incoming/quotadir/.undeletable 			$domain/anon_ftp/incoming/quotadir/.quota >> $product_log 2>&1
		rmdir $domain/anon_ftp/incoming/quotadir >> $product_log 2>&1
	done
	sw_engine_pleskrun $PRODUCT_ROOT_D/admin/plib/scripts/upgrade_proftpd_107.php && suc || warn "fail\nupgrade anon_ftp quota to new format"
}

upgrade_transfer_proftpd_logs_11_6_1()
{
	transfer_log_with_compat_symlink "${PRODUCT_ROOT_D}/var/log/xferlog" "${PRODUCT_LOGS_D}/xferlog"
	transfer_log_with_compat_symlink "${PRODUCT_ROOT_D}/var/log/ftp_tls.log" "${PRODUCT_LOGS_D}/ftp_tls.log"

	if ls ${PRODUCT_ROOT_D}/var/log/ftp_tls.log.*.gz > /dev/null 2>&1; then
		mv -f ${PRODUCT_ROOT_D}/var/log/ftp_tls.log.*.gz "${PRODUCT_LOGS_D}"
	fi

	if ls ${PRODUCT_ROOT_D}/var/log/xferlog.processed* > /dev/null 2>&1; then
		mv -f ${PRODUCT_ROOT_D}/var/log/xferlog.processed* "${PRODUCT_LOGS_D}"
	fi
}

tomcat_reconfigure_server_safe()
{
	echo_try "reconfigure Tomcat services"
	local ports=

	# needed for proper work tomcatmng
	conf_setval "${prod_conf_t}" CATALINA_HOME "${CATALINA_HOME}"

	for name in warp coyote; do
		db_select "SELECT val FROM misc WHERE val = '${name}_connector_port'" 			|| {
				p_echo "Failed to get Tomcat parameter '${name}_connector_port' from the database"
				return 1
			}
		if [ -n "$db_select_output" ]; then
			ports="--${name}-connector-port $db_select_output"
		fi
	done

	"$UTILS_DIR/tomcatmng" --install-psa --replace-if-exists $ports 		|| {
			p_echo "Failed to configure Tomcat PSA service"
			return 1
		}

	db_select "SELECT d.name
				FROM domains d, DomainServices ds
				WHERE d.id = ds.dom_id
					AND type='tomcat'
					AND ds.status = 0" 		|| {
			p_echo "Failed to get active Tomcat domains"
			return 1
		}
	local domains="`echo $db_select_output | tr ' ' ','`"

	if [ -n "$domains" ]; then
		"$PLESK_LIBEXEC_DIR/tomcat_upgrade_helper" --reconfigure-domains "$domains" 			|| {
				p_echo "Failed to reconfigure Tomcat settings for active domains"
				return 1
			}
	fi

	suc
	return 0
}

tomcat_reconfigure_server()
{
	tomcat_reconfigure_server_safe || warn "Tomcat server reconfiguration"
}
# vim:syntax=sh

set_swkey_params()
{
	SWKEY_REPOSITORY="/etc/sw/keys"
	KEYUPGRADE_PATH="/usr/lib/plesk-9.0/key-upgrade"
	KEYUPGRADE8_PATH="/usr/lib/plesk-9.0/key-upgrade8"
}

forced_key_upgrade_cmd()
{
	set_swkey_params
	# set to run key upgrade script
	p_echo "Starting forced Plesk Key upgrade task"
	nohup /bin/sh -c "$KEYUPGRADE_PATH force 2>&1 | logger -p daemon.err -t plesk-key-upgrader" 		>/dev/null 2>&1 &
}

# vim:ft=sh
# Usage:  pleskrc <service> <action>
pleskrc()
{
	[ 2 -le $# ] || die "Not enough arguments"

	local service_name=$1
	local action=$2
	local ret=0
	local inten
	shift
	shift

	# Now check redefined functions
	if test "$machine" = "linux" && is_function "${service_name}_${action}_${machine}_${linux_distr}"; then
		"${service_name}_${action}_${machine}_${linux_distr}" "$@"
		return $?
	elif is_function "${service_name}_${action}_${machine}"; then
		"${service_name}_${action}_${machine}" "$@"
		return $?
	elif is_function "${service_name}_${action}"; then
		"${service_name}_${action}" "$@"
		return $?
	fi

	# Not redefined - call default action
	eval "service=\$${service_name}_service"
	[ -n "$service" ] || die "$action $service_name service (Empty service name for '$service_name')"

	inten="$action service $service"
	[ "$action" = "status" -o "$action" = "exists" ] || echo_try "$inten"

	service_ctl "$action" "$service" "$service_name"

	ret="$?"
	if [ "$action" != "status" -a "${action}" != "exists" ]; then
		if [ "$ret" -eq 0 ]; then
			suc
		else
			if [ -x "/bin/systemctl" ]; then
				/bin/systemctl -l status "${service}.service" | awk 'BEGIN {s=0} s==1 {s=2} /^$/ {s=1} s==2 {print}' >> "$product_log"
			fi
			warn "$inten"
		fi
	fi

	return $ret
}

# NOTE:
#	Function service_ctl is just helper for pleskrc().
#	Do not call it directly, use pleskrc()!!!
service_ctl()
{
	local action=$1
	local service=$2
	local service_name=$3

	if [ "$action" != "exists" ]; then
		_service_exec $service exists;
		if [ "$?" != "0" ]; then
			warn "attempt to ${inten} - control script doesn't exist or isn't executable"
			return 1
		fi
	fi

	case "$action" in
		start)
			pleskrc "$service_name" status || _service_exec "$service" "$action"
			;;
		stop)
			! pleskrc "$service_name" status || _service_exec "$service" "$action"
			;;
		restart)
			if pleskrc "$service_name" status; then
				_service_exec "$service" "$action"
			else
				_service_exec "$service" start
			fi
			;;
		reload)
			! pleskrc "$service_name" status || _service_exec "$service" "$action"
			;;
		status)
			_service_exec "$service" status
			;;
		try-restart)
			if [ -x "/bin/systemctl" ]; then
				_service_exec "$service" "$action"
			else
				! pleskrc "$service_name" status || _service_exec "$service" "restart"
			fi
			;;
		try-reload)
			! pleskrc "$service_name" status || _service_exec "$service" "reload"
			;;
		reload-or-restart)
			if [ -x "/bin/systemctl" ]; then
				_service_exec "$service" "$action"
			elif pleskrc "$service_name" status; then
				_service_exec "$service" "reload"
			else
				_service_exec "$service" "start"
			fi
			;;
		*)
			_service_exec "$service" "$action"
			;;
	esac >> "$product_log"
}

_service_exec()
{
	local service=$1
	local action=$2

	local action_cmd
	local sysvinit_service="/etc/init.d/$service"

	if [ -x "/bin/systemctl" ]; then
		case "${action}" in
			exists)
				if /bin/systemctl list-unit-files | awk 'BEGIN { rc = 1 } $1 == "'$service'.service" { rc = 0;} END { exit rc }'; then
					return 0 # systemd unit
				elif [ -x "$sysvinit_service" ]; then
					return 0 # sysvinit compat
				fi
				return 1 # not found
				;;
			status)
				action="is-active"
				;;
			reload|graceful)
				action='reload-or-try-restart'
				;;
		esac
		/bin/systemctl "$action" "${service}.service"
	elif  [ -x "/sbin/initctl" -a -e "/etc/init/$service.conf" ]; then  # upstart (ubuntu)
		if [ "$action" = "status" ]; then
			/sbin/initctl status "$service" | grep -qE ' ([0-9]+)$' && return 0 || return 1
		elif [ "$action" = "exists" ]; then
			return 0
		else
			/sbin/initctl "$action" "$service"
		fi
	else
		if [ -x "/usr/sbin/invoke-rc.d" ]; then
			action_cmd="/usr/sbin/invoke-rc.d $service"
		elif [ -x "/sbin/service" ]; then
			action_cmd="/sbin/service $service"
		elif [ -x "/usr/sbin/service" ]; then
			action_cmd="/usr/sbin/service $service"
		else
			action_cmd="$sysvinit_service"
		fi

		if [ "$action" = "exists" ]; then
			[ -x "$sysvinit_service" ] && return 0 || return 1
		else
			$action_cmd $action 2>/dev/null
		fi
	fi
}

is_function()
{
	local type_output=$(type -t "$1")
	test "X${type_output}" = "Xfunction"
}

apparmor_install_profile()
{
	local profile="$1"
	local action="${2:-install}"

	return 0
}
form_undotted_version()
{
# Backward compatible with versions < 11.1.11
	local major=`echo $1 | awk -F. '{ print $1 }'`
	local minor=`echo $1 | awk -F. '{ print $2 }'`
	local patch=`echo $1 | awk -F. '{ print $3 }'`
	if [ -z "$major" -o -z "$minor" -o -z "$patch" ]; then
		return 1
	fi

	if [ $major -lt 11 ] ; then
		echo "${major}${minor}"
	elif [ $major -eq 11 -a $minor -lt 1 ] || [ $major -eq 11 -a $minor -eq 1 -a $patch -lt 11 ]; then
		printf "%d%02d" $major $minor
	else
		printf "%03d%03d%03d" $major $minor $patch
	fi
}

set_core_version()
{
	local prod_version="$1"
	local version
	local core_version

	local versions_list="10.0.0_10.8.0 10.0.1_10.9.0 10.1.0_10.10.0 10.1.1_10.10.1 10.2.0_10.11.0 10.3.1_10.12.0 10.4.4_10.13.4"

	for pair in $versions_list; do
		eval `echo "$pair" | awk -F '_' '{print "version="$1; print "core_version="$2}'`
		if [ "$version" = "$prod_version" ]; then
			prev_version="$core_version"
		fi
	done
}

get_prev_version()
{
	local oldest_known_product_version="100"
	local known_old_product_versions="100 101 104 107 108 109 1010 1011 1012 1013 1100 1101"

	# Debian package system reports previous version as script argument
	if [ -z "${prev_version}" -o -z "${prev_build}" ]; then

		if [ -f "$1/core.version" ]; then
		# Check version from core
			[ -n "$prev_version" ] || prev_version=`cat $1/core.version | awk '{ print $1 }'`
			prev_build=`cat $1/core.version | awk '{ print $4 }'`
		elif [ -f $1/version ]; then
		# version file must exists in product directory
			[ -n "$prev_version" ] || prev_version=`cat $1/version | awk '{ print $1 }'`
			prev_build=`cat $1/version | awk '{ print $4 }'`
			set_core_version "$prev_version"
		elif [ -f $1/version.upg ]; then
		# version file must exists in product directory
			[ -n "$prev_version" ] || prev_version=`cat $1/version.upg | awk '{ print $1 }'`
			prev_build=`cat $1/version.upg | awk '{ print $4 }'`
			set_core_version "$prev_version"
		else
		    p_echo "    Unable to find version file $1. "
		    p_echo "    Please, restore version file or remove old ${PRODUCT_NAME}, if unnecessary."
		    exit 1
		fi
	fi

	prev_short_version=`form_undotted_version ${prev_version}`
	if [ "X${prev_version}" = "X" -o "X${prev_build}" = "X" -o -z "$prev_short_version" ]; then
		p_echo "    Unable to determine version or build number of ${PRODUCT_NAME} installed. It is impossible"
		p_echo "    to upgrade ${PRODUCT_NAME} to version $product_version."
		p_echo "    To try to install a new version without saving data remove '$1' directory"
		p_echo "    manually. Please backup all data before removing it if necessary."
		exit 1
	fi

	# check, that it is known version of product
	if [ $prev_short_version -le '1101' ] ; then
		for i in $known_old_product_versions ; do
			if [ "X${prev_short_version}" = "X${i}" ]; then
				return 0
			fi
		done
	else
		# We support all (?!) versions >= 11.1.11 (for now).
		return 0
	fi

	# version is unknown for upgrader. let the user to know the propblem
	p_echo "    Unable to upgrade. You have $prev_version version installed,"
	p_echo "    but the upgrader is incompatible with this version."
	p_echo "    Please consider using migration instead of an upgrade."
	exit 1
}

check_product_root()
{
	# check product root
	p_echo
	pnnl_echo "===> Checking for previous installation ... "

	do_upgrade=0

	if [ ! -s "$PRODUCT_ROOT_D/version" -a ! -s "$PRODUCT_ROOT_D/version.upg" -a ! -s "$PRODUCT_ROOT_D/core.version" ]; then
		p_echo "not found."
	else
		p_echo "found."
		get_prev_version ${PRODUCT_ROOT_D}
		do_upgrade=1
	fi
}


check_localtime_file()
{
	[ -f /etc/localtime ] && return 0
	report_problem 		"There is no /etc/localtime file in system." 		"Run tzsetup as root to set up your timezone"
	return 1
}

check_tmp_permissions()
{
	local tmp_perm_mask=777
	[ $((0`stat -c %a /tmp` & 0$tmp_perm_mask)) -eq $((0$tmp_perm_mask)) ] && return 0

	if [ -f /tmp/pp-bootstrapper-ez-templates-prep-install.flag ]; then
		chmod 1777 /tmp
		p_echo " /tmp file access permissions were adjusted to 1777."
		p_echo
		return 0
	else
		report_problem 			"/tmp should have at least $tmp_perm_mask file access permissions." 			"Run 'chmod 1777 /tmp' as root to set up proper file access permissions."
		return 1
	fi
}

preinstall_checks()
{
	check_localtime_file || exit 1
	check_tmp_permissions || exit 1
}

# echo message to product log and console (always visible)
pp_echo()
{
    if [ -n "$product_log" ] ; then
        echo "$@" >> "$product_log" 2>&1
    fi
    echo "$@" >&2
}

# echo message to product log, unless debug
p_echo()
{
    if [ -n "$PLESK_INSTALLER_DEBUG" -o -n "$PLESK_INSTALLER_VERBOSE" -o -z "$product_log" ] ; then
        echo "$@" >&2
    else
        echo "$@" >> "$product_log" 2>&1
    fi
}

# echo message to product log without new line, unless debug
pnnl_echo()
{
    if [ -n "$PLESK_INSTALLER_DEBUG" -o -n "$PLESK_INSTALLER_VERBOSE" -o -z "$product_log" ] ; then
        echo -n "$*" >&2
    else
        echo -n "$*" >> "$product_log" 2>&1
    fi
}

die()
{
	PACKAGE_SCRIPT_FAILED="$*"

	report_problem 		"ERROR while trying to $*" 		"Check the error reason(see log file: ${product_log}), fix and try again"

	selinux_close

	exit 1
}

simply_die()
{
	report_problem "$@"
	exit 1
}

# This function needs for some print warning when some services cannot restart
# if called with arg, then we add arg to variable SR_WARNING and print warning
# message for arg, else print warning message for SR_WARNING
service_restart_warning()
{
	if [ "X$1" != "X" ]; then
		if [ "X${SR_WARNING}" = "X" ]; then
			SR_WARNING=$1
		else
			for i in ${SR_WARNING}; do
				if [ "$i" = "$1" ]; then
					break
				fi
			done
			if [ "$1" != "$i" ]; then
				SR_WARNING=$SR_WARNING" "$1
			fi
		fi
		p_echo "WARNING!"
		p_echo "Some problems occured when $1 restart, check it after installation/upgrade complete"
	        p_echo
	 	p_echo "Continue..."
		p_echo
	else
		if [ "X${SR_WARNING}" != "X" ]; then
			p_echo "WARNING!"
		        p_echo "Some problems occured when following services restart:"
			for i in ${SR_WARNING}; do
				p_echo $i
			done
			p_echo "Don't forget to check it after installation/upgrade complete"
        		p_echo
	       	        p_echo "More information can be found in log file: ${product_log}"
			p_echo
		fi
	fi
}

warn()
{
	local inten
	inten="$1"
	p_echo
	p_echo "WARNING!"
	pnnl_echo "Some problems are found during $inten"
	p_echo "(see log file: ${product_log})"
	p_echo
	p_echo "Continue..."
	p_echo

	product_log_tail | send_error_report_with_input "Warning: $inten"

	[ -n "$PLESK_INSTALLER_DEBUG" -o -n "$PLESK_INSTALLER_VERBOSE" ] || 	product_log_tail
}

# Use this function to report failed actions.
# Typical report should contain
# - reason or problem description (example: file copying failed)
# - how to resolve or investigate problem (example: check file permissions, free disk space)
# - how to re-run action (example: perform specific command, restart bootstrapper script, run installation again)
report_problem()
{
	[ -n "$product_problems_log" ] || product_problems_log="/dev/stderr"

	p_echo
	if [ "0$problems_occured" -eq 0 ]; then
		echo "***** $process problem report *****" >> "$product_problems_log" 2>&1
	fi
	for problem_message in "$@"; do
		p_echo "$problem_message"
		echo "$problem_message" >> "$product_problems_log" 2>&1
	done
	p_echo

	product_log_tail | send_error_report_with_input "Problem: $@"

	[ -n "$PLESK_INSTALLER_DEBUG" -o -n "$PLESK_INSTALLER_VERBOSE" ] || 		product_log_tail

	problems_occured=1
}

comm_err()
{
  p_echo "       Unfortunately, this situation has not yet been"
  p_echo "       resolved."
  p_echo "       Please, visit $support_contact for assistance."
  p_echo
  p_echo "Exiting..."
  p_echo

  exit 1
}

gid_err()
{
  gid="$1"

  p_echo
  p_echo "ERROR: It seems there is a group with GID=$gid"
  p_echo "       in this system, but $PRODUCT_NAME needs the same GID"
  p_echo "       for its operation."

  send_error_report "Error: group with GID=$gid already exists."
  comm_err
}

group_err()
{
  group="$1"
  gid="$2"

  p_echo
  p_echo "ERROR: It seems that there is group $group"
  p_echo "       in your system. $PRODUCT_NAME uses the same group"
  p_echo "       name but with another group ID ($gid)."

  send_error_report "Error: group $group exists, but it doesn't have GID=$gid."
  comm_err
}

uid_err()
{
  uid="$1"

  p_echo
  p_echo "ERROR: It seems there is a user with UID=$uid"
  p_echo "       in this system, but $PRODUCT_NAME needs the same UID"
  p_echo "       for its operation."

  send_error_report "Error: user with UID=$uid already exists."
  comm_err
}

logname_err()
{
  user="$1"
  uid="$2"

  p_echo
  p_echo "ERROR: It seems that there is a $user in your system"
  p_echo "       with a UID other than that used by $PRODUCT_NAME ($uid)"
  p_echo "       or the $user is in the primary group, but this"
  p_echo "       group is not the one that $PRODUCT_NAME uses."

  send_error_report "Error: user $user exists, but it doesn't have UID=$uid or it belongs to wrong primary group."
  comm_err
}

echo_try()
{
	msg="$*"
	pnnl_echo " Trying to $msg... "
}

suc()
{
	p_echo "done"
}

send_error_report()
{
	echo -n | send_error_report_with_input "$@"
}

# do not call it w/o input! Use send_error_report in these cases.
send_error_report_with_input()
{
	get_product_versions
	{
		echo "$@"
		echo ""
		if [ -n "$error_report_context" ]; then
			echo "Context: $error_report_context"
			echo ""
		fi
		if [ -n "$RP_LOADED_PATCHES" ]; then
			echo "Loaded runtime patches: $RP_LOADED_PATCHES"
			echo ""
		fi
		cat -
	} | $PRODUCT_ROOT_D/admin/bin/send-error-report --version "$product_this_version" install >/dev/null 2>&1
}

set_error_report_context()
{
	error_report_context="$*"
}

pre_accumulate_error()
{
	if [ -n "$ACCUMULATE_ERROR_RETURN_CODE_VAR" -o -n "$ACCUMULATE_ERROR_FAILED_ACTIONS_VAR" ]; then
		echo "Unable to init accumulate_error-stuff twice. calling post_accumulate_error() might be required." >&2
		exit 1
	fi
	ACCUMULATE_ERROR_RETURN_CODE_VAR="$1"
	ACCUMULATE_ERROR_FAILED_ACTIONS_VAR="$2"
}

accumulate_error()
{
	if [ -z "$ACCUMULATE_ERROR_RETURN_CODE_VAR" -o -z "$ACCUMULATE_ERROR_FAILED_ACTIONS_VAR" ]; then
		echo "accumulate_error-stuff is not set up. call pre_accumulate_error() first." >&2
		exit 1
	fi
	local action="$1"
	local actions
	eval ${ACCUMULATE_ERROR_RETURN_CODE_VAR}=1
	eval actions='$'${ACCUMULATE_ERROR_FAILED_ACTIONS_VAR}
	eval "$ACCUMULATE_ERROR_FAILED_ACTIONS_VAR=\"\${actions:+$actions, }$action\""
}
get_ip_address_list()
{
	if test -x /sbin/ip ; then
		/sbin/ip addr list | grep -v 'inet6' | grep -v '127\.[0-9]*\.[0-9]*\.[0-9]*' | sed -n -e 's/^ *inet \([0-9\.]*\).*$/\1/ p'
	else
		$ifconfig | grep "$inet_str" | grep -v "inet6" | sed -e "s/$inet_str.\([0-9.]*\).*/\1/g" | awk '{print $1}' - | grep -v '^127\.[0-9]*\.[0-9]*\.[0-9]*$'
	fi
}

get_domain()
{
	host="`${get_hostname}`"
	local try_dot=`echo "$host" | sed -e 's/^[^.]*\..*$/\./'`

	case "$try_dot" in
			.)	domain="`echo "$host" | sed -e 's/^[^.]*\.//'`"
				host="`echo $host | sed -e 's/\..*//'`"
				;;
			*)	domain="`$get_domainname`"
				;;
	 esac
	 domain=`echo "$domain" | sed 's/\.$//'`

	 if [ "X$host" = "X" -o "X${domain}" = "X" -o "$domain" = "(none)" ]; then
		p_echo "Warning!"
		p_echo "Unable to detect the fully qualified domain name of the host."
		p_echo "Please make sure that FQDN of the host is correctly specified in /etc/hosts and goes right after the host's IP address"
		p_echo "(depending on your OS, you might need to set the host name in /etc/HOSTNAME or /etc/hostname)."
		p_echo ""
		p_echo "Setting it to localhost.localdomain"
		host="localhost"
		domain="localdomain"
	 fi

 	 fullhost="${host}.${domain}"
	 return 0;
}

get_config_parameters()
{
	#list of addresses
	local ipaddress_list="`get_ip_address_list`"

	get_domain
	detect_vz

	if [ "X$PLESK_DISABLE_HOSTNAME_CHECKING" = "X" -a "X$PLESK_VZ" != "X1" ]; then
		check_hostname "$fullhost" || die "check the hostname" 
	else
		ipaddress="`echo "$ipaddress_list" | head -1`"
	fi

	start_after_install=1

	p_echo
}
reexec_with_clean_env()
{
	# Usage: call this function as 'reexec_with_clean_env "$@"' at the start of a script.
	#        Don't use with scripts that require sensitive environment variables.
	#        Don't put the call under any input/output redirection.
	# Purpose: make sure the script is executed with a sane environment.

	export LANG=C LC_MESSAGES=C LC_ALL=C
	export PATH=/sbin:/bin:/usr/sbin:/usr/bin:/usr/local/sbin:/usr/local/bin
	umask 022

	[ -z "$PLESK_INSTALLER_ENV_CLEANED" ] || { unset PLESK_INSTALLER_ENV_CLEANED; return 0; }
	[ -n "$BASH" ] || exec /bin/bash "$0" "$@"

	# N.B.: the following code requires Bash. On Dash it would cause syntax error upon parse w/o eval.
	eval '
	local extra_vars=()                     # list of variables to preserve
	for var in "${!PLESK_@}"; do            # enumerate all PLESK_* variables
		extra_vars+=("$var=${!var}")
	done
	extra_vars+=("PLESK_INSTALLER_ENV_CLEANED=1")

	# Exec self with clean env except for extra_vars, shell opts, and arguments.
	exec /usr/bin/env -i "${extra_vars[@]}" /bin/bash ${-:+-$-} "$0" "$@" || {
		echo "Failed to reexec self ($0) with clean environment" >&2
		exit 91		# Just some relatively unique error code
	}
	'
}

# vim:ft=sh

mk_backup()
{
	local target dup opts
	target="$1"
	dup="$2"
	opts="$3"

	if [ -L "$target" ]; then
		rm "$target"
	elif [ -$opts "$target" ]; then
		if [ ! -$opts "$target.$product_suffo" ]; then
			case "$dup" in
				mv)
					mv -f $target $target.$product_suffo || die "mv -f $target $target.$product_suffo"
					;;
				cp)
					cp -fp $target $target.$product_suffo || die "cp -fp $target $target.$product_suffo"
					;;
				*)
					p_echo " mk_backup: wrong option -- must be 'cp' or 'mv'"
					die "mk_backup"
					;;
			esac
		else
			case "$dup" in
				mv)
					mv -f $target $target.$product_suff || die "mv -f $target $target.$product_suff"
					;;
				cp)
					cp -fp $target $target.$product_suff || die "cp -fp $target $target.$product_suff"
					;;
				*)
					p_echo " mk_backup: wrong option -- must be 'cp' or 'mv'"
					die "mk_backup"
					;;
			esac
		fi
	else
		case "$opts" in
			f|d)
				;;
			*)
				p_echo " mk_backup: wrong option -- must be 'f' or 'd'"
				die "mk_backup"
				;;
		esac
	fi
}

# accumulates chown and chmod
set_ac()
{
	local u_owner g_owner perms node
	u_owner="$1"
	g_owner="$2"
	perms="$3"
	node="$4"

	# A very small optimization - replacing of two execs by one,
	#    it works only if the following conditions are observed:
	#       - u_owner is username (not UID);
	#       - g_owner is group (not GID);
	#       - perms is in octal mode.
	# If some conditions aren't observed,
	#    optimization doesn't work,
	#    but it doesn't break function
	[ "$(stat -c '%U:%G 0%a' $node)" != "$u_owner:$g_owner $perms" ] || return 0
	chown $u_owner:$g_owner $node || die "chown $u_owner:$g_owner $node"
	chmod $perms $node || die "chmod $perms $node"
}

start_product_now()
{
	define_start_script
	p_echo
	if [ "X$start_after_install" = "X1" ]; then
		local inten="start ${PRODUCT_NAME}"
		echo_try $inten
		p_echo

# [dpkg etc issue, bug #138010]
		[ -e "${START_SH}.dpkg-dist" ] && mv -f "${START_SH}.dpkg-dist" "${START_SH}"

		$START_SH stop  >> $product_log 2>&1
		/etc/init.d/sw-cp-server stop
		$START_SH start
	fi
}

start_product_now_final()
{
	unset PLESK_KEYUPDATE_DISABLED
	start_product_now
}

set_default_locale_params()
{
	local default_locale_file="$PRODUCT_ROOT_D/admin/htdocs/locales/default"

	if [ -f $default_locale_file ]; then
		test -r "$default_locale_file" || die "set up default locale: cannot read '$default_locale_file'"
		default_locale="`cat $default_locale_file | sed 's/^\([^:]*\):.*/\1/'`"
		default_locale_name="`cat $default_locale_file | sed 's/^[^:]*:\([^:]*\):.*/\1/'`"
		default_locale_country="`cat $default_locale_file | sed 's/^[^:]*:[^:]*:[^:]*:\([^:]*\);.*/\1/'`"
	else
		default_locale="en-US"
		default_locale_name="ENGLISH"
		default_locale_country="UNITED STATES"
	fi
}

set_services_info()
{
	detect_vz
	if [ "1" = "${PLESK_VZ}" ]; then
		p_echo 'Setting the services information on VZ mode.'
		cp -f ${PRODUCT_ETC_D}/service/vzmode/*.xml ${PRODUCT_ETC_D}/service
	fi
}

detect_vz()
{
	[ -z "$PLESK_VZ_RESULT" ] || return $PLESK_VZ_RESULT

	PLESK_VZ_RESULT=1
	PLESK_VZ=0
	PLESK_VE_HW_NODE=0
	PLESK_VZ_TYPE=

	local issue_file="/etc/issue"
	local vzcheck_file="/proc/self/status"
	[ -f "$vzcheck_file" ] || return 1

	local env_id=`sed -ne 's|^envID\:[[:space:]]*\([[:digit:]]\+\)$|\1|p' "$vzcheck_file"`
	[ -n "$env_id" ] || return 1
	if [ "$env_id" = "0" ]; then
		# Either VZ/OpenVZ HW node or unjailed CloudLinux
		PLESK_VE_HW_NODE=1
		return 1
	fi

	if grep -q "CloudLinux" "$issue_file" >/dev/null 2>&1 ; then
		return 1
	fi

	if [ -f "/proc/vz/veredir" ]; then
		PLESK_VZ_TYPE="vz"
	elif [ -d "/proc/vz" ]; then
		PLESK_VZ_TYPE="openvz"
	fi

	PLESK_VZ=1
	PLESK_VZ_RESULT=0
	return 0
}

# detects lxc and docker containers
detect_lxc()
{
	[ -z "$PLESK_LXC_RESULT" ] || return $PLESK_LXC_RESULT
	PLESK_LXC_RESULT=1
	PLESK_LXC=0
	if  { [ -f /proc/1/cgroup ] && grep -q 'docker\|lxc' /proc/1/cgroup; } || 		{ [ -f /proc/1/environ ] && cat /proc/1/environ | tr \\0 \\n | grep -q "container=lxc"; };
	then
		PLESK_LXC_RESULT=0
		PLESK_LXC=1
	fi
	return "$PLESK_LXC_RESULT"
}

call_optional_function()
{
	export LANG=C LC_MESSAGES=C LC_ALL=C
	local type_output="`type \"$1\" 2>/dev/null | head -n 1`"
	case "$type_output" in
		*function)
			"$@"
			;;
		*)
			return 0
			;;
	esac
}

add_string()
{
	local pattern string config
	pattern="$1"
	string="$2"
	config="$3"

	if ! grep -q "^$pattern\$" "$config"; then
		echo "$string" >>"$config"
	fi
}

sh_get_args()
{
	echo 'while true; do case "$1" in '"$1"'*) break;; esac; shift; done'
}

get_phpini_var()
{
	[ -z "$1" ] && return
	cat $php_ini | perl -n -e 'next if(/^;/);  print $5 if(/^(\s*)'$1'(\s*)=(\s*)("*)(.*?)("|\s|;)/);'
}

# see also: common/php/plib/PhpSettings/Config.php
phpini_value_is_boolean()
{
    value=`echo "$1" | tr "[:upper:]" "[:lower:]"`
    for i in '1' 'on' 'true' 'yes' '0' 'off' 'false' 'no' 'none'; do
        [ "$value" = "$i" ] && return 0
    done
    return 1
}

phpini_value_is_integer()
{
    expr match "$1" '[[:digit:]]\+[KMGT]\?$' 2>&1 > /dev/null
}

phpini_quote_value()
{
    if   phpini_value_is_boolean "$1" ; then
        echo $1 |  tr "[:upper:]" "[:lower:]"
    elif phpini_value_is_integer "$1" ||          expr match "$1" '\".*\"$' 2>&1 >/dev/null ||          test -z "$1" ; then
        echo $1
    else
        echo "\"$1\""
    fi
}

set_phpini_var()
{
	local variable=$1
	local value=`phpini_quote_value $2`

	if [ -z "$variable" ]; then
		warn "set_phpini_var() error: variable name is empty."
		return
	fi

	last_entry_line=`grep -n "^[^a-zA-Z]*$variable" $php_ini | tail -n 1 | awk -F ':' '{print $1}'`;
	if [ -z "$last_entry_line" -o 0$last_entry_line -eq 0 ]; then
		echo "$1 = $value" >> $php_ini
		result=$?
	else
		tmp_file=`mktemp /tmp/php-iniXXXXXX`
		$cp_preserve $php_ini $tmp_file
		cat $php_ini | awk -v variable="$variable" -v value="$value" '
		{
			line+=1;
			if ($0 ~ /^[ \t]*'$variable'[ \t]*=/) {
				printf("%s = %s\n",variable,value);
				done=1;
			} else {
				print $0;
			}
			if (!done && (line>'$last_entry_line')) {
				printf("%s = %s\n",variable,value);
				done=1;
			}
		}' > $tmp_file && mv -f $tmp_file $php_ini >> $product_log 2>&1
		result=$?
	fi

	if [ 0$result -eq 0 ]; then
		echo "$variable=$value has been added to $php_ini" >> $product_log 2>&1
	else
		warn "set_phpini_var() error: can't add $variable=$value to $php_ini";
	fi
}

### the function similar to awk -F'$fs' 'print $N'
get_narg_fs()
{
	local IFS="$2"
	get_narg $3 $1
}

get_narg()
{
	shift $1 2>/dev/null || return 0
	echo $1
}

get_random_number()
{
	perl -e '
	 my $res = int(rand($ARGV[0]));
	 if (exists $ARGV[1]) {
	  $res = $res + $ARGV[1];
	 }
	 print $res;
	' "$@"
}

get_random_string()
{
	local str_length="$1"
	local str_symbols="$2"
	if [ -x "$PRODUCT_ROOT_D/admin/sbin/random_str" -a -z "$str_symbols" ]; then
		"$PRODUCT_ROOT_D/admin/sbin/random_str" "$str_length"
	else
		# random_str utility may be unavailable in pre phase
		if [ -z "$str_length" ]; then
			str_length="14"
		fi
		if [ -z "$str_symbols" ]; then
			str_symbols="A-Za-z0-9_"
		fi

		< /dev/urandom tr -dc "$str_symbols" 2>/dev/null | head -c "$str_length" 2>/dev/null
	fi
}

fix_etc_hosts()
{
	if ! grep -q -E '^127\..*[[:space:]]localhost([[:space:]]|$)' /etc/hosts; then
		awk '
			BEGIN{ set = 0; }
			/^127\.0\.0\.1[[:space:]]/{ print $0, " localhost"; set = 1; next; }
			{ print }
			END{ if (set == 0) { print "127.0.0.1 localhost"; } }
		' /etc/hosts >/etc/hosts.tmp 		&& mv /etc/hosts.tmp /etc/hosts 		|| report_problem "insert localhost entry in /etc/hosts"
	fi
}

sequence()
{
	if type seq >/dev/null 2>&1; then
		seq $*
	elif type jot >/dev/null 2>&1; then
		jot $*
	else
		die "Unable to find seq or jot command"
	fi
}

get_ini_conf_var()
{
	local conf="$1"
	local section="$2"
	local param="$3"

	[ -n "$conf" -a -n "$param" ] || die "get_ini_conf_var(): required parameters missing"

	local section_empty=0
	[ -n "$section" ] || section_empty=1

	perl -n -e 'BEGIN { $insect='$section_empty' }
				next if (/^\s*;/);
				$insect=0 if (/^\s*\[.*\]/);
				$insect=1 if (/^\s*\['$section'\]/);
				$val = $2, $val =~ s/\s+$//, print $val . "\n" 
					if ($insect && /^\s*('$param')\s*=\s*([^;\n]*)(;.*)?$/);' $conf | head -n 1
}

sw_engine_pleskrun()
{
	if [ -x $PRODUCT_ROOT_D/bin/sw-engine-pleskrun ]; then
		$PRODUCT_ROOT_D/bin/sw-engine-pleskrun "$@"
	elif [ -x "/usr/bin/sw-engine" ]; then
		local args
		if [ -e "$PRODUCT_ROOT_D/admin/conf/php.ini" ]; then
			args="-c $PRODUCT_ROOT_D/admin/conf/php.ini"
		else
			args="-d psasem.semfile=\"$PRODUCT_ROOT_D/var/psasem.sem\""
		fi
		/usr/bin/sw-engine $args "$@"
	else
		echo "No Plesk PHP interpreter found"
		return 1
	fi
}

install_default_license_key()
{
	pnnl_echo "install default license key ..."

# key from 8.x?
	if 	db_test_table "key_history" && 		db_test "SELECT COUNT(*) FROM key_history WHERE filename = '/etc/psa/psa.key' LIMIT 1" '$1!="0"'; then
			p_echo "not needed, license from 8.x found"
			return;
	fi

# check OF-key
	if sw_engine_pleskrun -r 'exit(of_get_key_by_product("plesk-unix") === false ? 1 :0);'; then
			p_echo "not needed, OF license found"
			return
	fi

# check Unified key
	if sw_engine_pleskrun -r 'exit(of_get_key_by_product("plesk-unified") === false ? 1 :0);'; then
			p_echo "not needed, unified license found"
			return
	fi

	local dest_dir="/etc/sw/keys/keys"
	local key=plesk-default-key.xml

    [ ! -s "$dest_dir/$key" ] || return 0

    cp -f "${PRODUCT_ROOT_D}/tmp/$key" "$dest_dir/"
    chown root:swkey-data "$dest_dir/$key"
    chmod 0660 "$dest_dir/$key"
    suc
}

register_sw_engine_service()
{
	register_service sw-engine
}

implode()
{
	local IFS="$1"
	shift
	echo "$*"
}

configure_ssl_ciphers_protocols()
{
	[ -x "$PRODUCT_ROOT_D/bin/http2_pref" -a -x "$PRODUCT_ROOT_D/admin/sbin/sslmng" ] || return 0

	local inten="configure SSL ciphers and protocols for: $*"
	echo_try "$inten"

	if [ -n "$*" ]; then
		if [ "$*" = "nginx" ] && is_product_installation; then
			$PRODUCT_ROOT_D/bin/http2_pref --configure >>$product_log 2>&1 && suc && return 0
		fi
		$PRODUCT_ROOT_D/admin/sbin/sslmng --services="$*" >>$product_log 2>&1 && suc && return 0

		warn "$inten"
		return 1
	fi

	if is_product_installation ; then
		$PRODUCT_ROOT_D/bin/http2_pref --configure >>$product_log 2>&1 || true
	fi
	$PRODUCT_ROOT_D/admin/sbin/sslmng >>$product_log 2>&1 && suc && return 0

	warn "$inten"
	return 1
}
label_file_once()
{
	local _fname _mess _messLines _fileHead _tmpFile

	test $# -eq 2 || die 'label_file_once() not enough arguments'
	
	_fname="${1}"
	_mess="${2}"

	inten="add header to file ${_fname}"
	echo_try $inten

	if [ ! -f "${_fname}" ]; then
		p_echo "can not add header to file ${_fname} : file does not exist"
		return
	fi
	_messLines=`printf "${_mess}" | wc -l | awk '{print $NF}'`

	_tmpFile=`basename "${_fname}"`
	_tmpFile=`mktemp "/tmp/${_tmpFile}.XXXXXX"` || die "can not create temporary file"
	head -n ${_messLines} "${_fname}" > "${_tmpFile}"

	if printf "${_mess}" | diff -q - "${_tmpFile}" 1>/dev/null; then
		p_echo "file ${_fname} already contains required header"
		rm -f "${_tmpFile}"
		return
	fi
	rm -f "${_tmpFile}"

	# preserve permissions, not optimal, need to rework
	cp -p "${_fname}" "${_fname}.tmp"

	printf "${_mess}" | cat - "${_fname}" > "${_fname}.tmp"
	cp -f "${_fname}.tmp" "${_fname}"
	rm -f "${_fname}.tmp"
	p_echo "completed successfully"
}

lookup_command()
{
	test $# -eq 1 || die 'lookup_command() not enough arguments'
	local _path=/bin:/usr/bin:/usr/local/bin:/sbin:/usr/sbin:/usr/local/sbin:${PATH}
	
	local utility_path=`env PATH=${_path} which "$1" 2>/dev/null`
	if [ -z "$utility_path" ]; then
		p_echo "Unable to find utility '$1'in ${_path}"
		return
	fi
	echo $utility_path
}
set_xinetd_params()
{
    xinetd_conf="/etc/xinetd.conf"
    xinetd_service="xinetd"
    xinetd_dir="/etc/xinetd.d"
}


# vim: ft=sh
lockfile()
{
	local tempfile="$1.$$"
	local lockfile="$1.lock"

	echo $$ > $tempfile 2>/dev/null || {
		p_echo "Unable to create lock file in `dirname $tempfile`"
		return 1
	}

	ln $tempfile $lockfile >/dev/null 2>&1 && {
		rm -f $tempfile
		return 0
	}

	kill -0 `cat $lockfile` >/dev/null 2>&1 && {
		rm -f $tempfile
		return 1
	}

	p_echo "Removing stale lock file $lockfile"
	rm -f $lockfile
	ln $tempfile $lockfile >/dev/null 2>&1 && {
		rm -f $tempfile
		return 0
	}

	rm -f $tempfile
	return 1
}

unlockfile()
{
	rm -f "$1.lock"
}
#-*- vim:syntax=sh

product_log_name_ex()
{
	local aux_descr="$1"
	local action="${CUSTOM_LOG_ACTION_NAME-installation}"

	if [ -n "$aux_descr" ]; then
		aux_descr="_${aux_descr}"
	fi

	if [ -n "$CUSTOM_LOG_NAME" ]; then
		echo "${CUSTOM_LOG_NAME}${action:+_$action}${aux_descr}.log"
	else
		echo "plesk_17.5.3${action:+_$action}${aux_descr}.log"
	fi
}

product_log_name()
{
	product_log_name_ex
}

product_problems_log_name()
{
	product_log_name_ex "problems"
}

problems_log_tail()
{
	[ -f "$product_problems_log" ] || return 0
	tac "$product_problems_log" | awk '/^START/ { exit } { print }' | tac
}

product_log_tail()
{
	[ -f "$product_log" ] || return 0
	{
		tac "$product_log" | awk '/^START/ { exit } { print }' | tac
	} 2>/dev/null
}

cleanup_problems_log()
{
	[ -f "$product_problems_log" ] || return 0
	touch "$product_problems_log.tmp"
	chmod 0600 "$product_problems_log.tmp"
	awk 'BEGIN 						{ st = "" } 
		 /^START/ 					{ st=$0; next } 
		 /^STOP/ && (st ~ /^START/) { st=""; next } 
		 (st != "") 				{ print st; st="" } 
		 							{ print }
		' "$product_problems_log" > "$product_problems_log.tmp" && 	mv -f "$product_problems_log.tmp" "$product_problems_log" || 	rm -f "$product_problems_log.tmp"
	
	if [ ! -s "$product_problems_log" ]; then 
		rm -f "$product_problems_log"
	fi
}

set_log_action_name()
{
	CUSTOM_LOG_ACTION_NAME="$1"
}

mktemp_log()
{
	local logname="$1"
	local dir="$2"

	if [ "${logname:0:1}" != "/" ]; then
		logname="$dir/$logname"
	fi
	dir="`dirname $logname`"
	if [ ! -d "$dir" ]; then
		mkdir -p "$dir" || { echo "Unable to create log directory : $dir"; exit 1; }
		if [ "$EUID" -eq "0" ]; then
			set_ac root root 0700 "$dir"
		fi
	fi

	if [ "${logname%XXX}" != "$logname" ]; then
		mktemp "$logname"
	else
		echo "$logname"
	fi
}

log_is_in_dev()
{
	test "${1:0:5}" = "/dev/"
}

start_writing_logfile()
{
	local logfile="$1"
	local title="$2"
	! log_is_in_dev "$logfile" || return 0
	echo "START $title" >> "$logfile" || { echo "Cannot write installation log $logfile" >&2; exit 1; }
	[ "$EUID" -ne "0" ] || set_ac root root 0600 "$logfile"
}

create_product_log_symlink()
{
	local logfile="$1"
	local prevdir="$2"

	local prevlog="$prevdir/`basename $logfile`"
	[ -e "$prevlog" ] || ln -sf "$logfile" "$prevlog"
}

log_start()
{
	true product_log_name product_problems_log_name mktemp_log

	local title="$1"
	local custom_log="$2"
	local custom_problems_log="$3"

	local product_log_dir="/var/log/plesk/install"

	product_log="$product_log_dir/`product_log_name`"
	product_problems_log="$product_log_dir/`product_problems_log_name`"
	problems_occured=0

	# init product log
	[ ! -n "$custom_log" ] || product_log="$custom_log"
	product_log=`mktemp_log "$product_log" "$product_log_dir"`

	# init problems log
	if [ -n "$custom_problems_log" ]; then
		product_problems_log=`mktemp_log "$custom_problems_log" "$product_log_dir"`
	elif [ -n "$custom_log" ]; then
		product_problems_log="$product_log"
	else
		product_problems_log=`mktemp_log "$product_problems_log" "$product_log_dir"`
	fi

	# write starting message into logs
	start_writing_logfile "$product_log" "$title"
	if [ "$product_log" != "$product_problems_log" ]; then
		start_writing_logfile "$product_problems_log" "$title"
	fi

	# create compat symlinks if logs are written to default localtions
	if [ -z "$custom_log" -a -z "$CUSTOM_LOG_NAME" ]; then
		create_product_log_symlink "$product_log" "/tmp"
		[ ! -z "$custom_problems_log" ] || create_product_log_symlink "$product_problems_log" "/tmp"
	fi

	is_function profiler_setup && profiler_setup "$title" || :
}

log_stop()
{
	local title="$1"
	local subject="$2"

	if [ "$product_log" = "$product_problems_log" ] || 			log_is_in_dev "$product_problems_log"; then
		[ -e "$product_log" ] && echo "STOP $title" >>"$product_log"
		is_function profiler_stop && profiler_stop || :
		return
	fi

	if [ -z "$subject" ]; then
		subject="[${title}]"
	fi

	# check if problems are non-empty, check for problems_occured
	local status
	local problem_lines="`problems_log_tail | wc -l`"
	if [ "$problem_lines" -eq 0 ]; then
		status="completed successfully"
	else
		if [ $problems_occured -ne 0 ]; then
			status="failed"
		else
			status="completed with warnings"
		fi
	fi

	if [ -e "$product_log" ]; then
		p_echo
		p_echo "**** $subject $status."
		p_echo
	fi

	if [ "$problem_lines" -ne 0 ]; then
		[ ! -e "$product_log" ] || problems_log_tail >>"$product_log" 2>&1
		problems_log_tail
	fi

	[ ! -e "$product_log" ] || echo "STOP $title" >>"$product_log"
	if [ $problems_occured -ne 0 ]; then
		echo "STOP $title: PROBLEMS FOUND" >>"$product_problems_log"
	else
		[ ! -s "$product_problems_log" ] || echo "STOP $title: OK" >>"$product_problems_log"
	fi

	if [ "X${PLESK_INSTALLER_KEEP_PROBLEMS_LOG}" = "X" ]; then
		cleanup_problems_log
	fi

	# remove symlink to problems log if the log was removed
	local linkpath="/tmp/`basename $product_problems_log`"
	if [ -L "$linkpath" -a ! -e "$linkpath" ]; then
		rm -f "$linkpath"
	fi

	is_function profiler_stop && profiler_stop || :
}

report_context_action()
{
	local action
	get_product_versions
	if [ "X$do_upgrade" = "X1" -o "$product_prev_version" != "$product_this_version" ]; then
		action="Doing $product_name upgrade from $product_prev_version to $product_this_version"
	else
		action="Doing $product_name $product_this_version installation"
	fi
	echo "$action"
}

set_maintenance_mode_flag()
{
	touch /var/lock/parallels-panel-maintenance-mode.flag
}

reset_upgrade_failure_flag()
{
	rm -f /var/lock/parallels-panel-maintenance-mode.flag
	rm -f /var/lock/parallels-panel-upgrade-failure.flag
}

reset_maintenance_mode_flag()
{
	rm -f /var/lock/parallels-panel-maintenance-mode.flag
}


generate_password_file()
{
	local passwd_file="$1"
	local title="$2"
	local user="$3"
	local group="$4"
	local mode="$5"

	if [ "$#" -ne 5 ]; then 
		die "Some arguments was not defined for this function"
	fi

	local passwd
	local inten="generate random password for $title"

	echo_try $inten

	if [ ! -s "$passwd_file" ]; then 
		passwd=`get_random_string` || die $inten
		echo "$passwd" > "$passwd_file" || die $inten
		set_ac $user $group $mode $passwd_file || die $inten
	else
		set_ac $user $group $mode $passwd_file || die $inten
	fi
}

get_pid()
{
	local i

	local ex_f="$1"
	local opt="$2"
	local owner="$3"

	local min_num="1"

	# Use pidof by default, bug 121868, except for FreeBSD - 140182
	if type pidof >/dev/null 2>&1 && [ "$os" != "BSD" ]; then
		for pid in `pidof -o $$ -o $PPID -o %PPID -x $ex_f`; do
			# Check for owner
			[ "$opt" = "true" -a "$owner" != "`ps -p $pid -o ruser=`" ] && continue
			min_num=$pid
			break
		done
		common_var=$min_num
		return $min_num
	fi

	case "$opt" in
		false)
			for i in `$ps_long | grep $ex_f | grep -v grep | grep -v httpsdctl | grep -v apachectl | awk '{print $2}' -`; do
				min_num=$i
				break
			done
			;;
		true)
			for i in `$ps_long | grep $ex_f | grep -v grep | grep -v httpsdctl | grep -v apachectl | grep "$owner" | awk '{print $2}' -`; do
				min_num=$i
				break
			done
			;;
		*)
			p_echo "get_pid: wrong parameter"
			die "get_pid $ex_f $opt $owner"
			;;
	esac

	common_var=$min_num
	return $min_num
}

update_python_modules()
{
	local opt_compile=
	local opt_optimize=
	local opt_remove=

	while [ "$#" -gt 0 ]; do
		case "$1" in
			--install)
				opt_compile="yes"
				opt_optimize="yes"
				shift
				;;
			--compile)
				opt_compile="yes"
				shift
				;;
			--optimize)
				opt_optimize="yes"
				shift
				;;
			--deinstall|--remove)
				opt_remove="yes"
				shift
				;;
			*)
				break
				;;
		esac
	done

	# Sanity checks
	[ -z "$opt_remove" -a -z "$opt_compile" -a -z "$opt_optimize" ] && simply_die "update_python_modules(): no arguments specified"
	[ -n "$opt_remove" -a -n "$opt_compile$opt_optimize" ] && simply_die "update_python_modules(): cannot both install and remove compiled modules"

	if [ -n "$opt_remove" ]; then
		echo_try "remove compiled Python modules"
	else
		echo_try "compile Python modules"
	fi

	for path in "$@"; do
		{
			[ -z "$opt_compile"  ] || $PYTHON_BIN -m compileall $path
			[ -z "$opt_optimize" ] || $PYTHON_BIN -O -m compileall $path
			[ -z "$opt_remove"   ] || {
				if [ -f "$path" ] && [ -n "`echo \"$path\" | sed -n -e '/.py$/p'`" ]; then
					rm -f "${path}c" "${path}o"
				else
					[ ! -d "$path" ] || path="$path/"
					find $path* -name '*.py[co]' -type f -delete
					rmdir --ignore-fail-on-non-empty $path
				fi
			}
		} >> $product_log 2>&1
	done

	suc
}

#
# Support for runtime patching of shell scripts (including utilities and package scripts).
#

# --- Service functions ---

# Load and apply a patch in a relatively safe way
rp_safe_load_patch()
{
	local patch_file="$1"
	echo_try "load shell patch '$patch_file'"
	/bin/sh -n "$RP_BASEDIR/$patch_file" && 
	{
		. "$RP_BASEDIR/$patch_file"
		RP_LOADED_PATCHES="$RP_LOADED_PATCHES $patch_file"
	} &&
	suc
}

# Apply patches specific to the current context (e.g., depending on utility basename or package name)
# This is currently not implemented. This may be overriden by "spark".
rp_patch_runtime_context_specific()
{
	:
}

# --- Main entry points ---

rp_patch_runtime()
{
	# List of loaded patch files
	RP_LOADED_PATCHES=

	local RP_BASEDIR="$PRODUCT_BOOTSTRAPPER_DIR/rp"
	[ -d "$RP_BASEDIR" ] || return 0

	if [ -r "$RP_BASEDIR/spark" ]; then
		rp_safe_load_patch "spark"
	fi

	call_optional_function rp_patch_runtime_context_specific "$@"
}

check_hostname()
{
    local hostname="$1"
    local output

	echo_try "resolve hostname '${hostname}' and validate its IP address"
	output="`perl -e '
		use strict;
		use Socket;
		use Errno;

		my $host = shift || "localhost";
		my $port = 715;

		my $proto = getprotobyname("tcp");
		if (!$proto) {
			print "System  error: $!";
			exit 1;
		}
		my $addr = inet_aton($host);
		$addr = inet_aton($1) if !$addr & $host =~ /^(.*)\.localdomain$/;
		if(!$addr) {
			print "Unable to resolve host name.\n";
			print "Please specify the host name in the /etc/hosts file or specify DNS configuration in the /etc/resolv.conf on your system.";
			exit 1;
		}

		if (!socket(SOCK, PF_INET, SOCK_STREAM, $proto)) {
			print "System error: $!";
			exit 1;
		}

		my $sin = sockaddr_in($port, $addr);

		if (!bind(SOCK, $sin) && $! != Errno::EADDRINUSE) {
			print "Can not bind a socket to address $host:$port - $!\n";
			print "Please, check your network configuration";
			exit 1;
		}

		printf inet_ntoa($addr);
		close(SOCK);
		' $hostname `"

		if [ $? != 0 ]; then
			p_echo
			p_echo "${output}"
			return 1
		fi

		ipaddress=${output}
		suc
		return 0
}

txtdb_set_params()
{
	db_file=""
	db_owner=""
	db_group=""
	db_perms="0644"
	db_delim=':'
	db_key_field="1"
	"$@"
	[ -n "$db_file" -a -n "$db_delim" ] || die "set txtdb parameters. Check arguments for 'txtdb_*' calls."
}

txtdb_fix_perms()
{
	[ -n "$db_owner" -a -n "$db_group" ] && chown "$db_owner":"$db_group" "$db_file"
	[ -n "$db_perms" ] && chmod "$db_perms" "$db_file"
	:
}

# Create new empty DB or replace an existing one
txtdb_truncate()
{
	local set_params_func="$1"
	local db_file db_owner db_group db_perms db_delim db_key_field
	txtdb_set_params "$set_params_func"

	rm -f "$db_file" && touch "$db_file" && txtdb_fix_perms
}

# Check whether DB file already exists and is writable
txtdb_exists()
{
	local set_params_func="$1"
	local db_file db_owner db_group db_perms db_delim db_key_field
	txtdb_set_params "$set_params_func"

	[ -w "$db_file" ]
}

txtdb_do_update()
{
	local set_params_func="$1"
	local awk_script="$2"
	local key_value="$3"
	local data_value="$4"
	local db_file db_owner db_group db_perms db_delim db_key_field
	txtdb_set_params "$set_params_func"

	awk -F "$db_delim" -v KEY="$key_value" -v DATA="$data_value" -v KEY_FIELD="$db_key_field" 		"$awk_script" "$db_file" > "$db_file.new" 2>> $product_log && 		mv -f "$db_file.new" "$db_file" && txtdb_fix_perms
}

txtdb_do_select()
{
	local set_params_func="$1"
	local awk_script="$2"
	local key_value="$3"
	local data_value="$4"
	local db_file db_owner db_group db_perms db_delim db_key_field
	txtdb_set_params "$set_params_func"

	awk -F "$db_delim" -v KEY="$key_value" -v DATA="$data_value" -v KEY_FIELD="$db_key_field" 		"$awk_script" "$db_file" 2>> $product_log
}

# Insert or replace existing record identified by key field
txtdb_replace()
{
	local db_file db_owner db_group db_perms db_delim db_key_field
	txtdb_set_params "$1"

	# Note that records order is not preserved
	local replace_with="$2"
	local key_to_replace=`get_narg_fs "$replace_with" "$db_delim" "$db_key_field"`
	txtdb_do_update "$1" 'END { print DATA } ( $KEY_FIELD != KEY )' "$key_to_replace" "$replace_with"
}

# Select entire record by key field
txtdb_select()
{
	local key_to_select="$2"
	txtdb_do_select "$1" '( $KEY_FIELD == KEY )' "$key_to_select"
}


#
# Unified per-component cumulative upgrade support.
#

# --- Upgrader utility functions ---

cu_get_continue_stage()
{
	local continue_from="$1"
	get_narg_fs "$continue_from" '-' 1
}

cu_get_continue_id()
{
	local continue_from="$1"
	echo "`  get_narg_fs "$continue_from" '-' 2  `-`  get_narg_fs "$continue_from" '-' 3  `"
}

cu_get_current_continue_from()
{
	echo "$CU_LAST_STAGE-$CU_LAST_ACTION_ID"
}

# Remove extraneous delimiters (usually spaces) between array elements
normalize_array()
{
	echo $@
}

# Check that first argument is in array (other arguments)
is_in_array()
{
	local needle="$1"
	shift # everything else is haystack
	for elem in $@; do
		[ "$elem" = "$needle" ] && return 0
	done
	return 1
}

# Remove all instances of first argument from array (other arguments). Echoes normalized array with space delimiters.
remove_from_array()
{
	local what="$1"
	shift # everything else is array
	local new_array=
	for elem in $@; do
		[ "$elem" = "$what" ] || new_array="$new_array $elem"
	done
	echo $new_array
}

array_length_helper()
{
	echo $#
}

# Get number of elements in a given array
array_length()
{
	array_length_helper $@
}

# --- Chained upgrade actions checks ---

# cu_do <upgrade action> [<arguments>]: Execute an action or echo information about it
cu_do()
{
	if [ -n "$CU_DO_GET_ACTION_NAME" ]; then
		echo "$1"
		return
	fi

	if [ -n "$CU_CONTINUE_FROM_ID" ]; then
		# Run only actions starting from CU_CONTINUE_FROM_ID
		# WARNING: this will not work properly if there is a subshell execution somewhere along the chain!
		[ "$CU_LAST_ACTION_ID" = "$CU_CONTINUE_FROM_ID" ] && CU_CONTINUE_FROM_ID= || return 0
	fi

	if "$@"; then
		cu_remove_failed_action "$CU_LAST_ACTION_ID"
		return 0
	else
		cu_add_failed_action "$CU_LAST_ACTION_ID"
		return 1
	fi
}

# Echo resulting action name (function name) or empty if skipped
cu_get_action_name()
{
	local CU_DO_GET_ACTION_NAME="yes"
	"$@"
}

# Execute action only if it wasn't applied before (by timestamp, e.g. use `date '+%Y%m%d%H%M%S'` or even `date '+%Y/%m/%d %H:%M:%S'`)
cu_check_timestamp()
{
	local timestamp="`echo \"$1\" | tr -dc '[:digit:]'`"
	shift

	true cu_get_action_name
	CU_LAST_ACTION_ID="$timestamp-`cu_get_action_name "$@"`"

	# Skip timestamp check if this action should be done again (during repair)
	if [ -z "$CU_FAILED_ACTIONS" ] || ! is_in_array "$CU_LAST_ACTION_ID" "$CU_FAILED_ACTIONS"; then
		[ "$timestamp" -gt "$CU_START_TIMESTAMP" ] || return 0
	fi

	if [ "$timestamp" -gt "$CU_LATEST_TIMESTAMP" ]; then
		CU_LATEST_TIMESTAMP="$timestamp"
	fi

	"$@"
}

# Execute action only if it wasn't applied before (by version)
cu_check_version()
{
	true cu_get_action_name
	CU_LAST_ACTION_ID="$CU_UPGRADE_VERSION_FROM-`cu_get_action_name "$@"`"

	# Skip version check if this action should be done again (during repair)
	if [ -z "$CU_FAILED_ACTIONS" ] || ! is_in_array "$CU_LAST_ACTION_ID" "$CU_FAILED_ACTIONS"; then
		[ "$CU_UPGRADE_VERSION_FROM" -ge "$CU_START_VERSION_FROM" ] || return 0
	fi

	"$@"
}

# Execute action only if it belongs to a current stage
cu_check_stage()
{
	local stage="$1"
	shift

	[ "$stage" = "$CU_STAGE" ] || return 0

	"$@"
}

# Check that all previous cu_check_timestamp actions contain timestamps less than a given one
cu_verify_latest_timestamp_lt()
{
	local timestamp="`echo \"$1\" | tr -dc '[:digit:]'`"
	[ "$CU_LATEST_TIMESTAMP" -lt "$timestamp" ] || simply_die "Upgrader actions for '$CU_STAGE' stage of '$CU_LAST_DB_UPGRADER_NAME' upgrader are not valid. Timestamp values MUST be less than $timestamp. Please fix upgrader definition!"
}

# --- Upgrader bookkeeping functions ---

# Start a block of upgrade actions that should be executed when upgrading from a given version
cu_start_upgrade_from_version()
{
	CU_UPGRADE_VERSION_FROM="$1"
}

# Remember failed upgrade action (to be able to rerun it afterwards)
cu_add_failed_action()
{
	local action_id="$1"
	[ -n "$action_id" ] && CU_FAILED_ACTIONS="$CU_FAILED_ACTIONS $action_id" && CU_HAS_NEW_FAILED_ACTIONS="yes"
}

cu_remove_failed_action()
{
	local action_id="$1"
	[ -n "$CU_FAILED_ACTIONS" -a -n "$action_id" ] && CU_FAILED_ACTIONS="` remove_from_array "$action_id" "$CU_FAILED_ACTIONS" `"
}

# Store configuration data using supplied callback. Duplicate to default location if necessary.
cu_do_store_config_data()
{
	local upgrader_name="$1"
	local store_config_data_func="$2"

	$store_config_data_func "$upgrader_name"
	# Duplicate into default storage, if custom one was used. This is mostly for humans.
	[ "$store_config_data_func" = "cu_txtdb_store_config_data" ] || cu_txtdb_store_config_data "$upgrader_name"
}

# Handler for cases when critical action fails and there is no rollback procedure defined
cu_critical_failure_handler()
{
	[ -n "$CU_LAST_DB_UPGRADER_NAME" -a -z "$CU_LAST_DB_CONTINUE_FROM" ] || {
		p_echo "Can't handle upgrader critical failure - upgrader state is broken. Probably a programming error."
		return
	}

	p_echo "Updating '$CU_LAST_DB_UPGRADER_NAME' configuration status after a critical failure during upgrade stage '$CU_LAST_STAGE' on '$CU_LAST_ACTION_ID' action"

	local cu_db_current_timestamp cu_db_continue_from cu_db_failed_actions_list cu_db_latest_timestamp
	# Update only continue_from and failed_actions_list fields.
	cu_db_current_timestamp="$CU_LAST_DB_CURRENT_TIMESTAMP"
	cu_db_continue_from="`cu_get_current_continue_from`"
	cu_db_failed_actions_list="$CU_FAILED_ACTIONS"
	cu_db_latest_timestamp="$CU_LAST_DB_LATEST_TIMESTAMP"

	cu_do_store_config_data "$CU_LAST_DB_UPGRADER_NAME" "$CU_LAST_STORE_CONFIG_DATA_FUNC"

	CU_LAST_DB_UPGRADER_NAME=
}

# Check that following upgrader run will actually have any actions executed
cu_check_upgrade_required()
{
	local upgrader_func="$1"
	local CU_DO_GET_ACTION_NAME="yes"
	local upgrade_actions="`$upgrader_func`"
	[ -n "$upgrade_actions" ]
}

# --- Upgrader set up functions ---

# Set up installed (product/package) version and installed configuration timestamp
cu_set_initial_versions()
{
	CU_START_VERSION_FROM="${1:-starting_component_version_not_specified}"
	CU_START_TIMESTAMP="${2:-0}"
}

# Reset upgrader global variables
cu_set_globals()
{
	local upgrader_name="$1"
	local stage="$2"
	local store_config_data_func="$3"

	CU_DO_GET_ACTION_NAME=
	CU_LATEST_TIMESTAMP=0
	CU_LAST_ACTION_ID=

	# Initial DB record values. Used by cu_critical_failure_handler()
	CU_LAST_DB_UPGRADER_NAME="$upgrader_name"
	CU_LAST_DB_CURRENT_TIMESTAMP="$cu_db_current_timestamp"
	CU_LAST_DB_CONTINUE_FROM="$cu_db_continue_from"
	CU_LAST_DB_FAILED_ACTIONS_LIST="$cu_db_failed_actions_list"
	CU_LAST_DB_LATEST_TIMESTAMP="$cu_db_latest_timestamp"
	CU_LAST_STAGE="$stage"
	CU_LAST_STORE_CONFIG_DATA_FUNC="$store_config_data_func"
}

# --- Default implementations to load/store configuration data ---

cu_txtdb_config_data_params()
{
	db_file="$PLESK_DB_DIR/components-configuration-state.db"
	db_delim='#'
}

# Path to default upgrader configuration db changed in Plesk 11.5.26. This function moves existing db.
cu_txtdb_try_upgrade_location()
{
	local old_db_file="$PRODUCT_ROOT_D/var/components-configuration-state.db"
	local db_file db_owner db_group db_perms db_delim db_key_field
	txtdb_set_params cu_txtdb_config_data_params

	if [ -s "$old_db_file" -a ! -e "$db_file" ]; then
		mv -f "$old_db_file" "$db_file"
	fi
}

cu_txtdb_ensure_config_data_exists()
{
	cu_txtdb_try_upgrade_location
	txtdb_exists cu_txtdb_config_data_params >/dev/null 2>&1 || txtdb_truncate cu_txtdb_config_data_params
}

cu_txtdb_load_config_data()
{
	local upgrader_name="$1"

	cu_txtdb_ensure_config_data_exists
	local record="`txtdb_select cu_txtdb_config_data_params "$upgrader_name"`"

	cu_db_current_timestamp="`  get_narg_fs "$record" '#' 2`"
	cu_db_continue_from="`      get_narg_fs "$record" '#' 3`"
	cu_db_failed_actions_list="`get_narg_fs "$record" '#' 4`"
	cu_db_latest_timestamp="`   get_narg_fs "$record" '#' 5`"
}

cu_txtdb_store_config_data()
{
	local upgrader_name="$1"
	local record="$upgrader_name#$cu_db_current_timestamp#$cu_db_continue_from#$cu_db_failed_actions_list#$cu_db_latest_timestamp"
	cu_txtdb_ensure_config_data_exists
	txtdb_replace cu_txtdb_config_data_params "$record"
}

# No-op config loader. For use with cu_run_upgrader --init for the first stage.
cu_dummy_load_config_data()
{
	:
}

# --- Main code for generic cumulative upgrader ---

cu_run_upgrader()
{
	local opt_upgrader_name="$1"
	local opt_upgrader_func="$2"
	shift 2

	local opt_prepare_func=	opt_rollback_func= opt_commit_func= opt_start_version= opt_stage= opt_do_repair= opt_do_update_version_in_db= opt_do_init=
	local opt_load_config_data_func=cu_txtdb_load_config_data opt_store_config_data_func=cu_txtdb_store_config_data

	local inten="perform upgrade for $opt_upgrader_name"

	while [ "$#" -gt 0 ]; do
		case "$1" in
			--prepare)
				opt_prepare_func="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--rollback)
				opt_rollback_func="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--commit)
				opt_commit_func="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--load-config-data)
				opt_load_config_data_func="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--store-config-data)
				opt_store_config_data_func="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--from-version)
				opt_start_version="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--stage)
				opt_stage="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--repair)
				opt_do_repair="yes"
				shift
				;;
			--update-version)
				opt_do_update_version_in_db="yes"
				shift
				;;
			--init)
				opt_do_init="yes"
				shift
				;;
			*)
				die "$inten: unknown key '$1'"
				;;
		esac
	done

	# Sanity checks
	[ -n "$opt_upgrader_name"          ] || die "$inten: empty component upgrader name argument"
	[ -n "$opt_upgrader_func"          ] && is_function "$opt_upgrader_func"          || die "$inten: invalid component upgrader function argument"
	[ -n "$opt_load_config_data_func"  ] && is_function "$opt_load_config_data_func"  || die "$inten: invalid value for --load-config-data argument"
	[ -n "$opt_store_config_data_func" ] && is_function "$opt_store_config_data_func" || die "$inten: invalid value for --store-config-data argument"
	[ -z "$opt_prepare_func"           ] || is_function "$opt_prepare_func"           || die "$inten: invalid value for --prepare argument"
	[ -z "$opt_rollback_func"          ] || is_function "$opt_rollback_func"          || die "$inten: invalid value for --rollback argument"
	[ -z "$opt_commit_func"            ] || is_function "$opt_commit_func"            || die "$inten: invalid value for --commit argument"
	[ -n "$opt_do_init" -a \( -n "$opt_do_repair" -o -n "$opt_prepare_func" -o -n "$opt_rollback_func" -o -n "$opt_commit_func" \) ] && die "$inten: excessive options prohibited for use with --init"

	# Upgrader setup
	local CU_STAGE="$opt_stage"
	local cu_db_current_timestamp cu_db_continue_from cu_db_failed_actions_list cu_db_latest_timestamp
	$opt_load_config_data_func "$opt_upgrader_name"
	cu_set_initial_versions "$opt_start_version" "$cu_db_current_timestamp"
	cu_set_globals "$opt_upgrader_name" "$opt_stage" "$opt_store_config_data_func"

	local CU_CONTINUE_FROM_ID= CU_FAILED_ACTIONS= CU_HAS_NEW_FAILED_ACTIONS=
	if [ -z "$opt_do_init" ]; then
		if [ -n "$cu_db_continue_from" ]; then
			if [ -z "$opt_do_repair" ]; then
				pp_echo "Unable to upgrade '$opt_upgrader_name' to the latest version, since its configuration is in broken state. Run bootstrapper repair to fix."
				return 2
			fi

			local cu_continue_from_stage="` cu_get_continue_stage "$cu_db_continue_from" `"
			CU_CONTINUE_FROM_ID="` cu_get_continue_id "$cu_db_continue_from" `"
			if [ "$cu_continue_from_stage" != "$CU_STAGE" ]; then
				p_echo "Skipping stage '$CU_STAGE' of upgrade for '$opt_upgrader_name' as previous upgrade attempt was not finished on another stage."
				return 3
			fi
		elif [ -n "$cu_db_failed_actions_list" ]; then
			if [ -z "$opt_do_repair" ]; then
				p_echo "Before start of upgrade for '$opt_upgrader_name' (stage '$CU_STAGE') following actions are registered as failed: $cu_db_failed_actions_list."
			else
				CU_FAILED_ACTIONS="` normalize_array "$cu_db_failed_actions_list" `"
			fi
		fi

		if ! cu_check_upgrade_required "$opt_upgrader_func" ; then
			p_echo "Upgrade or repair for '$opt_upgrader_name' (stage '$CU_STAGE') is not required"
			[ -n "$opt_prepare_func" ] && $opt_prepare_func "$opt_upgrader_name" "$CU_STAGE" noop-stage
			[ -n "$opt_commit_func" ] && $opt_commit_func "$opt_upgrader_name" "$CU_STAGE" noop-stage
			return 0
		fi

		# Preparation step
		[ -n "$opt_prepare_func" ] && $opt_prepare_func "$opt_upgrader_name" "$CU_STAGE"

		# Register rollback actions.
		# Note that rollback scope currently spans to the end of upgrade procedure (this function).
		if [ -n "$opt_rollback_func" ]; then
			transaction_add_rollback_action "$opt_rollback_func"
			# Rollback function contract: everything changed in this upgrader pass will be 
			# rolled back. Component version should not change. Therefore there is no need 
			# to update anything in DB.
		else
			transaction_add_rollback_action "cu_critical_failure_handler"
		fi

		# Run upgrade actions
		$opt_upgrader_func

		# Commit (post) step - if we are here upgrade has completed (but there may be failed actions)
		[ -n "$opt_commit_func" ] && $opt_commit_func "$opt_upgrader_name" "$CU_STAGE"
	else
		# This is just an initialization run to set proper upgrader configuration on component installation
		p_echo "Initializing '$opt_upgrader_name' upgrader configuration (stage '$CU_STAGE')"
		local CU_DO_GET_ACTION_NAME="yes"
		$opt_upgrader_func >/dev/null 2>&1
	fi

	# Update DB record if there are changes. Update current_timestamp only if explicitly allowed (e.g. this is the last stage).
	cu_db_continue_from=

	if [ -n "$opt_do_repair" -o -z "$CU_LAST_DB_FAILED_ACTIONS_LIST" ]; then
		cu_db_failed_actions_list="$CU_FAILED_ACTIONS"
	else
		# Append failed actions from this stage. Don't care for duplicates - they will be removed when a given action succeeds.
		cu_db_failed_actions_list="` normalize_array "$CU_LAST_DB_FAILED_ACTIONS_LIST" "$CU_FAILED_ACTIONS" `"
	fi

	if [ -z "$cu_db_latest_timestamp" -a "$CU_LATEST_TIMESTAMP" -gt 0 ] || [ "0$cu_db_latest_timestamp" -lt "$CU_LATEST_TIMESTAMP" ]; then
		cu_db_latest_timestamp="$CU_LATEST_TIMESTAMP"
	fi

	[ -n "$opt_do_update_version_in_db" -a "0$cu_db_latest_timestamp" -gt 0 ] && cu_db_current_timestamp="$cu_db_latest_timestamp"

	if [    "$cu_db_current_timestamp" != "$CU_LAST_DB_CURRENT_TIMESTAMP" 		 -o "$cu_db_latest_timestamp" != "$CU_LAST_DB_LATEST_TIMESTAMP" 		 -o "$cu_db_failed_actions_list" != "$CU_LAST_DB_FAILED_ACTIONS_LIST" 		 -o "$cu_db_continue_from" != "$CU_LAST_DB_CONTINUE_FROM" ]; then
		cu_do_store_config_data "$opt_upgrader_name" "$opt_store_config_data_func"
	fi

	# If this is just an initialization, we don't need to disarm rollback, etc.
	[ -z "$opt_do_init" ] || return 0

	# Upgrade is complete - disarm rollback actions
	if [ -n "$opt_rollback_func" ]; then
		transaction_remove_rollback_action "$opt_rollback_func"
	else
		transaction_remove_rollback_action "cu_critical_failure_handler"
	fi

	if [ -n "$cu_db_failed_actions_list" ]; then
		p_echo "After end of upgrade for '$opt_upgrader_name' (stage '$CU_STAGE') following actions are registered as failed: $cu_db_failed_actions_list."
	fi

	[ ! -n "$CU_HAS_NEW_FAILED_ACTIONS" ]
}

cu_upgrader_conf_init()
{
	local opt_upgrader_name="$1"
	local opt_upgrader_func="$2"
	shift 2

	local opt_stages= opt_load_config_data_func= opt_store_config_data_func=

	local inten="initialize upgrader configuration for $opt_upgrader_name"

	while [ "$#" -gt 0 ]; do
		case "$1" in
			--stages)
				opt_stages="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--load-config-data)
				opt_load_config_data_func="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			--store-config-data)
				opt_store_config_data_func="$2"
				[ "$#" -ge 2 ] && shift 2 || break
				;;
			*)
				die "$inten: unknown key '$1'"
				;;
		esac
	done

	# Sanity checks
	[ -n "$opt_upgrader_name"          ] || die "$inten: empty component upgrader name argument"
	[ -n "$opt_stages"                 ] || die "$inten: empty value for --stages argument, should be space separated list of stages"

	# Preparations
	local num_stages="`array_length $opt_stages`"
	local cur_stage_idx=1

	# Run all stages through cu_run_upgrader. First stage resets configuration, last one upgrades current timestamp.
	for stage in $opt_stages; do
		local store_option="${opt_store_config_data_func:+--store-config-data $opt_store_config_data_func}"

		local load_option="${opt_load_config_data_func:+--load-config-data $opt_load_config_data_func}"
		if [ "$cur_stage_idx" -eq 1 ]; then
			load_option="--load-config-data cu_dummy_load_config_data"
		fi

		local update_option=
		if [ "$cur_stage_idx" -eq "$num_stages" ]; then
			update_option="--update-version"
		fi

		cu_run_upgrader "$opt_upgrader_name" "$opt_upgrader_func" --init --from-version 0 --stage "$stage" $load_option $store_option $update_option

		cur_stage_idx=$(( cur_stage_idx + 1 ))
	done
}

get_userID()
{
# try to get UID
	common_var=`id -u "$1" 2>/dev/null` 

# if id returns 0 the all is ok
	test "$?" -eq "0" -a -n "$common_var"
}

get_groupID()
{
# try to get GID, id -g doesn't show groups without users
	common_var=`getent group $1 2>/dev/null | awk -F':' '{print $3}'`

# We have non-empty value if success
	test -n "$common_var"
}

group_check()
{
	local group grp_res
	group="$1"

	grp_res=`getent group "$group" 1>/dev/null 2>&1 ; echo $?`

	echo "$grp_res"
}

add_user_to_group()
{
	local user group grp_res existing newlist

	user="$1"
	group="$2"

	# check exists group
	p_echo " Checking for the group '$group'..."
	grp_res=`group_check "$group"`

	if [ "X$grp_res" != "X0" ]; then
			p_echo " Group '$group' not exists"
			p_echo " It is necessary to add group '$group'"
			err
	fi

	inten="add supplementary group '$group' for user '$user'"
	echo_try "$inten"

	existing=`id -Gn "$user"|sed 's|[[:space:]]\+|,|g'`
	if
		test "`id -gn "$user"`" = "$group" 		|| echo "$existing" | grep -q "\\<$group\\>"
	then
		p_echo " already there"
		return
	fi

	if test -z "$existing"; then
		newlist="$group"
	else
		newlist="$existing,$group"
	fi

	usermod -G "$newlist" "$user" 2>>"$product_log" && suc || die "$inten"
}

group_op()
{
	group="$1"
    gid="$2"
	id_force="$3"
	what_done=0
	local inten

    if [ "X$gid" = "Xsearch" ]; then
        gid=`get_user_group_id "$group" "gid" 30 400` 
    fi

	inten="add group '$group'"

	p_echo " Checking for the group '$group'..."
	grp_res=`group_check "$group"`

	case "$id_force" in
		true)
			case "$grp_res" in
				0)
				    case "$machine" in
				    *)
					etc_grp_res=`getent group 2>/dev/null | egrep "^$group:"`
					case "$etc_grp_res" in
					*:*:*:*)
						chk_res=`echo "$etc_grp_res" | grep ":$gid:"`
						case "$chk_res" in
							*:*:*:*)
								p_echo " Group '$group' already exists"
								;;
							*) group_err "$group" "$gid"
								;;
						esac
						;;
					*)
						p_echo " Group '$group' already exists"
						;;
					esac
					;;

				    esac
				    ;;
				*)
					gid_res=`getent group 2>/dev/null | egrep "^[^:]*:[*x]*:$gid:" | head -1`
					case "$gid_res" in
						*:*:*:*)
							gid_err "$gid"
							;;
						*)
							echo_try "$inten"
							groupadd -g "$gid" "$group" >> $product_log 2>&1 && suc ||
							die $inten
							what_done=1
							;;
					esac
					;;
			esac
			;;
		false)
			case "$grp_res" in
				0)
					p_echo " Group '$group' already exists"
					;;
				*)
					echo_try "$inten"
					groupadd "$group" >> $product_log 2>&1 && suc ||
					die $inten
					what_done=1
					;;
			esac
			;;
		*)
			p_echo "group_op: wrong last parameter -- must be 'true' or 'false'"
			p_echo
			p_echo "Exiting..."
			p_echo
			exit 1
			;;
	esac

	p_echo
	common_var=$what_done
	if [ "X$what_done" = "X1" ]; then
		groups_created="$groups_created $group";
	fi
	return $what_done
}

user_op()
{
	user="$1"
	uid="$2"
	group="$3"
	u_desc="$4"
	u_home="$5"
	u_shell="$6"
	id_force="$7"
	gid_force="$8"
	what_done=0

    if [ "X$uid" = "Xsearch" ]; then
        uid=`get_user_group_id "$user" "uid" 30 400` 
    fi

	p_echo " Checking for the user '$user'..."
	id_res=`id "$user" 1>/dev/null 2>&1 ; echo $?`

	case "$id_force" in
		true)
			case "$id_res" in
				0)
					chk_res=`id "$user" 2>&1 | egrep '(^uid=.*gid=.*)|(^id.*user)'`
					if [ "X${gid_force}" = "Xfalse" ]; then
						chk_uid=`echo "$chk_res" | egrep "uid=""$uid""\(""$user""\)"`
					else
						chk_uid=`echo "$chk_res" | egrep "uid=""$uid""\(""$user""\)" | egrep "gid=[0-9]*\(""$group""\)"`
					fi
					case "$chk_uid" in
						uid=*gid=*)
							p_echo " User '$user' already exists"
							;;
						*) logname_err "$user" "$uid"
							;;
					esac
					chsh -s "$u_shell" "$user"
					;;
				*)
					uid_res=`getent passwd 2>/dev/null | egrep "^[^:]*:[^:]*:""$uid"":"`
					case "$uid_res" in
						*:*:*:*)
							uid_err "$uid"
							;;
						*)
							pnnl_echo " Trying to add user '$user'... "
							useradd $useradd_options -u "$uid" -g "$group" -d "$u_home" -s "$u_shell" -c "$u_desc" "$user" >> $product_log 2>&1 &&
							suc || die "add user '$user'"
							what_done=1
							;;
					esac
					;;
			esac
			;;
		false)
			case "$id_res" in
				0)
					p_echo " User '$user' already exists"
					;;
				*)
					pnnl_echo " Trying to add user '$user'... "
					useradd $useradd_options -g "$group" -d "$u_home" -s "$u_shell" -c "$u_desc" "$user" >> $product_log 2>&1 &&
					suc || die "add user '$user'"
					what_done=1
					;;
			esac
			;;
		*)
			p_echo "user_op: wrong last parameter -- must be 'true' or 'false'"
			p_echo
			p_echo "Exiting..."
			p_echo
			exit 1
			;;
	esac

	p_echo
	common_var=$what_done
	if [ "X$what_done" = "X1" ]; then
		users_created="$users_created $user";
	fi
	return $what_done
}

get_user_group_id()
{
    local name="$1"
    local src_name="$2"
    local min_num="$3"
    local max_num="$4"
    local count="$min_num"

    case "$src_name" in
       uid)
            local src="passwd"
            local type="users"

            if get_userID "$name"; then
               echo "$common_var"
               return 0
            fi
       ;;
       gid)
            local src="group"
            local type="groups"
            if get_groupID "$name"; then
               echo "$common_var"
               return 0
            fi
       ;;
       *)
            p_echo "$src_name for $name was not set"
            inten="detect uid/gid. Source file was not detect."
            die "$inten"
       ;;
    esac

    if [ $min_num -le 0 -o $min_num -ge $max_num ]; then 
		p_echo "get_user_group_id(): wrong parameters"
		exit 1
    fi 

	nums=`getent $src 2>/dev/null | awk -F: '{print $3}' - | xargs echo -n " "`

    while [ "$count" -le "$max_num" ]; do
         echo " $nums " | grep " $count " > /dev/null 2>&1 || break
         count=`expr $count + 1`

         if [ "$count" -ge "$max_num" ]; then 
            p_echo "$src_name for $name was not set"
			inten="get free $src_name. Free $src_name not found."
			die "$inten"
         fi
    done  
    
    echo "$count"
    return 0
}

add_login_shell()
{
	local shell

	shell="$1"

	p_echo " Checking that $shell registered as login shell..."

	case $linux_distr in 
	    debian)
		add-shell $shell
	    ;;
	    *)
		if grep -q "^$shell\$" /etc/shells; then
			p_echo "$shell already registered as a login shell"
		else
			echo "$shell" >> /etc/shells 2>> $product_log || die "register login shell $shell"
		fi
	    ;;
	esac

	p_echo
}

# prep_install

prep_install()
{
	mk_backup $inetd_conf cp f
}
# vim:ft=sh

initial_conf()
{
	DEMO_VERSION="no"
	PRODNAME="psa"
	PRODUCT_NAME="psa"
	product_full="Plesk Onyx"
	product=${PRODNAME}
	PRODUCT_FULL_NAME="Plesk Onyx"

	product_etc="/etc/${PRODNAME}"
	prod_conf_t="/etc/psa/psa.conf"
	prodkey="$product_etc/$PRODNAME.key"

	minimal_changes=""

	EXTERNAL_PACKAGES=""
	EXTERNAL_PACKAGES_DIR=""

	BUILDER_UID="10007"

	PERL5LIB=/opt/psa/local/share/perl/5.20.2:/opt/psa/local/lib/x86_64-linux-gnu/perl/5.20.2
	export PERL5LIB

	support_contact="http://www.parallels.com/support"
	sales_email="sales@parallels.com"

	product_version="17.5.3"
	product_db_version="017005003"
	conceived_os_vendor=Debian
	conceived_os_version="8.0"
	osrels="debian8.0"

	clients_group="${product}cln"
	clients_GID=10001

	services_group="psaserv"
	services_GID=10003

	product_suff="saved_by_${product}".`date "+%m.%d;%H:%M"`
	product_suffo="saved_by_${product}"

	# plesk default password
	if [ "X$DEMO_VERSION" = "Xyes" ]; then
		PRODUCT_DEFAULT_PASSWORD="plesk"
	else
		PRODUCT_DEFAULT_PASSWORD="setup"
	fi
}

read_conf()
{
	[ -n "$prod_conf_t" ] || prod_conf_t=/etc/psa/psa.conf

	if [ -s $prod_conf_t ]; then
		tmp_var=`perl -e 'undef $/; $_=<>; s/#.*$//gm;
		         s/^\s*(\S+)\s*/$1=/mg;
		         print' $prod_conf_t`
		eval $tmp_var
	else
		if [ "X$do_upgrade" = "X1" ]; then
			p_echo "Unable to find product configuration file: $prod_conf_t. Default values will be used."
			return 1
		fi
	fi
	return 0
}

get_my_cnf_param()
{
	local r=

	local my_cnf
	find_my_cnf "non-fatal" && 		r=`perl -e '$p="'"$1"'";
		undef $/; $_=<>; s/#.*$//gm;
		/\[mysqld\](.*?)\[/sg;
		$_=substr($1, rindex $1,"$p") and
		/$p\s*=(.*)/m and print $1
		' ${my_cnf}`
	echo $r
}

get_mysql_socket()
{
	# Marked as local as it's not used anywhere else now.
	local mysql_socket="/var/run/mysqld/mysqld.sock"

	local mysqlsock=`get_my_cnf_param  socket`
	local MYSQL_SOCKETS="/var/lib/mysql/mysql.sock /tmp/mysql.sock /var/run/mysqld/mysqld.sock"

	for i in $mysql_socket $mysqlsock $MYSQL_SOCKETS; do
		if [ -S "$i" ]; then
			# This is used internally by mysqld_safe. Maybe this whole function isn't required nowadays.
			# See also: http://dev.mysql.com/doc/refman/5.0/en/problems-with-mysql-sock.html
			MYSQL_UNIX_PORT=$i
			export MYSQL_UNIX_PORT
			mysql_socket="$i"
			break
		fi
	done
}

conf_update()
{
	local def="$1.default"
	umask 0022

	[ -s "$def" ] || return 1
	[ -s "$1" ] || echo '#' > "$1"

	perl -e '
		open (my $config, $ARGV[0]) or die "Cannot open config $ARGV[0]\n";
		while(<$config>){
			s/#.*$//;
			next if /^\s*$/;
			/(\w+)\s+(.+)$/;
			$cf{$1}=$2;
		}
		close $config;

		$cf{PRODUCT_ROOT_D}=$ENV{PRODUCT_ROOT_D} if $ENV{PRODUCT_ROOT_D};

		open (my $default, $ARGV[1]) or die "Cannot open default config $ARGV[1]\n";
		while(<$default>){
			s/^\s*(\w+)(\s+)(.+)$/"$1$2".( (exists($cf{$1}))? $cf{$1}:$3 )/e;
			print;
		}
		close $default;
		' "$1" "$def" > "$1.new"

	if [ -s "$1.new" ]; then
		if cmp -s $1.new $1; then
			rm -f $1.new
			return 1
		else
			rm -f $1.old
			ln $1  $1.old
			mv -f $1.new $1
			chmod 644 $1    # bug 64313, 65043
			p_echo "config updated"
			return 0
		fi
	fi
	return 2
}

# setup new value for parameter
# $1 config file name $2 paramater name, $3 parameter value
conf_setval()
{
	local filename="$1"
	local varname="$2"
	local varvalue="$3"

	oldval="`conf_getvar $filename $varname`"
	[ "$oldval" != "$3" ] || return 0

	cat "$1" | awk -v varname="$varname" -v varvalue="$varvalue" 		'BEGIN { f = 0 }
		{ if ($1 == varname) { f = 1; print varname "\t" varvalue } else { print $0 } }
		END { if (f == 0) { print "\n" varname "\t" varvalue } }' 			> "$filename.new" && 	mv -f "$filename.new" "$filename" && 	chmod 644 "$filename"
}

# A set of functions for work with config variables
# $1 is config file name, $2 is variable name, $3 is variable value

conf_getvar()
{
	cat $1 | perl -n -e '$p="'$2'"; print $1 if m/^$p\s+(.*)/'
}

conf_rmvar()
{
	cat $1 | perl -n -e '$p="'$2'"; print $_ if(! m/^$p\s+(.*)/)' >$1.new
	if cmp -s $1.new $1; then
		rm -f $1.new
		return 0
	fi
	rm -f $1.old
	ln $1  $1.old
	mv -f $1.new $1
	chmod 644 "$1"
}

conf_setvar()
{
	conf_setval "$@"
}

 #default values

product_default_conf()
{

PRODUCT_ROOT_D=/opt/psa
PRODUCT_RC_D=/etc/init.d
PRODUCT_ETC_D=/opt/psa/etc
PLESK_LIBEXEC_DIR=/usr/lib/plesk-9.0
HTTPD_VHOSTS_D=/var/www/vhosts
HTTPD_CONF_D=/etc/apache2
HTTPD_INCLUDE_D=/etc/apache2/conf-enabled
HTTPD_BIN=/usr/sbin/apache2
HTTPD_LOG_D=/var/log/apache2
HTTPD_SERVICE=apache2
QMAIL_ROOT_D=/var/qmail
PLESK_MAILNAMES_D=/var/qmail/mailnames
RBLSMTPD=/usr/sbin/rblsmtpd
NAMED_RUN_ROOT_D=/var/named/run-root
NAMED_OPTIONS_CONF=
NAMED_ZONES_CONF=
WEB_STAT=/usr/bin/webalizer
MYSQL_VAR_D=/var/lib/mysql
MYSQL_BIN_D=/usr/bin
MYSQL_SOCKET=/var/run/mysqld/mysqld.sock
PGSQL_DATA_D=/var/lib/postgresql//main
PGSQL_CONF_D=/etc/postgresql//main
PGSQL_BIN_D=/usr/lib/postgresql/9.4/bin
DUMP_D=/var/lib/psa/dumps
DUMP_TMP_D=/tmp
MAILMAN_ROOT_D=/usr/lib/mailman
MAILMAN_VAR_D=/var/lib/mailman
PYTHON_BIN=/usr/bin/python
CATALINA_HOME=/var/lib/tomcat7
DRWEB_ROOT_D=/opt/drweb
DRWEB_ETC_D=/etc/drweb
GPG_BIN=/usr/bin/gpg
TAR_BIN=/bin/tar
AWSTATS_ETC_D=/etc/awstats
AWSTATS_BIN_D=/usr/lib/cgi-bin
AWSTATS_TOOLS_D=/usr/share/awstats/tools
AWSTATS_DOC_D=/usr/share/awstats
OPENSSL_BIN=/usr/bin/openssl
LIB_SSL_PATH=/lib/libssl.so
LIB_CRYPTO_PATH=/lib/libcrypto.so
CLIENT_PHP_BIN=/opt/psa/bin/php-cli
SNI_SUPPORT=true
APS_DB_DRIVER_LIBRARY=/usr/lib/x86_64-linux-gnu/libmysqlserver.so.2
IPv6_DISABLED=false
SA_MAX_MAIL_SIZE=256000

}

#-*- vim:ft=sh

register_service() {

	[ -n "$1" ] || die "register_service: service name not specified"
	local inten="register service $1"
	echo_try "$inten"

	{
		if [ -x "/bin/systemctl" ]; then
			/bin/systemctl enable "$1.service"
			/bin/systemctl --system daemon-reload >/dev/null 2>&1
		fi


		if  [ -x "/sbin/initctl" -a -e "/etc/init/$1.conf" ]; then  # upstart (ubuntu)
			[ ! -e "/etc/init/$1.override" ] || sed -i "/^manual$/d" "/etc/init/$1.override"
		else
			# update-rc.d for Debian/Ubuntu (not SuSE)
			/usr/sbin/update-rc.d "$1" defaults
		fi



		local rs_db="$PRODUCT_ROOT_D/admin/sbin/register_service_db"
		[ ! -x "$rs_db" ] || "$rs_db" -a "$@"
	}

	suc
}

unregister_service() {

	[ -n "$1" ] || die "unregister_service: service name not specified"
	local inten="unregister service $1"
	echo_try $inten

	{
		local rs_db="$PRODUCT_ROOT_D/admin/sbin/register_service_db"
		[ ! -x "$rs_db" ] || "$rs_db" -r "$1"

		/bin/systemctl disable "$1.service"
		/bin/systemctl --system daemon-reload >/dev/null 2>&1


		if  [ -x "/sbin/initctl" -a -e "/etc/init/$1.conf" ]; then  # upstart (ubuntu)
			grep -q '^manual$' "/etc/init/$1.override" 2>/dev/null || echo "manual" >> "/etc/init/$1.override"
		else
			/usr/sbin/update-rc.d -f "$1" remove
		fi

	} >> $product_log 2>&1

	suc
}
# vim:ft=sh

selinux_is_active()
{
	if [ -z "$SELINUX_ENFORCE" ]; then
		selinux_getenforce
	fi

	case "$SELINUX_ENFORCE" in
	Enforcing|Permissive) return 0;;
	*) return 1;;
	esac
}

selinux_support_is_installed()
{
	# This function checks if Plesk SELinux support component is installed
	set_selinux_params
	[ -s "$selinux_module" ]
}

selinux_configuration_is_required()
{
	# All public functions that modify SELinux state should check that this is true!
	selinux_is_active && selinux_support_is_installed
}

selinux_get_mount_dir()
{
	unset SELINUX_MOUNT_DIR

	if awk '$2 == "/selinux"{exit(1)}' /proc/mounts && mkdir -p /selinux; then
		SELINUX_MOUNT_DIR=/selinux
	else
		SELINUX_MOUNT_DIR="`mktemp -d /tmp/selinuxXXXXXX`"
	fi >>"$product_log" 2>&1
}

selinux_getenforce()
{
	if [ "$1" = "--check" -a -n "$SELINUX_ENFORCE" ]; then
		return
	fi
	unset SELINUX_ENFORCE

	if ! ( command -v selinuxenabled >/dev/null 2>&1 && selinuxenabled ); then
		SELINUX_ENFORCE=Disabled
		return
	fi

	if awk '$3 == "selinuxfs"{exit(1)}' /proc/mounts; then
		selinux_get_mount_dir
		mount -t selinuxfs none "$SELINUX_MOUNT_DIR"
	fi

	if ! command -v getenforce >/dev/null 2>&1; then
		SELINUX_ENFORCE=Disabled
		return
	fi

	SELINUX_ENFORCE="`getenforce`"
	if test $? -ne 0; then
		SELINUX_ENFORCE=Disabled
		return
	fi
}

selinux_init()
{
	unset SELINUX_ENFORCE
	if selinux_is_active; then
		setenforce Permissive
	else
		unset SELINUX_ENFORCE
	fi
}

selinux_close()
{
	if [ -z "$SELINUX_ENFORCE" -o "$SELINUX_ENFORCE" = "Disabled" ]; then
		return
	fi

	setenforce "$SELINUX_ENFORCE"
}


relabel_named_directories()
{
	selinux_configuration_is_required || return 0

	if [ -d "$NAMED_RUN_ROOT_D/proc" ]; then
		# Looks like /proc exists in chroot only on RedHat/CentOS 5
		selinux_relabel_dir -e "$NAMED_RUN_ROOT_D/proc" "$NAMED_RUN_ROOT_D"
	else
		selinux_relabel_dir "$NAMED_RUN_ROOT_D"
	fi
}

relabel_plesk_directories()
{
	selinux_configuration_is_required || return 0

	local ret=0
	local verbose_mode=
	if [ "$1" = "--verbose" ]; then
		verbose_mode="yes"
		shift
	fi

	set_horde_params
	if selinux_is_active; then
		for dir in "$PRODUCT_ROOT_D" "$horde_logdir" 					"$QMAIL_ROOT_D/alias" "$QMAIL_ROOT_D/bin" "$QMAIL_ROOT_D/boot" "$QMAIL_ROOT_D/control" 					"$QMAIL_ROOT_D/plugins" "$QMAIL_ROOT_D/popuser" "$QMAIL_ROOT_D/queue" "$QMAIL_ROOT_D/users" 					/var/db/kav /var/db/Quarantine "$PLESK_DB_DIR" 					"$POSTFIX_LIBEXEC_DIR" 					/var/drweb "$DRWEB_ROOT_D" "$PLESK_LIBEXEC_DIR" 					"/etc/nginx" "/usr/sbin/nginx" /var/lib/nginx 					"/var/log/nginx" "/var/run/nginx.pid" 					"/var/lib/plesk/mail" /opt/kav "/usr/lib/x86_64-linux-gnu/php/modules" "/var/run" 					"$@"; do
			[ -z "$verbose_mode" ] || pp_echo "  Restoring SELinux context on '$dir'"
			selinux_relabel_dir "$dir" || ret=1
		done

		[ -z "$verbose_mode" ] || pp_echo "  Restoring SELinux context on '$NAMED_RUN_ROOT_D'"
		relabel_named_directories || ret=1
	fi
	[ -z "$do_repair" ] || return $ret
}

selinux_relabel_dir()
{
	selinux_configuration_is_required || return 0

	if ! command -v restorecon >/dev/null 2>&1; then
		return
	fi

	local ret=0
	if ! restorecon -i -R "$@" >>"$product_log" 2>&1; then
		report_problem "Error while setting SELinux types. Command was: restorecon -i -R $*"
		ret=1
	fi

	[ -z "$do_repair" ] || return $ret
}

# vim:ft=sh

set_selinux_params()
{
	selinux_module="$PRODUCT_ROOT_D/etc/plesk.pp"
}


necessary_system_accounts_prep_install_psaadm()
{
# for admin
	p_echo " Checking for the system groups and users necessary for admin server..."
	group_op "$admin_group" "search" false
	user_op "$admin_user" "search" "$admin_group" "Admin Server" "$dummy_home" "$dummy_shell" false
}

necessary_system_accounts_prep_install_psaserv()
{
# for psaservices (only after creating apache and psaftp)
	group_op "$services_group" "$services_GID" false
	add_user_to_group "$admin_user" "$services_group"
}

necessary_system_accounts_prep_install_psacln()
{
# clients
	found=`get_user_group_id "clients" "gid" $clients_GID $max_suexec_GID`
	group_op $clients_group $found false
}

necessary_system_accounts_prep_install()
{
	p_echo
	p_echo "===> Checking for the necessary system accounts"
	common_var=0

	necessary_system_accounts_prep_install_psaadm
	necessary_system_accounts_prep_install_psaserv
	necessary_system_accounts_prep_install_psacln
}

necessary_system_accounts_post_install()
{
	# these users may not exist before packages isntallation
	add_user_to_group sw-cp-server "$admin_group"
	add_user_to_group "$admin_user" lock-manager
}
# vim:ft=sh:

#set_params

set_common_params()
{
	common_var=0

	PATH=/sbin:/bin:/usr/sbin:/usr/bin:/usr/local/sbin:/usr/local/bin
	LANG=C
	export PATH LANG
	unset GREP_OPTIONS
	umask 022
	ulimit -n 65535 2>/dev/null

	K_HUP="/bin/kill -HUP"
	K_KILL="/bin/kill -KILL"
	K_TERM="/bin/kill -TERM"
	K_USR2="/bin/kill -USR2"
	K_TEST="/bin/kill -0"

	users_created=""
	groups_created=""

	certificate_file="$PRODUCT_ETC_D/httpsd.pem"
	services="/etc/services"
	mtab="/etc/mtab"
	get_hostname="hostname"
	get_domainname="domainname"

	#default parameters
	tar="tar"
	crontab="/usr/bin/crontab"

	cp_preserve="cp -p"
	SYSTEM_RC_D=/etc/init.d
	PLESK_LIBEXEC_DIR="/usr/lib/plesk-9.0"
	PLESK_DB_DIR="/var/lib/plesk"
	POSTFIX_LIBEXEC_DIR="/usr/lib/postfix"
	PRODUCT_BOOTSTRAPPER_DIR="/opt/psa/bootstrapper/pp17.5.3-bootstrapper"
	AUTOGENERATED_CONFIGS="#ATTENTION!\n#\n#DO NOT MODIFY THIS FILE BECAUSE IT WAS GENERATED AUTOMATICALLY,\n#SO ALL YOUR CHANGES WILL BE LOST THE NEXT TIME THE FILE IS GENERATED.\n"
	AUTOGENERATED_CONFIGS_UPGRADE="#ATTENTION!\n#\n#DO NOT MODIFY THIS FILE BECAUSE IT WAS GENERATED AUTOMATICALLY,\n#SO ALL YOUR CHANGES WILL BE LOST AFTER YOU UPGRADE PARALLELS PLESK PANEL.\n"
	PRODUCT_LOGS_D="/var/log/plesk"

	set_common_params_linux 

	rp_patch_runtime
}

set_common_params_linux()
{
	get_hostname="hostname -f"
	fstab="/etc/fstab"
	cp_preserve="cp --preserve=all --remove-destination"
	machine="linux"
	sendmail="/usr/sbin/sendmail"
	ps="ps axw"
	ps_long="ps axuw"
	false_shell="/bin/false"
	dummy_home="/"
	compress="gzip -9 -c"
	uncompress="gunzip -c"
	uudecode="uudecode -o /dev/stdout"
	ifconfig="/sbin/ifconfig -a"
	inet_str="inet addr"

	useradd_options="-r"
	if [ -f /etc/SuSE-release ]; then
		linux_distr="suse"
	elif [ -f /etc/debian_version ]; then
		linux_distr="debian"
		get_domainname="dnsdomainname"
	else
		linux_distr="redhat"
	fi

	sndml_ini="/etc/init.d/sendmail"
	mail_local="/usr/libexec/mail.local"
	if [ -x /sbin/nologin ]; then
		dummy_shell="/sbin/nologin"
	else
		dummy_shell="/bin/false"
	fi
	bash_shell="/bin/bash"
	rbash_shell="/bin/rbash"
	uudecode_full="/usr/bin/uudecode"
	# FIXME: remove or keep? can't find use of this variable
	# named_osrelease=`perl -F"/[.-]/" -n -a  -e 'printf "%02u%02u%02u\n", $F[0],$F[1],$F[2]' /proc/sys/kernel/osrelease`

	return 0
}

define_start_script()
{
	if [ -f "$PRODUCT_RC_D/${product}.sh" ]; then
		START_SH="$PRODUCT_RC_D/${product}.sh"
	else
		START_SH="$PRODUCT_RC_D/${product}"
	fi
}

set_local_params()
{
	for i in set_admin_params lock_keyupdate set_apache_params set_awstats_params set_courier_imap_params set_syslog_params set_logrotate_params set_mysqld_params set_qmail_params set_webalizer_params ; do
		$i
		local res="$?"
		if [ "0$res" -gt 0 ]; then
			warn "$i function completed with errors (error code $res)"
		fi

	done
}

# vim:ft=sh

set_file_swap_params()
{
	local pleskswaprc='/etc/pleskswaprc'
	[ ! -f "$pleskswaprc" ] || . /etc/pleskswaprc
	[ -n "$PLESK_SWAP_PATH" ] || PLESK_SWAP_PATH='/pleskswap'
	[ -n "$PLESK_SWAP_SIZE" ] || PLESK_SWAP_SIZE='1G'
	[ -n "$PLESK_REQUIRED_MEMORY" ] || PLESK_REQUIRED_MEMORY='1G'

	FSTAB='/etc/fstab'
}

file_swap_is_required()
{
	local total_mem_mib=$(LC_ALL=C LANG=C free -m -t | perl -nale 'print $F[1] if /^Total:/')
	local required_mem_mib="`units2units $PLESK_REQUIRED_MEMORY M`"
	[ "$total_mem_mib" -lt "$required_mem_mib" ] || return 1
	pp_echo "Total amount of memory is less than minimal required size (${total_mem_mib}M < ${required_mem_mib}M)"
	return 0
}

file_swap_is_switched_off()
{
	case "${PLESK_SWAP:-}" in
		0|false|disable)
			pp_echo "Swapfile creation is disabled: envirinment vaiable \$PLESK_SWAP='$PLESK_SWAP'."
			return 0
		;;
	esac

	if [ -f "/etc/pleskswapdisable" ]; then
		pp_echo "Swapfile creation is disabled: file '/etc/pleskswapdisable' is present."
		return 0
	fi

	detect_vz
	if [ "$PLESK_VZ" = "1" ]; then
		pp_echo "Swapfile creation is disabled: installation into Virtuozzo container."
		return 0
	fi

	detect_lxc
	if [ "$PLESK_LXC" = "1" ]; then
		pp_echo "Swapfile creation is disabled: installation into Docker/LXC container."
		return 0
	fi

	return 1
}

file_swap_enable()
{
	if file_swap_status; then
		echo "Error: Plesk swapfile is already enabled." >&2
		return 1
	fi

	local swap_size_mb="`units2units $PLESK_SWAP_SIZE M`"

	pp_echo "===> Enable swapfile in $PLESK_SWAP_PATH"
	dd if=/dev/zero of="$PLESK_SWAP_PATH" bs=1M count="$swap_size_mb" status=none || return 1
	chmod 0600 "$PLESK_SWAP_PATH" || return 1
	mkswap "$PLESK_SWAP_PATH" || return 1
	if ! grep -qw "^$PLESK_SWAP_PATH" "${FSTAB}"; then
		cp -f "${FSTAB}" "${FSTAB}.saved_by_plesk"
		echo "$PLESK_SWAP_PATH none swap sw 0 0" >> "${FSTAB}"
	fi
	if swapon "$PLESK_SWAP_PATH"; then
		rm -f "${FSTAB}.saved_by_plesk"
		return 0
	else
		[ ! -f "${FSTAB}.saved_by_plesk" ] || mv -f "${FSTAB}.saved_by_plesk" "${FSTAB}"
		return 1
	fi
}

file_swap_status()
{
	[ -f "$PLESK_SWAP_PATH" ] || return 1
	grep -qw "^$PLESK_SWAP_PATH" "${FSTAB}" || return 1
	return 0
}
# -*- vim:syntax=sh

set_syslog_params()
{
	syslog_conf_ng="/etc/syslog-ng/syslog-ng.conf"

	syslog_conf=""
	for config in rsyslog.conf rsyslog.early.conf syslog.conf; do
		[ -f "/etc/$config" ] && syslog_conf="$syslog_conf /etc/$config"
	done

	syslog_service=""
	syslog_binary=""

	# Make sure the sequence of services is correlate with binaries
	local syslog_services="syslog sysklogd rsyslog syslog-ng"
	local syslog_binaries="syslogd syslogd rsyslogd syslog-ng"

	for service in $syslog_services; do
		[ -f "/lib/systemd/system/${service}.service" ] && 			syslog_service="$service" && break
	done

	for binary in $syslog_binaries; do
		for bin_path in /sbin /usr/sbin; do
			[ -x "$bin_path/${binary}" ] && 				syslog_binary="$bin_path/${binary}" && break
		done
		[ -n "$syslog_binary" ] && break
	done

}
## @@constructor set_syslog_params

true syslog_status_linux_debian
syslog_status_linux_debian()
{
	get_pid "$syslog_binary" false
	local pid=$common_var
	if test "$pid" -ne 1; then
		# running
		return 0
	fi
	return 1
}

true syslog_reload_linux_suse
true syslog_reload

syslog_reload_linux_suse()
{
	if [ "$syslog_service" = "rsyslog" -a "$syslog_binary" = "/sbin/rsyslogd" ]; then
		# Suse 13.1 man 8 rsyslogd:
		# So it is advised to use HUP only for closing files, and a "real restart" (e.g. /etc/rc.d/rsyslogd restart) to activate configuration changes.
		service_ctl restart $syslog_service syslog
	else
		service_ctl reload $syslog_service syslog
	fi
}

syslog_reload()
{
	# set_syslog_params must be called in outed function

	if [ "X$syslog_service" = "Xrsyslog" ]; then
		# Bug 142129
		# rsyslog service registration is necessary
		# it is workaround on default rsyslog service registration behaviour
		register_service rsyslog
		# then we restart it
	fi

	detect_vz
	if [ "$syslog_service" = "syslog" -o "$syslog_service" = "rsyslog" ] && [ "$PLESK_VZ" = "1" ]; then
		# 146355 - rsyslog/syslog service returns false on VZ
		service_ctl restart $syslog_service syslog
	elif [ "X$syslog_service" = "Xrsyslog" ]; then
		service_ctl restart $syslog_service syslog
	else
		service_ctl reload $syslog_service syslog
	fi
}

units2units() {
	local bytes
	local n="${1%%[^0-9]*}"
	case "${1,,}" in
		*[0-9]) bytes="$1" ;;
		*k|*kib) bytes="$(( $n * 1024 ** 1 ))" ;;
		*m|*mib) bytes="$(( $n * 1024 ** 2 ))" ;;
		*g|*gib) bytes="$(( $n * 1024 ** 3 ))" ;;
		*t|*tib) bytes="$(( $n * 1024 ** 4 ))" ;;
		*kb)     bytes="$(( $n * 1000 ** 1 ))" ;;
		*mb)     bytes="$(( $n * 1000 ** 2 ))" ;;
		*gb)     bytes="$(( $n * 1000 ** 3 ))" ;;
		*tb)     bytes="$(( $n * 1000 ** 4 ))" ;;
		*) echo "units2units: incorrect value '$1'" >&2; exit 1 ;;
	esac
	case "${2,,}" in
        k|kib) echo $(( $bytes / 1024 ** 1 )) ;;
        m|mib) echo $(( $bytes / 1024 ** 2 )) ;;
        g|gib) echo $(( $bytes / 1024 ** 3 )) ;;
        t|tib) echo $(( $bytes / 1024 ** 4 )) ;;
        kb)    echo $(( $bytes / 1000 ** 1 )) ;;
        mb)    echo $(( $bytes / 1000 ** 2 )) ;;
        gb)    echo $(( $bytes / 1000 ** 3 )) ;;
        tb)    echo $(( $bytes / 1000 ** 4 )) ;;
        "")    echo $bytes ;;
        *)     echo "Unknown unit: $2" >&2; exit 1 ;;
    esac
}

# vim:ft=sh

# Prints string in format "<OS_NAME> <OS_VERSION>"
get_runtime_os_info()
{
# TODO: check and improve (if needed) runtime os detection.
	local os_name os_version
# arch?
	local getver='sed -e s/[^0-9.]*\([0-9.]*\).*/\1/g'

	if [ -e '/etc/debian_version' ]; then
		if [ -e '/etc/lsb-release' ]; then
### Ubuntu
			. /etc/lsb-release
			os_name=$DISTRIB_ID
			os_version=$DISTRIB_RELEASE
		else
###	Debian
			os_name='Debian'
			os_version=$(head -1 /etc/debian_version)
		fi
	elif [ -e '/etc/redhat-release' ]; then
		local header=$(head -n 1 /etc/redhat-release)
		if expr match "$header" "Red Hat" > /dev/null; then
			os_name='RedHat'
		elif expr match "$header" "CentOS" > /dev/null; then
			os_name='CentOS'
		elif expr match "$header" "CloudLinux" > /dev/null; then
			os_name='CloudLinux'
		else
			os_name="$(echo $header | sed -e 's/ release.*//i')"
		fi

		os_version=$(echo $header | ${getver})
	fi
	echo "$os_name $os_version"
}

# append record into upgrade_history before installation
# in case if previous version is not reflected there.
pre_update_history()
{
	local query
	local inten="insert pre-upgrade info into history table"

	get_product_versions

	query="SELECT * FROM upgrade_history WHERE version_info='${product_prev_version}' LIMIT 1"

	if ! db_test "$query" '1'; then
		query="INSERT INTO upgrade_history (upgrade_date, version_info) VALUES (now(),
			'${product_prev_version}')"
		db_do --inten "$inten" "$query"
	fi
}

update_history()
{
	local mu_version="$1"
	local query
	local inten="insert upgrade info into history table"

	get_product_versions

	local os_info="`get_runtime_os_info`"

	if [ -n "$mu_version" ]; then
		query="SELECT * FROM upgrade_history WHERE version_info='${product_this_version}' 			AND os='${os_info}' AND mu='${mu_version}' LIMIT 1"
	else
		query="SELECT * FROM upgrade_history WHERE version_info='${product_this_version}' 			AND os='${os_info}' LIMIT 1"
	fi

	if db_test "$query" '1'; then
		# record already exists
		return 0
	fi

	if [ -n "$mu_version" ]; then
		query="INSERT INTO upgrade_history (upgrade_date, version_info, os, mu) VALUES (now(),
			'${product_this_version}', '${os_info}', '${mu_version}')"
	else
		query="INSERT INTO upgrade_history (upgrade_date, version_info, os) VALUES (now(),
			'${product_this_version}', '${os_info}')"
	fi

	db_do --inten "$inten" "$query"
}

# -*- vim:syntax=sh

packagemng_set_dirty_flag()
{
	if [ -f ${PRODUCT_ROOT_D}/admin/bin/packagemng ]; then
		${PRODUCT_ROOT_D}/admin/bin/packagemng --set-dirty-flag 1>>$product_log 2>&1
	fi
}

get_product_versions()
{
	local prod_root_d="/opt/psa"
	
	product_name="psa"
	product_this_version="17.5.3"
	product_this_version_tag=""
	if [ -z "$product_prev_version" ]; then
		if [ -r "$prod_root_d/version.upg" ]; then
			product_prev_version=`awk '{ print $1 }' "$prod_root_d/version.upg"`
		elif [ -r "$prod_root_d/version" ]; then
			product_prev_version=`awk '{ print $1 }' "$prod_root_d/version"`
		else
			product_prev_version="$product_this_version"
		fi
	fi
}

# Product is being upgraded to a new version
is_product_upgrade()
{
	get_product_versions
	[ "$product_prev_version" != "$product_this_version" ]
}

# Clean installation of the product is being performed
is_product_installation()
{
	[ "X$do_upgrade" != "X1" -a ! -s "/opt/psa/version.upg" ]
}

# Product management node is already installed (at least files are in place, maybe scripts were not finished yet)
is_management_node_installed()
{
	[ -s "$PRODUCT_ROOT_D/version" -o -s "$PRODUCT_ROOT_D/version.upg" -o -s "$PRODUCT_ROOT_D/core.version" -o -s "$PRODUCT_ROOT_D/.revision" ]
}

set_logrotate_params()
{
	# primary Plesk logrotate config file
	logrotate_conf="${product_etc}/logrotate.conf"

	# directory for logrotate configs of virtual hosts
	logrotate_include_d="${PRODUCT_ETC_D}/logrotate.d"
}
# vim:ft=sh:

set_module_params()
{
	packname="psa-$module"
	scriptname="psa-$module"

	# This is the version like 10203
	module_version_n=$((`echo "$module_version" | sed -e 's/^\(.*\)\.\(.*\)\.\(.*\)$/\1*10000+\2*100+\3/'`))

	build="1705181214.14"

	product_rc_d="/etc/init.d"

	module_db_sql="$PRODUCT_ETC_D/modules/$module/${module}_db.sql"

	module_var="$PRODUCT_ROOT_D/var/modules/$module"

	sedflags="-r"
}

autoinstall_extensions()
{
	for e in $PRODUCT_ROOT_D/var/modules/install/*.zip; do
		[ -e "$e" -a ! -L $e ] || continue
		local inten="install or update `basename $e` extension"
		echo_try "$inten"
		if ! $PRODUCT_ROOT_D/bin/extension --upgrade "$e"; then
			warn "$inten"
			continue
		fi
		rm -f $e
		suc
	done
}
# -*- vim:ft=sh

# MySQL server configuration

generate_mysql_credentials()
{
	local args="$*"
	mysql_defaults=`echo $args | cut -d: -f 1`

	mysql_user=`echo $args | cut -d: -f 2`
	[ -n "$mysql_user" ] && mysql_user="--user=$mysql_user"

	mysql_passwd=`echo $args | cut -d: -f 3-`
	[ -n "$mysql_passwd" ] && mysql_passwd="$mysql_passwd"
}

# This function must be called *ONLY* for psa installation
setup_admin_login()
{
	true mysql_quote_string

	unset MYSQL_PWD
	local mysql_db_name="mysql"
	local sl
	local inten="define valid mysql credentials"
	local i

	echo_try "$inten"
	get_admin_passwd

# --no-defaults:admin:$admin_passwd		admin with pass
# --no-defaults:root:					root without pass
# :root:								root with pass in defaults
# --no-defaults:admin:					admin withiyt pass
# --no-defaults:admin:$PRODUCT_DEFAULT_PASSWORD		admin with default pass
# :admin:								admin with pass in defaults
# ::									user and pass in defaults
# --defaults-file=/root/.my.cnf			hspc defaults (paranoid)
#
	for sl in `sequence 20`; do
		for i in	"--no-defaults:admin:$admin_passwd" 					"--no-defaults:root:" 					":root:" 					"--no-defaults:admin:" 					"--no-defaults:admin:$PRODUCT_DEFAULT_PASSWORD" 					":admin:" 					"::" 					"--defaults-file=/root/.my.cnf::"; do
			generate_mysql_credentials "$i"
			if mysql_test_connection; then
				suc
# 	create/Update admin superuser
				echo "grant all privileges on *.* to 'admin'@'localhost' identified by '`mysql_quote_string $admin_passwd`' with grant option;flush privileges" | mysql_direct mysql
				update_psa_shadow

#	backup private my.cnf if it contains password (vz/hspc related)
				rm -f /.my.cnf
				if grep -q '\s*password\s*=' /root/.my.cnf 2>/dev/null; then
					p_echo "WARNING: You current personal mysql config /root/.my.cnf has been renamed into /root/.my.cnf.bak"
					mv -f /root/.my.cnf /root/.my.cnf.bak
				fi

				set_mysql_auth
				return 0
			fi
		done
		p_echo "One more attempt to connect"
		try_reset_mysql_auth
		sleep "$sl"
	done
	die "$inten. If you are installing Plesk on an already configured MySQL server, you need to specify the administrator's credentials to succeed with the installation." 		"To do this, you need to create a file - /root/.my.cnf with the 'client' section where you need to provide user and its password" 		"(\"user = \$admin_name\" and \"password = \$admin_pass\")." 		"After installation is finished, the file /root/.my.cnf will be renamed to /root/.my.cnf.bak"
}

update_psa_shadow()
{
	echo "$admin_passwd" > $mysql_passwd_file || die
	chown $admin_user:$admin_group ${mysql_passwd_file} || die
	chmod 600 ${mysql_passwd_file} || die
}

encrypt_psa_shadow()
{
	local init_conf_bin="$PRODUCT_ROOT_D/bin/init_conf"
	[ -x "$init_conf_bin" ] && 	PSA_PASSWORD="$admin_passwd" "$init_conf_bin" --set-admin-password -passwd '' -encrypted-password >> "$product_log" 2>&1
	if [ "$?" -eq 0 ]; then
		admin_passwd=
		set_mysql_auth
	else
		warn "encrypting admin password"
	fi
}

set_mysql_auth()
{
	# This function requires set_mysql_server_params and set_mysql_client_params (or set_mysqld_params)
	# to be called before

	local inten="set up mysql authentication"
	get_admin_passwd

	pleskrc mysql start

	mysql_host=
	mysql_user="--user=admin"
	mysql_passwd="$admin_passwd"
	unset mysql_defaults

	if [ -z "$MYSQL_AUTH_SKIP_CONNECTION_CHECK" ]; then
		mysql_test_connection 60 || die "$inten"
	fi
	suc
}

get_admin_passwd()
{
    [ -z "$admin_passwd" ] || return 0

    if [ -f "$mysql_passwd_file" ]; then
		admin_passwd=`cat "$mysql_passwd_file"`
		return 0
    fi

	admin_passwd=`get_random_string 12 'A-Za-z0-9_'`

	[ -n "$admin_passwd" ] || admin_passwd="$PRODUCT_DEFAULT_PASSWORD"
}

mysql_quote_string()
{
	echo "$1" | perl -pe 's|\\|\\\\|g'
}

reset_admin_passwd()
{
	true mysql_quote_string

	[ -n "$admin_passwd" ] || die "admin_passwd is not set yet"
	# I bet someone somewhere replaces 'localhost' with '%' (for weird automation purposes), so be nice
	echo "SET PASSWORD FOR 'admin'@'localhost' = PASSWORD('`mysql_quote_string $admin_passwd`')" | mysql_direct mysql || 	echo "SET PASSWORD FOR 'admin'@'%'         = PASSWORD('`mysql_quote_string $admin_passwd`')" | mysql_direct mysql
}

mysql_test_connection()
{
	inten="establish test connection"
	echo_try $inten
	attempts=${1:-1}
	for i in `sequence $attempts`; do
		echo "" | mysql_direct mysql >> "$product_log" 2>&1
		if [ "$?" -eq "0" ]; then
			p_echo "connected"
			return 0
		fi
		[ "$attempts" -eq "1" ] || sleep 2
	done

	p_echo "failed"
	return 1
}

configure_mysql_innodb()
{
	local my_cnf="$1"

	if grep -q '^skip-innodb' "$my_cnf"; then
		sed -i -e '/^skip-innodb/ d' "$my_cnf"
		need_mysql_restart=yes
	fi
}

#bug #46837. Disable "LOAD DATA LOCAL INFILE" command.
configure_mysql_infile()
{
	local my_cnf="$1"
	local awk_script

	if ! (test -f $my_cnf && grep -q '^\[mysqld\]' "$my_cnf"); then
		echo "[mysqld]" >> $my_cnf
		echo "local-infile=0" >> $my_cnf
		need_mysql_restart=yes
		return
	fi

	awk_script='/^\[mysqld\]$/{print; print "local-infile=0"; next;} {print}'
	if ! grep -q '^local-infile' "$my_cnf"; then
		awk "$awk_script" "$my_cnf" >"$my_cnf.tmp" && mv "$my_cnf.tmp" "$my_cnf" || die "edit $my_cnf"
		need_mysql_restart=yes
	fi
}

configure_mysql_address()
{
	local my_cnf="$1"
	local address="$2"

	if ! grep -q '^bind-address.*=.*$address' "$my_cnf"; then
		sed -e "/^bind-address.*=.*/d" -i "$my_cnf" || 			die "remove 'bind-address' directive from '$my_cnf'"
		sed -e "s|^\(\[mysqld\].*\)|\1\nbind-address = $address|" -i "$my_cnf" || 			die "configure MySQL server via '$my_cnf' to listen on address $address"
		need_mysql_restart=yes
	fi
}

check_ipv6_network_available()
{
	ping6 -c1 -q "::1" >/dev/null 2>&1
}

configure_mysql_address_local()
{
	local ipv6_supported_version="5.5.3"
	local current_version="`mysql_raw_anydb -e \"select version();\"`"
	local my_cnf="$1"
	local address="127.0.0.1"

	# if we cannot detect mysql-server version
	# use ipv4 only address
	if [ -n "$current_version" ]; then
		mysql_compare_versions "$current_version" "$ipv6_supported_version"
		if [ $? -ne 1 ] && check_ipv6_network_available; then
			address="::ffff:127.0.0.1"
		fi
	fi

	configure_mysql_address "$my_cnf" "$address"
}

configure_mysql_disable_old_passwords()
{
	local my_cnf="$1"
	local awk_script='/^[[:space:]]*old_passwords/ { print "# Forced OLD_PASSWORD format is turned OFF by Plesk\n#" $0; next } { print }'

	if awk "$awk_script" "$my_cnf" > "$my_cnf.tmp" || die "edit $my_cnf"; then
		if diff -q "$my_cnf" "$my_cnf.tmp" 1>/dev/null ; then
			rm -f "$my_cnf.tmp"
		else
			mv "$my_cnf.tmp" "$my_cnf" || die "disable old_passwords in $my_cnf"
			need_mysql_restart=yes
		fi
	fi
}

configure_mysql_no_strict_mode()
{
	local my_cnf="$1"

	local mysql_db_name="mysql"
	db_select "SELECT @@GLOBAL.sql_mode"
	local sql_mode="$db_select_output"
	[ -n "$sql_mode" ] || return 0

	local ok_sql_mode="`echo "$sql_mode" |
						tr ',' '\n' |
						egrep -v 'STRICT_TRANS_TABLES|STRICT_ALL_TABLES|NO_ZERO_DATE' |
						xargs | tr ' ' ','`"
	[ "$ok_sql_mode" != "$sql_mode" ] || return 0

	sed -e "/^sql_mode.*=.*/d" -i "$my_cnf" || 		die "remove 'sql_mode' directive from '$my_cnf'"
	sed -e "s|^\(\[mysqld\].*\)|\1\nsql_mode=$ok_sql_mode|" -i "$my_cnf" || 		die "configure MySQL server via '$my_cnf' with no strict mode"
	need_mysql_restart=yes
}

find_my_cnf()
{
	local non_fatal="$1"
	local cnf_files="/etc/my.cnf /etc/mysql/my.cnf /var/db/mysql/my.cnf"

	for my_cnf in $cnf_files; do
		if [ -f ${my_cnf} ]; then
			break
		fi
	done

	[ -f "$my_cnf" -o -n "$non_fatal" ] || die "find MySQL server configuration file. " 		"If you use thirdparty MySQL server build, make sure you don't use implicit default configuration " 		"and have MySQL configuration file in one of the following locations: $cnf_files"
	[ -f "$my_cnf" ]
}

run_configure_mysql_funcs()
{
	local need_mysql_restart=
	local my_cnf=
	find_my_cnf

	for func in "$@"; do
		eval "$func $my_cnf"
	done

	if [ -n "$need_mysql_restart" ]; then
		pleskrc mysql restart
	fi
}

mysql_install_init()
{

	register_service "$mysql_service"

	run_configure_mysql_funcs 		configure_mysql_innodb 		configure_mysql_infile 		configure_mysql_disable_old_passwords 		tune_vz_mysql

	setup_admin_login
	mysql_fix_remove_bad_users

	run_configure_mysql_funcs 		configure_mysql_address_local 		configure_mysql_no_strict_mode
}

mysql_fix_remove_bad_users()
{
	local mysql_db_name="mysql"
	db_do "drop database if exists test"
	db_do "delete from user where User=''"
	db_do "delete from user where User='root'"
	db_do "flush privileges"
}

mysql_compare_versions()
{
	local l_greater=2
	local l_equal=0
	local l_less=1

	local lhs_major="`echo $1 | cut -d'.' -f1`"
	local lhs_minor="`echo $1 | cut -d'.' -f2`"
	local lhs_patch="`echo $1 | cut -d'.' -f3 | grep -o -e '^[0-9]\+'`"

	local rhs_major="`echo $2 | cut -d'.' -f1`"
	local rhs_minor="`echo $2 | cut -d'.' -f2`"
	local rhs_patch="`echo $2 | cut -d'.' -f3 | grep -o -e '^[0-9]\+'`"

	# TODO(galtukhov): rewrite string comparision as python one-liner

	if [ "$lhs_major" -gt "$rhs_major" ]; then
		return $l_greater
	elif [ "$lhs_major" -lt "$rhs_major" ]; then
		return $l_less
	else
		if [ "$lhs_minor" -gt "$rhs_minor" ]; then
			return $l_greater
		elif  [ "$lhs_minor" -lt "$rhs_minor" ]; then
			return $l_less
		else
			if [ "$lhs_patch" -gt "$rhs_patch" ]; then
				return $l_greater
			elif [ "$lhs_patch" -lt "$rhs_patch" ]; then
				return $l_less
			else
				return $l_equal
			fi
		fi
	fi
}

enable_innodb_engine()
{
	local have_innodb="$1"

	if [ -z "${have_innodb}" ]; then
		local mysql_db_name=mysql
		local db_select_output=
		db_select "SHOW VARIABLES LIKE 'have_innodb'"
		have_innodb=`get_narg 2 ${db_select_output}`
		[ -n "$have_innodb" ] || have_innodb="NO"
	fi

	echo_try "allow INNODB engine"
	if [ "${have_innodb}" = "YES" ]; then
		p_echo "Innodb is already enabled - nothing to be done"
		return 0
	fi

	# Tables are not converted back to INNODB, only engine support is switched on!
	find_my_cnf
	! grep -q '^skip-innodb' "$my_cnf" || sed -i -e '/^skip-innodb/ d' "$my_cnf"
	suc

	pnnl_echo "Restarting mysql"
	pleskrc mysql restart || die "restart MySQL"
}

# -*- vim:ft=sh

# mysql

set_mysqld_params()
{
	mysqld_user="mysql"
	mysqld_UID=3306
	mysqld_group="mysql"
	mysqld_GID=3306

	product_db_sql="$PRODUCT_BOOTSTRAPPER_DIR/db/${PRODNAME}_db.sql"

	set_mysql_server_params
	set_mysql_client_params
}

## @@constructor set_mysqld_params

set_mysql_server_params()
{
	get_mysql_socket

	if [ -x "${PRODUCT_RC_D}/mysql" ]; then
	    mysql_service="mysql"
	elif [ -x "${PRODUCT_RC_D}/mysql.sh" ]; then
	    mysql_service="mysql.sh"
	elif [ -x "${PRODUCT_RC_D}/mysqld" ]; then
	    mysql_service="mysqld"
	elif [ -x "${PRODUCT_RC_D}/mysql" ]; then
	    mysql_service="mysql"
	elif [ -n "/lib/systemd/system" -a -f "/lib/systemd/system/mariadb.service" ]; then
	    mysql_service="mariadb"
	elif [ -n "/lib/systemd/system" -a -f "/lib/systemd/system/mysqld.service" ]; then
	    # Percona server
	    mysql_service="mysqld"
	elif [ "X$DEMO_VERSION" = "Xyes" ]; then
	    mysql_service="mysqld"
	else
	    die "detect MySQL service name"
	fi
}

set_mysql_client_params()
{
	mysql_client="$MYSQL_BIN_D/mysql"

	# Override this variable as needed
	mysql_db_name="$PRODNAME"
	mysql_passwd_file="$product_etc/.${PRODNAME}.shadow"

	mysql_args="-N"
	mysql_args_raw="-Br"
}

#Invoke mysql
mysql()
{
	mysql_anydb -D$mysql_db_name "$@"
}

mysql_anydb()
{
	(
		export MYSQL_PWD="$mysql_passwd"
		$mysql_client $mysql_host $mysql_user $mysql_args "$@" 2>>"$product_log"
		local status=$?

		if [ $status -gt 0 ]; then
			$mysql_client $mysql_host $mysql_user $mysql_args -D$mysql_db_name $mysql_args_raw -e "SHOW ENGINE innodb status" >>"$product_log" 2>&1
		fi
		unset MYSQL_PWD
		return $status
	)
}

# Invoke mysql without any wrapper or something else
mysql_direct()
{
	# a bit dirty but it works properly for passwords with spaces, quotes and double quotes:
	if [ -n "$mysql_passwd" ]; then
		(
			export MYSQL_PWD="$mysql_passwd"
			$mysql_client $mysql_host $mysql_defaults $mysql_user $mysql_args "$@" 2>>"$product_log"
			rc=$?
			unset MYSQL_PWD
			return $rc
		)
	else
		$mysql_client $mysql_host $mysql_defaults $mysql_user $mysql_args "$@" 2>>"$product_log"
	fi
}

# Invoke mysql in raw mode
mysql_raw()
{
	mysql $mysql_args_raw "$@"
}

mysql_raw_anydb()
{
	mysql_anydb $mysql_args_raw "$@"
}

check_db()
{
	local inten="find $PRODNAME database"
	echo_try $inten

#	If database not exists
	if ! db_test_database "psa"; then
		p_echo "not found"
		# DB should be installable (schema file must be present)
		[ -f "$product_db_sql" ] || die "$inten: file not found: $product_db_sql"
		return 1
	fi

#	Database found, but version undefined
	if ! db_select "select val from misc where param='version'" || [ -z "$db_select_output" ]; then
		p_echo "DATABASE ERROR!!!"
		p_echo "Database $PRODNAME found, but version undefined"
		die $inten
	fi

	prev_db_version="$db_select_output"

# Now check psa database version
	p_echo "version is $prev_db_version"
	if [ "$prev_db_version" -le "011010" ] ; then
		prev_db_short_version=`expr ${prev_db_version} / 10`
	else
		# support upgrade on patch version
		prev_db_short_version="$prev_db_version"
	fi

	return 0
}

product_install_db_local()
{
	p_echo "===> Installing $PRODNAME database"

	db_install "$PRODNAME" --name "psa" --skip-when check_db --initializer product_db_initializer_local 		|| die "install initial $PRODNAME database"
}

product_db_initializer_local()
{
	product_db_installer

	# DB initialization specific to this machine
	p_echo "===> Initializing $PRODNAME database"

	db_init_fullhostname
	db_init_spamassassin
	db_init_default_locale
	db_init_updater_state
}

product_db_installer()
{
	# DB schema installation and initialization not bound to this machine
	mysql < "$product_db_sql" || die "Cannot initialize $PRODNAME database from $product_db_sql"
	set_db_version
}

db_init_fullhostname()
{
	local inten="insert Hostname into the MySQL database"
	local query="replace into misc (param,val) values('FullHostName','$fullhost')"
	db_do --inten "$inten" "$query"
}

db_init_spamassassin()
{
	detect_vz
	if [ "$PLESK_VZ" = "1" ]; then
		db_do "replace into ServiceNodeConfiguration (\`serviceNodeId\`, \`section\`, \`name\`, \`value\`) values (1, 'mailServer', 'spamfilter_max_children', '1')" # default is 5
	fi
}

db_init_default_locale()
{
	local default_locale default_locale_name default_locale_country

	p_echo 'Setting the default locale'
	set_default_locale_params

	db_do "replace into misc (param, val) values('def_locale', '$default_locale')"
	db_do "replace into locales(id) values('$default_locale')"
	db_do "alter table clients alter locale set default '$default_locale'"

	p_echo "The default locale is set to $default_locale ($default_locale_name, $default_locale_country)"
}

db_init_updater_state()
{
	# By default show updater UI in Panel
	local disable_updater="false"
	mysql <<EOF || die "init updater state"
		REPLACE INTO misc(param, val) VALUES('disable_updater', '$disable_updater');
EOF
}

set_db_version()
{
	# Install or update db version
	echo_try "set psa database version to $product_db_version"
	db_do "replace into misc (param,val) values('version', '$product_db_version')"
	suc
}

dump_db()
{
	local databases=""
	for db in "$@"; do
		if db_test_database "$db"; then
			databases="$databases $db"
		fi
	done

	[ -n "$databases" ] || return 1
	# https://bugs.mysql.com/bug.php?id=68376 (doesn't work on mysql < 5.5) :
	# local ignore_event_opts="--events --ignore-table=mysql.event"
	(
		export MYSQL_PWD="$mysql_passwd"
		${MYSQL_BIN_D}/mysqldump $mysql_user --quote-names $MYSQLDUMP_OPTIONS --databases $databases
		rc=$?
		unset MYSQL_PWD
		return $rc
	)
}

requires_configured_mysql()
{
	# Initialize and configure local MySQL server if it is not usable yet.
	# This function is for use in components that require MySQL, particularly on SN.

	set_mysqld_params

	if [ ! -s "$mysql_passwd_file" ]; then
		mysql_install_init
	else
		set_mysql_auth
	fi
}

uses_mysql_client()
{
	# Initialize local MySQL authentication parameters and test connection.
	# After that db_*() and mysql*() functions may be used.
	# Any MySQL-specific overrides (e.g., local mysql_db_name="whatever") should appear
	# *after* calling this function.
	# This function is for use in component installers that require access to MySQL DB.

	set_mysql_server_params
	set_mysql_client_params
	set_mysql_auth
}
# -- start of mysqlnd old passwords upgrade --

# The reason why this code is not in the db/upgrade_db_1100.sh file is the following:
# It turned out that during the upgrade from 8.6 to 11.0 sw-engine may be installed to
# satisfy bootstrapper's dependency; new sw-engine with the mysqlnd extenstion refused to work 
# with the passwords in the old format. So passwords upgrade should be done prior to the database upgrade.
# This is vital as there are PHP-scripts (working with the database) invoked during the upgrade.

upgrade_mysqlnd_old_passwd()
{
	# It's probably better to extract all child functions into .in file
	run_configure_mysql_funcs configure_mysql_disable_old_passwords

	db_test "SELECT length(PASSWORD('test'))" '$1 == 41' || die "disable old_passwords=1 in my.cnf"
	check_mysql_users_with_old_passwd_format || return 0

	inten="upgrade MySQL password format"
	reset_admin_passwd || die "$inten for admin"
	reset_sitebuilder_passwd || die "$inten for Sitebuilder user"
	reset_apsc_passwd || die "$inten for APS controller user"
	reset_horde_passwd || die "$inten for Horde webmail user"
	reset_atmail_passwd || die "$inten for AtMail webmail user"
	reset_mysql_dbusers_passwd || die "$inten for Panel DB users"
	reset_phpmyadmin_passwd || die "$inten for phpMyAdmin control user"
	reset_siteeditor_passwd || die "$inten for SiteEditor user"

	check_mysql_users_with_old_passwd_format && warn_mysql_old_passwd_users && return 1 || :
}

check_mysql_users_with_old_passwd_format()
{
	local mysql_db_name=mysql
	local user="$1"
	local query="SELECT User FROM user WHERE length(Password) != 41"
	if [ -n "$user" ]; then
		query="${query} AND User = '$user'"
	fi
	db_test "${query} LIMIT 1" '1'
}

reset_user_password()
{
	true mysql_quote_string

	local user="$1"
	local passwd="$2"
	local hosts="${3:-localhost}"

	local inten="reset database user password for '${user}@${host}'"
	echo_try "${inten}"

	if [ "$hosts" = "127.0.0.1" -o "$hosts" = "::1" ] ; then
		hosts="${hosts} localhost"
	fi
	hosts="${hosts} %"

	for host in $hosts; do
		if echo "SET PASSWORD FOR '$user'@'$host' = PASSWORD('`mysql_quote_string $passwd`')" | mysql_direct mysql; then
			suc
			return 0
		fi
	done

	warn "$inten - database user does not exist."
	return 1
}

is_local_mysql_instance()
{
	local host="$1"
	[ -z "$host" -o "$host" = "localhost" -o "$host" = "127.0.0.1" -o "$host" = "::1" ]
}

check_db_accessible()
{
	local mysql_db_name="$1"
	local mysql_user="--user=$2"
	local mysql_passwd="$3"

	#INFO: we should not die here
	echo "" | mysql_direct 
}

reset_sitebuilder_passwd()
{
	local no_force="$1"
	local sb_conf="/usr/local/sb/config"
	[ -f "$sb_conf" ] || return 0

	local sb_host=`    get_ini_conf_var "$sb_conf" "database" "host"`
	local sb_username=`get_ini_conf_var "$sb_conf" "database" "username"`
	local sb_password=`get_ini_conf_var "$sb_conf" "database" "password"`
	local sb_dbname=`  get_ini_conf_var "$sb_conf" "database" "dbname"`

	is_local_mysql_instance "$sb_host" || return 0

	if [ -n "$no_force" ]; then
		check_db_accessible "$sb_dbname" "$sb_username" "$sb_password" && return 0 || :
	fi

	if reset_user_password "$sb_username" "$sb_password" "$sb_host"; then
		check_db_accessible "$sb_dbname" "$sb_username" "$sb_password"
	else
		# https://jira.plesk.ru/browse/PPP-4676
		return 0
	fi
}

reset_siteeditor_passwd()
{
	local no_force="$1"

	local se_conf="/usr/local/siteeditor/config"
	[ -f "${se_conf}" ] || return 0

	local se_host=`    get_ini_conf_var "${se_conf}" "database" "host"`
	local se_username=`get_ini_conf_var "${se_conf}" "database" "username"`
	local se_password=`get_ini_conf_var "${se_conf}" "database" "password"`
	local se_dbname=`  get_ini_conf_var "${se_conf}" "database" "dbname"`

	is_local_mysql_instance "$se_host" || return 0

	if [ -n "$no_force" ]; then
		! check_db_accessible "$se_dbname" "$se_username" "$se_password" || return 0
	fi

	reset_user_password "${se_username}" "${se_password}" "${se_host}" && 	check_db_accessible "$se_dbname" "$se_username" "$se_password"
}

reset_apsc_passwd()
{
	check_apsc_installed || return 0
	read_apsc_connection_params_db
	check_mysql_users_with_old_passwd_format "$apsc_db_user" || return 0

	is_local_mysql_instance "$apsc_db_host" || return 0

	if reset_user_password "$apsc_db_user" "$apsc_db_passwd" "$apsc_db_host"; then
		check_db_accessible "$apsc_db_name" "$apsc_db_user" "$apsc_db_passwd" || return 1
	fi
}

get_webmail_config_param()
{
	local webmail_config="$1"
	local config_var="$2"
	local param="$3"
	local default_val="$4"
	local result=`sw_engine_pleskrun -n -r 'include "'$webmail_config'"; echo $'${config_var}${param}';' 2>/dev/null`
	[ -n "$result" ] || result="$default_val"
	echo "$result"
}

get_horde_config_param()
{
	get_webmail_config_param "/usr/share/psa-horde/config/conf.php" "conf" "$1" "$2"
}

get_roundcube_config_param()
{
	#INFO:  this function now is able to extract settings from config.inc.php, now it just db_dsnw;
	#	if need another parameters need to reimplement this function - parse defaults.inc.php, parse config.inc.php and merge data
	get_webmail_config_param "/usr/share/psa-roundcube/config/config.inc.php" "config" "$1"
}

get_atmail_config_param()
{
	get_webmail_config_param "/var/www/atmail/libs/Atmail/Config.php" "pref" "$1" "$2"
}

reset_horde_passwd()
{
	local no_force="$1"

	set_horde_params

	local horde_shadow="$horde_passwd_file"
	[ -f "$horde_shadow" ] || horde_shadow="/etc/psa/.webmail.shadow"
	local hor_host=`  get_horde_config_param "['sql']['hostspec']" "localhost"`
	local hor_user=`  get_horde_config_param "['sql']['username']" "horde"`
	local hor_dbname=`get_horde_config_param "['sql']['database']" "horde"`
	local hor_passwd=`get_horde_config_param "['sql']['password']"`
	[ -n "$hor_passwd" ] || hor_passwd=`cat "$horde_shadow" 2>/dev/null`
	[ -n "$hor_passwd" -a -n "$hor_user" -a -n "$hor_dbname" ] || return 0

	is_local_mysql_instance "$hor_host" || return 0

	if [ -n "$no_force" ]; then
		! check_db_accessible "$hor_dbname" "$hor_user" "$hor_passwd" || return 0
	fi

	reset_user_password "$hor_user" "$hor_passwd" "$hor_host" && 	check_db_accessible "$hor_dbname" "$hor_user" "$hor_passwd"
}

parse_mysql_dsn_get_user()
{
	echo $1 | sed -e 's|^mysql://\([^:]\+\):\([^@]\+\)@\([^/]\+\)/\(.*\)$|\1|'
}

parse_mysql_dsn_get_passwd()
{
	echo $1 | sed -e 's|^mysql://\([^:]\+\):\([^@]\+\)@\([^/]\+\)/\(.*\)$|\2|'
}

parse_mysql_dsn_get_host()
{
	echo $1 | sed -e 's|^mysql://\([^:]\+\):\([^@]\+\)@\([^/]\+\)/\(.*\)$|\3|'
}

parse_mysql_dsn_get_db()
{
	echo $1 | sed -e 's|^mysql://\([^:]\+\):\([^@]\+\)@\([^/]\+\)/\(.*\)$|\4|'
}

reset_roundcube_passwd()
{
	local no_force="$1"

	local roundcube_dsn=`get_roundcube_config_param "['db_dsnw']"`
	local roundcube_user=`parse_mysql_dsn_get_user "$roundcube_dsn"`
	local roundcube_passwd=`parse_mysql_dsn_get_passwd "$roundcube_dsn"`
	local roundcube_host=`parse_mysql_dsn_get_host "$roundcube_dsn"`
	local roundcube_db=`parse_mysql_dsn_get_db "$roundcube_dsn"`

	[ -n "$roundcube_passwd" -a -n "$roundcube_user" -a -n "$roundcube_db" ] || return 0

	is_local_mysql_instance "$roundcube_host" || return 0

	if [ -n "$no_force" ]; then
		! check_db_accessible "$roundcube_db" "$roundcube_user" "$roundcube_passwd" || return 0
	fi

	reset_user_password "$roundcube_user" "$roundcube_passwd" "$roundcube_host" && 	check_db_accessible "$roundcube_db" "$roundcube_user" "$roundcube_passwd"
}

reset_atmail_passwd()
{
	local no_force="$1"

	local atmail_shadow="/etc/psa-webmail/atmail/.atmail.shadow"
	local at_host=`  get_atmail_config_param "['sql_host']" "localhost"`
	local at_user=`  get_atmail_config_param "['sql_user']" "atmail"`
	local at_dbname=`get_atmail_config_param "['sql_table']" "atmail"`
	local at_passwd=`get_atmail_config_param "['sql_pass']"`
	[ -n "$at_passwd" ] || at_passwd=`cat "$atmail_shadow" 2>/dev/null`
	[ -n "$at_passwd" -a -n "$at_user" -a -n "$at_dbname" ] || return 0

	is_local_mysql_instance "$at_host" || return 0

	if [ -n "$no_force" ]; then
		! check_db_accessible "$at_dbname" "$at_user" "$at_passwd" || return 0
	fi

	reset_user_password "$at_user" "$at_passwd" "$at_host" && 	check_db_accessible "$at_dbname" "$at_user" "$at_passwd"
}

reset_mysql_dbusers_passwd()
{
	local mysql_db_name=mysql
	db_do --inten "Convert MySQL DB users passwords to new format" 		"UPDATE mysql.user m 
				INNER JOIN psa.db_users u ON (m.User = u.login) 
				INNER JOIN psa.accounts a ON (u.account_id = a.id) 
				INNER JOIN psa.data_bases d ON (u.db_id = d.id) 
		SET m.Password = PASSWORD(a.password) 
		WHERE d.type = 'mysql' AND a.type = 'plain'; 
		FLUSH PRIVILEGES;"
}

reset_phpmyadmin_passwd()
{
	local no_force="$1"

	local pmahost="localhost"
	local controlpass pmadb
	set_phpmyadmin_params

	local pma_config="$PHPMYADMIN_CONFIG.save"
	[ -f "$pma_config" ] || pma_config="$PHPMYADMIN_CONFIG"
	pma__set_values_from_config "$pma_config" 'controluser' 'controlpass' 'pmadb'

	[ -n "$controlpass" -a -n "$controluser" -a -n "$pmadb" ] || return 0

	if [ -n "$no_force" ]; then
		check_db_accessible "$pmadb" "$controluser" "$controlpass" || return 0
	fi

	reset_user_password "$controluser" "$controlpass" "$pmahost" && 	check_db_accessible "$pmadb" "$controluser" "$controlpass"
}

warn_mysql_old_passwd_users()
{
	local mysql_db_name=mysql
	local query="SELECT concat(User, '@', Host) FROM user WHERE length(Password) != 41"

	if [ -n "$controluser" ]; then
		# There is an error in old phpMyAdmin upgrade routine: DB was recreated every time
		query="$query AND (User not like 'pma\_%' OR User = '$controluser')"
	fi

	query="$query AND (User != 'debian-sys-maint')"

	db_select "$query"
	local problem_users="$db_select_output"

	[ -n "$problem_users" ] || return 0
	report_problem "Passwords for some MySQL users were not upgraded." 		"All MySQL DB users that may be used by Panel MUST have passwords in new 41-character format in order for Panel to function properly." 		"Currently following users have passwords in old format: `echo -n $problem_users | sed -e 's/ /, /g'`." 		"Please ensure that MySQL server is not forced to use old password hashing algorithm and upgrade password format manually for users from the list above. Consult with MySQL server manual on detailed steps." 		"You may skip certain users if you are sure they weren't created and aren't used by Panel."
}

# -- end of mysqlnd old passwords upgrade --

# Reset mysqld authentication in case we are unable to guess password

create_reset_mysql_auth_sql_file()
{
	true mysql_quote_string

	(
		umask 0027
		local sql="`mktemp /tmp/plesk-reset-mysql-auth.XXXXXX.sql`"
		cat > "$sql" <<-EOT
			-- WARNING: each statement in this file MUST be on one line, no comments on statement lines.
			-- DB must be selected to create procedures.
			USE mysql;

			-- Remove password strength validation.
			CREATE PROCEDURE plesk_uninstall_validate_password() BEGIN IF EXISTS (SELECT * FROM mysql.plugin WHERE name = 'validate_password') THEN UNINSTALL PLUGIN validate_password; END IF; END;
			CALL plesk_uninstall_validate_password();
			DROP PROCEDURE IF EXISTS plesk_uninstall_validate_password;

			-- Create or reset admin superuser.
			GRANT ALL PRIVILEGES ON *.* TO 'admin'@'localhost' IDENTIFIED BY '`mysql_quote_string "$admin_passwd"`' WITH GRANT OPTION;
			FLUSH PRIVILEGES;

			-- Shut down mysqld.
			SHUTDOWN;

			EOT
		chgrp "$mysqld_group" "$sql" >&2
		chmod 640 "$sql" >&2
		echo "$sql"
	)
}

try_reset_mysql_auth()
{
	# Don't try this more than once. It won't succeed anyway.
	[ -z "$PLESK_INSTALLER_TRIED_MYSQL_AUTH_RESET" ] || return 0
	PLESK_INSTALLER_TRIED_MYSQL_AUTH_RESET="yes"

	local inten="reset authentication for MySQL server"
	echo_try "$inten"

	pleskrc mysql stop

	get_admin_passwd

	# Run mysqld in "reset authentication" mode.
	p_echo "Starting mysqld in authentication reset mode."
	local init_sql="`create_reset_mysql_auth_sql_file`"
	PATH="$PATH:/usr/libexec:/libexec" mysqld --user "$mysqld_user" --init-file "$init_sql" >> "$product_log" 2>&1 &
	local pid="$!"

	# Normally the command above should stop by itself. Wait for it.
	local tries=60 i=
	pnnl_echo "Waiting for mysqld to stop ($tries seconds at most)"
	for i in `sequence $tries`; do
		kill -0 "$pid" 2>/dev/null || break
		pnnl_echo "."
		sleep 1
	done
	pnnl_echo " "

	# If it didn't stop, likely error occurred before SHUTDOWN query. Request shutdown via signal.
	kill -s TERM "$pid" 2>/dev/null 		&& p_echo "We had to send mysqld signal to shutdown. This likely means that authentication reset failed (see mysqld log for details, search for 'init_file')." 		|| p_echo "Looks like mysqld stopped normally."

	# SIGTERM might not have been honored by mysqld, but I've seen no such behavior.
	pnnl_echo "Waiting for mysqld return code (if this hangs, try 'kill -9 $pid')... "
	wait "$pid"
	local rc="$?"
	if [ "$rc" -ne 0 ]; then
		p_echo "mysqld failed with code $rc, authentication reset failed (see mysqld log for details)."
	else
		suc
	fi

	rm -f "$init_sql"

	pleskrc mysql restart
}
# -*- vim:ft=sh

# MySQL service action handlers

true mysql_start_linux_suse
mysql_start_linux_suse()
{
	local rc
	
	inten="start service mysql"
	echo_try "$inten"

	service_ctl start $mysql_service mysql
	rc="$?"

	# bug 52690. MySQL init script reports failure if protected mysqld is running (true for SuSE >= 11.3)
	if [ "$rc" -ne 0 ]; then
		local mysqld_bin="/usr/sbin/mysqld"
		killall -TERM mysqld >> $product_log 2>&1
		if [ -x "$mysqld_bin" ]; then
			for i in 2 4 8 16 32; do
				get_pid "$mysqld_bin" false
				local pid="$common_var"
				if test "$pid" -eq 1; then
					break
				fi
				killall -TERM mysqld >> $product_log 2>&1
				sleep $i
			done
		fi
		service_ctl start $mysql_service mysql
		rc="$?"
	fi

	[ "$rc" -eq 0 ] && suc || warn "$inten"
	return $rc
}

###	FIXME: probably need var service_restart warn
true mysql_stop
mysql_stop()
{
	local op_result i

	inten="stop MySQL server"
	echo_try $inten

	service_ctl stop $mysql_service mysql
	op_result=$?

	if [ "X$linux_distr" = "Xdebian" ]; then
		# Debian has well designed mysql stopping code
		[ "$op_result" -eq 0 ] || die $inten
		suc
		return 0
	fi

	for i in 2 4 6 8 16; do
		if ! mysql_status ; then
			suc
			return 0
		fi

		# I just want to be sure that mysql really stopped
		killall -TERM mysqld mysql safe_mysqld mysqld_safe >> $product_log 2>&1

		sleep $i
	done

	die "$inten"
}

true mysql_status
mysql_status()
{
	local file

    #Check with native script first
	#debian script always return 0. bug #111825
	[ "X$linux_distr" = "Xdebian" ] && msqld_status_supported="no"
	
	if [ -z "$msqld_status_supported" ]; then
		msqld_status_supported="yes"
	fi

	if [ "$msqld_status_supported" = "yes" ]; then
		service_ctl status $mysql_service mysql && return 0
	fi

	if [  "$msqld_status_supported" = "no" ]; then
		# MySQL AB packages
		file="/usr/sbin/mysqld"
	fi

    if [ -x "$file" ]; then
		#standard build and debian
		get_pid "$file" false
		pid=$common_var
		if test "$pid" -ne 1; then
			echo "$file (pid $pid) is running..." >>$product_log 2>&1
			return 0
		else
			echo "$file is stopped" >>$product_log 2>&1
			return 1
		fi
	fi

	return 1
}


__tune_vz_mysql_sig_handler() {
	local rc=$?
	trap - EXIT
	rm -f -- "$1"
	exit $rc
}

__tune_vz_mysql()
{
	local myCnf myCnfTemp ||:

	myCnf=$1
	if test -z "$myCnf"; then
		warn "E:usage: $PROG /path/to/my.cnf"
		return 1
	fi

	if ! test -s "$myCnf" ; then
		warn "E: $myCnf doesn't exist or empty"
		return 100
	fi
	
	myCnfTemp=`mktemp ${myCnf}.XXXXXXX`
	if test $? -ne 0 ; then
		warn "E:Unable to create a temporary configuration file"
		return 2
	fi
	
	$cp_preserve $myCnf $myCnfTemp || {
		warn "E:Unable to create a temporary configuration file"
		return 2
	}

	trap "__tune_vz_mysql_sig_handler \"$myCnfTemp\""  HUP PIPE INT TERM EXIT

	local rc ||:

	local db_select_output
	local mysql_db_name=mysql
	db_select "SHOW VARIABLES LIKE 'have_innodb'"
	local have_innodb=`get_narg 2 ${db_select_output}`
	[ -n "$have_innodb" ] || have_innodb="NO"

	db_select "SHOW VARIABLES LIKE 'have_bdb'"
	local have_bdb=`get_narg 2 ${db_select_output}`
	[ -n "$have_bdb" ] || have_bdb="NO"

	cat "$myCnf" | gawk 				-v have_innodb="${have_innodb}" 				-v have_bdb="${have_bdb}" 				'
function getsize(sizestr)
{
  if (match(sizestr, "[^[:digit:]]")) {
    size=substr(sizestr, 0, RSTART);
    mult=substr(sizestr, RSTART, 1);
    if ("K" == mult) {
      mult=1024;
    } else if ("M" == mult) {
      mult=1024*1024;
    } else if ("G" == mult) {
      mult=1024*1024*1024;
    }
    size = size * mult;
  } else {
    size = sizestr;
  }
  return size;
}

# add unspecified variables
function finalize_section()
{
  if (have_bdb && 0 == skipbdb) {
    printf("skip-bdb\n\n");
    changed=1;
  }
  if (have_innodb && 0 == ibp_size) {
    print "innodb_buffer_pool_size=2M"
    changed=1;
  }
  if (have_innodb && 0 == iamp_size) {
    print "innodb_additional_mem_pool_size=500K"
    changed=1;
  }
  if (have_innodb && 0 == ilb_size) {
    print "innodb_log_buffer_size=500K"
    changed=1;
  }
  if (have_innodb && 0 == thread_conc_set) {
    print "innodb_thread_concurrency=2"
    changed=1;
  }
}

# decrease value of "varname" if it exceed the upper bound limit.
# If value fit (less or equal), leave it unmodified.
function adjustSetting(varstr, varname, minsize, sizestr)
{
  if (match(varstr,"^[[:space:]]*(set-variable[[:space:]]*=[[:space:]]*)?" varname "[[:space:]]*=[[:space:]]*")) {
    if (have_innodb) {
      size = getsize(substr(varstr, RSTART+RLENGTH));
      if (minsize < size) {
        changed = 1;
        print "#", $0;
        print varname "=" sizestr;
      } else {
        print $0
      }
    } else {
      # we do not have innodb so we need to disable this setting
      print "#", $0;
      changed = 1;
    }
    return 1;
  }
  return 0;
}

function reset_variables()
{
  skipbdb=0;   # already have skip-bdb
  ibp_size=0;  # already set innodb_buffer_pool_size
  iamp_size=0; # already set innodb_additional_mem_pool_size
  ilb_size=0;  # already set innodb_log_buffer_size
  thread_conc_set=0; # -*- innodb_thread_concurrency
}

BEGIN {
  insect=0;    # within one of server sections
  changed=0;   # already changed smth
  have_bdb = (tolower(have_bdb) == "yes") ;
  have_innodb = (tolower(have_innodb) == "yes") ;
  reset_variables()
}
/^[[:space:]]*\[.+\]/ {
  if (0 != insect) {
# need to close a previous section
    finalize_section();
    # section is ended. Reset variables
    reset_variables()
    insect    = 0;
  }
  if ( match($0, "^[[:space:]]*\\[(mysqld|server|mysqld_safe|safe_mysqld)\\]")) {
    insect = 1;
#    print "# in server section";
  }
}

(match($0, "^[[:space:]]*(set-variable[[:space:]]*=)?[[:space:]]*") && 0 != insect) {
  varstr=substr($0, RSTART+RLENGTH);
  if (adjustSetting(varstr, "innodb_buffer_pool_size", 2 * 1024 * 1024, "2M")) { ibp_size=1; next; }
  if (adjustSetting(varstr, "innodb_additional_mem_pool_size", 500 * 1024, "500K")) { iamp_size=1; next; }
  if (adjustSetting(varstr, "innodb_log_buffer_size", 500 * 1024, "500K")) { ilb_size=1; next; }
  if (adjustSetting(varstr, "innodb_thread_concurrency", 2, "2")) { thread_conc_set=1; next; }
}

(adjustSetting($0, "innodb_buffer_pool_size", 2 * 1024 * 1024, "2M")) { ibp_size=1; next; }
(adjustSetting($0, "innodb_additional_mem_pool_size", 500 * 1024, "500K")) { iamp_size=1; next; }
(adjustSetting($0, "innodb_log_buffer_size", 500 * 1024, "500K")) { ilb_size=1; next; }
(adjustSetting($0, "innodb_thread_concurrency", 2, "2")) { thread_conc_set=1; next; }

/^[[:space:]]*skip-bdb[[:space:]]*(#.*)?$/ { skipbdb=1; }

{ print $0; }

END {
  if (0 != insect) {
    finalize_section();
  }

  if (0 == changed) {
    exit 110;
  }
}
' > "$myCnfTemp"
	rc=$?

	if test $rc -eq 110; then
		# nothing was changed
		p_echo "I: nothing to change in $myCnf"
		return 5
	fi

	if test $rc -ne 0; then
		warn "E: unable to process ${myCnf}: rc=$rc"
		return 2
	fi
	
	local myCnfBak ||:
	myCnfBak="${myCnf}.bak"

	if ! ln -f "$myCnf" "$myCnfBak" ; then
		warn "E: Unable to create a backup file"
		return 3
	fi

	if ! mv -f "$myCnfTemp" "$myCnf"; then
		warn "E: Unable to overwrite the configuration file"
		return 4
	fi
	
	trap - EXIT
	return 0
}

tune_vz_mysql()
{
	# check if we in VZ environment
	detect_vz
	test "x$PLESK_VZ" = "x1" || return 0;
	# run in subshell to isolate TRAP usage
	( __tune_vz_mysql "$1" )
	local rc="$?"
	! [ "$rc" -eq 0 ] || need_mysql_restart=yes		# Configuration file was updated
	! [ "$rc" -eq 5 ] || rc=0						# Everything is OK, no changes are required
	return "$rc"
}
#-*- vim:syntax=sh

set_named_params()
{
	# set up default values
	bind_UID=53
	bind_GID=53
	bind_user="bind";
	bind_group="bind";

	# get UID of named user, if exists
	if get_userID $bind_user; then
		bind_UID=$common_var;
	fi

	# get GID of named group, if exists
	if get_groupID $bind_group; then
		bind_GID=$common_var;
	fi

	# path to directory of internal named
	NAMED_ROOT_D="${PRODUCT_ROOT_D:?'PRODUCT_ROOT_D is undefined'}/named"

	# path to directory of named pid file
	bind_run="${NAMED_RUN_ROOT_D:?'NAMED_RUN_ROOT_D is undefined'}/var/run/named"

	named_service="bind9"
	named_log="/dev/null"

	# path to named config file
	named_conf="/etc/named.conf"
	rndc_conf="/etc/rndc.conf"
	rndc_namedb_conf="/etc/namedb/rndc.conf"
	rndc_bind_conf="/etc/bind/rndc.conf"

	#140025. Restrict CPU cores for Bind
	bind_number_of_workers=2
}

set_named_accounts()
{
      p_echo " Checking for the system groups and users necessary for named (bind)..."
      group_op "$bind_group" "$bind_GID" false
      user_op "$bind_user" "$bind_UID" "$bind_group" "Bind Sandbox" "$dummy_home" "$dummy_shell" false
}

set_named_perms()
{
	inten="set up default permissions"
	echo_try $inten

	# set up permissions of named files directory, if exists
	if [ -d $NAMED_ROOT_D ]; then
		chown -R 0:0 "$NAMED_ROOT_D" >> "$product_log" 2>&1 || die $inten
		chmod -R 444 "$NAMED_ROOT_D" || die $inten
		chmod 555 "$NAMED_ROOT_D" "$NAMED_ROOT_D"/* >> "$product_log" 2>&1 || die $inten

		chmod 555 "$NAMED_ROOT_D"/bin/* >> "$product_log" 2>&1 || die $inten
		chmod 555 "$NAMED_ROOT_D"/include/*  >> "$product_log" 2>&1 || die $inten
		chmod 555 "$NAMED_ROOT_D"/man/* "$NAMED_ROOT_D"/sbin/* >> "$product_log" 2>&1 || die $inten
	fi

	# set up permissions of named run-root directory
	[ -d $bind_run ] || mkdir -p $bind_run
	chown -R $bind_user $bind_run >> "$product_log" 2>&1 || die $inten
	chmod 1755 $bind_run >> "$product_log" 2>&1 || die $inten

	chown ${bind_user}:${bind_group} "${NAMED_RUN_ROOT_D}/var" || die $inten
	chmod 755 "${NAMED_RUN_ROOT_D}/var" >> "$product_log" 2>&1 || die $inten
	chown ${bind_user}:${bind_group} "${NAMED_RUN_ROOT_D}/var/run" || die $inten
	chmod 755 "${NAMED_RUN_ROOT_D}/var/run" >> "$product_log" 2>&1 || die $inten

	suc
}

sysconfig_named_debian()
{
	local named_sysconf_file named_sysconf_files config_rebuild

	named_sysconf_files=/etc/default/bind9
	# use .systemd syffix for systemd specific sysconfigs
	named_sysconf_files+=" /etc/default/bind9.systemd"

	for named_sysconf_file in $named_sysconf_files; do
		config_rebuild=0

		if [ -f "${named_sysconf_file}" ]; then

			# check presence of required run-root option
			cat ${named_sysconf_file} | sed -e 's|#.*$||g' | grep -E "^[[:space:]]*OPTIONS\=.*[[:space:]]\-t[[:space:]]*${NAMED_RUN_ROOT_D}" > /dev/null 2>&1
			[ $? -eq 0 ] || config_rebuild=1

			# check presence of required config file option
			cat ${named_sysconf_file} | sed -e 's|#.*$||g' | grep -E "^[[:space:]]*OPTIONS\=.*[[:space:]]\-c[[:space:]]*${named_conf}" > /dev/null 2>&1
			[ $? -eq 0 ] || config_rebuild=1

			# check presence of bind user option
			cat ${named_sysconf_file} | sed -e 's|#.*$||g' | grep -E "^[[:space:]]*OPTIONS\=.*[[:space:]]\-u[[:space:]]*${bind_user}" > /dev/null 2>&1
			[ $? -eq 0 ] || config_rebuild=1

			# check presence of workers number
			cat ${named_sysconf_file} | sed -e 's|#.*$||g' | grep -E "^[[:space:]]*OPTIONS\=.*[[:space:]]\-n[[:space:]]" > /dev/null 2>&1
			[ $? -eq 0 ] || config_rebuild=1

		else
			config_rebuild=1
		fi

		if [ 0${config_rebuild} -gt 0 ]; then

			[ ! -f ${named_sysconf_file} ] || mk_backup ${named_sysconf_file} mv f

			if expr match "${named_sysconf_file}" ".*\.systemd" >/dev/null 2>&1; then
				# For systemd we don't use ${OPTIONS}, and add -f (foreground)
				echo "OPTIONS=\"-f -t ${NAMED_RUN_ROOT_D}  -c ${named_conf} -u ${bind_user} -n ${bind_number_of_workers}\"" > ${named_sysconf_file}
			else
				echo "OPTIONS=\"\${OPTIONS} -t ${NAMED_RUN_ROOT_D}  -c ${named_conf} -u ${bind_user} -n ${bind_number_of_workers}\"" > ${named_sysconf_file}
			fi
		fi
	done
}

sysconfig_named()
{
	register_service "$named_service"
	sysconfig_named_debian

	return 0
}

configure_named_ipv6 ()
{
	realconf=`readlink $named_conf`
	if [ x"$realconf" = x"" ]; then
		realconf=$named_conf
	fi
        tmp_conf=`mktemp $realconf.XXXXXX`

        if os_supports_ipv6; then
                listen_cmd='/options {/a		listen-on-v6 { any; };'
        else
                listen_cmd=""
        fi
        sed -e "/listen-on-v6/d" -e "$listen_cmd" $realconf > $tmp_conf
	chmod 644 $tmp_conf
        mv $tmp_conf $realconf
}

apparmor_named_install()
{


	apparmor_install_profile "named"
}

install_named()
{
	install_named_own

	pleskrc named stop
	configure_named_ipv6

	if [ "$do_upgrade" != 1 ]; then
		"$UTILS_DIR/dnsmng" --dns-recursion localnets
	fi

	pleskrc named start
}

debian_fix_named_iniscript()
{
	start_script="/etc/init.d/bind9"
	if [ -f $start_script ]; then
		# replace pidfile path to the one in chroot; ensure named is dead after 'rndc stop'
		# be carefull not to make replacements more than one time
		sed -e 's|/var/run/bind/run/named.pid|/var/named/run-root/var/run/named/named.pid|g' 			-e 's|\([^a-zA-Z0-9_.]\)/var/run/named/named.pid|\1/var/named/run-root/var/run/named/named.pid|g' 			-e '/killall\|rndc stop -p/!    s|\(/usr/sbin/rndc stop\)|\1 ; sleep 1 ; killall -9 named|g' 			-e '/killall/!                  s|\(/usr/sbin/rndc stop -p\s*\)| ( \1 ; sleep 1 ; killall -9 named ) 2>/dev/null |g' 			< "$start_script" > "$start_script.tmp" &&
		mv -f "$start_script.tmp" "$start_script" &&
		chmod 755 "$start_script"
	fi
}

install_named_own()
{
	p_echo
	p_echo
	cd "$PRODUCT_ROOT_D" || die "cd $PRODUCT_ROOT_D"
	p_echo "===> Installing and configuring Named Server"
	p_echo

	set_named_params
	set_named_accounts

	get_pid named false
	old_named_pid=$common_var
	if [ $old_named_pid -gt 1 ]; then
		p_echo "Stop running named daemon"
		pleskrc named stop
	fi

	if [ ! -d "$NAMED_RUN_ROOT_D" ]; then
		inten="create named run-root directory"
		echo_try $inten
		mkdir -p "$NAMED_RUN_ROOT_D" && suc || die "$inten"
	fi

	# it is needed write localhost.rev?
	exclude_localhost_rev=
	if [ -f "$NAMED_RUN_ROOT_D/var/localhost.rev" ]; then
		exclude_localhost_rev="--exclude var/localhost.rev"
	fi

	inten="untar named run-root directory structure"
	echo_try $inten
	$tar xpf "$PRODUCT_ROOT_D/tmp/run-root.tar" --no-same-owner -C "$NAMED_RUN_ROOT_D" $exclude_localhost_rev || die "$inten"
	(
		cd $NAMED_RUN_ROOT_D
		mkdir -m 755 -p dev
		
		
		mknod -m 644 dev/random c 1 8
		mknod -m 644 dev/urandom c 1 9
		chown root:root dev/random dev/urandom

	) >>"$product_log" 2>&1
	relabel_named_directories
	suc

	# cp named.conf if not exists or not contain string 'directory "/var";'
	grep_res=`egrep 'directory[[:space:]]*"/var";' "${NAMED_RUN_ROOT_D}${named_conf}" 2>/dev/null`
	if [ $? -ne 0 ]; then
		inten="create named.conf"
		echo_try "$inten"
		cp "${NAMED_RUN_ROOT_D}${named_conf}.default" "${NAMED_RUN_ROOT_D}${named_conf}" && suc || die "$inten"
	fi

	copy_named_localtime

	mk_backup "$named_conf" mv f

	ln -s "$NAMED_RUN_ROOT_D$named_conf" "$named_conf"

	if [ -d /etc/namedb ]; then
		mk_backup "$rndc_namedb_conf" mv f
		ln -s "$NAMED_RUN_ROOT_D$rndc_conf" "$rndc_namedb_conf"

		[ -d "$NAMED_RUN_ROOT_D/etc/namedb" ] || mkdir "$NAMED_RUN_ROOT_D/etc/namedb"

		[ "X$NAMED_RUN_ROOT_D$rndc_conf" != "X$NAMED_RUN_ROOT_D$rndc_namedb_conf" ] && 		    [ ! -f $NAMED_RUN_ROOT_D$rndc_namedb_conf ] && 		    ln -s "$NAMED_RUN_ROOT_D$rndc_conf" "$NAMED_RUN_ROOT_D$rndc_namedb_conf"
	fi

# make named startup script happy

	# Debian
	if [ -d /etc/bind ]; then
		mk_backup "$rndc_bind_conf" mv f
		ln -s "$NAMED_RUN_ROOT_D$rndc_conf" "$rndc_bind_conf"
		[ -d "$NAMED_RUN_ROOT_D/etc/bind" ] || mkdir "$NAMED_RUN_ROOT_D/etc/bind"
		ln -s "$NAMED_RUN_ROOT_D$rndc_conf" "$NAMED_RUN_ROOT_D$rndc_bind_conf"
		debian_fix_named_iniscript

	fi

	mk_backup "$rndc_conf" mv f
	ln -s "$NAMED_RUN_ROOT_D$rndc_conf" "$rndc_conf"


	#cd $NAMED_RUN_ROOT_D/var || die "cd $NAMED_RUN_ROOT_D/var"
	#sh make-localhost "$host" "$domain" "$fullhost" || die "sh make-localhost $host $domain $fullhost"

	set_named_perms
	sysconfig_named
}

copy_named_localtime()
{
	inten="copy localtime file to named sandbox"

	cp -f /etc/localtime "$NAMED_RUN_ROOT_D/etc/localtime"
	chown -R 0:0 "$NAMED_RUN_ROOT_D/etc/localtime" >> "$product_log" 2>&1 || die $inten
	chmod -R 444 "$NAMED_RUN_ROOT_D/etc/localtime" || die $inten
}


true named_status_linux_debian
named_status_linux_debian()
{
    get_pid "/usr/sbin/named" false
    local pid=$common_var
    if test "$pid" -ne 1; then
# running
		return 0
    fi
    return 1
}
# vim:ft=sh

upgrade_named_restrict_number_of_workers_11_5_29()
{
	set_named_params
	sysconfig_named
	! pleskrc named status || pleskrc named restart
}

pam_params()
{
   local calltype
   calltype=$1
   initial_conf
   local product="psa"
   pam_config_dir="/opt/psa/pam_plesk_config"
   [ "x" = "x1" ] && pam_extra_args="-m /lib/security/pam_plesk.so"
   if [ "x$calltype" != "x1" ]; then # install, not uninstall
       [ "x" = "x1" ] && pam_extra_args="$pam_extra_args --fixup-ftpd"
       [ "x" = "x1" ] && pam_extra_args="$pam_extra_args --fixup-des"
   fi
   pam_inst_script="pam_plesk_install"
   pam_uninst_script="pam_plesk_remove"

   pam_extra_args="$pam_extra_args --fixup-systemd"
   return 0
}

install_pam()
{
  inten="install PAM module"
  pam_params
  echo_try "$inten"

  "$pam_config_dir/$pam_inst_script" -qq 	--install-log "$product_log" 	--problem-log "$product_problems_log" $pam_extra_args
 
  if [ "X$?" = "X0" ]; then
     suc
  else
     warn "install PAM module not successed" 
  fi

  return 0
}
phpmyadmin_config_backup()
{
	set_phpmyadmin_params
	if [ -f "$PHPMYADMIN_CONFIG" ]; then
		cp -f "$PHPMYADMIN_CONFIG" "$PHPMYADMIN_CONFIG.save"
		set_ac root root 0640 "$PHPMYADMIN_CONFIG.save"
	fi
}

phpmyadmin_rm_config_backup()
{
	rm -f "$PHPMYADMIN_CONFIG.save"
}

install_phpmyadmin()
{
	p_echo "===> Installing phpMyAdmin"
	set_phpmyadmin_params
	rename_old_configs
	pma__configure && phpmyadmin_rm_config_backup
	suc
}

set_phpmyadmin_params()
{
	PHPMYADMIN_D="$PRODUCT_ROOT_D/admin/htdocs/domains/databases/phpMyAdmin"
	PHPMYADMIN_CONFIG="$PHPMYADMIN_D/libraries/config.default.php"
}

get_pma_config_params()
{
	controluser=''
	uses_mysql_client
	pma__set_default_values
	pma__read_values_from_config
	[ -n "$pmadb" -a -n "$controluser" -a -n "$controlpass" ] || pma__set_random_values
}

rename_old_configs()
{
	for i in $PHPMYADMIN_D/config.*; do
		[ -f "$i" -a "`echo $i | sed 's|config.sample.inc.php||'`" = "$i" ] && mv $i $i.old
	done;
}

pma__read_values_from_config()
{
	local pma_config

	if [ -f "$PHPMYADMIN_CONFIG.save" ]; then
		pma_config="$PHPMYADMIN_CONFIG.save"
	else
		pma_config="$PHPMYADMIN_CONFIG"
	fi

	# looking for values of following variables in config.default.php.save
	pma__set_values_from_config "$pma_config" 		'controluser' 'controlpass' 'pmadb' 'bookmarktable' 'column_info' 		'history' 'pdf_pages' 'recent' 'favorite' 'relation' 'table_coords' 		'table_info' 'table_uiprefs' 'tracking' 'userconfig' 'users' 'usergroups' 		'navigationhiding' 'savedsearches' 'central_columns' 'designer_settings' 'export_templates'
}

pma__write_values_to_config()
{
	# write new config.default.php
	(
		umask 0027
		sed -e "`pma__sed_script`" < "$PHPMYADMIN_CONFIG.tpl" > "$PHPMYADMIN_CONFIG" || return 1
		set_admin_params
		set_ac root "$admin_group" 0640 "$PHPMYADMIN_CONFIG"
	)
}

pma__configure()
{
	local controluser controlpass pmadb
	local bookmarktable column_info history pdf_pages recent favorite relation table_coords table_info table_uiprefs tracking userconfig users usergroups navigationhiding savedsearches central_columns designer_settings export_templates

	get_pma_config_params

	# Make sure that phpmyadmin script is idempotent, when phpmyadmin upgrade
	pma__create_pmadb

	pma__write_values_to_config
}

pma__set_values_from_config()
{
	eval `pma__read_values "$@"`
}

pma__read_values()
{
	local PHPMYADMIN_CONFIG="$1"
	shift

	[ -f $PHPMYADMIN_CONFIG ] && sw_engine_pleskrun -r '
			try {
				error_reporting(0);
				$GLOBALS["db_host"] = "localhost";
				class vz {
					public static function in_vzcp_proxy() { return false; }
				}
				function in_vzcp_proxy() { return false; }
				require_once("'$PHPMYADMIN_CONFIG'");
				for ($i = 1; $i < $argc; $i++)
					if (isset($cfg["Servers"][1][$argv[$i]]))
						echo $argv[$i] . "=\"" . $cfg["Servers"][1][$argv[$i]] . "\"\n";
			} catch (Exception $e) {}' -- "$@" 2>/dev/null || true
}

pma__set_default_values()
{
	controluser="pma"
	pmadb="phpmyadmin"

	####
	# WARNING:
	# If you will modify the values you have to add templates (@@PMA_BOOKMARK@@, etc.)
	# in create_tables.sql.tpl and other SQL templates derived from phpMyAdmin SQL scripts.
	####
	bookmarktable="pma_bookmark"
	column_info="pma_column_info"
	history="pma_history"
	pdf_pages="pma_pdf_pages"
	recent="pma_recent"
	favorite="pma_favorite"
	relation="pma_relation"
	table_coords="pma_table_coords"
	table_info="pma_table_info"
	table_uiprefs="pma_table_uiprefs"
	tracking="pma_tracking"
	userconfig="pma_userconfig"
	users="pma_users"
	usergroups="pma_usergroups"
	navigationhiding="pma_navigationhiding"
	savedsearches="pma_savedsearches"
	central_columns="pma_central_columns"
	designer_settings="pma_designer_settings"
	export_templates="pma_export_templates"
}

pma__set_random_values()
{
	local pma__random_suffix

	# generate random controluser and pmadb
	while true; do
		pma__random_suffix=`get_random_string 12`
		controluser="pma_$pma__random_suffix" # maximum length is 16 characters
		pmadb="phpmyadmin_$pma__random_suffix"
		# check controluser existance
		# if user already exists then continue
		! db_test "SELECT COUNT(*) FROM mysql.user WHERE User='$controluser' LIMIT 1;" '$1!="0"' || continue
		# check pmadb existance
		# if db already exists then continue
		! db_test_database "$pmadb" || continue
		break
	done
	# generate random controlpass
	controlpass=`get_random_string 16`
}

pma__create_pmadb()
{
	local create_tables_sql="$PHPMYADMIN_D/sql/create_tables.sql.tpl"
	local upgrade_column_info_sql="$PHPMYADMIN_D/sql/upgrade_column_info_4_3_0+.sql.tpl"
	local inten="create pmadb"
	echo_try $inten

	# create pmadb
	sed -e "`pma__sed_script`" < "$create_tables_sql" | mysql_direct >> "$product_log" 2>&1 || warn "$inten"

	local mysql_db_name=$pmadb
	if ! db_test_column "pma_column_info" "input_transformation"; then
		# update pmadb
		sed -e "`pma__sed_script`" < "$upgrade_column_info_sql" | mysql_direct >> "$product_log" 2>&1 || warn "$inten"
	fi
	suc
}

pma__sed_script()
{
	echo "	s|@@""PMA_TEMP_DIR""@@|$PRODUCT_ROOT_D/var/phpMyAdmin|g
	s|@@""CONTROLUSER""@@|$controluser|g
	s|@@""CONTROLPASS""@@|$controlpass|g
	s|@@""PMADB""@@|$pmadb|g
	s|@@""PMA_BOOKMARK""@@|$bookmarktable|g
	s|@@""PMA_COLUMN_INFO""@@|$column_info|g
	s|@@""PMA_HISTORY""@@|$history|g
	s|@@""PMA_PDF_PAGES""@@|$pdf_pages|g
	s|@@""PMA_RECENT""@@|$recent|g
	s|@@""PMA_FAVORITE""@@|$favorite|g
	s|@@""PMA_RELATION""@@|$relation|g
	s|@@""PMA_TABLE_COORDS""@@|$table_coords|g
	s|@@""PMA_TABLE_INFO""@@|$table_info|g
	s|@@""PMA_TABLE_UIPREFS""@@|$table_uiprefs|g
	s|@@""PMA_TRACKING""@@|$tracking|g
	s|@@""PMA_USERCONFIG""@@|$userconfig|g
	s|@@""PMA_USERS""@@|$users|g
	s|@@""PMA_USERGROUPS""@@|$usergroups|g
	s|@@""PMA_NAVIGATIONHIDING""@@|$navigationhiding|g
	s|@@""PMA_SAVEDSEARCHES""@@|$savedsearches|g
	s|@@""PMA_CENTRAL_COLUMNS""@@|$central_columns|g
	s|@@""PMA_DESIGNER_SETTINGS""@@|$designer_settings|g
	s|@@""PMA_EXPORT_TEMPLATES""@@|$export_templates|g 	"
}

upgrade_phpmyadmin_to_4_7_plus()
{
	local inten="upgrade pmadb table to version 4.7.0+"
	echo_try $inten

	set_phpmyadmin_params
	get_pma_config_params

	local mysql_db_name=$pmadb
	local upgrade_tables_sql="$PHPMYADMIN_D/sql/upgrade_tables_4_7_0+.sql.tpl"
	sed -e "`pma__sed_script`" < "$upgrade_tables_sql" | mysql_direct >> "$product_log" 2>&1 || warn "$inten"

	suc
}

# vim:ft=sh

mail_restore_all()
{
	# temporary solution to close #143977 and #144076: three keys to turn on
	if [ ! -e "$PRODUCT_ROOT_D/version" -a ! -e "$PRODUCT_ROOT_D/core.version" ]; then
		p_echo "$PRODUCT_ROOT_D/version isn't yet on its place. Deferring mail_restore execution to a later stage" >&2
		return
	fi
	if [ ! -x "/usr/lib/plesk-9.0/mailsrv_entities_dump" ]; then
		p_echo "/usr/lib/plesk-9.0/mailsrv_entities_dump isn't yet on its place. Deferring mail_restore execution to a later stage" >&2
		return
	fi
	if [ ! -x "/usr/lib/plesk-9.0/mail_restore" ]; then
		p_echo "/usr/lib/plesk-9.0/mail_restore isn't yet on its place. Deferring mail_restore execution to a later stage" >&2
		return
	fi

	local args="--fast"
	if [ "$1" = "full" ]; then
		args="--with-spam"
	fi

	# FIXME: There's a possible race during upgrade procedure. Noone
	# managed to catch the bug in debug mode, so be stupid:
	sync; sync; sleep 3

	inten="execute mail_restore to synchronize mail server settings and Plesk Database"
	echo_try $inten
	/usr/lib/plesk-9.0/mail_restore $args >>$product_log 2>&1
	# mail_restore_all should return proper exit code.
	return "$?"
}

true exim_status_linux_debian
exim_status_linux_debian()
{
	get_pid /usr/lib/exim/exim3 false
	local pid=$common_var

	if test "$pid" -ne 1; then
		#running
		return 0;
	fi
	return 1
}

qmail_admin_alias()
{
	local userdata="`getent passwd $1`"
	local uid="`echo $userdata | awk -F':' '{print $3}'`"
	local gid="`echo $userdata | awk -F':' '{print $4}'`"
	local home="`echo $userdata | awk -F':' '{print $6}'`"

	if [ -z "$uid" -o -z "$gid" -o ! -d "$home" ]; then
		report_problem "Userinfo for $1 is incomplete. Qmail alias for this user will not be created" 			"Manually create .qmail file in $1 home directory"
		return 1
	fi

	[ ! -s "$home/.qmail" ] || return 0
	echo "&root" > "$home/.qmail"
	chmod 0600 "$home/.qmail"
	chown $uid:$gid "$home/.qmail"
}

set_postfix_auth()
{
	local run="$PRODUCT_ROOT_D/admin/sbin/mailmng-server --set-mail-params"

	if db_test "SELECT value FROM ServiceNodeConfiguration WHERE name='relay' AND value='open' AND serviceNodeId=1 AND section='mailServer'" 1 ; then
		run="$run --relayclient"
	fi

	if db_test "SELECT * FROM ServiceNodeConfiguration WHERE name='disable_smtp_auth' and value=0 AND serviceNodeId=1 AND section='mailServer'" 1 ; then
		run="$run --smtpauth"
	fi

	if db_test "SELECT * FROM ServiceNodeConfiguration WHERE name='disable_pop_auth' and value=0 AND serviceNodeId=1 AND section='mailServer'" 1 ; then
		run="$run --popauth"
	fi

	db_select "SELECT value FROM ServiceNodeConfiguration WHERE name='poplock_time' AND serviceNodeId=1 AND section='mailServer'"
	if [ -n "$db_select_output" ]; then
		# this will also set correct crontab for poplock.db cleanup
		run="$run --poplock-time=$db_select_output"
	fi

	$run >> $product_log 2>&1
}

start_postfix()
{
	local context="$1"

	[ "$context" = "bootstrapper-post" ] || return 0

	# If obsoleted, sendmail package on SuSE invokes try-restart for sendmail daemon. 
	# Postfix provides compatibility with sendmail and is started in daemon mode instead with wrong (sendmail.service) cgroup.
	# Since postfix process will have wrong cgroup, it will be considered dead by systemd, and will not be stopped/started by it.
	# We cannot: 1) use conflict instead of obsolete against sendmail package due to VZ;
	#            2) stop sendmail in bootstrapper prep-install due to uncertain ordering in VZ context.
	# Therefore we should just kill postfix and restart it normally.
	# Note that we have postfix_status() overridden, therefore ask both it (via pleskrc) and systemctl (via service_ctl).
	if pleskrc postfix status && ! service_ctl status $postfix_service postfix ; then
		p_echo "Trying to stop postfix not detected by systemd as active. Probably it is running but has wrong cgroups."
		/usr/sbin/postfix stop >> $product_log 2>&1
		p_echo "Done. Now we will attempt to start it via systemctl."
	fi

	pleskrc postfix start
	pleskrc pc_remote start
}


# vim:ft=sh

set_postfix_sendmail_wrapper_params()
{
	sendmail_vendor="$sendmail.postfix"
	sendmail_wrapper="$PLESK_LIBEXEC_DIR/postfix-sendmail-wrapper"
	sendmail_postfix_divert_package="plesk-mail-pc-driver"
	postdrop="/usr/sbin/postdrop"
	postdrop_group="postdrop"
}

set_qmail_sendmail_wrapper_params()
{
	sendmail_vendor="/var/qmail/bin/sendmail"
	sendmail_wrapper="$PLESK_LIBEXEC_DIR/qmail-sendmail-wrapper"
}

upgrade_sendmail_wrapper_12_0_5()
{
	# This upgrade action removes sendmail-wrapper (then called psa-sendmail) installed by 12.0.5
	# Installation (if needed) will be done later in Postfix/QMail driver bootstrapper post-install

	local old_sendmail_wrapper="$PLESK_LIBEXEC_DIR/psa-sendmail"

	local old_sendmail_vendor="$sendmail.os.orig"

	[ -L "$sendmail" -a "`readlink -f \"$sendmail\"`" = "$old_sendmail_wrapper" ] || return 0
	[ -f "$old_sendmail_vendor" ] || return 0
	# The latter is only for Postfix (due to the check right above)

	{
		rm -f "$sendmail"
		dpkg-divert --remove --rename "$sendmail" && dpkg-statoverride --remove "$sendmail"
	} >> $product_log 2>&1
}

install_sendmail_wrapper()
{
	local smtp_server="$1"
	local inten="install sendmail wrapper"
	echo_try "${inten}"

	case "${smtp_server}" in
		qmail)
			set_qmail_sendmail_wrapper_params
			# For Debian and Ubuntu sendmail link is in psa-qmail package
			;;
		postfix)
			set_postfix_sendmail_wrapper_params
			install_sendmail_wrapper_postfix_deb
			;;
		msmtp)
			:
			;;
		*)
			warn "${inten}: unknown SMTP server '${smtp_server}'"
			return 1
			;;
	esac

	suc
}

install_sendmail_wrapper_postfix_deb()
{
	{
		dpkg-statoverride --remove "${postdrop}" || :
		dpkg-statoverride --update --add root ${postdrop_group} 2550 "${postdrop}"
		# We need to be careful with 'dpkg-divert --rename' and 'dpkg-statoverride' used together
		# because statoverriding twice after renaming will change symlink's target, not vendor's sendmail
		if [ -e "${sendmail_vendor}" -a -L "${sendmail}" ]; then
			rm -f "$sendmail"
			dpkg-divert --remove --rename "$sendmail" && dpkg-statoverride --remove "$sendmail"
		fi
		dpkg-statoverride --update --add root ${postdrop_group} 750 "${sendmail}"
		dpkg-divert --package "$sendmail_postfix_divert_package" --divert "${sendmail_vendor}" --rename "${sendmail}"
		ln -s "${sendmail_wrapper}" "${sendmail}"
	} >> "${product_log}" 2>&1
}

upgrade_sendmail_wrapper_12_0_9()
{
	# fix protected column
	local handler_db_path="/var/lib/plesk/mail/handlers/handlers.db"
	local query="ALTER TABLE handlers ADD COLUMN protected BOOLEAN NOT NULL default 0"
	if [ -w "${handler_db_path}" ]; then
		local sqlite3_binary=`lookup_command sqlite3`
		if [ -n "${sqlite3_binary}" ]; then
			${sqlite3_binary} "${handler_db_path}" "$query" || true
		fi
	fi

	# add check-quota into before-sendmail queue
	"$PLESK_LIBEXEC_DIR/mail_mailbox_restore" --handlers-only

	set_postfix_params
	set_qmail_params
	if pleskrc postfix exists >/dev/null 2>&1; then
		pleskrc postfix reload
	elif pleskrc qmail exists >/dev/null 2>&1; then
		pleskrc qmail reload
	fi
}

remove_unused_postfix_milters_12_0_18()
{
	local commtouch_marker_file='/etc/init.d/ctmilter_initd'
	local commtouch_milter_pattern='unix:\/var\/spool\/postfix\/ctmilter\/ctmilter.sock'

	# edit postfix configuration to remove non-smtpd milter
	[ -s "/etc/postfix/main.cf" ] || return 0

	set_postfix_params
	local filtered_value="`read_postconf_value 'non_smtpd_milters' | sed -e '/^\(inet:127.0.0.1:12768\|inet:localhost:12768\)$/d'`"
	write_postconf_value 'non_smtpd_milters' "$filtered_value" || return 1

	if [ ! -e "$commtouch_marker_file" ]; then
		for milters in non_smtpd_milters smtpd_milters; do
			filtered_value="`read_postconf_value "$milters" | sed -e "/^${commtouch_milter_pattern}$/d"`"
			write_postconf_value "$milters" "$filtered_value" || return 1
		done
	fi

	pleskrc postfix reload
}

set_mail_params()
{
	HANDLERS_USER="mhandlers-user"
	MAIL_USERGROUP="popuser"
}

mailname_to_lowercase_11_1_15()
{
	local res=0

	echo_try "rename mailboxes to lower case format"

	$PRODUCT_ROOT_D/admin/sbin/mailmng-service --stop-service
	if ! $PRODUCT_ROOT_D/bin/repair --convert-mailboxes-to-lowercase -no-mchk; then
		warn "rename mailboxes to lower case format. Please run $PRODUCT_ROOT_D/bin/repair --convert-mailboxes-to-lowercase manually."
		res=1
	else
		suc
	fi
	$PRODUCT_ROOT_D/admin/sbin/mailmng-service --start-service

	deferred_mail_restore_all

	return $res
}

## vim:filetype=sh

upgrade_drweb_storage()
{
	# should we move files from old quarantine into new?
	[ -e "/etc/drweb/drweb_handler.conf" ] || return 0
	sed -i 		-e 's|^\s*Spool\s*=.*|Spool = /var/spool/drweb/spool|g' 		-e 's|^\s*SpoolFilesMode\s*=.*|SpoolFilesMode = 0664|g' 		-e 's|^\s*Quarantine\s*=.*|Quarantine = /var/spool/drweb/infected|g' 		-e 's|^\s*QuarantineFilesMode\s*=.*|QuarantineFilesMode = 0660|g' 		/etc/drweb/drweb_handler.conf || warn "update drweb handler config"

	# as far as we change dir location and config, drweb daemon reload is required
	if [ -x "/opt/psa/admin/sbin/drwebmng" ] ; then
		/opt/psa/admin/sbin/drwebmng --reload || service_restart_warning drweb
	fi
}


upgrade_gl_rules_17_0_12()
{
	# Add office365 services into GreyListing whitelist
	$PLESK_LIBEXEC_DIR/mail_grey_restore >> $product_log 2>&1
}
## vim:filetype=sh

upgrade_greylisting_permissions()
{
	local gl_db_dir="/var/lib/plesk/mail/greylist"

	[ -d "$gl_db_dir" ] || return 0
	chown -R root:popuser ${gl_db_dir} || pp_echo 'Failed to set correct ownership for grey listing databases'
	chmod 770 ${gl_db_dir} || pp_echo 'Failed to set correct grey listing databases directory permissions'
	chmod 660 ${gl_db_dir}/* || pp_echo 'Failed to set correct grey listing databases permissions'

	#may be call restore if there is(are) failures?
}


# vim:ft=sh

upgrade_rsyslog_maillog_configuration_12_1_6()
{
	true
}

upgrade_outgoing_mail_rate_limiting_handlers_12_0_7()
{
	local mhc="$UTILS_DIR/mail_handlers_control"
	[ -x "$mhc" ] || return 0

	# Since we dropped "in" limits, we should remove all old mail rate limiting handlers
	{
		$mhc --remove --queue before-queue    --type global --name limit-in
		$mhc --remove --queue before-sendmail --type global --name limit-in
		$mhc --remove --queue before-remote   --type global --name limit-out
	} >> $product_log 2>&1

	# as well as install new ones and recreate backend DB
	deferred_mail_outgoing_restore_full
}

upgrade_drop_sender_ip_mail_handler_12_0_10()
{
	local mhc="$UTILS_DIR/mail_handlers_control"
	[ -x "$mhc" ] || return 0

	$mhc --remove --queue before-queue --type global --name sender-ip >> $product_log 2>&1
	:
}
## vim:filetype=sh

upgrade_quota_handler_address_11_1_13()
{
	local common_mess="update check-quota handler recipients address column in handlers DB"
	echo_try ${common_mess}

	local handler_db_path=/var/lib/plesk/mail/handlers/handlers.db

	if [ ! -w "${handler_db_path}" ]; then
		# This is more probably an upgrade from earlier versions than a broken machine
		p_echo "Unable to ${common_mess} - missing DB '${handler_db_path}'. Will call mail_restore later."
		deferred_mail_restore_all
		return 0
	fi

	local sqlite3_binary=`lookup_command sqlite3`
	if [ -z "${sqlite3_binary}" ]; then
		pp_echo "Unable to ${common_mess} - missing sqlite3 binary"
		return 1
	fi

	local query='UPDATE `handlers` SET `address`="all-recipients" WHERE `name`="check-quota"'
	${sqlite3_binary} "${handler_db_path}" "$query"
	if [ $? -ne 0 ]; then
		pp_echo "Unable to ${common_mess} - execution of DB query '$query' failed"
		return 1
	fi

	suc
}

#-*- vim:syntax=sh

maillog_create_compat_symlink()
{
	mkdir -p "$PRODUCT_ROOT_D/var/log"
	[ -e "$PRODUCT_ROOT_D/var/log/maillog" -a ! -L "$PRODUCT_ROOT_D/var/log/maillog" ] || ln -sf "$mail_log" "$PRODUCT_ROOT_D/var/log/maillog"
}

select_maillog()
{
	local mail_log inten
	mail_log="/var/log/maillog"
	inten="set maillog file to $mail_log"

	set_syslog_params



	touch $mail_log
	chmod 0640 $mail_log

	local rsyslog_mail_log="${mail_log}"

	for config in $syslog_conf; do
		if [ -f "$config" ]; then
			if [ "${config}" = "/etc/rsyslog.conf" ]; then
				set_maillog $config "${rsyslog_mail_log}" "${inten}" >> $product_log 2>&1
			else
				set_maillog $config "${mail_log}" "${inten}" >> $product_log 2>&1
			fi
		fi
	done

	[ -f $syslog_conf_ng ] && set_maillog_ng "$mail_log" "$inten"  >> "$product_log" 2>&1

	pleskrc syslog reload
	maillog_create_compat_symlink
}

_set_maillog_ng_change()
{
    local conf_file mail_log inten bak_ext

    conf_file=$1
    mail_log=$2
    inten=$3
    bak_ext=$4

    echo_try "$inten"

    local mloc_cmd mloc_arg
    local ws
    ws=$(echo -ne " \t")

    ## Set log to $mail_log
    mloc_cmd="-e"
    mloc_arg='s|^\(['"$ws"']*destination['"$ws"']\+mail['"$ws"']*{[^"]*"\)[^"]*\(.*\)$|\1'"$mail_log"'\2|'

    ## By default SuSE-9.3 and SuSE-10 syslog-ng settings
    ## doesn't allow any 'mail' log messages to be put into
    ## /var/log/messages. We wish to allow mail.warn to pass
    ## through the filter, so we need to tune the conditions...
    ## So we change
    ## filter f_messages   { not facility(news, mail) and not filter(f_iptables); };
    ## to
    ## filter f_messages   { not (facility(news) or filter(f_iptables)) or filter(f_mailwarn); };
    ##
    ## I'm not sure if the whole filter is well optimized though
    ##
    ## Here I don't try to perform a sofisticated replace. The proper way
    ## would require a deep analisys of the current configuration and
    ## I don't see any sense to do it at this time, probably once in a future...

    local mfmt_cmd mfmt_arg
    mfmt_cmd=-e
    mfmt_arg='s|^\(['"$ws"']*filter['"$ws"']\+f_messages['"$ws"']*{\).*|\1 not (facility(news) or filter(f_iptables)) or filter(f_mailwarn); };|'

    ## now execute the entire command *IN-PLACE*
    ## all modern sed's (well at least on supported systems
    ## do support this option.

    ## One HAVE NOT TO quite mail_log_expr and mf_expr below!
    sed $mloc_cmd "$mloc_arg" $mfmt_cmd "$mfmt_arg" "-i$bak_ext" "$conf_file" || die "$intent"
}

set_maillog_ng() {
    local mail_log intent
    mail_log=$1
    intent=$2

	if [ -f "${syslog_conf_ng}.in" ]; then
	    _set_maillog_ng_change "${syslog_conf_ng}.in" "$mail_log" "$intent" ".bak" && 		/sbin/SuSEconfig --module syslog-ng
	else
# Modest SuSE 1.20 doens't rule syslog through SuSE-config, bug 118238
		_set_maillog_ng_change "${syslog_conf_ng}" "$mail_log" "$intent" ".bak"
	fi

}

set_maillog()
{
	local syslog_conf mail_log inten mail_log_str log_str num
	syslog_conf=$1
	mail_log=$2
	inten=$3

	mail_log_str="-$mail_log"

	echo_try "$inten"

	log_str="^mail.*[[:space:]]*$mail_log*"
	grep "$log_str" $syslog_conf > /dev/null
	if [ "$?" -ne 0 ]; then
		grep '^mail\.\*' $syslog_conf > /dev/null
		if [ "$?" -eq 0 ]; then
			## if line "mail.*       ..." is exist then
			## replace this with new line
			grep -q '^mail\.\*.*;' $syslog_conf
			if [ $? -eq 0 ]; then
				sed -e "s|^mail\.\*.*\(;.*\)$|mail.*					$mail_log_str\1|" 					< $syslog_conf > $syslog_conf.tmp
			else
				sed -e "s|^mail\.\*.*$|mail.*						$mail_log_str|" 					< $syslog_conf > $syslog_conf.tmp
			fi
			mv -f $syslog_conf.tmp $syslog_conf
		else
			## if line "mail.*       ..." is NOT exist then
			## search "*.       ..." line
			num=`awk '{if ((my==0) && (index($1, "*.") == 1)) {my=1; print FNR;}}' < $syslog_conf`
			if [ "0$num" -gt "0" ]; then
				## if line "*.       ..." is exist then
				## disable all lines beginning with "mail."
				## and insert new line "mail.*      ..." before this
				sed -e 's/^\(mail\.\)/#\1/' 					-e ''${num}'i					mail.*						'$mail_log_str'' 					< $syslog_conf > $syslog_conf.tmp && 				mv -f $syslog_conf.tmp $syslog_conf || die "$inten"
			else
				## if line "*.       ..." is NOT exist then
				## disable all lines beginning with "mail."
				## and insert new line "mail.*      ..." at the end of file
				sed -e 's/^\(mail\.\)/#\1/'	< $syslog_conf > $syslog_conf.tmp && 				echo "mail.*						$mail_log_str" >> $syslog_conf.tmp && 				mv -f $syslog_conf.tmp $syslog_conf || die "$inten"
			fi
		fi
	fi

	sed -e 's|\(^.*\)maili\none\;\(.*\)|\1\2|g' < $syslog_conf > $syslog_conf.tmp &&           mv -f $syslog_conf.tmp $syslog_conf || die "$inten"
	echo 'mail.none			-/var/log/messages' >> /var/log/messages 

	suc
}
#qmail

set_qmail_params()
{
    QMAIL_DIR="/var/qmail"
    QMAIL_USERS="alias:2021:nofiles:false:alias  qmaild:2020:nofiles:false:  qmaill:2022:nofiles:false:  qmailp:2023:nofiles:false:  qmailq:2520:qmail:  qmailr:2521:qmail:  qmails:2522:qmail:"
    QMAIL_GROUPS="qmail:2520  nofiles:2020"

    #variable from psa.conf but used in mail-qc-driver
    #so we need define this before install psa base package
    if [ -z "$QMAIL_ROOT_D" ]; then 
	    QMAIL_ROOT_D="/var/qmail"
    fi

    mm_mailgroup=12

    qmail_service="qmail"
    qmail_extra_cmd=""
}


upgrade_qmail_reduce_suids_11_1_12()
{
	if ! /opt/psa/admin/sbin/mailmng-server --features | egrep -q "SMTP_Server_package.*qmail" 2>/dev/null; then
		p_echo "Skip qmail specific upgrade actions"
		return
	fi

	upgrade_qmail_permissions
	drop_unused_mailman_wrapper
}

upgrade_qmail_permissions()
{
	local inten="setup correct owner and permissions on poplock and mail authentication DB"
	echo_try "$inten"

	local res=0

	local perm_err="set correct permissions on"
	local ownership_err="set correct ownership for"
	
	if [ -d "/var/lib/plesk/mail/poplock" ]; then
		if ! chown -R root:nofiles /var/lib/plesk/mail/poplock; then
			warn "${ownership_err} poplock DB"
			res=1
		fi

		if ! chmod 770 /var/lib/plesk/mail/poplock; then
			warn "${perm_err} poplock DB directory"
			res=1
		fi

		if ! chmod 660 /var/lib/plesk/mail/poplock/poplock.db; then
			warn "${perm_err} poplock DB file"
			res=1
		fi
	fi

	if [ -d "/var/lib/plesk/mail/auth" ]; then
		if ! chown -R root:nofiles /var/lib/plesk/mail/auth; then
			warn "${ownership_err} auth DB"
			res=1
		fi

		if ! chmod 750 /var/lib/plesk/mail/auth; then
			warn "${perm_err} auth DB directory"
			res=1
		fi

		if ! chmod 0640 /var/lib/plesk/mail/auth/passwd.db; then
			warn "${perm_err} auth DB file"
			res=1
		fi

		if ! chmod 440 /var/lib/plesk/mail/auth/passwd_db_key; then
			res=1
			warn "${perm_err} auth passwd file "
		fi
	fi

	if [ $res = 0 ]; then
		suc
	else
		warn "$inten"
	fi
}

drop_unused_mailman_wrapper()
{
	[ -x /usr/lib/plesk-9.0/mailman_lists_dump ] && /usr/lib/plesk-9.0/mailman_lists_dump
}


# db_backup [ dump_prefix [ db1 [ db2 [...] ] ] ]
db_backup()
{
	local timestamp=`date +%Y%m%d-%H%M%0S`
	local dfile_prefix="${1:-mysql.preupgrade.${prev_version}-${product_version}}"
	dfile="$DUMP_D/${dfile_prefix}.${timestamp}.dump.gz"

	local databases
	[ -n "$*" ] && shift
	if [ -n "$*" ]; then
		databases="$*"
	else
		local apsc_db_name=`read_apsc_connection_param_db "database"`
		databases="mysql $PRODNAME $apsc_db_name horde atmail"
	fi

	if test -f $dfile; then
		# dump names should not collide, but better be safe than sorry
		local i
		i=1
		while test -f "$dfile.$i"; do
			i=$(($i+1))
		done
		mv -v "$dfile" "$dfile.$i" >> "$product_log" || die "save previous backup $dfile"
	fi

	inten="backup MySQL database"
	echo_try "$inten"
	if ! test -d "$DUMP_D"; then
		mkdir -p -m 750 "$DUMP_D"
	fi
	touch "$dfile" || die "$inten"
	chown 0:0 "$dfile" || die "$inten"
	chmod 600 "$dfile" || die "$inten"
	dump_db $databases | gzip > "$dfile" || die "$inten"
	suc
	p_echo " MySQL databases are dumped to $dfile"
}
safe_prep_db_upgrade()
{
	p_echo
	p_echo "===> Upgrading database"
	p_echo

	[ -s "$PRODUCT_ROOT_D/version" -o -s "$PRODUCT_ROOT_D/core.version" ] && get_prev_version "$PRODUCT_ROOT_D"

	uses_mysql_client

	check_db
	safe_prep_db_backup_all

	enable_innodb_engine
	upgrade_apsc_prep_stage
	core_run_upgrade_prep_stage
}

safe_prep_db_backup_all()
{
	# backup database and register rollback action
	export MYSQLDUMP_OPTIONS="--add-drop-database"
	db_backup
	unset MYSQLDUMP_OPTIONS
	SAFE_PREP_DB_BACKUP="$dfile"
	transaction_add_rollback_action "safe_prep_db_upgrade_rollback"
}

safe_prep_db_upgrade_rollback()
{
	if [ -n "$SAFE_PREP_DB_BACKUP" ]; then
		p_echo
		p_echo "===> Restoring database from backup $SAFE_PREP_DB_BACKUP"
		p_echo
		zcat "$SAFE_PREP_DB_BACKUP" | mysql_anydb
		if [ $? -ne 0 ]; then
			p_echo
			p_echo "    [ERROR] Database restoring failed"
			p_echo "    Try to restore database from $SAFE_PREP_DB_BACKUP manually"
			p_echo
			return
		fi
		SAFE_PREP_DB_BACKUP=
	fi
}

safe_prep_db_install()
{
	requires_configured_mysql

	product_install_db_local
}

safe_prep_db()
{
	p_echo
	p_echo "===> Performing safe prep-install database actions"
	p_echo

	# perform database-related prep actions
	if [ "$do_upgrade" -ne 0 ]; then
		safe_prep_db_upgrade
	else
		safe_prep_db_install
	fi
}

transaction_begin()
{
	[ -n "$TRANSACTION_STARTED" ] && die "Another transaction in progress!"
	TRANSACTION_STARTED="true"
	TRANSACTION_ROLLBACK_FUNCS=
	TRANSACTION_COMMIT_FUNCS=
	local transaction_autocommit="$1"
	if [ -n "$transaction_autocommit" ]; then
		trap "transaction_commit" PIPE EXIT
		trap "transaction_rollback" HUP INT QUIT TERM
	else
		trap "transaction_rollback" HUP PIPE INT QUIT TERM EXIT
	fi
}

transaction_rollback()
{
	[ -z "$TRANSACTION_STARTED" ] && die "Transaction is not started!"
	# perform rollback actions
	local f
	for f in ${TRANSACTION_ROLLBACK_FUNCS}; do
		"$f"
	done
	TRANSACTION_STARTED=
	TRANSACTION_ROLLBACK_FUNCS=
	TRANSACTION_COMMIT_FUNCS=
	trap - HUP PIPE INT QUIT TERM EXIT
	exit 1
}

transaction_commit()
{
	[ -z "$TRANSACTION_STARTED" ] && die "Transaction is not started!"
	# perform commit actions
	local f
	for f in ${TRANSACTION_COMMIT_FUNCS}; do
		"$f"
	done
	TRANSACTION_STARTED=
	TRANSACTION_ROLLBACK_FUNCS=
	TRANSACTION_COMMIT_FUNCS=
	trap - HUP PIPE INT QUIT TERM EXIT
}

transaction_add_rollback_action()
{
	[ -z "$TRANSACTION_STARTED" ] && die "Transaction is not started!"
	# LIFO rollback order
	[ -z "$TRANSACTION_ROLLBACK_FUNCS" ] 		&& TRANSACTION_ROLLBACK_FUNCS="$1" 		|| TRANSACTION_ROLLBACK_FUNCS="$1 $TRANSACTION_ROLLBACK_FUNCS"
}

transaction_remove_rollback_action()
{
	[ -z "$TRANSACTION_STARTED" ] && die "Transaction is not started!"
	# Preserves order of remaining records
	TRANSACTION_ROLLBACK_FUNCS=`remove_from_array "$1" "$TRANSACTION_ROLLBACK_FUNCS"`
}

transaction_add_commit_action()
{
	[ -z "$TRANSACTION_STARTED" ] && die "Transaction is not started!"
	# FIFO commit order
	[ -z "$TRANSACTION_COMMIT_FUNCS" ] 		&& TRANSACTION_COMMIT_FUNCS="$1" 		|| TRANSACTION_COMMIT_FUNCS="$TRANSACTION_COMMIT_FUNCS $1"
}

core_upgrader_repair_arg()
{
	if [ -n "$1" -o -n "$do_repair" ]; then
		echo "--repair"
	fi
}

core_upgrader_action_name()
{
	if [ -n "`core_upgrader_repair_arg $1`" ]; then
		echo "upgrade and repair"
	else
		echo "upgrade"
	fi
}

core_run_upgrade_prep_stage()
{
	local action_name="`core_upgrader_action_name $1`"
	local optional_repair_arg="`core_upgrader_repair_arg $1`"
	pp_echo "===> Cumulative Plesk database $action_name (revertable stage) has been started."
	cu_run_upgrader "core" core_upgrader 		--load-config-data core_cu_load_config_data --store-config-data core_cu_store_config_data 		--prepare core_upgrader_prepare --rollback core_upgrader_rollback --commit core_upgrader_commit 		--from-version "$prev_db_short_version" --stage "prep" $optional_repair_arg
	local ret="$?"
	if [ "$ret" -eq 0 ]; then
		pp_echo "===> Cumulative $action_name of Plesk database (revertable stage) has been completed."
	else
		pp_echo "===> Plesk database was not upgraded completely. See installation log for details."
		# reflect error in exit-code in repair mode:
		[ ! "$1" = "repair" ] || return "$ret"
	fi
}

core_run_upgrade_post_stage()
{
	local action_name="`core_upgrader_action_name $1`"
	local optional_repair_arg="`core_upgrader_repair_arg $1`"
	pp_echo "===> Cumulative Plesk $action_name (final stage) has been started."
	cu_run_upgrader "core" core_upgrader 		--load-config-data core_cu_load_config_data --store-config-data core_cu_store_config_data 		--prepare core_upgrader_prepare --rollback core_upgrader_rollback --commit core_upgrader_commit 		--from-version "$prev_short_version" --stage "post" $optional_repair_arg --update-version
	local ret="$?"
	if [ "$ret" -eq 0 ]; then
		pp_echo "===> Cumulative $action_name of Plesk (final stage) has been completed."
	else
		pp_echo "===> Plesk was not upgraded completely. See installation log for details."
		# reflect error in exit-code in repair mode:
		[ ! "$1" = "repair" ] || return "$ret"
	fi
}

core_init_upgrader_conf()
{
	echo_try "initialize Plesk DB and files upgrader configuration"
	cu_upgrader_conf_init "core" core_upgrader 		--load-config-data core_cu_load_config_data --store-config-data core_cu_store_config_data 		--stages "prep post"
	suc
}

core_read_misc_attr()
{
	db_select "SELECT val FROM misc WHERE param = '$1'"
	echo "$db_select_output"
}

core_cu_load_config_data()
{
	local upgrader_name="$1"

	[ "$upgrader_name" = "core" ] || return 0
	db_test_table "misc" || return 0

	cu_db_current_timestamp="`  core_read_misc_attr 'upgrader_config_current_timestamp'`"
	cu_db_continue_from="`      core_read_misc_attr 'upgrader_config_continue_from'`"
	cu_db_failed_actions_list="`core_read_misc_attr 'upgrader_config_failed_actions_list'`"
	cu_db_latest_timestamp="`   core_read_misc_attr 'upgrader_config_latest_timestamp'`"
}

core_cu_store_config_data()
{
	local upgrader_name="$1"

	[ "$upgrader_name" = "core" ] || return 0
	db_test_table "misc" || die "save core upgrader state to 'misc' table"

	db_do "REPLACE INTO misc (param, val) VALUES 		('upgrader_config_current_timestamp', '$cu_db_current_timestamp'), 		('upgrader_config_continue_from', '$cu_db_continue_from'), 		('upgrader_config_failed_actions_list', '$cu_db_failed_actions_list'), 		('upgrader_config_latest_timestamp', '$cu_db_latest_timestamp')"
}

core_upgrader_prepare()
{
	local upgrader_name="$1"
	local upgrader_stage="$2"
	local upgrader_aux="$3"

	# DB is backed up earlier via safe_prep_db_backup_all()

	if [ "$upgrader_stage" = "prep" ]; then
		pp_echo "===> Preparing Plesk database upgrade (revertable stage)."

		get_old_domain
	else
		pp_echo "===> Preparing Plesk upgrade (final stage)."
	fi

	db_added_columns=
	[ "$upgrader_aux" = "noop-stage" ] || stop_before_upgrade
}

core_upgrader_commit()
{
	local upgrader_name="$1"
	local upgrader_stage="$2"

	if [ "$upgrader_stage" = "prep" ]; then
		set_db_version
	fi
}

core_upgrader_rollback()
{
	# Empty rollback function is required to suppress default upgrader behavior
	# (switching configuration into broken state). Actual rollback will be done
	# in safe_prep_db_upgrade_rollback().
	:
}

core_upgrader()
{
	local db="cu_check_stage prep"
	local files="cu_check_stage post"
	local ver="cu_check_version"
	local ts="cu_check_timestamp"
	local do="cu_do"

	# Some upgrade actions might rely on MySQL passwords being upgraded. Note that rollback won't work for Billing.
	$db    $ts '2011/11/15 17:13:14' $do upgrade_mysqlnd_old_passwd

	cu_start_upgrade_from_version 1100
	$db    $ver $do add_db_changes 11000
	$db    $ver $do db_smb_upgrade_1100
	$db    $ver $do db_upgrade_heterogenous_1100
	$db    $ver $do db_upgrade_apsmail_1100
	$db    $ver $do db_upgrade_service_nodes_1100
	$db    $ver $do db_upgrade_mail_server_settings_1100
	$db    $ver $do db_upgrade_webusers_miva_1100
	$files $ver $do pbm_create_certificate # We should generate certificate on upgrade since backup was merged into core
	$files $ver $do upgrade_adminlogs_permissions_1100

	cu_start_upgrade_from_version 1101
	$db    $ver $do db_upgrade_longtasks_1101
	$db    $ver $do db_upgrade_db_users_1101

	cu_start_upgrade_from_version 011001011
	$db    $ver $do db_upgrade_add_indexes_for_performance_11_1_11
	$db    $ver $do db_upgrade_change_sys_users_home_to_utf8
	$files $ver $do deferred_mail_restore_all

	cu_start_upgrade_from_version 011001012
	$db    $ver $do add_db_changes 11.1.12
	$db    $ver $do db_upgrade_service_node_properties_11_1_12
	$db    $ver $do db_upgrade_custom_buttons_11_1_12
	$files $ver $do upgrade_greylisting_permissions
	$files $ver $do upgrade_drweb_storage
	$files $ver $do upgrade_qmail_reduce_suids_11_1_12

	cu_start_upgrade_from_version 011001013
	$db    $ver $do db_upgrade_service_node_configuration_11_1_13
	$files $ver $do upgrade_quota_handler_address_11_1_13

	cu_start_upgrade_from_version 011001014
	$db    $ver $do db_upgrade_dns_templates_11_1_14

	cu_start_upgrade_from_version 011001015
	$files $ts '2012/12/06 13:03:47' $do db_upgrade_subscription_properties_11_1_15
	$files $ts '2013/07/09 17:45:00' $do mailname_to_lowercase_11_1_15
	$files $ts '2012/12/12 14:44:47' $do register_sw_engine_service
	$files $ts '2012/12/14 09:26:01' $do upgrade_pci_compliance_11_1_15

	cu_start_upgrade_from_version 011001016
	$db    $ts '2012/12/19 18:57:31' $do add_db_changes 11.1.16

	cu_start_upgrade_from_version 011001017
	$db    $ts '2013/01/15 17:41:18' $do db_upgrade_forwarding_11_1_17
	$db    $ts '2013/01/21 16:34:00' $do db_upgrade_seo_safe_domain_aliases_11_1_17
	$files $ts '2013/01/21 17:11:13' $do upgrade_postfix_allow_authenticated_dnsbled_clients_11_1_17
	$db    $ts '2013/01/25 09:37:00' $do db_upgrade_db_user_unique_key_11_1_17

	cu_start_upgrade_from_version 011001018
	$db    $ts '2013/01/29 15:00:00' $do add_db_changes 11.1.18
	$db    $ts '2013/02/04 19:10:05' $do db_upgrade_php_handler_id_11_1_18
	$files $ts '2013/05/24 18:13:37' $do upgrade_skel_remove_statistics_11_1_18

	cu_start_upgrade_from_version 011005019
	$db    $ts '2013/02/22 09:43:00' $do add_db_changes 11.5.19
	$db    $ver $do						 db_upgrade_smb_11_5_19
	$db    $ts '2013/02/14 17:51:00' $do db_upgrade_drop_php_version_11_5_19
	$db    $ts '2013/02/25 14:43:00' $do db_upgrade_webmails_11_5_19

	cu_start_upgrade_from_version 011005021
	$db    $ts '2013/03/07 12:57:01' $do db_upgrade_enable_dns_connector_11_5_21
	$db    $ts '2013/03/19 15:19:00' $do add_db_changes 11.5.21
	$db    $ts '2013/03/25 14:24:01' $do db_upgrade_set_force_update_applications_notification_11_5_21
	$db    $ts '2013/03/28 09:40:01' $do db_add_public_ip_11_5_21

	cu_start_upgrade_from_version 011005022
	$files $ts '2013/04/01 17:38:44' $do db_upgrade_drop_ip_address_id_11_5_22
	$files $ts '2013/06/05 18:07:46' $do upgrade_vhost_directories_11_1_18
	$db    $ts '2013/04/08 18:18:44' $do db_upgrade_set_property_unpaid_website_status_for_plans_11_5_22
	$db    $ts '2013/04/08 18:18:44' $do db_upgrade_set_property_unpaid_website_status_for_subscriptions_11_5_22

	cu_start_upgrade_from_version 107
	# Moved from 107 and 1012 because of the BL (#80977) and backend (#136372) dependencies. Should be after upgrade_vhost_directories_11_1_18.
	$files $ver $do upgrade_proftpd_107

	cu_start_upgrade_from_version 011005023
	$files $ts '2013/04/18 15:25:35' $do fix_file_sharing_permissions_from_10_0_0
	$files $ts '2013/04/18 19:44:02' $do upgrade_courier_to_4_12_pop_before_smtp_11_5_23

	cu_start_upgrade_from_version 011005024
	$db    $ts '2013/04/26 14:00:44' $do db_upgrade_correcting_property_unpaid_website_status_for_plans_11_5_24
	$db    $ts '2013/04/26 14:00:44' $do db_upgrade_correcting_property_unpaid_website_status_for_subscriptions_11_5_24
	$db    $ts '2013/04/26 16:58:00' $do db_upgrade_maintenance_mode_id_11_5_24
	$db    $ts '2013/04/30 13:45:00' $do db_upgrade_service_node_configuration_11_5_24
	$files $ts '2013/04/30 12:37:00' $do db_upgrade_secret_keys_11_5_24
	$files $ts '2013/05/14 14:11:38' $do upgrade_domains_awstats_conf_11_5_24

	cu_start_upgrade_from_version 011005029
	$db    $ts '2013/06/20 10:30:00' $do add_db_changes 11.5.29
	$db    $ver $do						 db_smb_upgrade_11_5_29
	$db    $ts '2013/06/25 13:39:00' $do db_upgrade_correcting_aps_tokens_11_5_29
	$files $ts '2013/07/22 14:44:38' $do upgrade_named_restrict_number_of_workers_11_5_29

	cu_start_upgrade_from_version 011005030
	$db    $ts '2013/07/22 15:01:00' $do db_upgrade_db_user_unique_key_11_1_17

### Below month numbers are increased by 10 due to published screw-up in 11.5.30 upgrader (see r327259)
	cu_start_upgrade_from_version 011005099
	$db    $ts '2013/18/20 13:33:24' $do add_db_changes 11.5.99
	$db    $ts '2013/18/20 13:34:26' $do db_upgrade_external_id_11_5_99
	$db    $ts '2013/18/20 13:35:28' $do db_upgrade_move_configuration_11_5_99
	$files $ts '2013/18/27 11:44:17' $do upgrade_postfix_remove_deprecated_headers_11_5_30
	$db    $ts '2013/19/02 11:36:45' $do db_upgrade_iis_app_pools_cpu_limit_11_5_99
	$db    $ts '2013/19/09 13:00:00' $do db_upgrade_sdk_add_services_11_5_99
	$files $ver $do deferred_relabel_plesk_directories_with_mailnames

	cu_start_upgrade_from_version 011006000
	$files $ts '2013/19/10 10:13:12' $do upgrade_panel_log_11_6_0
	$files $ts '2013/19/10 18:00:01' $do upgrade_httpsd_access_log_11_6_0
	$files $ts '2013/19/13 15:14:03' $do upgrade_transfer_rotated_logs_11_6_0

	cu_start_upgrade_from_version 011006001
	$files $ts '2013/19/17 12:35:41' $do upgrade_transfer_pmm_logs_11_6_1
	$files $ts '2013/19/18 09:30:17' $do upgrade_transfer_httpsd_access_log_processed_11_6_1
	$files $ts '2013/19/18 11:41:47' $do upgrade_transfer_proftpd_logs_11_6_1
	$db    $ts '2013/19/18 18:30:00' $do add_db_changes 11.6.1
	$files $ts '2013/19/23 14:27:14' $do deferred_mail_spam_restore_full

	cu_start_upgrade_from_version 011006002
	$files $ts '2013/21/05 10:30:04' $do upgrade_transfer_maillog_11_6_2
	$db    $ts '2013/20/01 13:00:00' $do db_upgrade_web_server_configurations_11_6_2
	$db    $ts '2013/20/01 15:50:00' $do add_db_changes 11.6.2
	$db    $ts '2013/20/07 18:30:00' $do db_upgrade_mail_description_11_6_2
	$db    $ts '2013/20/09 15:30:00' $do db_upgrade_clients_description_11_6_2

	cu_start_upgrade_from_version 011006003
	$db    $ts '2013/20/14 12:40:00' $do add_db_changes 11.6.3
	$db    $ts '2013/20/18 16:50:00' $do db_upgrade_domains_description_11_6_3
	$files $ts '2013/20/22 12:30:00' $do deferred_httpd_reconfigure_all

	cu_start_upgrade_from_version 011006004
	$files $ts '2013/21/05 12:05:30' $do reconfigure_chained_ssl_certificates_11_6_4

	cu_start_upgrade_from_version 012000005
	$files $ts '2013/21/28 16:37:15' $do deferred_mail_outgoing_restore_full
	$files $ts '2013/21/29 18:09:00' $do upgrade_cgi_bin_mode_defaults_12_0_5

	cu_start_upgrade_from_version 012000006
	$db    $ts '2013/21/29 12:30:00' $do db_upgrade_service_plan_domain_12_0_6
	$db    $ts '2013/22/03 12:30:00' $do db_upgrade_iis_app_pools_12_0_6
	$db    $ts '2013/22/06 18:00:00' $do db_upgrade_subscription_outgoing_messages_overlimit_percent_to_default_12_0_6
	$db    $ts '2013/22/06 18:00:00' $do db_upgrade_plan_outgoing_messages_overlimit_percent_to_default_12_0_6
	$db    $ts '2013/22/10 16:00:00' $do add_db_changes 12.0.6
	$files $ts '2013/22/13 16:22:59' $do upgrade_sendmail_wrapper_12_0_5

	cu_start_upgrade_from_version 012000007
	$db    $ts '2013/22/20 16:00:00' $do add_db_changes 12.0.7
	$files $ts '2013/22/20 16:27:50' $do upgrade_outgoing_mail_rate_limiting_handlers_12_0_7
	$db    $ts '2013/22/25 11:00:00' $do db_upgrade_sessions_12_0_7
### End of +10 to month number

	cu_start_upgrade_from_version 012000008
	$db	$ts    '2014/01/15 15:17:00' $do add_db_changes 12.0.8
	$db	$ts    '2014/01/23 17:30:00' $do db_upgrade_plan_properties_default_values_12_0_8
	$db	$ts    '2014/01/23 17:30:00' $do db_upgrade_subscription_properties_default_values_12_0_8

	cu_start_upgrade_from_version 012000009
	$db    $ts '2014/02/13 10:31:51' $do add_db_changes 12.0.9
	$db    $ts '2014/01/28 17:00:00' $do db_upgrade_plan_properties_default_values_12_0_9
	$db    $ts '2014/01/28 17:00:00' $do db_upgrade_subscription_properties_default_values_12_0_9
	$db    $ts '2014/01/28 17:00:00' $do db_upgrade_mbox_outgoing_messages_to_default_12_0_9
	$db    $ts '2014/01/28 17:00:00' $do db_upgrade_domain_outgoing_messages_to_default_12_0_9
	$files $ts '2014/02/04 12:57:05' $do upgrade_sendmail_wrapper_12_0_9
	$db    $ts '2014/02/04 17:00:00' $do db_upgrade_wordpress_instances_12_0_9
	$db    $ts '2014/02/12 17:00:00' $do db_upgrade_sessions_12_0_9

	cu_start_upgrade_from_version 012000010
	$db    $ts '2014/02/27 17:40:00' $do add_db_changes 12.0.10
	$files $ts '2014/02/25 17:48:22' $do upgrade_drop_sender_ip_mail_handler_12_0_10
	$db    $ts '2014/03/05 14:00:00' $do db_upgrade_enable_sendmail_default_values_12_0_10

	cu_start_upgrade_from_version 012000012
	$db    $ts '2014/03/18 16:30:00' $do db_upgrade_default_php_handler_12_0_12
	$files $ts '2014/03/27 12:58:00' $do upgrade_postfix_12_0_12

	cu_start_upgrade_from_version 012000013
	$db    $ts '2014/04/03 14:40:20' $do add_db_changes 12.0.13

	cu_start_upgrade_from_version 012000015
	$db    $ts '2014/04/15 15:48:00' $do add_db_changes 12.0.15

	cu_start_upgrade_from_version 012000017
	$db    $ts '2014/05/11 18:00:00' $do add_db_changes 12.0.17
	$db    $ts '2014/05/11 15:00:00' $do db_upgrade_wordpress_instances_12_0_17

    cu_start_upgrade_from_version 012000018
    $files $ts '2014/07/22 17:30:00' $do remove_unused_postfix_milters_12_0_18

	cu_start_upgrade_from_version 012001004
	$db    $ts '2014/08/08 13:00:00' $do db_upgrade_iis_app_pools_12_1_4
	$db    $ts '2014/08/18 17:28:30' $do db_upgrade_features_new_12_1_4
	$db    $ts '2014/08/26 11:10:00' $do db_upgrade_unified_db_12_1_4

	cu_start_upgrade_from_version 012001005
	$db    $ts '2014/09/01 15:00:00' $do db_upgrade_unified_db_12_1_5
	$files $ts '2014/09/03 19:10:00' $do db_upgrade_subscriptions_12_1_5 # This function is executed on post stage to fix column attributes for upgrade from 11.0.9 (there is post there)
	$files $ts '2014/09/05 14:00:00' $do db_upgrade_generate_service_instance_references_12_1_5 # This function is executed on post stage because rows are inserted on post stage on upgrades from 10.x
	$db    $ts '2014/09/11 17:00:00' $do db_drop_pmm_tables_12_1_5

	cu_start_upgrade_from_version 012001006
	$files $ts '2014/09/15 11:23:12' $do upgrade_rsyslog_maillog_configuration_12_1_6
	$db    $ts '2014/09/18 15:43:21' $do db_upgrade_smb_unified_12_1_6
	$files $ts '2014/09/23 11:33:13' $do db_remove_tables_12_1_6

	cu_start_upgrade_from_version 012001007
	$db    $ts '2014/10/02 12:57:35' $do db_drop_storefront_12_1_7
	$db    $ts '2014/10/03 17:40:00' $do add_db_changes 12.1.7

	cu_start_upgrade_from_version 012001008
	$files $ts '2014/10/14 12:19:57' $do refresh_trusted_ips_12_1_8
	$db    $ts '2014/10/16 17:55:00' $do add_db_changes 12.1.8
	$db    $ts '2014/10/20 10:28:15' $do db_upgrade_backups_scheduled_12_1_8

	cu_start_upgrade_from_version 012001013
	$files $ts '2015/01/12 14:57:11' $do regenerate_postfix_poplock_cron_task_12_1_13

	cu_start_upgrade_from_version 012001018
	$db    $ts '2015/03/26 13:25:22' $do php_upgrade_system_handlers_pre_12_1_18
	$files $ts '2015/03/26 14:01:54' $do php_upgrade_system_handlers_post_12_1_18

	cu_start_upgrade_from_version 012001020
	$files $ts '2015/04/15 14:03:10' $do register_local_postgres_12_1_20

	cu_start_upgrade_from_version 012001022
	$files $ts '2015/05/19 16:53:10' $do configure_ssl_ciphers_protocols

	cu_start_upgrade_from_version 012001024
	$files $ts '2015/06/10 12:53:10' $do upgrade_transfer_rc_actions_log

	cu_start_upgrade_from_version 012005030
	$files $ver                      $do move_out_custom_vhost_templates custom_templates_before_17.0

	cu_start_upgrade_from_version 013000001
	$files $ts '2015/10/28 18:05:40' $do upgrade_dump_tmp_d_13_0_2

	cu_start_upgrade_from_version 017000011
	$files $ts '2016/04/25 17:20:00' $do upgrade_gl_rules_17_0_12

	cu_start_upgrade_from_version 017000013
	$files $ts '2016/06/23 10:50:00' $do remove_webmail_configs

	cu_start_upgrade_from_version 017000015
	$files $ts '2016/09/09 12:47:00' $do replace_gitman_to_git

	cu_start_upgrade_from_version 017000017
	$files $ts '2017/03/09 12:43:00' $do update_registered_services_db_for_resctrl

	cu_start_upgrade_from_version 017005002
	$files $ts '2017/03/09 13:22:00' $do update_dump_d_mode_17_5_3

	# DO NOT MODIFY ANYTHING BELOW THIS LINE! (without approval of review from the entire Linux team)
	call_optional_function dev_core_upgrader
	cu_verify_latest_timestamp_lt '2017/03/25 11:00:00'
}

###
### HOW TO ADD UPGRADE ACTION
###
### Short version:
###  To the end of core_upgrader() function (above) add a line in the format:
###   STAGE $ts 'TIMESTAMP' $do UPGRADE_ACTION
###  Where:
###   STAGE is either '$db' or '$files' without quotes. If your upgrade action affects only Plesk DB and doesn't require any
###       new files (e.g. PHP scripts) to run, then use '$db'. Otherwise use '$files'.
###   TIMESTAMP is a current (at the moment you write your code) timestamp in the format 'YYYY/MM/DD hh:mm:ss'.
###       You can use output of `date '+%Y/%m/%d %H:%M:%S'` command.
###   UPGRADE_ACTION is a call to the function that implements your upgrade action. While you can supply arguments here, it is
###       advised not to. Function name should reflect what you are upgrading and *from* which version. Usually if you use
###       '$db' as STAGE then function name should start with 'db_upgrade_', otherwise with 'upgrade_'.
###
###  If you use 'add_db_changes <VERSION>' as UPGRADE_ACTION add your update SQL to plesk-install/psa_db_upgrade_<VERSION>.sql,
###  don't forget to add the aforementioned file to plesk-install/Makefile.
###
###
### Slightly longer version:
###  'cu_start_upgrade_from_version <version>' calls mark groups of upgrade actions for upgrade from a given version. They
###      affect only upgrade actions with '$ver' filter, but they also serve a role of self-documentation and therefore
###      should be added for any product version change.
###  '$db' marks upgrade actions that will only be executed in prep-install bootstrapper stage. In case of critical failure
###      all changes made to Plesk DB here will be automatically reverted. You should do only changes to Plesk DB during
###      this stage; don't use anything not packaged with bootstrapper here (you may use db_*() and add_db_changes() functions).
###  '$files' marks upgrade actions that will only be executed in post-install bootstrapper stage. During this stage all
###      required '$db' actions have already been completed (unless some of them were optional), all new packaged files
###      are already in place and there are no old removed packaged files. You can do pretty much everything at this stage.
###  '$ver' marks that whether or not this upgrade action should be called will be checked based on previous DB or product
###      version (for use with '$db' and '$files' respectively). This requires cu_start_upgrade_from_version() call somewhere
###      above.
###  '$ts TIMESTAMP' marks that whether or not this upgrade action should be called will be checked based on core component
###      configuration version (which is a timestamp). Any upgrade actions with a TIMESTAMP newer than configuration version
###      will be executed. This approach allowes smooth upgrade between builds within a same product version. Use it.
###      TIMESTAMP is a string with a current timestamp in any sensible format which allows sane ordering when punctuation is
###      stripped. Output of the following command is recommended for use as TIMESTAMP: `date '+%Y/%m/%d %H:%M:%S'`.
###      Usually timestamps will increase from previous upgrade action to the next, but it is not a requirement.
###  '$do' is just a required word that should preceed upgrade action call. The call itself may include parameters, but it's
###      advised not to, since it's much better to have upgrade actions uniquely identifiable by name.
###
###  Upgrade action should be more or less self-contained. Never do several unrelated things in a single upgrade action.
###  Never separate related upgrade actons across several calls in core_upgrader() unless there is a good reason to.
###
###  Upgrade actions can be either critical or optional. Optional upgrade actions should indicate failure with non-zero return
###  code. Critical upgrade actions should indicate failure with die() or similar calls (that terminate execution). Note that
###  using most of db_*() functions makes an upgrade action a critical one. Critical failure of a '$db' upgrade action will
###  trigger DB rollback. Critical failure of a '$files' upgrade action will switch core configuration into a broken unusable
###  state, from which one can recover only by successfully passing all required upgrade actions during bootstrapper repair
###  procedure. "Optional" failure of any upgrade action will add it to the list of failed ones and continue with upgrade.
###  All failed actions are rerun during bootstrapper repair procedure.
###


#vhosts

install_vhosts()
{
	local inten="create Virtual Hosts directory"
# Paranoia
	echo_try $inten
	mkdir -p "$HTTPD_VHOSTS_D/default"
	mkdir -p "$HTTPD_VHOSTS_D/.skel"
	suc

	inten="sync default virtual host"
	echo_try $inten

	rsync -rlptgoD --ignore-existing /usr/share/plesk-vhost/ $HTTPD_VHOSTS_D/default/
	suc
# Dir check for awoid overwrite in future
	if [ ! -d "$HTTPD_VHOSTS_D/.skel/0" -a -x "$PRODUCT_ROOT_D/admin/sbin/skelmng" ]; then
		$PRODUCT_ROOT_D/admin/sbin/skelmng --install-default --id 0 >> "$product_log" 2>&1
	fi
}

chown_webstat()
{
	# select only domains with hosting
	query="select d.id from domains d, hosting h where d.id=h.dom_id";

	for dom_id in `echo $query | mysql`; do

		query1="select s.login from sys_users s, hosting h where s.id=h.sys_user_id and h.dom_id='$dom_id'";
		login=`echo $query1 | mysql`;

		query1="select name from domains where id='$dom_id'";
		dom_name=`echo $query1 | mysql`;

		if [ -d "$HTTPD_VHOSTS_D/$dom_name/httpdocs/webstat" ]; then
#			chown $login:$clients_group `get_ls full "$HTTPD_VHOSTS_D/$dom_name/httpdocs/webstat"`
			find "$HTTPD_VHOSTS_D/$dom_name/httpdocs/webstat" -exec chown $login:$clients_group {} \;
		fi

	done >> "$product_log" 2>&1
}

upgrade_vhost_directories_11_1_18()
{
	echo_try "upgrade vhost directories for domains"
	$PRODUCT_ROOT_D/bin/repair --update-vhosts-structure -no-reconfigure && suc || warn "fail\nCannot upgrade vhost directories for domains. Please run $PRODUCT_ROOT_D/bin/repair --update-vhosts-structure manually."
	deferred_httpd_reconfigure_all
}

upgrade_skel_remove_statistics_11_1_18()
{
	echo_try "drop unused 'statistics' directory from vhost skeletons"
	local unmodified_skel_stat_md5sum="3f5517860e8adfa4b05c9ea6268b38eb"
	local skel_stat_md5sums="`find -L $HTTPD_VHOSTS_D/.skel/*/statistics -type f 2>/dev/null | xargs --no-run-if-empty md5sum | cut -d ' ' -f 1 | sort | uniq`"
	if [ -z "$skel_stat_md5sums" -o "$skel_stat_md5sums" = "$unmodified_skel_stat_md5sum" ]; then
		rm -rf $HTTPD_VHOSTS_D/.skel/*/statistics >>"$product_log" 2>&1
	else
		pnnl_echo " 'statistics' directory seems customized, so it was preserved. Please remove it later manually. "
	fi
	suc
}


#webalizer

set_webalizer_params()
{
	WEBALIZER_ROOT_D="$PRODUCT_ROOT_D/webalizer"
}

## @@constructor set_webalizer_params

set_watchdog_params()
{
	watchdog_wd_bin="$PRODUCT_ROOT_D/admin/bin/modules/watchdog/wd"
	watchdog_configure_bin="$PRODUCT_ROOT_D/libexec/modules/watchdog/configure/configure"
	watchdog_wdservice_bin="$PRODUCT_ROOT_D/admin/sbin/modules/watchdog/wdservice"
	watchdog_monit_bin="$PRODUCT_ROOT_D/admin/bin/modules/watchdog/monit"

	watchdog_monitrc="$PRODUCT_ETC_D/modules/watchdog/monitrc"
	watchdog_wdcollect_inc="$PRODUCT_ETC_D/modules/watchdog/wdcollect.inc.php"
}

wd_adapt()
{
	set_watchdog_params
    if [ -x "$watchdog_wd_bin" ] ; then
        "$watchdog_wd_bin" --adapt >>$product_log 2>&1
    fi
}

wd_configure()
{
	set_watchdog_params
	if [ -x "$watchdog_configure_bin" ] ; then
		"$watchdog_configure_bin" >>$product_log 2>&1
	fi
}

wd_start()
{
	set_watchdog_params
	if [ -x "$watchdog_wdservice_bin" ]; then
		"$watchdog_wdservice_bin" monit start "$watchdog_monitrc" || true
		"$watchdog_wdservice_bin" wdcollect start "$watchdog_wdcollect_inc" || true
	fi
}

# vim:ft=sh

read_postconf_value()
{
	local param="$1"
	/usr/sbin/postconf -h "$param" 2>>"$product_log" | sed -e 's/\(\s*,\s*\|\s\+\)/\n/g' | egrep -v '^\s*$'
}

write_postconf_value()
{
	local param="$1"
	local value="$2"
	local conf_value="`echo "$value" | awk 'BEGIN { ORS=""; PORS="" } { print PORS; print; PORS=", " }'`"
	/usr/sbin/postconf -e "$param=$conf_value" >>"$product_log" 2>&1
}


set_postfix_params()
{
	postfix_service="postfix"
	pc_remote_service="pc-remote"
}

true postfix_status
postfix_status()
{
	# here be dragons.
	# the practical experience shows that simple checking of status of
	# Postfix "master" process is not enough. So we read Postfix master
	# process pid file if any, then try to look for a process with
	# name ``qmgr'' and parent pid being equal to
	# the pid read from the pidfile. If pgrep finds such a process
	# it returns 0, if not its exit status is non-zero.
	# pgrep is portable enough to prefer it to "hand-made" alternatives
	# such as famous ``ps | grep $name | grep -v grep...'' pipes
	# bug 147822. do not interrupt installation for FreeBSD

	[ -f "/var/spool/postfix/pid/master.pid" ] || return 1

	local ppid

	read ppid </var/spool/postfix/pid/master.pid 2>/dev/null
	if [ $? -ne 0 -o -z "$ppid" ]; then
		# not found or other error
		return 1;
	fi
	pgrep -P $ppid qmgr >/dev/null 2>/dev/null
}
# vim:ft=sh

upgrade_postfix_allow_authenticated_dnsbled_clients_11_1_17()
{
	[ -s "/etc/postfix/main.cf" ] || return 0

	set_postfix_params

	local filtered_value="`read_postconf_value 'smtpd_client_restrictions' | sed -e '/^permit_mynetworks$/d' -e '/^permit_sasl_authenticated$/d'`"
	local extended_value="permit_mynetworks
permit_sasl_authenticated
$filtered_value"

	write_postconf_value 'smtpd_client_restrictions' "$extended_value" || return 1
	pleskrc postfix reload
}
# vim:ft=sh


upgrade_postfix_remove_deprecated_headers_11_5_30()
{
	local filtered_value
	local inten
	local res

	[ -s "/etc/postfix/main.cf" ] || return 0

	set_postfix_params


	inten="drop header inclusion 'X-No-Auth: unauthenticated sender' from postfix"
	echo_try $inten
	filtered_value="`read_postconf_value 'smtpd_sender_restrictions' | sed -e '/^check_client_access\s\+pcre:.*non_auth.re/d'`"
	write_postconf_value 'smtpd_sender_restrictions' "$filtered_value" && suc || warn $inten

	inten="drop header inclusion 'X-No-Relay: not in my network' from postfix"
	echo_try $inten
	filtered_value="`read_postconf_value 'smtpd_recipient_restrictions' | sed -e '/^check_client_access\s\+pcre:.*no_relay.re/d'`"
	write_postconf_value 'smtpd_recipient_restrictions' "$filtered_value" && suc || warn $inten

	pleskrc postfix reload
}

# vim:ft=sh

upgrade_postfix_12_0_12()
{
	# Remove old plesk transports
	local master_cf_path=/etc/postfix/master.cf
	[ -s "$master_cf_path" ] || return 0

	cat $master_cf_path | perl -ne '
		@line = split /\s+/, $_, 8;
		next if ($line[1] == "unix" && ($line[0] =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}-/ || $line[0] =~ /^-[0-9a-fA-F-]+/));
        print $_;
    ' | perl -p -00e0 > $master_cf_path.tmp && mv -f $master_cf_path.tmp $master_cf_path

	# Call for mail_transport_restore

	/usr/lib/plesk-9.0/remote_mail_restore/mail_transport_restore
}

regenerate_postfix_poplock_cron_task_12_1_13()
{
	if /opt/psa/admin/sbin/mailmng-server --features | egrep -q "SMTP_Server_package.*postfix" 2>/dev/null; then
		remove_from_crontab "/usr/lib/plesk-9.0/postfix-poplockdb-clean"
		set_postfix_auth
	fi
}
register_webmail()
{
		local name="$1"
		local inten="register webmail '$1'"
		
		[ -x "$webmail_tool" ] || die "$inten : '$webmail_tool' is not executable"
		[ -z "$name" ] && die $inten

		$webmail_tool --install --name="$name" || warn $inten
		deferred_httpd_reconfigure
}

install_db_file()
{
	local file="$1"
	local mysql_db_name="$2"
	local inten="$3"
	echo_try $inten
	[ -f "$file" ] || die "Unable to $inten: $file not found"
	cat $file | mysql >>"$product_log" 2>&1 || die "Unable to $inten, see $product log for details"
	suc
}

remove_webmail_configs()
{
	local apache_conf_d="$HTTPD_CONF_D/plesk.conf.d"
	local nginx_conf_d="/etc/nginx/plesk.conf.d"
	rm -rf 		$apache_conf_d/roundcube.conf* 		$apache_conf_d/horde.conf* 		$apache_conf_d/atmail.conf* 		$apache_conf_d/atmailcom.conf* 		$apache_conf_d/webmails/* 		$nginx_conf_d/webmail.conf* 		$nginx_conf_d/webmails/*
	deferred_httpd_reconfigure_all
}
set_webmail_params()
{
	webmail_conf_d="/etc/psa-webmail"
	webmail_tool="${PRODUCT_ROOT_D}/admin/bin/webmailmng"
}

set_horde_params()
{
	horde_datadir="/usr/share/psa-horde"
	horde_confdir="/etc/psa-webmail/horde"
	horde_sysconfd="$horde_confdir/horde"
	horde_logdir="/var/log/psa-horde"
	horde_passwd_file="/etc/psa-webmail/horde/.horde.shadow"
	horde_title="Horde Web Based mail client"

	imp_datadir="$horde_datadir/imp"
	imp_sysconfd="$horde_confdir/imp"
	
	turba_datadir="$horde_datadir/turba"
	turba_sysconfd="$horde_confdir/turba"

	kronolith_datadir="$horde_datadir/kronolith"
	kronolith_sysconfd="$horde_confdir/kronolith"

	ingo_datadir="$horde_datadir/ingo"
	ingo_sysconfd="$horde_confdir/ingo"

	mnemo_datadir="$horde_datadir/mnemo"
	mnemo_sysconfd="$horde_confdir/mnemo"
	
	passwd_datadir="$horde_datadir/passwd"
	passwd_sysconfd="$horde_confdir/passwd"

	horde_user="horde_sysuser"
	horde_group="horde_sysgroup"

	horde_php_ini="/etc/psa-webmail/horde/horde/php.ini"
}


horde_run_migration_script()
{
	local app=$1
	local migrate_cmd="${psa_pear_dir}/pear/horde-db-migrate --config=${psa_pear_conf} ${app}"
	local horde_conf=${horde_sysconfd}/conf.php

	if [ -z "${app}" ]; then
		#we can not exec script if cache driver is sql, cause cache table may be not installed yet
		#if app is specified it means that framework is already configured, so cache tables is already created
		sed -i -e "s|\$conf\['cache'\]\['driver'\].*|\$conf['cache']['driver'] = 'File';|g" ${horde_conf}
	fi

	horde_run_php_script ${migrate_cmd} || warn "install or upgrade horde database, please re-run ${migrate_cmd}"

	if [ -z "${app}" ]; then
		sed -i -e "s|\$conf\['cache'\]\['driver'\].*|\$conf['cache']['driver'] = 'File';|g" ${horde_conf}
	fi
}

horde_run_php_script()
{
	env PHP_PEAR_SYSCONF_DIR=$psa_pear_dir sw-engine -d "include_path=${psa_pear_include_dir}:." $@ >> $product_log 2>&1
}

horde_convert_shares_to_sql()
{
	local app_name=$1
	local script_name=${psa_pear_dir}/pear/${app_name}-convert-datatree-shares-to-sql

	if ! db_test_table horde_datatree; then
		p_echo "Skip convert datatree to mysql - table horde_datatree is missed"
		return
	fi

	db_select "SELECT datatree_id FROM horde_datatree WHERE group_uid = 'horde.shares.${app_name}'"
	[ -n "${db_select_output}" ] || return 0


	if [ ! -f "${script_name}" ]; then
		warn "convert datatree shares to sql - missed convertion script"
		return
	fi

	# NOTE: using echo instead of printf is breaks correct execution on DEB-like OSes
	printf '1\ny' | horde_run_php_script "${script_name}" || warn "execute '${script_name}'"
}

horde_convert_shares_to_sqlng()
{
	local app_name=$1
	local script_name=${psa_pear_dir}/pear/${app_name}-convert-sql-shares-to-sqlng

	# NOTE: using echo instead of printf is breaks correct execution on DEB-like OSes
	printf 'y' | horde_run_php_script "${script_name}" || warn "execute '${script_name}'"
}

upgrade_horde_DB_to_latest_h3()
{
	local mysql_db_name="horde"

	#H5 or above detect
	! db_test_table horde_core_schema_info || return 0

	if ! db_test_table horde_syncml_map; then
        	install_db_file "$horde_datadir/.upgrade/3.1_to_3.2.mysql.sql"                 	$mysql_db_name                 	"upgrade horde database 3.1 -> 3.2"
	fi

	local inten="upgrade horde db sequences"
	horde_run_php_script "$horde_datadir/.upgrade/2008-08-29_fix_mdb2_sequences.php" && suc || warn "$inten"

	if db_test_column horde_signups signup_email; then
		install_db_file "$horde_datadir/.upgrade/3.3_to_3.3.5.sql" 					$mysql_db_name                 	"upgrade horde database 3.3 -> 3.3.5" 
	fi
}

upgrade_imp_DB_to_latest_h3()
{
	local mysql_db_name="horde"

	#H5 or above detect
	! db_test_table imp_schema_info || return 0

	if ! db_test_table imp_sentmail; then
		install_db_file "$imp_datadir/.upgrade/imp.sql" 			$mysql_db_name                 	"upgrade imp database 4.1.x -> 4.2.x"
	fi
}

upgrade_ingo_DB_to_latest_h3()
{
	local mysql_db_name="horde"

	#H5 or above detect
	! db_test_table ingo_schema_info || return 0

	if ! db_test_table ingo_shares_users; then
		install_db_file "$ingo_datadir/.upgrade/ingo.sql" 			$mysql_db_name                 	"upgrade ingo database 1.1.x -> 1.2"
	fi

	if db_test_table ingo_shares_users; then
		install_db_file "$ingo_datadir/.upgrade/1.2_to_1.2.1.sql" 			$mysql_db_name                 	"upgrade ingo database 1.2 -> 1.2.1"
	fi

	if db_test_column ingo_shares_groups group_uid; then
		install_db_file "$ingo_datadir/.upgrade/1.2.1_to_1.2.2.sql" 			$mysql_db_name                 	"upgrade ingo database 1.2.1 -> 1.2.2"
	fi
}

upgrade_kronolith_DB_to_latest_h3()
{
	local mysql_db_name="horde"

	#H5 or above detect
	! db_test_table kronolith_schema_info || return 0

	if ! db_test_table kronolith_shares_groups; then
		install_db_file "$kronolith_datadir/.upgrade/2.1_to_2.2.sql" 			$mysql_db_name                 	"upgrade kronolith database 2.1 -> 2.2"
	fi

	if db_test_column kronolith_shares share_owner; then
		install_db_file "$kronolith_datadir/.upgrade/2.2_to_2.3.sql" 			$mysql_db_name                 	"upgrade kronolith database 2.2 -> 2.3"
	fi
}

upgrade_mnemo_DB_to_latest_h3()
{
	local mysql_db_name="horde"

	#H5 or above detect
	! db_test_table mnemo_schema_info || return 0

	if ! db_test_table mnemo_shares; then
		install_db_file "$mnemo_datadir/.upgrade/2.1_to_2.2.sql" 			$mysql_db_name                 	"upgrade mnemo database 2.1 -> 2.2"
	fi

	if db_test_column mnemo_shares share_owner; then
		install_db_file "$mnemo_datadir/.upgrade/2.2_to_2.2.1.sql" 			$mysql_db_name                 	"upgrade mnemo database 2.2 -> 2.2.1"
	fi

	if db_test_column mnemo_shares_groups group_uid; then
		install_db_file "$mnemo_datadir/.upgrade/2.2.1_to_2.2.2.sql" 			$mysql_db_name                 	"upgrade mnemo database 2.2.1 -> 2.2.2"
	fi

}


upgrade_turba_DB_to_latest_h3()
{
	local mysql_db_name="horde"

	#H5 or above detect
	! db_test_table turba_schema_info || return 0

	if ! db_test_table turba_shares; then
		install_db_file "$turba_datadir/.upgrade/2.1_to_2.2_add_sql_share_tables.sql" 			$mysql_db_name                 	"upgrade turba database 2.1 -> 2.2"
	fi

	if ! db_test_column turba_objects object_firstname; then
		local inten="change turba address book scheme"
		horde_run_php_script "$turba_datadir/.upgrade/2.1_to_2.2_sql_schema.php" && suc || warn "$inten"
	fi

	if db_test_column turba_shares share_owner; then
		install_db_file "$turba_datadir/.upgrade/2.2.1_to_2.3.sql" 		$mysql_db_name                 "upgrade turba database 2.2.1 -> 2.3"
	fi

}

generate_horde_password()
{
	generate_password_file "$horde_passwd_file" "$horde_title" root "$horde_group" 640
	horde_password=`cat $horde_passwd_file` || die $inten
}

generate_horde_php_ini()
{
	${PRODUCT_ROOT_D}/admin/sbin/phpinimng --type=cgi ${horde_php_ini} < ${horde_php_ini}.tpl
}

switch_ingo_backend()
{
	set_horde_params

	local backend="$1"
	if [ -z "$backend" ]; then
		if is_courier_imap_active ; then
			backend="imap"
		elif is_dovecot_active ; then
			backend="sieve"
		fi
	fi

	local inten="switch Horde Ingo backend to '$backend'"
	echo_try "$inten"
	[ "$backend" = "imap" -o "$backend" = "sieve" ] || {
		warn "$inten. The backend is unknown or failed to detect suitable backend."
		return 1
	}

	local current_backend="$ingo_sysconfd/backends.local.php"
	[ -L "$current_backend" ] || {
		p_echo "Ingo doesn't seem to be installed. Skip."
		return 0
	}

	local target_backend="backends.local.$backend.php"
	[ -f "$ingo_sysconfd/$target_backend" ] || {
		warn "$inten. Target backend configuration file '$ingo_sysconfd/$target_backend' doesn't exist."
		return 1
	}

	ln -sf "$target_backend" "$current_backend" >> $product_log 2>&1 && suc
}

update_roundcube_db_settings()
{
	local cfg_db_dsnw="\$config\['db_dsnw'\]"
	sed -e "s#${cfg_db_dsnw}.*#${cfg_db_dsnw} = 'mysql://roundcube:${roundcube_password}@localhost/roundcubemail';#" 		-i ${roundcube_conf}
}

roundcube_fix_perms_on_private_files()
{
	# update perms on private roundcube files (for upgrade)
	set_ac root "$roundcube_group" 0640 ${roundcube_conf}
	set_ac root root 0640 ${roundcube_passwd_file}

	# remove redundant files from previous roundcube versions:
	rm -f "${roundcube_php_ini}.tpl"

	# fix permissions on log and temp dirs (at least for debian-like oses):
	set_ac "$roundcube_user" "root" 0750 "$roundcube_logs_d"
	set_ac "$roundcube_user" "root" 0750 "$roundcube_temp_d"

	# fix permissions on log files
	for l in `ls "${roundcube_logs_d}"`; do
		chown ${roundcube_user}:${roundcube_group} "${roundcube_logs_d}/$l"
	done
}

roundcube_fix_db_grants()
{
	local mysql_db_name='mysql'
	db_do "GRANT SELECT, INSERT, UPDATE, DELETE, CREATE, DROP, INDEX, ALTER 
			ON \`roundcubemail\`.* TO 'roundcube'@'localhost'; 
			FLUSH PRIVILEGES;"
}

roundcube_generate_php_ini()
{
	${PRODUCT_ROOT_D}/admin/sbin/phpinimng --type=cgi "${roundcube_php_ini}" < "${roundcube_php_ini_tpl}"
	# get rid of warnings/errors about unspecified timezone:
	egrep -q '^\s*date.timezone\s*=\s*\S' "${roundcube_php_ini}" || echo 'date.timezone = UTC' >> "${roundcube_php_ini}"
}

roundcube_get_current_files_version()
{
	local version_file="$roundcube_home/program/include/iniset.php"
	sed -ne 's/^\s*define(.RCMAIL_VERSION.,\s*.\([0-9.]\+[a-z-]*\).);.*$/\1/p' "$version_file" 2>>"$product_log"
}

roundcube_touch_errors_log()
{
	# make fail2ban happy: create 'errors'-log after install for plesk-roundcube jail
	touch "$roundcube_logs_d/errors"
	chown "$roundcube_user:$roundcube_group" "$roundcube_logs_d/errors"
}
# vim:set ft=sh:

set_roundcube_params()
{
	roundcube_passwd_file="/usr/share/psa-roundcube/config/.roundcube.shadow"
	roundcube_title="Roundcube Web Based mail client"
	roundcube_conf="/usr/share/psa-roundcube/config/config.inc.php"
	roundcube_sql_d="/usr/share/psa-roundcube/SQL"
	roundcube_user="roundcube_sysuser"
	roundcube_group="roundcube_sysgroup"
	roundcube_home="/usr/share/psa-roundcube"
	roundcube_logs_d="/var/log/plesk-roundcube"
	roundcube_temp_d="/var/tmp/plesk-roundcube"

	roundcube_php_ini="/etc/psa-webmail/roundcube/php.ini"
	roundcube_php_ini_tpl="/usr/share/psa-roundcube/config/php.ini.tpl"

	roundcube_preupgrade_version_file="/etc/psa-webmail/roundcube/version.upg"
}

# vim:syntax=sh

set_nginx_params()
{
	nginx_service=nginx
	nginx_rc_config="/etc/default/nginx"
	nginx_user="nginx"
}


reconfigure_chained_ssl_certificates_11_6_4()
{
	local inten="reinstall chained SSL certificates"
	echo_try "${inten}"

	if ! $PRODUCT_ROOT_D/bin/repair --reconfigure-ssl-certificates >>$product_log 2>&1; then
		warn "${inten}"
		return 1
	else
		suc
	fi
}


fix_file_sharing_permissions_from_10_0_0()
{
	set_apache_params

	local file_sharing_passwd_file="${HTTPD_VHOSTS_D}/fs-passwd/dav.passwd"

	[ -f "${file_sharing_passwd_file}" ] || return 0

	chown psaadm:${apache_group} "${file_sharing_passwd_file}" 
	chmod 640 "${file_sharing_passwd_file}"
}

install_ssh_banner_config()
{
	local config="/root/.plesk_banner"
	[ ! -e "$config" ] || return 0

	cat > "$config" <<-EOF
		# This script is indirectly called by Bash on login shell startup.
		# It outputs Plesk shell banner with login instructions.
		# If you don't need it, you may comment out the following line.
		$PLESK_LIBEXEC_DIR/ssh-banner || :
		EOF
}

install_ssh_banner_hook()
{
	local into_config="$1"
	local invoker="[ ! -s ~/.plesk_banner ] || . ~/.plesk_banner"

	if ! grep -qxF "$invoker" "$into_config"; then
		sed -e "/~\/.plesk_banner/ d" -i "$into_config" || die "remove '.plesk_banner' from '$into_config'"
		echo >> "$into_config" && echo "$invoker" >> "$into_config" || die "add '.plesk_banner' into '$into_config'"
	fi
}

install_ssh_banner()
{
	local inten="install Plesk shell banner for root"
	echo_try "$inten"

	install_ssh_banner_config

	local hooked=
	for bash_config in .bash_profile .bash_login .profile; do
		local config="/root/$bash_config"
		[ -e "$config" ] || continue
		install_ssh_banner_hook "$config"
		hooked="yes"
	done

	if [ -z "$hooked" ]; then
		local config="/root/.bash_profile"
		touch "$config"
		install_ssh_banner_hook "$config"
	fi

	suc
}
upgrade_adminlogs_permissions_1100()
{
    chown psaadm:0 ${PRODUCT_ROOT_D}/admin/logs
    chmod 750 ${PRODUCT_ROOT_D}/admin/logs
}

transfer_log_with_compat_symlink()
{
	local old_path="$1"
	local new_path="$2"
	local merge="$3"
	local skip_link="$4"

	if [ -f "$old_path" -a ! -L "$old_path" ]; then
		if [ ! -f "$new_path" -o ! -s "$new_path" ] ; then
# simply move file into new place
			mv -f "$old_path" "$new_path" || { warn "move '$old_path' to '$new_path'"; return 1; }
		elif [ "$merge" = 'merge' ]; then
			local new_path_tmp=`mktemp "${new_path}.XXXX"`
			mv -f "$old_path" "$new_path_tmp" || { warn "move '$old_path' to temporary '$new_path_tmp'"; return 1; }
			cat "$new_path" >> "$new_path_tmp" || { warn "cat '$new_path' to temporary '$new_path_tmp'"; return 1; }
			mv -f "$new_path_tmp" "$new_path" || { warn "move temporary '$new_path_tmp' to '$new_path'"; return 1; }
		else
# preserve old log with suffix in target directory
			mv -f "$old_path" "$new_path.$product_suff"  || { warn "move '$old_path' to '$new_path.$product_suff'"; return 1; }
		fi
	elif [ "${skip_link}" = "skip_link_if_absent" ]; then
		return 0
	fi

	[ ! "${skip_link}" = "skip_link_always" ] || return 0

# (re-)create compatibility symlink
	if [ ! -e "$old_path" -o -L "$old_path" ]; then
		ln -sf "$new_path" "$old_path" || { warn "create symlink '$old_path' to '$new_path'"; return 1; }
	fi
}

upgrade_panel_log_11_6_0()
{
	local old_panel_log="${PRODUCT_ROOT_D}/admin/logs/panel.log"
	local new_panel_log="${PRODUCT_LOGS_D}/panel.log"

	transfer_log_with_compat_symlink "$old_panel_log" "$new_panel_log"
}

upgrade_httpsd_access_log_11_6_0()
{
	local old_path="${PRODUCT_ROOT_D}/admin/logs/httpsd_access_log"
	local new_path="${PRODUCT_LOGS_D}/httpsd_access_log"

	transfer_log_with_compat_symlink "$old_path" "$new_path"
}

upgrade_transfer_rotated_logs_11_6_0()
{
	for log in health-alarm.log panel.log sitebuilder.log; do
		if ls ${PRODUCT_ROOT_D}/admin/logs/${log}.*.gz > /dev/null 2>&1; then
			mv -f ${PRODUCT_ROOT_D}/admin/logs/${log}.*.gz "${PRODUCT_LOGS_D}"
		fi
	done
}

upgrade_transfer_httpsd_access_log_processed_11_6_1()
{
	if ls ${PRODUCT_ROOT_D}/admin/logs/httpsd_access_log.processed* > /dev/null 2>&1; then
		mv -f ${PRODUCT_ROOT_D}/admin/logs/httpsd_access_log.processed* "${PRODUCT_LOGS_D}"
	fi
}

upgrade_transfer_maillog_11_6_2()
{
	transfer_log_with_compat_symlink "${PRODUCT_ROOT_D}/var/log/maillog" "/var/log/maillog" 'merge'
	select_maillog

# transfer processed and rotated logs
	if ls ${PRODUCT_ROOT_D}/var/log/maillog.processed* > /dev/null 2>&1; then
		mv -f ${PRODUCT_ROOT_D}/var/log/maillog.processed* "/var/log"
	fi
}

upgrade_transfer_rc_actions_log()
{
	local log="${PRODUCT_ROOT_D}/tmp/rc_actions.log"
	if [ -s "$log" ]; then
		mv -f "$log" "${PRODUCT_LOGS_D}/"  > /dev/null 2>&1;
	fi
}
# vim:ft=sh

deferred_httpd_reconfigure()
{
	touch "/var/lock/plesk_deferred_httpd_reconfigure"
}

deferred_httpd_reconfigure_all()
{
	touch "/var/lock/plesk_deferred_httpd_reconfigure_all"
}

deferred_watchdog_reconfigure()
{
	touch "/var/lock/plesk_deferred_watchdog_reconfigure"
}

deferred_relabel_plesk_directories_with_mailnames()
{
	touch "/var/lock/plesk_deferred_relabel_plesk_directories_with_mailnames"
}

deferred_mail_restore_all()
{
	touch "/var/lock/plesk_deferred_mail_restore_all"
}

deferred_mail_spam_restore_full()
{
	touch "/var/lock/plesk_deferred_mail_spam_restore_full"
}

deferred_mail_outgoing_restore_full()
{
	touch "/var/lock/plesk_deferred_mail_outgoing_restore_full"
}

deferred_webserver_apache_configure()
{
	touch "/var/lock/plesk_deferred_webserver_apache_configure"
}

run_now_webserver_apache_configure()
{
	[ ! -x "$PLESK_LIBEXEC_DIR/services/webserver.apache" ] || $PLESK_LIBEXEC_DIR/services/webserver.apache configure
}

deferred_ftp_proftpd_configure()
{
	touch "/var/lock/plesk_deferred_ftp_proftpd_configure"
}

run_now_ftp_proftpd_configure()
{
	[ ! -x "$PLESK_LIBEXEC_DIR/services/ftp.proftpd" ] || $PLESK_LIBEXEC_DIR/services/ftp.proftpd configure
}

deferred_awstats_configure()
{
	touch "/var/lock/plesk_deferred_awstats_configure"
}

run_now_awstats_configure()
{
	[ ! -x "$PLESK_LIBEXEC_DIR/statistics/awstats" ] || $PLESK_LIBEXEC_DIR/statistics/awstats configure
}

deferred_super_server_action_configure()
{
	touch "/var/lock/plesk_deferred_super_server_action_configure"
}

run_now_notify_imap_pop3_service_about_mailserver_change()
{
	set_courier_imap_params
	if pleskrc courier_authdaemon exists; then
		pleskrc courier_authdaemon condrestart
	fi
	# Dovecot, on the other hand, will notice authentication DB (mail server) change
	# in a couple of minutes when authentication worker respawns. No need to restart entire service.
}

### perform actions :
perform_deferred_actions()
{
	local flag_files="relabel_plesk_directories_with_mailnames 		relabel_plesk_directories 		super_server_action_configure 		watchdog_configure 		watchdog_reconfigure 		watchdog_start 		ftp_proftpd_configure 		webserver_apache_configure 		httpd_reconfigure_all 		httpd_reconfigure 		apache_restart 		nginx_restart 		mail_restore_all 		mail_mailbox_restore_full 		mail_spam_restore_full 		mail_outgoing_restore_full 		notify_imap_pop3_service_about_mailserver_change 		apply_ssl_settings_to_imap_pop3 		apply_ssl_settings_to_smtp 		awstats_configure 		hmonitor_start_notificationd 		mail_restore_imap_pop3 	"
	local res=0
	for flag in $flag_files; do
		if [ -e "/var/lock/plesk_deferred_$flag" ]; then
			local local_res=""
			case $flag in
				watchdog_configure)
					wd_configure && wd_adapt
					# ignore watchdog failed actions (distupgrade)
					local_res=0
				;;
				watchdog_reconfigure)
					wd_adapt	
				;;
				watchdog_start)
					wd_start
				;;
				ftp_proftpd_configure)
					run_now_ftp_proftpd_configure
				;;
				webserver_apache_configure)
					run_now_webserver_apache_configure
				;;
				awstats_configure)
					run_now_awstats_configure
				;;
				httpd_reconfigure_all)
					httpdmng_reconfigure all -service-node local
				;;
				httpd_reconfigure)
					httpdmng_reconfigure server
				;;
				relabel_plesk_directories)
					if [ -s "/opt/psa/version.upg" ]; then
						relabel_plesk_directories
					else
						relabel_plesk_directories "$HTTPD_VHOSTS_D" "$PLESK_MAILNAMES_D"
					fi
				;;
				relabel_plesk_directories_with_mailnames)
					relabel_plesk_directories "$PLESK_MAILNAMES_D"
				;;
				mail_restore_all)
					mail_restore_all
				;;
				mail_mailbox_restore_full)
					local mail_mailbox_restore="$PLESK_LIBEXEC_DIR/mail_mailbox_restore"
					[ ! -x "$mail_mailbox_restore" ] || "$mail_mailbox_restore" >>"$product_log" 2>&1
				;;
				mail_spam_restore_full)
					local mail_spam_restore="$PLESK_LIBEXEC_DIR/mail_spam_restore"
					[ ! -x "$mail_spam_restore" ] || "$mail_spam_restore" >>"$product_log" 2>&1
				;;
				mail_outgoing_restore_full)
					local mail_outgoing_restore="/usr/lib/plesk-9.0/remote_mail_restore/mail_outgoing_restore"
					[ ! -x "$mail_outgoing_restore" ] || "$mail_outgoing_restore" >>"$product_log" 2>&1
				;;
				notify_imap_pop3_service_about_mailserver_change)
					run_now_notify_imap_pop3_service_about_mailserver_change
				;;
				apply_ssl_settings_to_imap_pop3)
					configure_ssl_ciphers_protocols dovecot courier-imap
				;;
				apply_ssl_settings_to_smtp)
					configure_ssl_ciphers_protocols postfix qmail msmtp
				;;
				apache_restart)
					set_apache_params
					pleskrc apache restart
				;;
				nginx_restart)
					set_nginx_params
					local nginx_ctl="$PRODUCT_ROOT_D/admin/sbin/nginx_control"
					if [ -x "${nginx_ctl}" -a -x "${PRODUCT_RC_D}/${nginx_service}" ]; then
						"$nginx_ctl" --restart >>"$product_log" 2>&1 || warn "restart nginx web-server"
					fi
				;;
				super_server_action_configure)
					set_xinetd_params
					register_service "$xinetd_service"
				;;
				hmonitor_start_notificationd)
 					healthmonitor_start_notificationd
				;;
				mail_restore_imap_pop3)
					local mail_imap_restore="/usr/lib/plesk-9.0/remote_mail_restore/mail_imap_restore"
					[ ! -x "$mail_imap_restore" ] || "$mail_imap_restore" >>"$product_log" 2>&1
				;;
				*)
				;;
			esac

			local temp_res=$?
			if [ -z "$local_res" ]; then
				local_res="$temp_res"
			fi
			if [ "0" = "$local_res" ]; then
				remove_deferred_action_flag $flag
			else
				res=1
				p_echo "Warning : failed to perform deferred action : $flag"
			fi
		fi
	done

	drop_upgrade_flag	
	
	return $res
}

remove_deferred_action_flag()
{
	local flag="$1"
	rm -f "/var/lock/plesk_deferred_$flag"

	# check for depending actions:
	case "$flag" in
		httpd_reconfigure_all)
			remove_deferred_action_flag httpd_reconfigure
		;;
		httpd_reconfigure)
			remove_deferred_action_flag apache_restart 
			remove_deferred_action_flag nginx_restart
		;;
		watchdog_configure)
			remove_deferred_action_flag watchdog_reconfigure
		;;
		relabel_plesk_directories_with_mailnames)
			remove_deferred_action_flag relabel_plesk_directories
		;;
		mail_restore_all)
			remove_deferred_action_flag mail_restore_imap_pop3
		;;
		*)
		;;
	esac
}

dump_incomplete_deferred_actions()
{
	ls -1 /var/lock/plesk_deferred_* 2>/dev/null || :
}

# vim:ft=sh:

os_supports_ipv6()
{
	local ipv6_addresses_file="/proc/net/if_inet6"

	test -f "$ipv6_addresses_file" -a -n "`cat \"${ipv6_addresses_file}\"`"
}

upgrade_pci_compliance_11_1_15()
{
	[ -s "$PRODUCT_ROOT_D/conf/cipher.lst" ] || return 0
	[ ! -e "/etc/sw-cp-server/conf.d/pci-compliance.conf" ] || return 0

	{
		echo -n "ssl_ciphers "
		cat $PRODUCT_ROOT_D/conf/cipher.lst | tr -d '\n'
		echo ";"
		echo "ssl_prefer_server_ciphers on;"
	} > /etc/sw-cp-server/conf.d/pci-compliance.conf
	/etc/init.d/sw-cp-server reload || :
}

replace_gitman_to_git()
{
	local inten="upgrade gitman extension to git"
	echo_try "$inten"
	if ! $PRODUCT_ROOT_D/bin/extension --list  | grep -q '^gitman\s' 2>/dev/null; then
		p_echo "no gitman extension installed"
		return
	fi

	$PRODUCT_ROOT_D/bin/extension --install git && \ 
	$PRODUCT_ROOT_D/bin/extension --uninstall gitman && suc || warn "$inten" 
}
register_local_postgres_12_1_20()
{
	db_select "SELECT COUNT(*) FROM DatabaseServers WHERE type='postgresql' AND host='localhost' AND server_version != '';"
	if [ 0$db_select_output -gt 0 ]; then
		register_service postgresql
	fi
}

is_service_enabled_pylibplesk()
{
	local service="$1"
	local PYTHON_PLESKLIBDIR="$PRODUCT_ROOT_D/admin/lib/python"

	local is_enabled
	is_enabled="`"$PYTHON_BIN" -c 'import sys; 		sys.path[:0] = ["'"$PYTHON_PLESKLIBDIR"'"]; 		import plesk_service; 		print(plesk_service.is_registered("'"$service"'"));' 2>>"$product_log"`" || {
			warn "attempt to determine whether service $service is enabled"
			return 2
		}

	[ "$is_enabled" = "True" ]
}

reregister_enabled_phpfpm_services()
{
	local inten="add missing PHP-FPM services into registered services database"
	echo_try "$inten"

	local phpfpm_services="`"$PRODUCT_ROOT_D/admin/sbin/php_handlers_control" --list-json |
		"$PYTHON_BIN" -c 'import sys, json; 			handlers = json.load(sys.stdin)["php"]; 			print("\n".join(filter(None, map(lambda h: h.get("service"), handlers))));' 2>>"$product_log"`"
	
	for service in $phpfpm_services; do
		is_service_enabled_pylibplesk "$service" || continue
		"$PRODUCT_ROOT_D/admin/sbin/phpinimng" --enable-service --service-name "$service" >>"$product_log" 2>&1 ||
			warn "attempt to re-enable service $service. System resource limits may not be applied to it."
	done

	suc
}

update_registered_services_db_for_resctrl()
{
	reregister_enabled_phpfpm_services
}

set_common_pear_params()
{
	psa_pear_dir="/usr/share/psa-pear"
	psa_pear_conf="/usr/share/psa-pear/pear.conf"
	psa_pear_bin="${psa_pear_dir}/pear/pear"
	psa_pear_include_dir="/usr/share/psa-pear/pear/php"
}

pear_register_packages()
{
	local action=$1
	local pkg_descr_dir=$2

	[ -n "${pkg_descr_dir}" -a -d ${pkg_descr_dir} ] || die "act on PEAR packages - description directory is not specified or does not exist"

	case "${action}" in
		install)
			action=upgrade
		;;
		uninstall)
		;;
		*)
			die "act on PEAR packages - unknown command '${action}'"
		;;
	esac

	if [ "${action}" = "upgrade" ]; then
		if ! ${psa_pear_bin} -c "$psa_pear_conf" "${action}" --offline --nodeps --register-only ${pkg_descr_dir}/*.xml < /dev/null >> $product_log 2>&1; then
			warn "Registering of pear packages failed"
			return 1
		fi
	else
		for i in ${pkg_descr_dir}/*.xml; do
			local pkg_name
			#on uninstall we should use channel/app-name, for example - horde/Horde_Data
			pkg_name="horde/"`basename "${i}" | cut -d '.' -f 1`

			if ! ${psa_pear_bin} -c "$psa_pear_conf" "${action}" --offline --nodeps --register-only "${pkg_name}" < /dev/null >> $product_log 2>&1; then
				warn "${action} package ${i} in PEAR"
				return 1
			fi
		done
	fi
}

# vim:syntax=sh

healthmonitor_start_notificationd()
{
	notification_service="psa-health-monitor-notificationd"
	pleskrc notification start
}

# vim:ft=sh:

bootstrapper_log_title()
{
	local title="Bootstrapper $product_version $script_action"
	if [ -n "$script_target" ]; then
		title="$title for $script_target"
	fi
	title="$title AT `date`"
	echo "$title"
}

bootstrapper_log_subject()
{
	local subject="Product $script_action"
	if [ -n "$script_target" ]; then 
		subject="$subject for $script_target"
	fi
	echo "$subject"
}

lock_bootstrapper()
{
	lockfile "/var/lock/parallels-panel-bootstrapper-running" || {
		p_echo "Another instance of bootstrapper is already running. Please wait until it finishes."
		p_echo "If you are completely sure there is no other instance running, "
		p_echo "just remove /var/lock/parallels-panel-bootstrapper-running.lock file and retry."
		p_echo "Note: running more than one instance at once may badly damage your system."

		exit 1
	}
}

unlock_bootstrapper()
{
	unlockfile "/var/lock/parallels-panel-bootstrapper-running"
}

bootstrapper_transaction_begin()
{
	lock_bootstrapper
	log_start "`bootstrapper_log_title`"
	set_error_report_context "`bootstrapper_log_title`. `report_context_action`"

	set_common_params
	read_conf

	transaction_begin

	# LIFO rollback order
	transaction_add_rollback_action "unlock_bootstrapper"
	transaction_add_rollback_action "show_report_after_rollback"
	p_echo
	p_echo "**** Product $script_action started."

	# Detect upgrade (most of the scripts depend on it)
	check_product_root
}

bootstrapper_transaction_end()
{
	# FIFO commit order
	transaction_add_commit_action "show_report_after_commit"
	transaction_add_commit_action "unlock_bootstrapper"
	transaction_commit
}

show_report_after_rollback()
{
	# Report fatal error
	log_stop "`bootstrapper_log_title`" "`bootstrapper_log_subject`"

	# Show manual restart recomendation for bootstrapped post-install action only.
	if [ -f /tmp/pp-bootstrapper-mode.flag -a "X$script_action" = "Xpost-install" ]; then
		p_echo "To complete product installation one should run bootstrapper script:"
		p_echo "      $PRODUCT_BOOTSTRAPPER_DIR/bootstrapper.sh repair"
		p_echo
	fi
}

show_report_after_commit()
{
	# show possible warnings in problem report
	log_stop "`bootstrapper_log_title`" "`bootstrapper_log_subject`"
}

# include base and optional bootstrapper actions, define set_component_pkgname_list() (see build.mk)
set_component_pkgname_list()
{
	COMPONENT_PKGNAME_base=plesk-base
	COMPONENT_PKGNAME_core=plesk-core
	COMPONENT_PKGNAME_libpamplesk=libpam-plesk
	COMPONENT_PKGNAME_psavhost=psa-vhost
	COMPONENT_PKGNAME_managementnode=plesk-management-node
	COMPONENT_PKGNAME_l10n=plesk-l10n
	COMPONENT_PKGNAME_horde=psa-horde
	COMPONENT_PKGNAME_imp=psa-imp
	COMPONENT_PKGNAME_ingo=psa-ingo
	COMPONENT_PKGNAME_kronolith=psa-kronolith
	COMPONENT_PKGNAME_turba=psa-turba
	COMPONENT_PKGNAME_mnemo=psa-mnemo
	COMPONENT_PKGNAME_passwd=psa-passwd
	COMPONENT_PKGNAME_roundcube=plesk-roundcube
	COMPONENT_PKGNAME_mailpcdriver=plesk-mail-pc-driver
	COMPONENT_PKGNAME_mailqcdriver=plesk-mail-qc-driver
	COMPONENT_PKGNAME_mailmcdriver=plesk-mail-mc-driver
	COMPONENT_PKGNAME_spammng=psa-spamassassin
	COMPONENT_PKGNAME_courierimap=psa-courier-imap
	COMPONENT_PKGNAME_dovecot=plesk-dovecot
	COMPONENT_PKGNAME_drwebconfigurator=psa-drweb-configurator
	COMPONENT_PKGNAME_fail2banconfigurator=plesk-fail2ban-configurator
	COMPONENT_PKGNAME_firewall=psa-firewall
	COMPONENT_PKGNAME_configtroubleshooter=plesk-config-troubleshooter
	COMPONENT_PKGNAME_phpmyadmin=psa-phpmyadmin
	COMPONENT_PKGNAME_dnsbinddriver=plesk-dns-bind-driver
	COMPONENT_PKGNAME_kav8=psa-kav8
}

# vim:syntax=sh

declare_legacy_variables()
{
	echo_try "got legacy variables"
	[ -n "$prod_conf_t" ] || prod_conf_t="/etc/psa/psa.conf"

	if [ ! -s "$prod_conf_t" ]; then
		p_echo "product conf not exists, nothing to be done"
		return
	fi

	local var
	PLESK_LEGACY_VARS="$*"
	for var in $PLESK_LEGACY_VARS; do
		unset $var
		eval $var=`conf_getvar $prod_conf_t $var` # save the value
	done
	suc
}

dump_legacy_variables()
{
	echo_try "save legacy variables"
	[ -n "$prod_conf_t" ] || prod_conf_t="/etc/psa/psa.conf"

	if [ ! -s "$prod_conf_t" ]; then
		p_echo "product conf not exists, cannot dump anything"
		return
	fi

	local var
	local value
	for var in $PLESK_LEGACY_VARS; do
		value="`eval echo \\$${var}`"
 	[ -n "$value" -a -z "`conf_getvar $prod_conf_t $var`" ] || continue # variable is not defined or already in psa.conf
		echo "$var $value" >> $prod_conf_t
	done
	suc
}

product_logs_dir_fix_permissions()
{
	[ -d "$PRODUCT_LOGS_D" ] || mkdir -p "$PRODUCT_LOGS_D"
	chmod 0750 "$PRODUCT_LOGS_D"
	chown "psaadm:0" "$PRODUCT_LOGS_D"
}

base_rollback_product_config()
{
	if [ -n "$prod_conf_t_bak" -a -f "$prod_conf_t_bak" ]; then
		mv -f "$prod_conf_t_bak" "$prod_conf_t"
		unset prod_conf_t_bak
	fi
}

base_commit_product_config()
{
	if [ -n "$prod_conf_t_bak" -a -f "$prod_conf_t_bak" ]; then
		rm -f "$prod_conf_t_bak"
		unset prod_conf_t_bak
	fi
}

base_rollback_product_config_default()
{
	if [ -n "$prod_conf_default_t_bak" -a -f "$prod_conf_default_t_bak" ]; then
		mv -f "$prod_conf_default_t_bak" /etc/psa/psa.conf.default
		unset prod_conf_default_t_bak
	fi
}

base_commit_product_config_default()
{
	if [ -n "$prod_conf_default_t_bak" -a -f "$prod_conf_default_t_bak" ]; then
		rm -f "$prod_conf_default_t_bak"
		unset prod_conf_default_t_bak
	fi
}

base_bootstrapper_prep_install()
{
	# Create users and groups (rollback is not required)
	p_echo "Create user 'psaadm' and group 'psaadm'"
	group_op "psaadm" "search" false
	user_op "psaadm" "search" "psaadm" "psa user" "${PRODUCT_ROOT_D}/admin" "$dummy_shell" false

	p_echo "Create group swkey-data"
	group_op "swkey-data" "search" false
	add_user_to_group "psaadm" "swkey-data"

	set_qmail_params
	set_mail_params

	p_echo "Create Mail accounts"
	group_op "$MAIL_USERGROUP" "search" true
	user_op "$MAIL_USERGROUP" "search" "$MAIL_USERGROUP" "POP3 service user" "$QMAIL_DIR/$MAIL_USERGROUP" "$dummy_shell" true false
	user_op "$HANDLERS_USER" "search" "$MAIL_USERGROUP" "mail handlers user" "$dummy_home" "$dummy_shell" true false

	# Create or update product config
	prod_conf_t="/etc/psa/psa.conf"
	if [ -f "$prod_conf_t" ]; then
		prod_conf_t_bak="${prod_conf_t}.bak"
		cp -pf "$prod_conf_t" "$prod_conf_t_bak"
		transaction_add_rollback_action "base_rollback_product_config"
		transaction_add_commit_action "base_commit_product_config"
	fi

	# For backward compatibility
	[ -d "/etc/psa" ] || mkdir -p /etc/psa
	if [ -f "/etc/psa/psa.conf.default" ]; then
		cp -pf /etc/psa/psa.conf.default /etc/psa/psa.conf.default.bak
		prod_conf_default_t_bak="/etc/psa/psa.conf.default.bak"
		transaction_add_rollback_action "base_rollback_product_config_default"
		transaction_add_commit_action "base_commit_product_config_default"
	fi
	cp -f $PRODUCT_BOOTSTRAPPER_DIR/psa.conf.default /etc/psa/psa.conf.default
	chmod 755 /etc/psa

	declare_legacy_variables QMAIL_MAILNAMES_D CURL_CA_BUNDLE_FILE HTTPD_BIN_D LOGROTATE CGI_PHP_BIN FTPD_BIN_D FTPD_VAR_D FTPD_CONF FTPD_CONF_INC FTPD_SCOREBOARD
	# QMAIL_MAILNAMES_D -> PLESK_MAILNAMES_D
	[ ! -s "$prod_conf_t" ] || PLESK_MAILNAMES_D_SAVED="`conf_getvar $prod_conf_t QMAIL_MAILNAMES_D`"
	conf_update $prod_conf_t && p_echo "===> Updated $prod_conf_t"
	dump_legacy_variables
	[ -z "$PLESK_MAILNAMES_D_SAVED" ] || conf_setvar "$prod_conf_t" "PLESK_MAILNAMES_D" "$PLESK_MAILNAMES_D_SAVED"

	product_logs_dir_fix_permissions
}

# vim:ft=sh:

base_install_common()
{
	inten="generate default ${PRODUCT_NAME} SSL Certificate"
	echo_try $inten
	generate_default_certificate || die
	suc

# work around https://bugzilla.sw.ru/show_bug.cgi?id=435433
    chown root:swkey-data /etc/sw/keys/info /etc/sw/keys/lock/repository /etc/sw/keys/lock/ka
}

base_bootstrapper_post_install()
{
	base_install_common
}


# vim:ft=sh:

core_python_modules()
{
	local py_modules_d="/opt/psa/admin/lib/python 		$PRODUCT_ROOT_D/admin/share/pmmcli 	"

	if [ "$1" = "--install" ]; then
		# Make sure no old .pyc and .pyo files are left
		update_python_modules --remove $py_modules_d
	fi

	update_python_modules "$@" $py_modules_d
}

# vim:ft=sh:

core_groupadd_f()
{
    grep -q "^$1:" /etc/group || /usr/sbin/groupadd "$1"
}

maintainer_main_pre()
{
	set_local_params
	set_ftpd_params

	get_config_parameters

	# debian will try upgrade unconfigured package
	# it is not an upgrade, actually
	if [ "X$do_upgrade" = "X1" ]; then
		if [ ! -s ${PRODUCT_ROOT_D}/version ]; then
			do_upgrade=0
		fi
	fi

# Upgrade psa -> plesk-core
	if [ "X$do_upgrade" = "X0" -a -s ${PRODUCT_ROOT_D}/version ]; then
			do_upgrade=1
	fi

	preinstall_checks

	if [ "X$do_upgrade" = "X1" -a -s ${PRODUCT_ROOT_D}/version ]; then
		cp ${PRODUCT_ROOT_D}/version ${PRODUCT_ROOT_D}/version.upg
	fi
	necessary_system_accounts_prep_install
	core_groupadd_f psacln
	add_login_shell "$PRODUCT_ROOT_D/bin/chrootsh"
}

core_start_product_on_rollback()
{
	start_after_install=1 # for start_product_now
	start_product_now
}

sso_port_change_prep()
{
	# SSO port is changed in sw-sso 2.6, i.e. on upgrade to 10.2.0
	[ "X$do_upgrade" = "X1" ] || return 0
	[ ! -f /etc/sw-cp-server/applications.d/00-sso-cpserver.conf ] || return 0

	# signal to SSO that we upgrade plesk.conf in 10.2.0
	touch /tmp/pp-bootstrapper-upgrade-cpserver-configs.flag

	# signal to Billing the need to perform upgrade actions
	touch /tmp/pp-bootstrapper-change-sso-port.flag

	# actual upgrade for Plesk is done in db_upgrade_sso_port_1010
}

core_bootstrapper_prep_install()
{
	local is_upgrade=0

	selinux_init
	maintainer_main_pre

	# stop product
	define_start_script
	if [ -f $START_SH ]; then
		is_upgrade=1
		# schedule product start on rollback if product is started
		$START_SH status && transaction_add_rollback_action "core_start_product_on_rollback"
		stop_before_upgrade
	fi

	# Perform safe prep-install db actions
	safe_prep_db

	sso_port_change_prep

	if [ "$is_upgrade" -gt 0 ]; then
		pre_update_history
	fi

	selinux_close
}



# mode: shell-script
#-*- vim:syntax=sh

PRODUCT_ROOT_D=/opt/psa
export PRODUCT_ROOT_D

set_relay_opts()
{
	local controls="$PRODUCT_ROOT_D/var/root.controls.lock"
	if [ ! -s "$controls" ]; then
		echo "/server/mail.php:relay_open" > $controls
	fi
}

clean_backend_cache()
{
	db_do "DELETE FROM BackendCache"
	rm -f "$PRODUCT_ROOT_D/tmp/*.cache"
}

clean_upgrade_info()
{
	db_do "TRUNCATE smb_productUpgrades; TRUNCATE smb_componentUpdates;"
    db_do "DELETE FROM misc WHERE param = 'available_patch'"
}

smb_init_log()
{
	local panel_log="/var/log/plesk/panel.log"
	touch "$panel_log"
	chmod 0640 "$panel_log"
	chown psaadm:0 "$panel_log"
}

create_compat_logs_symlinks()
{
# panel.log
	[ -e "$PRODUCT_ROOT_D/admin/logs/panel.log" ] || ln -s "/var/log/plesk/panel.log" "$PRODUCT_ROOT_D/admin/logs/panel.log"

# httpsd_access_log
	[ -e "$PRODUCT_ROOT_D/admin/logs/httpsd_access_log" ] || ln -s "/var/log/plesk/httpsd_access_log" "$PRODUCT_ROOT_D/admin/logs/httpsd_access_log"
}

install_site_isolation_config()
{
	echo_try "install site isolation settings"
	if [ -e "$PRODUCT_ROOT_D/admin/conf/site_isolation_settings.ini" ]; then
		pnnl_echo "config found. Nothing to be done"
		return
	fi
	cp -f $PRODUCT_ROOT_D/admin/conf/site_isolation_settings.ini.default $PRODUCT_ROOT_D/admin/conf/site_isolation_settings.ini
	suc
}

configure_apache_for_filesharing()
{
	add_apache_module "dav"
	add_apache_module "dav_fs"
	add_apache_module "auth_digest"
    add_apache_module "authz_user"
	add_apache_module "dav_lock"
}

vz_disable_admin_cache()
{
	detect_vz
	[ "$PLESK_VZ" = "1" ] || return 0
	local php_ini="$PRODUCT_ROOT_D/admin/conf/php.ini"
	local enabled=`get_phpini_var 'opcache.enabled'`

	[ ! "$enabled" = "0" ] || return 0
	set_phpini_var 'opcache.enabled' '0'
}

db_generate_guids()
{
   local make_guids_script="$PRODUCT_ROOT_D/admin/plib/make_guids.php"
   if [ -f "$make_guids_script" ]; then
       sw_engine_pleskrun "$make_guids_script" >> "$product_log" 2>&1
   fi
}

run_bl_upgrade()
{
	local inten="execute post install/upgrade actions"
	if sw_engine_pleskrun $PRODUCT_ROOT_D/admin/plib/Upgrade/upgrade.php "$@" >> "$product_log" 2>&1; then
		suc
	else
		warn "$inten"
		# reflect error in exit-code in repair mode:
		[ ! "$1" = "--repair" ] || return 1
	fi
}

package_main()
{
	uses_mysql_client

	set_xinetd_params
	deferred_super_server_action_configure
	fix_etc_hosts
	set_services_info

	selinux_init

	if [ "X$do_upgrade" = "X0" ]; then
		# check for psa-to-core upgrade situation
		check_product_root
	fi

	get_config_parameters

	if [ "X$do_upgrade" = "X1" ]; then
		get_prev_version $PRODUCT_ROOT_D
	fi

	install_rbash

	if [ "X$do_upgrade" = "X0" ]; then
		generate_encryption_key
	fi

	install_cron_somely
	install_admin
	import_default_certificate
	install_chroot_base

	upgrade_apsc_post_stage

	if [ "X$do_upgrade" = "X1" ]; then
		core_python_modules --install
		core_run_upgrade_post_stage
	else
		update_registered_services_db_for_resctrl
	fi

	install_site_isolation_config
	packagemng_set_dirty_flag
	set_relay_opts

	if [ "XRELEASE" = "XCTP" ]; then
		setup_support_url "http://swdn.swsoft.com/en/feedback/ctp"
	fi

	label_file_once "${PRODUCT_ETC_D}/webalizer.conf" "${AUTOGENERATED_CONFIGS}"
	# this routine requires 5.0.0 database version, which always come after upgrader
	chown_webstat

	create_version_file
	vz_disable_admin_cache

	if [ "X$do_upgrade" != "X1" ]; then
		core_init_upgrader_conf
# create guid for admin
		db_generate_guids
# cert for pbm
		pbm_create_certificate
	fi

	run_bl_upgrade
	install_cron_backup
	configure_apache_for_filesharing

	clean_backend_cache

	if [ "X$do_upgrade" = "X1" ]; then
		if [ -e "$CATALINA_HOME/conf/server.xml" ]; then
			tomcat_reconfigure_server
        fi
	fi

	smb_init_log

	create_compat_logs_symlinks

	if [ "X$do_upgrade" != "X1" ]; then
		encrypt_psa_shadow
	fi


# During Plesk installation, sw-engine (psasem minit) will create
# $PRODUCT_ROOT_D/var with wrong ownership on debian/ubuntu
	if [ -d "${PRODUCT_ROOT_D}/var" ]; then
		chown psaadm:psaadm "${PRODUCT_ROOT_D}/var"
		chmod 0755 "${PRODUCT_ROOT_D}/var"
	fi

	# en_US.UTF-8 requred for Backup/Restore
	setup_utf8_locale

	clean_upgrade_info
	autoinstall_extensions

	if is_product_upgrade; then
		forced_key_upgrade_cmd
		clean_panel_sessions
	fi

	if is_product_installation; then
		configure_ssl_ciphers_protocols
		install_ssh_banner
	fi

	service_restart_warning
	selinux_close
}

setup_utf8_locale()
{


	if [ ! -x /usr/bin/locale -o ! -x /usr/sbin/locale-gen ]; then
		pnnl_echo "Warning: unable to allow UTF8 locale - package 'locales' is not installed!"
		return
	fi

	/usr/bin/locale -a | grep -q en_US.utf8
	if [ $? -eq 0 ]; then
		pnnl_echo "UTF8 locale is already allowed"
		return
	fi


	sed -i -e 's/.*en_US.UTF-8.*/en_US.UTF-8 UTF-8/' /etc/locale.gen
	/usr/sbin/locale-gen
}

maintainer_main_post()
{

	package_main

	if [ "X$do_upgrade" != "X1" ]; then
		register_sw_engine_service
		register_service sw-cp-server
		register_service psa
	fi
}

pmm_install_create_compat_logs_symlink()
{
	local old_pmm_logs="$PRODUCT_ROOT_D/PMM/logs"
	[ -e "$old_pmm_logs" -a ! -L "$old_pmm_logs" ] || ln -sfT "$PRODUCT_LOGS_D/PMM" "$old_pmm_logs"
}

create_version_file()
{
    echo ${product_version} ${conceived_os_vendor} ${conceived_os_version} 1705181214.14 > ${PRODUCT_ROOT_D}/core.version
}

sso_port_change_post()
{
	[ -f /tmp/pp-bootstrapper-upgrade-cpserver-configs.flag ] && rm -rf /tmp/pp-bootstrapper-upgrade-cpserver-configs.flag
	[ -f /tmp/pp-bootstrapper-change-sso-port.flag ] && rm -f /tmp/pp-bootstrapper-change-sso-port.flag

	# actual upgrade for Plesk is done in db_upgrade_sso_port_1010
}

core_bootstrapper_post_install()
{
	set_local_params
	necessary_system_accounts_post_install
	sso_port_change_post
	maintainer_main_post
	pmm_install_create_compat_logs_symlink

	update_history
}

libpamplesk_bootstrapper_post_install()
{
	install_pam
}



#-*- vim:syntax=sh

psavhost_bootstrapper_post_install()
{
	install_vhosts
}


managementnode_create_version_file()
{
	# Make core happy on upgrade
	if [ ! -s "${PRODUCT_ROOT_D}/core.version" -a -s "${PRODUCT_ROOT_D}/version" ]; then
		cp -f "${PRODUCT_ROOT_D}/version" "${PRODUCT_ROOT_D}/core.version"
	fi
	echo 17.5.3 ${conceived_os_vendor} ${conceived_os_version} 1705181214.14 > ${PRODUCT_ROOT_D}/version

	rm -f "${PRODUCT_ROOT_D}/version.rpm*"
}

managementnode_bootstrapper_post_install()
{
	start_after_install=1 # for start_product_now_final
	check_product_root
	install_vhosts
	managementnode_create_version_file
	install_default_license_key
	start_product_now_final
}

# vim: ft=sh

# mode: shell-script
# vim:syntax=sh

set_locales_params()
{
	active_locales_filename="${PRODUCT_ROOT_D}/var/psa_active_locales"
}

set_active_locales()
{
	local locales="$1"
	local loc_list=""
	local glue=""

	for i in $locales; do
		loc_list="$loc_list$glue'$i'"
		glue=", "
	done

	db_do "UPDATE locales SET active = 'true' WHERE id IN ($loc_list)"
	db_do "UPDATE locales SET active = 'false' WHERE id NOT IN ($loc_list)"
}

save_active_locales()
{
	db_select "SELECT id FROM locales WHERE active = 'true'"
	if [ -z "$db_select_output" ]; then
		rm -f "$active_locales_filename"
	else
		echo "$db_select_output" > "$active_locales_filename"
	fi
}

restore_active_locales()
{
	# Do nothing. We should not try restore unknown locales
	[ -f "$active_locales_filename" ] || return 0

	local locales="`cat $active_locales_filename`"
	[ -n "$locales" ] || return 0

	set_active_locales "$locales"
}

# mode: shell-script
# vim:syntax=sh

l10n_bootstrapper_prep_install()
{
	if [ "X$do_upgrade" = "X1" ]; then
		uses_mysql_client
		set_locales_params
		save_active_locales
	fi
}

# mode: shell-script
# vim:syntax=sh

l10n_register()
{
	if [ ! -x "$PRODUCT_ROOT_D/admin/sbin/localemng" ]; then
		p_echo "Localemng not found - locales cannot be registered"
		return
	fi

	# We cannot use filepp because of bootstrapper
	for locale in ar ca-ES cs-CZ da-DK de-DE el-GR es-ES fi-FI fr-FR he-IL hu-HU id-ID it-IT ja-JP ko-KR ms-MY nb-NO nl-NL pl-PL pt-BR pt-PT ro-RO ru-RU sv-SE th-TH tl-PH tr-TR uk-UA vi-VN zh-CN zh-TW; do
		$PRODUCT_ROOT_D/admin/sbin/localemng --register --locale="$locale"
	done
}

l10n_bootstrapper_post_install()
{
	l10n_register

	if [ "X$do_upgrade" = "X1" ]; then
		uses_mysql_client
		set_locales_params
		restore_active_locales
	fi
}


# vim:ft=sh:

horde_bootstrapper_prep_install()
{
	requires_configured_mysql
}


horde_install_framework()
{
	[ -z "$horde_packages_registered" ] || return 0

	if pear_register_packages install "${horde_datadir}/.xml"; then
		horde_packages_registered="True"
	else
		warn 'install Horde5 framework'
		unset horde_packages_registered
	fi
}

create_horde_user()
{
	local mysql_db_name="mysql"
	db_do "CREATE DATABASE IF NOT EXISTS horde"
	db_do "	SET SESSION sql_mode=REPLACE(@@SESSION.sql_mode, 'NO_AUTO_CREATE_USER', ''); 
			GRANT SELECT, INSERT, UPDATE, DELETE, CREATE, DROP, INDEX, ALTER ON horde.* 
				TO 'horde'@'localhost' IDENTIFIED BY '$horde_password'; 
			FLUSH PRIVILEGES;"
	# Note that GRANT will create user if it doesn't exist yet, otherwise just update permissions.
}

horde_upg_convert_passwd_file()
{
	local pass_file
	for pass_file in "/etc/psa/.webmail.shadow" "/etc/psa/webmail/horde/.horde.shadow"; do
		[ -s "$pass_file" ] || continue
		mv -f "$pass_file" $horde_passwd_file
		break
	done
}

horde_upg_remove_vhost_conf()
{
	#we should drop old horde's apache config, cause zzz_horde_vhost.conf includes nor existent directory from old package
	rm -f "${HTTPD_INCLUDE_D}/zzz_horde_vhost.conf"
}

horde_upg_fix_perms_on_private_files()
{
	[ -d "$horde_logdir" ] && set_ac "$horde_user" root 750 "$horde_logdir"
	[ -e "$horde_logdir/psa-horde.log" ] && set_ac "$horde_user" "$horde_group" 640 "$horde_logdir/psa-horde.log"
}

horde_cu_upgrade_only_actions()
{
	[ "$HORDE_CLEAN_INSTALL_FLAG" = "yes" ] && return

	"$@"
}

horde_upgrader()
{
	local install_or_upgrade="cu_check_stage files"
	local upgrade_only="cu_check_stage files horde_cu_upgrade_only_actions"
	local ts="cu_check_timestamp"
	local do="cu_do"

	cu_start_upgrade_from_version 011005023
	$upgrade_only $ts '2013/05/31 12:05:27' $do horde_upg_convert_passwd_file
	$upgrade_only $ts '2013/05/31 12:05:29' $do horde_upg_remove_vhost_conf

	$install_or_upgrade $ts '2013/05/31 12:06:37' $do generate_horde_password
	$install_or_upgrade $ts '2013/05/31 12:06:39' $do create_horde_user

	#1. register application/framework in local PEAR
	$install_or_upgrade $ts '2013/05/31 12:06:37' $do echo_try "register horde5 and its application in local PEAR"
	$install_or_upgrade $ts '2013/05/31 12:06:39' $do horde_install_framework
	$install_or_upgrade $ts '2013/05/31 12:06:41' $do imp_install_application
	$install_or_upgrade $ts '2013/05/31 12:06:43' $do turba_install_application
	$install_or_upgrade $ts '2013/05/31 12:06:45' $do ingo_install_application
	$install_or_upgrade $ts '2013/05/31 12:06:47' $do mnemo_install_application
	$install_or_upgrade $ts '2013/05/31 12:06:50' $do kronolith_install_application
	$install_or_upgrade $ts '2013/05/31 12:06:52' $do passwd_install_application
	$install_or_upgrade $ts '2013/05/31 12:06:54' $do suc

	#2. upgrade framework/application DB to latest H3 (actual only for upgrade from PPP < 11.5.23)
	$upgrade_only $ts '2013/05/31 12:13:37' $do echo_try "upgrade horde <= 3 to latest 3"
	$upgrade_only $ts '2013/05/31 12:13:39' $do upgrade_horde_DB_to_latest_h3
	$upgrade_only $ts '2013/05/31 12:13:41' $do upgrade_imp_DB_to_latest_h3
	$upgrade_only $ts '2013/05/31 12:13:45' $do upgrade_turba_DB_to_latest_h3
	$upgrade_only $ts '2013/05/31 12:13:53' $do upgrade_ingo_DB_to_latest_h3
	$upgrade_only $ts '2013/05/31 12:14:17' $do upgrade_mnemo_DB_to_latest_h3
	$upgrade_only $ts '2013/05/31 12:14:23' $do upgrade_kronolith_DB_to_latest_h3
	$upgrade_only $ts '2013/05/31 12:14:25' $do suc

	#3. run migrations scripts for framework/application
	$install_or_upgrade $ts '2013/05/31 12:28:34' $do echo_try "execute horde5 upgrade scripts"
	$install_or_upgrade $ts '2013/05/31 12:28:35' $do horde_run_migration_script
	$install_or_upgrade $ts '2013/05/31 12:28:36' $do horde_run_migration_script imp
	$install_or_upgrade $ts '2013/05/31 12:28:38' $do horde_run_migration_script turba
	$install_or_upgrade $ts '2013/05/31 12:28:39' $do horde_run_migration_script ingo
	$install_or_upgrade $ts '2013/05/31 12:28:41' $do horde_run_migration_script mnemo
	$install_or_upgrade $ts '2013/05/31 12:28:43' $do horde_run_migration_script kronolith
	$install_or_upgrade $ts '2013/05/31 12:28:46' $do suc

	# 4.
	$upgrade_only $ts '2013/05/31 12:30:44' $do echo_try "convert datatree shares to SQL engine"
	$upgrade_only $ts '2013/05/31 12:30:46' $do horde_convert_shares_to_sql kronolith
	$upgrade_only $ts '2013/05/31 12:30:47' $do horde_convert_shares_to_sql mnemo
	$upgrade_only $ts '2013/05/31 12:30:49' $do horde_convert_shares_to_sql turba
	$upgrade_only $ts '2013/05/31 12:30:51' $do suc

	# 5. Run horde via fastcgi
	cu_start_upgrade_from_version 011005027
	# permissions for $horde_passwd_file are fixed earlier in generate_horde_password.

	$upgrade_only $ts '2013/06/07 11:53:28' $do horde_upg_fix_perms_on_private_files
	$upgrade_only $ts '2013/05/31 12:56:49' $do deferred_httpd_reconfigure

        # 6. Upgrade Horde from 5.0.4 to 5.1.4
	$install_or_upgrade $ts '2013/09/13 12:06:37' $do echo_try "register horde5.1.4 and its application in local PEAR"
	$install_or_upgrade $ts '2013/09/13 12:06:39' $do horde_install_framework
	$install_or_upgrade $ts '2013/09/13 12:06:41' $do imp_install_application
	$install_or_upgrade $ts '2013/09/13 12:06:43' $do turba_install_application
	$install_or_upgrade $ts '2013/09/13 12:06:45' $do ingo_install_application
	$install_or_upgrade $ts '2013/09/13 12:06:47' $do mnemo_install_application
	$install_or_upgrade $ts '2013/09/13 12:06:50' $do kronolith_install_application
	$install_or_upgrade $ts '2013/09/13 12:06:52' $do passwd_install_application
	$install_or_upgrade $ts '2013/09/13 12:06:54' $do suc

	$install_or_upgrade $ts '2013/09/13 12:28:34' $do echo_try "execute horde5.1.4 upgrade scripts"
	$install_or_upgrade $ts '2013/09/13 12:28:35' $do horde_run_migration_script
	$install_or_upgrade $ts '2013/09/13 12:28:36' $do horde_run_migration_script imp
	$install_or_upgrade $ts '2013/09/13 12:28:38' $do horde_run_migration_script turba
	$install_or_upgrade $ts '2013/09/13 12:28:39' $do horde_run_migration_script ingo
	$install_or_upgrade $ts '2013/09/13 12:28:41' $do horde_run_migration_script mnemo
	$install_or_upgrade $ts '2013/09/13 12:28:43' $do horde_run_migration_script kronolith
	$install_or_upgrade $ts '2013/09/13 12:28:46' $do suc

        # 7. Upgrade Horde from 5.1.4 to 5.2.1
	$install_or_upgrade $ts '2014/05/11 15:01:37' $do echo_try "register horde5.2.1 and its application in local PEAR"
	$install_or_upgrade $ts '2014/05/11 15:01:39' $do horde_install_framework
	$install_or_upgrade $ts '2014/05/11 15:01:41' $do imp_install_application
	$install_or_upgrade $ts '2014/05/11 15:01:43' $do turba_install_application
	$install_or_upgrade $ts '2014/05/11 15:01:45' $do ingo_install_application
	$install_or_upgrade $ts '2014/05/11 15:01:47' $do mnemo_install_application
	$install_or_upgrade $ts '2014/05/11 15:01:50' $do kronolith_install_application
	$install_or_upgrade $ts '2014/05/11 15:01:52' $do passwd_install_application
	$install_or_upgrade $ts '2014/05/11 15:01:54' $do suc

	$install_or_upgrade $ts '2014/05/11 15:05:34' $do echo_try "execute horde5.2.1 upgrade scripts"
	$install_or_upgrade $ts '2014/05/11 15:05:35' $do horde_run_migration_script
	$install_or_upgrade $ts '2014/05/11 15:05:36' $do horde_run_migration_script imp
	$install_or_upgrade $ts '2014/05/11 15:05:38' $do horde_run_migration_script turba
	$install_or_upgrade $ts '2014/05/11 15:05:39' $do horde_run_migration_script ingo
	$install_or_upgrade $ts '2014/05/11 15:05:41' $do horde_run_migration_script mnemo
	$install_or_upgrade $ts '2014/05/11 15:05:43' $do horde_run_migration_script kronolith
	$install_or_upgrade $ts '2014/05/11 15:05:46' $do suc

	# (pay some attention on timestamps below)
	# The following actions are not needed for upgrade from horde 5.2.6 of plesk 12.0.18.
	# But they should be performed for upgrade from horde < 5.2.1.
	$upgrade_only $ts '2014/05/11 15:05:21' $do echo_try "convert Horde shares tables to sharesng"
	$upgrade_only $ts '2014/05/11 15:05:22' $do horde_convert_shares_to_sqlng kronolith
	$upgrade_only $ts '2014/05/11 15:05:23' $do horde_convert_shares_to_sqlng mnemo
	$upgrade_only $ts '2014/05/11 15:05:24' $do horde_convert_shares_to_sqlng turba
	$upgrade_only $ts '2014/05/11 15:05:25' $do suc

	# upgrade_only actions are not registered on clean installation.
	# so here comes a little hack to update current timestamp.
	$install_or_upgrade $ts '2015/07/28 14:14:46' $do true

	# WARNING: all $install_or_upgrade actions will be executed on clean installation as well!

	# DO NOT MODIFY ANYTHING BELOW THIS LINE! (without approval of review from the entire Linux team)
	call_optional_function dev_horde_upgrader
	cu_verify_latest_timestamp_lt '2015/08/14 12:00:00'
}

horde_run_upgrade()
{
	local optional_repair_arg=
	[ -z "$1" -a -z "$do_repair" ] || optional_repair_arg="--repair"

	local inten="upgrade${optional_repair_arg:+ and repair} Horde webmail configuration"
	echo_try "$inten"
	cu_run_upgrader "horde" horde_upgrader --stage "files" $optional_repair_arg --update-version
	if [ "$?" = "0" ]; then
		suc
	else
		warn "$inten"
		# return non-zero exit code on error in repair mode:
		[ -z "$optional_repair_arg" ] || return 1
	fi
}

horde_run_repair()
{
	set_apache_params
	set_horde_params
	set_common_pear_params
	local mysql_db_name="horde"

	if ! reset_horde_passwd no_force; then
		warn "fix credentials for Horde db"
		return 1
	fi

	horde_run_upgrade repair
}

horde_bootstrapper_post_install()
{
	uses_mysql_client

	set_webmail_params
	set_apache_params
	set_horde_params
	set_common_pear_params

	local mysql_db_name="horde"
	
	local HORDE_CLEAN_INSTALL_FLAG=""
	db_select "SHOW TABLES"
	[ -z "${db_select_output}" ] && HORDE_CLEAN_INSTALL_FLAG="yes"

	horde_run_upgrade

	register_webmail horde

	#TODO: should we clean cache?

	generate_horde_php_ini
}



#used by horde's bootstrapper
imp_install_application()
{
	pear_register_packages install "${imp_datadir}/.xml" || warn 'install H5-application imp'
}



#used by horde's bootstrapper
ingo_install_application()
{
	pear_register_packages install "${ingo_datadir}/.xml" || warn 'install H5-application ingo'
}



#used by horde's bootstrapper
kronolith_install_application()
{	
	pear_register_packages install "${kronolith_datadir}/.xml" || warn 'install H5-application kronolith'
}



#used by horde's bootstrapper
turba_install_application()
{
	pear_register_packages install "${turba_datadir}/.xml" || warn 'install H5-application turba'
}



#used by horde's bootstrapper
mnemo_install_application()
{
	pear_register_packages install "${mnemo_datadir}/.xml" || warn 'install H5-application mnemo'
}



#used by horde's bootstrapper
passwd_install_application()
{
        pear_register_packages install "${passwd_datadir}/.xml" || warn 'install H5-application passwd'
}

roundcube_bootstrapper_prep_install()
{
	requires_configured_mysql
}


roundcube_run_native_upgrader()
{
	# It should be safe to run native Roundcube upgrade script each time a package is updated
	# The script will update main configuration files and DB to the current version

	local upgrade_script="$roundcube_home/bin/update.sh"

	local old_version="`cat $roundcube_preupgrade_version_file 2>/dev/null`"
	local new_version="`roundcube_get_current_files_version`"
	if [ -z "$old_version" ]; then
		p_echo "Upgrade of $roundcube_title configs and DB is not required - no previous version"
		return 0
	fi

	if [ ! -x "$upgrade_script" ]; then
		p_echo "Upgrade of $roundcube_title configs and DB is not possible: '$upgrade_script' is not executable"
		return 1
	fi

	local inten="upgrade $roundcube_title configs and DB from '$old_version' to '$new_version'"
	echo_try "$inten"
	if yes 2>/dev/null | "$upgrade_script" "--version=$old_version" >>"$product_log" 2>&1 ; then
		suc
		rm -f "$roundcube_preupgrade_version_file"
		return 0
	else
		warn "$inten. Run $upgrade_script manually to complete upgrade."
		return 1
	fi
}

roundcube_upgrade_layout_09x()
{
	[ -e "$roundcube_home/config/db.inc.php.upg" -a -e "$roundcube_home/config/main.inc.php.upg" ] || return 0

	local inten="upgrade configuration layout from 0.9.x"
	echo_try $inten

# For roundcube upgrader work, we should have old db.inc.php and main.inc.php, and should not have config.inc.php
	if [ -e "$roundcube_conf" ]; then
		mv -f "$roundcube_conf" "${roundcube_conf}.save"
	fi

	cp -f "$roundcube_home/config/main.inc.php.upg" "$roundcube_home/config/main.inc.php"
	cp -f "$roundcube_home/config/db.inc.php.upg" "$roundcube_home/config/db.inc.php"

	if roundcube_run_native_upgrader; then
		rm -f "${roundcube_conf}.save" 			"$roundcube_home/config/db.inc.php.upg" "$roundcube_home/config/db.inc.php" 			"$roundcube_home/config/main.inc.php.upg" "$roundcube_home/config/main.inc.php"

		p_echo "Upgrade configuration layout from 0.9.x completed successfully"
		return 0
	else
		mv -f "${roundcube_conf}.save" "$roundcube_conf"
# On fresh upgrade we have and uncofigured dsn
		local roundcube_password=`cat $roundcube_passwd_file`
		update_roundcube_db_settings

		mv -f "$roundcube_home/config/db.inc.php.upg" "$roundcube_home/config/db.inc.php.09x"
		mv -f "$roundcube_home/config/main.inc.php.upg" "$roundcube_home/config/main.inc.php.09x"

		rm -f "$roundcube_home/config/db.inc.php" "$roundcube_home/config/main.inc.php"

		warn "$inten. Try merge changes from config/main.inc.php.09x and config/db.inc.php.09x manually."
		return 1
	fi
}

roundcube_set_smtp_settings_12()
{
	local inten="set SMTP settings for roundcube"
	local script="$roundcube_home/bin/set_smtp_settings_12.sh"
	if [ ! -x "$script" ]; then
		p_echo "Cannot $inten: upgrade script '$script' is not executeable"
		return 1
	fi

	echo_try "$inten"
	if $script; then
		suc
		return 0
	else
		warn "$inten. Run '$script' manually to complete upgrade."
		return 1
	fi
}

roundcube_upgrader()
{
	local files="cu_check_stage files"
	local ts="cu_check_timestamp"
	local do="cu_do"

	cu_start_upgrade_from_version 011005027
	$files $ts '2013/06/18 15:08:09' $do roundcube_fix_perms_on_private_files
	$files $ts '2013/06/18 15:09:09' $do deferred_httpd_reconfigure

	cu_start_upgrade_from_version 011006002	# Upgrade Roundcube from 0.8.6 to 0.9.4
	$files $ts '2013/10/08 11:31:58' $do roundcube_fix_db_grants
	$files $ts '2013/10/08 11:32:00' $do deferred_httpd_reconfigure

	cu_start_upgrade_from_version 012000008
	$files $ts '2014/01/22 13:42:29' $do roundcube_touch_errors_log

	cu_start_upgrade_from_version 012000017
	$files $ts '2014/04/24 21:34:33' $do roundcube_upgrade_layout_09x

	cu_start_upgrade_from_version 012000018
	$files $ts '2014/06/26 16:23:33' $do roundcube_set_smtp_settings_12

	# DO NOT MODIFY ANYTHING BELOW THIS LINE! (without approval of review from the entire Linux team)
	call_optional_function dev_roundcube_upgrader
	cu_verify_latest_timestamp_lt '2015/08/14 12:00:00'
}

roundcube_run_upgrade()
{
	local optional_repair_arg=
	[ -z "$1" -a -z "$do_repair" ] || optional_repair_arg="--repair"

	local inten="upgrade${optional_repair_arg:+ and repair} Roundcube webmail configuration"
	echo_try "$inten"
	cu_run_upgrader "roundcube" roundcube_upgrader --stage "files" $optional_repair_arg --update-version
	if [ "$?" = "0" ]; then
		suc
	else
		warn "$inten"
		# return non-zero exit code on error in repair mode:
		[ -z "$optional_repair_arg" ] || return 1
	fi
}

roundcube_run_repair()
{
	set_apache_params
	set_roundcube_params

	local ret=0

	if ! reset_roundcube_passwd no_force; then
		warn "fix credentials for Roundcube db"
		return 1;
	fi

	roundcube_run_upgrade repair || ret=1
	roundcube_run_native_upgrader || ret=1
	return $ret
}

roundcube_bootstrapper_post_install()
{
	set_webmail_params
	set_roundcube_params

	roundcube_run_native_upgrader
	register_webmail roundcube
	roundcube_generate_php_ini
}

init_postfix_on_sn_hack()
{
	# As the name suggests, this is essentially a hack for SN installation/upgrade 
	# and as such should be removed when possible (e.g. mail_restore for SN is implemented).
	local gl_data_db="/var/lib/plesk/mail/greylist/data.db"
	local passwd_db="/var/spool/postfix/plesk/passwd.db"

	if ! is_management_node_installed; then
		[ -f "$passwd_db"  ] || "$PLESK_LIBEXEC_DIR/mailsrv_conf_init" >> $product_log 2>&1
		[ -f "$gl_data_db" ] || "$PLESK_LIBEXEC_DIR/mail_grey_restore" >> $product_log 2>&1
		# Note: mail_grey_restore is expected to fail on MySQL access - we need only greylisting DB initialization
	fi
}

postfix_add_postfix_to_outgoing_antispam_whitelist()
{
	local mailmng_outgoing="${PRODUCT_ROOT_D}/admin/sbin/mailmng-outgoing"
	"$mailmng_outgoing" --show-whitelist | grep -q "postfix" >/dev/null 2>&1 		|| "$mailmng_outgoing" --add-to-whitelist --sysuser="postfix"
}

mailpcdriver_bootstrapper_post_install()
{
	# See bugs #53084 and #69422.
	[ -x "${PRODUCT_RC_D}/postfix" -a ! -f "${PRODUCT_RC_D}/qmail" ] || return 0
	set_postfix_params
	init_postfix_on_sn_hack
	install_sendmail_wrapper postfix
	start_postfix "bootstrapper-post"
	postfix_add_postfix_to_outgoing_antispam_whitelist >> "$product_log" 2>&1
}

qmail_add_user_to_outgoing_antispam_whitelist()
{
	local username="$1"

	local mailmng_outgoing="${PRODUCT_ROOT_D}/admin/sbin/mailmng-outgoing"
	"$mailmng_outgoing" --show-whitelist | grep -q "$username" >/dev/null 2>&1 		|| "$mailmng_outgoing" --add-to-whitelist --sysuser="$username"
}

mailqcdriver_bootstrapper_post_install()
{
	install_sendmail_wrapper qmail

	qmail_add_user_to_outgoing_antispam_whitelist "qmaild" >> "$product_log" 2>&1
	qmail_add_user_to_outgoing_antispam_whitelist "qmailr" >> "$product_log" 2>&1
}

mailmcdriver_bootstrapper_post_install()
{
	install_sendmail_wrapper msmtp
}

#-*- vim:ft=sh:

set_spammng_params()
{
	old_plesk_spamassassin_service='psa-spamassassin'
	old_spamd_pid="/var/run/spamd/spamd_full.pid"

	spamassassin_service='spamassassin'
}

spammng_enable_services()
{
	local service
	for service in spamassassin spamass-milter spamas-milter; do
		[ -f $SYSTEM_RC_D/$service ] || continue
		if [ -x "/usr/sbin/dpkg-divert" ]; then
			[ -n "`/usr/sbin/dpkg-divert --list $SYSTEM_RC_D/$service`" ] || continue
			rm -f $SYSTEM_RC_D/$service
			dpkg-divert --package psa-spamassassin --remove --rename $SYSTEM_RC_D/$service
		else
			chmod +x $SYSTEM_RC_D/$service
			register_service $service
		fi
	done
}

spammng_cleanup_old_psa_spamassassin_service()
{
	# the service was already stopped and deregistered in preinstall, now just cleanup potential leftovers (config files)
	rm -rf "/etc/psa-spamassassin"
    # Fake init script for fixing bug 132781; it is removed in bootstrapper_post_install.
    create_spamassasin_fake_init_script "$SYSTEM_RC_D/$old_plesk_spamassassin_service" 
	if [ -n "$old_plesk_spamassassin_unit" -a -f "$old_plesk_spamassassin_unit" ]; then
		rm -f "$old_plesk_spamassassin_unit"
	fi
	:
}

spamassassin_upgrader()
{
	local files="cu_check_stage files"
	local ts="cu_check_timestamp"
	local do="cu_do"

	$files $ts '2013/02/21 15:14:41' $do conf_rmvar "${prod_conf_t}" SPAMASSASSIN_MAX_CHILDREN

	cu_start_upgrade_from_version 011005019
	$files $ts '2013/02/21 15:15:00' $do spammng_cleanup_old_psa_spamassassin_service
	$files $ts '2013/02/21 15:15:07' $do spammng_enable_services
	$files $ts '2013/02/21 15:42:41' $do deferred_mail_spam_restore_full

	# DO NOT MODIFY ANYTHING BELOW THIS LINE! (without approval of review from the entire Linux team)
	call_optional_function dev_spamassassin_upgrader
	cu_verify_latest_timestamp_lt '2015/08/14 12:00:00'
}

create_spamassasin_fake_init_script()
{
    echo "#!/bin/sh" > $1
    echo "exit 0;" >> $1
    chmod 755 $1
}

spamassassin_run_upgrade()
{
	local optional_repair_arg=
	[ -z "$1" -a -z "$do_repair" ] || optional_repair_arg="--repair"

	local inten="upgrade${optional_repair_arg:+ and repair} SpamAssassin service configuration"
	echo_try "$inten"
	cu_run_upgrader "spamassassin" spamassassin_upgrader --stage "files" $optional_repair_arg --update-version
	if [ "$?" = "0" ]; then
		suc
	else
		warn "$inten"
		# return non-zero exit code on error in repair mode:
		[ -z "$optional_repair_arg" ] || return 1
	fi
}

spamassassin_run_repair()
{
	set_admin_params
	set_spammng_params
	set_mail_params

	spamassassin_run_upgrade repair
}

spammng_bootstrapper_post_install()
{
	set_spammng_params
    rm -f "$SYSTEM_RC_D/$old_plesk_spamassassin_service" 
}


#-*- vim:ft=sh

stop_and_unregister_courier_imap_lt_4()
{
	# stops old Courier-IMAP services (<= 3.0.8)
	pleskrc old_courier stop
	unregister_service "$old_courier_service" >>"$product_log" 2>&1
	:
}

cleanup_configuration_courier_imap_lt_4()
{
	# Old init script may not be removed during package upgrade on deb-based systems
	rm -f "$SYSTEM_RC_D/$old_courier_service" >>"$product_log" 2>&1
	# Remove any old .dist files so they are replaced for sure
	for conf in imapd imapd-ssl pop3d pop3d-ssl; do
		rm -f "$COURIERIMAP_CONFDIR/$conf.dist" >>"$product_log" 2>&1
	done
	# Move out files used by mkimapdcert (just a precaution)
	for conf in imapd.cnf pop3d.cnf; do
		mv -f "$COURIERIMAP_CONFDIR/$conf" "$COURIERIMAP_CONFDIR/$conf.saved_by_plesk" >>"$product_log" 2>&1
	done
	:
}

courier_imap_upgrade_certs()
{
	courier_imap_generate_certs
	:
}

courier_imap_sysconftool_upgrade()
{
	local sysconftool="/usr/lib/courier-authlib/sysconftool"
	local tool_ver="0.16"
	local inten="upgrade Courier-IMAP services configuration files via sysconftool"
	echo_try "$inten"; p_echo
	$sysconftool -r "$tool_ver" 		"$COURIERIMAP_CONFDIR/imapd.dist" "$COURIERIMAP_CONFDIR/imapd-ssl.dist" 		"$COURIERIMAP_CONFDIR/pop3d.dist" "$COURIERIMAP_CONFDIR/pop3d-ssl.dist" >>"$product_log" 2>&1 && suc
}

courier_imap_disable_pci_compliance_for_upgrade()
{
	local courier_pci_comp=
	{
		# Consider PCI compliance enabled if at least one file matching these globs exists
		ls $COURIERIMAP_CONFDIR/imapd.pci* || ls $COURIERIMAP_CONFDIR/imapd-ssl.pci* || 		ls $COURIERIMAP_CONFDIR/pop3d.pci* || ls $COURIERIMAP_CONFDIR/pop3d-ssl.pci* 
	} 1>/dev/null 2>&1 && 		courier_pci_comp="on"

	[ "$courier_pci_comp" = "on" -a -x "$PRODUCT_ROOT_D/admin/bin/pci_compliance_resolver" ] || return 0

	echo_try "disable PCI compliance for Courier-IMAP before upgrade"
	touch "$COURIERIMAP_CONFDIR/plesk-pci-compliance-was-enabled.flag"
	"$PRODUCT_ROOT_D/admin/bin/pci_compliance_resolver" --disable courier >>"$product_log" 2>&1 && suc || p_echo "failed"
	# This action is always successfull. Otherwise we would have to be sure courier_imap_reenable_pci_compliance() would fail too.
}

courier_imap_reenable_pci_compliance()
{
	[ -f "$COURIERIMAP_CONFDIR/plesk-pci-compliance-was-enabled.flag" ] || return 0

	local inten="reenable PCI compliance for Courier-IMAP after upgrade"
	echo_try "$inten"
	rm -f "$COURIERIMAP_CONFDIR/plesk-pci-compliance-was-enabled.flag"
	"$PRODUCT_ROOT_D/admin/bin/pci_compliance_resolver" --enable courier >>"$product_log" 2>&1 && suc || warn "$inten"
	# This action is always successfull. Reenabling already enabled courier will produce utility error.
}

courier_imap_upgrader()
{
	local prep="cu_check_stage prep"
	local post="cu_check_stage post"
	local bpos="cu_check_stage bootstrapper-post"
	local ts="cu_check_timestamp"
	local do="cu_do"

	# Upgrade from 3.0.8 to 4.12.0
	$prep $ts '2013/05/07 15:17:03' $do courier_imap_disable_pci_compliance_for_upgrade # should be in bootstrapper, but will work this way too
	$prep $ts '2013/05/06 16:03:58' $do stop_and_unregister_courier_imap_lt_4
	$prep $ts '2013/05/06 16:25:32' $do courier_imap_upgrade_certs
	$prep $ts '2013/05/22 15:28:58' $do cleanup_configuration_courier_imap_lt_4
	$post $ts '2013/05/06 17:33:42' $do courier_imap_sysconftool_upgrade
	$post $ts '2013/05/13 12:12:13' $do courier_imap_services_register
	$bpos $ts '2013/05/07 15:17:41' $do courier_imap_reenable_pci_compliance
	$post $ts '2013/05/07 15:30:49' $do deferred_mail_restore_all
	$bpos $ts '2013/05/13 17:57:44' $do courier_imap_finish_upgrade_to_4
# Note, when we upgrade courier from 3.0.8, old courier removes pid files. So we should restart
# courier in scope of bootstrapper *ONLY*, or courier initscripts will hangs on next stop, after transaction ends.
# Most probably, we can restart courier anywhere when upgrade from 9.5.x will be removed
#	 $bpos $ts '2013/05/14 15:39:23' $do courier_imap_services_action restart
	# Upgrade from 4.14.0 to 4.15.0
	$post $ts '2014/04/28 14:10:00' $do courier_imap_sysconftool_upgrade
	$prep $ts '2014/04/28 14:20:00' $do courier_imap_generate_dhparams

	$bpos $ts '2014/04/28 14:30:00' $do courier_imap_services_action restart
	# DO NOT MODIFY ANYTHING BELOW THIS LINE! (without approval of review from the entire Linux team)
	call_optional_function dev_courier_imap_upgrader
	cu_verify_latest_timestamp_lt '2015/08/14 12:00:00'
}

courier_imap_run_upgrade()
{
	local stage="$1"
	[ "$stage" = "prep" -o "$stage" = "post" -o "$stage" = "bootstrapper-post" ] || die "courier_imap_run_upgrade(): upgrade stage is not specified or invalid"

	local optional_repair_arg=
	[ -z "$2" -a -z "$do_repair" ] || optional_repair_arg="--repair"

	local update_version_arg=
	[ "$stage" = "bootstrapper-post" ] && update_version_arg="--update-version"

	local inten="upgrade${optional_repair_arg:+ and repair} Courier-IMAP services configuration ($stage stage)"
	echo_try "$inten"
	cu_run_upgrader "courier_imap" courier_imap_upgrader --stage "$stage" $optional_repair_arg $update_version_arg
	if [ "$?" = "0" ]; then
		suc
	else
		warn "$inten"
		# return non-zero exit code on error in repair mode:
		[ -z "$optional_repair_arg" ] || return 1
	fi
}

courier_imap_run_repair()
{
	local ret=0
	set_courier_imap_params

	courier_imap_run_upgrade prep repair || ret=1
	courier_imap_run_upgrade post repair || ret=1
	courier_imap_run_upgrade bootstrapper-post repair || ret=1

	courier_imap_services_action condrestart 'use_exit_code' || ret=1

	return $ret
}

courier_imap_services_register()
{
	register_service "$courier_imapd_service"
	register_service "$courier_imaps_service"
	register_service "$courier_pop3d_service"
	register_service "$courier_pop3s_service"
}

courier_imap_finish_upgrade_to_4()
{
	rm -f "$COURIERIMAP_CONFDIR/plesk-courier-imap-3.0.8-upgrade.flag" >>"$product_log" 2>&1
}

courier_imap_scheduled_services_start()
{
	local action="$1"
	local start_flag="$COURIERIMAP_CONFDIR/plesk-courier-imap-needs-start.flag"
	case "$action" in
		schedule|add)
			touch "$start_flag"
			;;
		unschedule|remove)
			rm -f "$start_flag" >>"$product_log" 2>&1
			;;
		execute|start)
			if [ -f "$start_flag" ]; then
				courier_imap_scheduled_services_start remove
				switch_ingo_backend imap
				courier_imap_services_action start
			fi
			;;
		*)
			die "courier_imap_scheduled_services_start(): Unknown action '$action'"
			;;
	esac
}

courierimap_bootstrapper_post_install()
{
	set_courier_imap_params

	courier_imap_run_upgrade bootstrapper-post

	courier_imap_scheduled_services_start execute
}

# vim:ft=sh:

dovecot_scheduled_service_start()
{
	local action="$1"
	local start_flag="$DOVECOT_CONFDIR/plesk-dovecot-needs-start.flag"
	case "$action" in
		schedule|add)
			touch "$start_flag"
			;;
		unschedule|remove)
			rm -f "$start_flag" >>"$product_log" 2>&1
			;;
		execute|start)
			if [ -f "$start_flag" ]; then
				dovecot_scheduled_service_start remove
				switch_ingo_backend sieve
				pleskrc dovecot start
			fi
			;;
		*)
			die "dovecot_scheduled_service_start(): Unknown action '$action'"
			;;
	esac
}

dovecot_bootstrapper_post_install()
{
	set_dovecot_params

	dovecot_scheduled_service_start execute
}


# vim:ft=sh:

drweb_force_installing_new_drweb32ini()
{
	# New DrWeb 6.0.2 packages will install new drweb32.ini only if it doesn't exist yet
	mv -f "$DRWEB_ETC_D/drweb32.ini" "$DRWEB_ETC_D/drweb32.ini.old_drweb5" || :
}

drweb_reinstall_bases()
{
	# Old drweb-bases 5.0.1 packages execute 'rm -f /var/drweb/bases/*.vdb /var/drweb/updates/*'
	# unconditionally in pre-uninstall (i.e., even on upgrade). Therefore we need to replicate 
	# bases installation. The code is taken as-is from drweb-bases post-install.
	(
		if [ ! -f /var/drweb/updates/timestamp ] ; then
			cd /var/drweb/updates && 			gzip -dc /opt/drweb/bases/bases.tar.gz | tar -xpf - && 			rm -f /var/drweb/bases/*.vdb /var/drweb/lib/drweb32.dll && 			cp -p -f /var/drweb/updates/*.vdb /var/drweb/bases/ && 			cp -p -f /var/drweb/updates/drweb32.dll /var/drweb/lib/ && 			chown drweb /var/drweb/bases/*.vdb && 			chgrp drweb /var/drweb/bases/*.vdb && 			chown drweb /var/drweb/lib/drweb32.dll && 			chgrp drweb /var/drweb/lib/drweb32.dll && 			rm -f /var/drweb/updates/*.vdb /var/drweb/updates/drweb32.dll
		fi
	) >>"$product_log" 2>&1
}

drweb_restart_service()
{
	# Since DrWeb may be dead after upgrade to 6.0.2, we need to restart it
	pleskrc drweb restart || :
}

drweb_upgrader()
{
	local bprep="cu_check_stage bootstrapper-prep"
	local bpost="cu_check_stage bootstrapper-post"
	local ts="cu_check_timestamp"
	local do="cu_do"

	# Upgrade from 5.0.1 to 6.0.2
	$bprep $ts '2013/07/09 12:30:16' $do drweb_force_installing_new_drweb32ini
	$bpost $ts '2013/07/09 12:30:32' $do drweb_reinstall_bases
	$bpost $ts '2013/07/09 12:30:48' $do selinux_relabel_dir "/var/drweb" "$DRWEB_ROOT_D"
	$bpost $ts '2013/07/09 12:31:00' $do drweb_restart_service

	# DO NOT MODIFY ANYTHING BELOW THIS LINE! (without approval of review from the entire Linux team)
	call_optional_function dev_drweb_upgrader
	cu_verify_latest_timestamp_lt '2015/08/14 12:00:00'
}

drweb_run_upgrade()
{
	local stage="$1"
	[ "$stage" = "bootstrapper-prep" -o "$stage" = "bootstrapper-post" ] || die "drweb_run_upgrade(): upgrade stage is not specified or invalid"

	local optional_repair_arg=
	[ -z "$2" -a -z "$do_repair" ] || optional_repair_arg="--repair"

	local update_version_arg=
	[ "$stage" = "bootstrapper-post" ] && update_version_arg="--update-version"

	local inten="upgrade${optional_repair_arg:+ and repair} Parallels Premium antivirus (DrWeb) service configuration ($stage stage)"
	echo_try "$inten"
	cu_run_upgrader "drweb" drweb_upgrader --stage "$stage" $optional_repair_arg $update_version_arg
	if [ "$?" = "0" ]; then
		suc
	else
		warn "$inten"
		# return non-zero exit code on error in repair mode:
		[ -z "$optional_repair_arg" ] || return 1
	fi
}

drweb_run_repair()
{
	set_drweb_params
	local ret=0
	drweb_run_upgrade bootstrapper-prep repair || ret=1
	drweb_run_upgrade bootstrapper-post repair || ret=1
	return $ret
}

drwebconfigurator_bootstrapper_prep_install()
{
	set_drweb_params

	[ -d "$DRWEB_ETC_D" ] || return 0
	drweb_run_upgrade bootstrapper-prep
}


#-*- vim:ft=sh:

config_drweb_key_expiration_notifications()
{
	local inten="Turn key expiration notifications for 2 days in drweb config files"
	echo_try "$inten"

	# Add New parameters for DrWeb upgrade

	add_string '[[:space:]]*NotifyPeriod[[:space:]]*=.*' 		"NotifyPeriod = 2" "$DRWEB_ETC_D/drweb32.ini"

	add_string '[[:space:]]*NotifyType[[:space:]]*=.*' 		"NotifyType = Everyday" "$DRWEB_ETC_D/drweb32.ini"

	# Change Default values for clear Drweb installation
	NotifyPeriod=2
	NotifyType=Everyday

	sed -e 's/^[[:space:]]*NotifyPeriod[[:space:]]*=[[:space:]]*14/NotifyPeriod = '$NotifyPeriod'/g' 		-e 's/^[[:space:]]*NotifyType[[:space:]]*=[[:space:]]*Ever[[:space:]]*$/NotifyType = '$NotifyType'/g' 	< $DRWEB_ETC_D/drweb32.ini > $DRWEB_ETC_D/drweb32.ini.tmp && 	mv -f $DRWEB_ETC_D/drweb32.ini.tmp $DRWEB_ETC_D/drweb32.ini || die "$inten"

	set_ac drweb drweb 644 $DRWEB_ETC_D/drweb32.ini || die "$inten"		
	suc
}

drweb_fix_logs_permissions()
{
    # Fixing permissions to avoid logrotate errors from bug #141554
    local drweb_logs_package_location="/var/drweb/log"
    [ ! -d "$drweb_logs_package_location" ] || chmod 0644 "$drweb_logs_package_location"
}

check_drweb32_ini()
{
	# Correct drweb32.ini has this revision
	local ok="Revision: 1.90."
	[ -e "$1" ] || return 1
	grep -q "$ok" "$1"
}

force_drweb32_ini_on_upgrade_pvc()
{
	# 1. Bootstrapper prep actions aren't executed for optional components when installed via templates
	# 2. Anyway drweb-common (with drweb32.ini) upgrades before drweb-configurator
	if check_drweb32_ini $DRWEB_ETC_D/drweb32.ini; then
		return
	fi

	if check_drweb32_ini $DRWEB_ETC_D/drweb32.ini.rpmnew; then
		mv -f $DRWEB_ETC_D/drweb32.ini $DRWEB_ETC_D/drweb32.ini.old_drweb5
		mv -f $DRWEB_ETC_D/drweb32.ini.rpmnew $DRWEB_ETC_D/drweb32.ini
		return
	fi

	# No new drweb32.ini found (i.e debian/ubuntu)
	mkdir -p /var/drweb/run
	set_ac drweb drweb 755 /var/drweb/run
	perl -pi -e 's,^\s*EnginePath\s*=(.*)$,EnginePath = /var/drweb/lib/drweb32.dll,g' $DRWEB_ETC_D/drweb32.ini
}

drweb_actions_for_install_or_upgrade()
{
	force_drweb32_ini_on_upgrade_pvc
	# FIXME: move some of these actions to drweb_upgrader since executing them one time should be enough
	config_drweb_key_expiration_notifications

	# register drweb handler for existing mailnames
	# this also turns on or off the service if required (esp. needed on clean install)
	/usr/lib/plesk-9.0/mail_drweb_restore >> "$product_log" 2>&1
	qmail_admin_alias "$drweb_user"

	deferred_watchdog_reconfigure

	selinux_relabel_dir /opt/psa/handlers

	drweb_fix_logs_permissions
}

drwebconfigurator_bootstrapper_post_install()
{
	set_drweb_params

	drweb_actions_for_install_or_upgrade

	drweb_run_upgrade bootstrapper-post
}



fetch_current_fail2ban_version()
{
	local ver=
	for python_bin in "/usr/bin/python" "$PYTHON_BIN"; do
		ver="`env PYTHONPATH=/usr/share/fail2ban $python_bin -c 'import common.version as f2b; print(f2b.version)' 2>/dev/null`"
		[ "$?" -ne 0 -o -z "$ver" ] || break
		ver=
	done
	echo "$ver"
}

set_fail2ban_params()
{
	fail2ban_service="fail2ban"

	### Due to a lack of better alternative (since this is used in bootstrapper), 
	### hardcode values from thirdparty/fail2ban/include.mk
	fail2ban_conf_d="/etc/fail2ban"
	fail2ban_jail_d="$fail2ban_conf_d/jail.d"
	fail2ban_filter_d="$fail2ban_conf_d/filter.d"
	fail2ban_action_d="$fail2ban_conf_d/action.d"
	fail2ban_fail2ban_d="$fail2ban_conf_d/fail2ban.d"

	fail2ban_upg_source_conf_d="$fail2ban_conf_d.previous"
	fail2ban_upg_source_version_f="$fail2ban_upg_source_conf_d/.version"

	fail2ban_upg_need_start_flag="$fail2ban_upg_source_conf_d/.needs-service-start"
}

fail2ban_fix_httpd_vhosts_d_path_in_jail()
{
	local jail="$1"
	local f2bmng="$PRODUCT_ROOT_D/admin/sbin/f2bmng"
	local httpd_vhosts_d_orig="/var/www/vhosts"

	echo_try "fix HTTPD_VHOSTS_D value in '$jail' Fail2Ban jail"

	local jail_config_json="`$f2bmng --get-jail \"$jail\"`" 2>> "$product_log"
	if [ -n "$jail_config_json" ]; then
		# FIXME: Technically, both pattern and replacement should be escaped for sed; also strip trailing slash
		# echo in dash interprets escape sequences, so use printf instead
		printf "%s" "$jail_config_json" | 			sed -e 's#\(\s"\|\\n\)'"$httpd_vhosts_d_orig"'/#\1'"$HTTPD_VHOSTS_D"'/#g' | 			$f2bmng --set-jail "$jail" >> "$product_log" 2>&1 			&& suc || { p_echo "failed"; return 1; }
	else
		p_echo "Skipped, no '$jail' configuration found"
	fi
}

fail2ban_conf_upgrade_prep()
{
	local previous_f2b_version="`fetch_current_fail2ban_version`"

	if [ -e "$fail2ban_upg_source_conf_d" ]; then
		p_echo "Removing old '$fail2ban_upg_source_conf_d'"
		rm -rf "$fail2ban_upg_source_conf_d"
	fi

	if [ -d "$fail2ban_conf_d" ]; then
		cp -ar "$fail2ban_conf_d" "$fail2ban_upg_source_conf_d"
		[ -z "$previous_f2b_version" ] || echo "$previous_f2b_version" > "$fail2ban_upg_source_version_f"
		cat > "$fail2ban_upg_source_conf_d/README.upgrade" <<-EOF
			This folder contains Fail2Ban configuration prior to upgrade from version '${previous_f2b_version:-<unknown>}'. 
			Actual current configuration is located at '$fail2ban_conf_d'.

			You may safely delete this folder and all of its content as long as you're satisfied
			with current Fail2Ban operation.
			EOF

		! pleskrc fail2ban status || echo "yes" > "$fail2ban_upg_need_start_flag"
	else
		[ -z "$previous_f2b_version" ] || 			p_echo "Previous Fail2Ban version is '$previous_f2b_version', but no configuration found in '$fail2ban_conf_d'. Configuration upgrade will be skipped."
	fi
}

fail2ban_conf_upgrade_failure_readme()
{
	# WARNING: this function MUST have stdin, or it'll hang
	cat > "$fail2ban_conf_d/README.upgrade" <<-EOF
		Fail2Ban or its Plesk-specific configuration was upgraded to the current version on 
		`date -R`. Unfortunately, automatic configuration upgrade either failed or was not
		possible.

		Original Fail2Ban configuration from before upgrade was copied to '$fail2ban_upg_source_conf_d'.
		You may use it to manually upgrade current configuration. Once you're satisfied with
		the Fail2Ban operation, you may simply remove this copy.
		EOF
	echo >> "$fail2ban_conf_d/README.upgrade"
	cat - >> "$fail2ban_conf_d/README.upgrade"
}

fail2ban_conf_upgrade_success_readme()
{
	cat > "$fail2ban_conf_d/README.upgrade" <<-EOF
		Fail2Ban or its Plesk-specific configuration was successfully upgraded to the current
		version on `date -R`.

		Original Fail2Ban configuration from before upgrade was copied to '$fail2ban_upg_source_conf_d'.
		If you experience problems with the current Fail2Ban configuration you may use the 
		previous one as a reference to fix the current configuration. Once you're satisfied 
		with the Fail2Ban operation, you may simply remove the copy.
		EOF
}

fail2ban_conf_remove_customizations()
{
	echo_try "remove local customization from '$fail2ban_conf_d' to prepare for upgrade"

	find "$fail2ban_conf_d" -name '*.local' -delete >> "$product_log" 2>&1
	find "$fail2ban_jail_d" "$fail2ban_filter_d" "$fail2ban_action_d" -mindepth 1 -name '*.d' -type d -empty -delete >> "$product_log" 2>&1
	find "$fail2ban_fail2ban_d" -name '*.d' -type d -empty -delete >> "$product_log" 2>&1

	suc
}

fail2ban_conf_remove_obsolete_configs()
{
	[ -x "/usr/bin/dpkg-query" ] || return 0

	# Drop obsolete configuration files on deb-based systems.
	# dpkg will still remember them but only till another upgrade, so that should be OK.
	/usr/bin/dpkg-query -f '${Conffiles}\n' -W fail2ban plesk-fail2ban-configurator | 		awk '$NF == "obsolete" && $1 ~ "^/etc/fail2ban/" {print $1}' | xargs -n 100 rm -f
}

fail2ban_conf_upgrade_post()
{
	local previous_f2b_version="`cat \"$fail2ban_upg_source_version_f\" 2>/dev/null`"
	local current_f2b_version="`fetch_current_fail2ban_version`"

	[ -d "$fail2ban_conf_d" ] || { p_echo "Skipped Fail2Ban upgrade - no target config directory"; return 0; }
	[ -d "$fail2ban_upg_source_conf_d" ] || { p_echo "Skipped Fail2Ban upgrade - no source config directory"; return 0; }

	[ -n "$previous_f2b_version" -a -n "$current_f2b_version" ] || { 
		fail2ban_conf_upgrade_failure_readme <<-EOF
			Upgrade was not preformed as source (${previous_f2b_version:-<unknown>}) or
			target (${current_f2b_version:-<unknown>}) configuration versions were not known.
			EOF
		return 0
	}

	fail2ban_conf_remove_obsolete_configs
	fail2ban_conf_remove_customizations

	local inten="upgrade Fail2Ban configuration from version '$previous_f2b_version' to version '$current_f2b_version'"
	echo_try "$inten"

	local f2bmng="$PRODUCT_ROOT_D/admin/sbin/f2bmng"
	local output=
	output=`$f2bmng --upgrade-configuration --source-basedir="$fail2ban_upg_source_conf_d" --source-version="$previous_f2b_version" 2>&1`
	if [ "$?" -eq 0 ]; then
		fail2ban_conf_upgrade_success_readme
		[ -z "$output" ] || {
			p_echo "Additional output from the upgrade utility:"
			echo "$output" >> "$product_log"
		}

		[ ! -s "$fail2ban_upg_need_start_flag" ] || pleskrc fail2ban start
		suc
	else
		fail2ban_conf_upgrade_failure_readme <<EOF
Upgrade was not performed as upgrade utility failed. See upgrade log below:

$output
EOF
		p_echo "Upgrade utility failed with following messages:"
		echo "$output" >> "$product_log"
		warn "$inten"
	fi
}

fail2ban_upgrader()
{
	local bprep="cu_check_stage bootstrapper-prep"
	local bpost="cu_check_stage bootstrapper-post"
	local ts="cu_check_timestamp"
	local do="cu_do"

	# These are main configuration upgrade actions. Bump their timestamp when significant changes to configuration are made.
	$bprep $ts '2014/05/20 13:39:01' $do fail2ban_conf_upgrade_prep
	$bpost $ts '2014/05/20 13:39:02' $do fail2ban_conf_upgrade_post

	# One-time jails initialization - if HTTPD_VHOSTS_D was altered, fix Apache-related jails.
	# Note that anything that customizes jails should go after fail2ban_conf_upgrade_*.
	$bpost $ts '2014/02/11 10:16:20' $do fail2ban_fix_httpd_vhosts_d_path_in_jail "plesk-apache"
	$bpost $ts '2014/02/11 10:16:21' $do fail2ban_fix_httpd_vhosts_d_path_in_jail "plesk-apache-badbot"

	# DO NOT MODIFY ANYTHING BELOW THIS LINE! (without approval of review from the entire Linux team)
	call_optional_function dev_fail2ban_upgrader
	cu_verify_latest_timestamp_lt '2015/08/14 12:00:00'
}

fail2ban_run_upgrade()
{
	local stage="$1"
	[ "$stage" = "bootstrapper-prep" -o "$stage" = "bootstrapper-post" ] || die "fail2ban_run_upgrade(): upgrade stage is not specified or invalid"

	local optional_repair_arg=
	[ -z "$2" -a -z "$do_repair" ] || optional_repair_arg="--repair"

	local update_version_arg=
	[ "$stage" = "bootstrapper-post" ] && update_version_arg="--update-version"

	local inten="upgrade${optional_repair_arg:+ and repair} Fail2Ban configuration ($stage stage)"
	echo_try "$inten"
	cu_run_upgrader "fail2ban" fail2ban_upgrader --stage "$stage" $optional_repair_arg $update_version_arg
	if [ "$?" = "0" ]; then
		suc
	else
		warn "$inten"
		# return non-zero exit code on error in repair mode:
		[ -z "$optional_repair_arg" ] || return 1
	fi
}

fail2ban_run_repair()
{
	set_fail2ban_params
	local ret=0
	fail2ban_run_upgrade bootstrapper-prep repair || ret=1
	fail2ban_run_upgrade bootstrapper-post repair || ret=1
	return $ret
}

# vim:ft=sh:


fail2banconfigurator_bootstrapper_prep_install()
{
	set_fail2ban_params

	# Note that cu_upgrader_conf_init shouldn't be called anywhere for fail2ban upgrader
	# That is, upgrade is attempted on new installations as well (e.g., upgrade from system fail2ban package)
	fail2ban_run_upgrade bootstrapper-prep
}

# vim:ft=sh:


fail2banconfigurator_bootstrapper_post_install()
{
	set_fail2ban_params

	fail2ban_run_upgrade bootstrapper-post
}

# vim:ft=sh:

# vim:ft=sh:

set_module_params_firewall()
{
	module="firewall"
	module_full="Firewall"
	module_description="The Firewall module helps protect your host and private network from unauthorized access."
	module_icon="/modules/$module/images/module-icon.png"

	module_version="1.0.2"

	iptables_bin_d=/sbin
	
	iptables_bin="$iptables_bin_d/iptables"
	ip6tables_bin="$iptables_bin_d/ip6tables"
    
	set_module_params
}


# vim:ft=sh:

firewall_has_state_match()
{
	local iptables="$1"
	local safeact="$PRODUCT_ROOT_D/admin/sbin/modules/$module/safeact"

	[ -x "$iptables" -a -x "$safeact" ] || return 1
	$safeact --try-enable-features --iptables="$iptables" --rule="-m state --state NEW,ESTABLISHED,RELATED,INVALID" >/dev/null 2>&1
}

firewall_update_default_input_policy()
{
	# Sometimes state match is not available in iptables (e.g., VZ). In this case INPUT should ACCEPT connections by default.
	local default_policy="deny"
	if [ -x "$iptables_bin" ]; then
		firewall_has_state_match "$iptables_bin" || default_policy="allow"
	fi
	if [ -x "$ip6tables_bin" ]; then
		firewall_has_state_match "$ip6tables_bin" || default_policy="allow"
	fi

	# Previously default policy was always 'deny'
	[ "$default_policy" != "deny" ] || return 0
	echo_try "switch default firewall INPUT policy to ACCEPT since 'state' match is not available"

	local new_rule='a:4:{s:4:"type";s:8:"catchall";s:5:"class";s:8:"catchall";s:9:"direction";s:5:"input";s:6:"action";s:5:"allow";}'
	local old_rule_pat='a:%:{s:4:"type";s:8:"catchall";s:5:"class";s:8:"catchall";s:9:"direction";s:5:"input";s:6:"action";s:4:"deny";%'
	local input_dir="0"
	db_do "UPDATE module_${module}_rules SET object='$new_rule' WHERE object LIKE '$old_rule_pat' AND direction=$input_dir" && suc
}

firewal_drop_unused_flags()
{
	rm -f $module_var/active.flag $module_var/ip_forward.active
}

firewall_set_enabled_in_db()
{
    local table="module_${module}_objects"

    db_select "SELECT object FROM $table where tag='confirmed' LIMIT 1"
    if [ -n "$db_select_output" ]; then
        db_do_if_not         "SELECT object FROM $table where tag='enabled' LIMIT 1" 1         "INSERT INTO $table (tag, object) VALUES ('enabled', 'b:1;')"
    fi
}

firewall_read_object()
{
	db_select "SELECT object FROM module_firewall_objects WHERE tag = '$1'"
	echo "$db_select_output"
}

firewall_cu_load_config_data()
{
	local upgrader_name="$1"
	local table="module_${module}_objects"

	[ "$upgrader_name" = "firewall" ] || return 0
	db_test_table "$table" || return 0

	cu_db_current_timestamp="`  firewall_read_object 'upgrader_config_current_timestamp'`"
	cu_db_continue_from="`      firewall_read_object 'upgrader_config_continue_from'`"
	cu_db_failed_actions_list="`firewall_read_object 'upgrader_config_failed_actions_list'`"
	cu_db_latest_timestamp="`   firewall_read_object 'upgrader_config_latest_timestamp'`"
}

firewall_cu_store_config_data()
{
	local upgrader_name="$1"
	local table="module_${module}_objects"

	[ "$upgrader_name" = "firewall" ] || return 0
	db_test_table "$table" || die "save $module_full upgrader state to '$table' table"

	db_do "REPLACE INTO $table (tag, object) VALUES 		('upgrader_config_current_timestamp', '$cu_db_current_timestamp'), 		('upgrader_config_continue_from', '$cu_db_continue_from'), 		('upgrader_config_failed_actions_list', '$cu_db_failed_actions_list'), 		('upgrader_config_latest_timestamp', '$cu_db_latest_timestamp')"
}

firewall_upgrader()
{
	local post="cu_check_stage post"
	local ts="cu_check_timestamp"
	local do="cu_do"

	# Upgrade to Plesk 11.5.30, Firewall 1.0.2
	$post $ts '2013/07/15 16:12:16' $do firewall_update_default_input_policy

	# upgrade to Plesk 11.6.5
	$post $ts '2013/11/06 18:50:10' $do firewal_drop_unused_flags
	$post $ts '2013/11/08 18:00:00' $do firewall_set_enabled_in_db

	# DO NOT MODIFY ANYTHING BELOW THIS LINE! (without approval of review from the entire Linux team)
	call_optional_function dev_firewall_upgrader
	cu_verify_latest_timestamp_lt '2015/08/14 12:00:00'
}

firewall_run_upgrade()
{
	local optional_repair_arg=
	[ -z "$1" -a -z "$do_repair" ] || optional_repair_arg="--repair"

	local inten="upgrade${optional_repair_arg:+ and repair} $module_full service configuration"
	echo_try "$inten"
	cu_run_upgrader "firewall" firewall_upgrader 		--load-config-data firewall_cu_load_config_data --store-config-data firewall_cu_store_config_data 		--stage "post" $optional_repair_arg --update-version
	if [ "$?" = "0" ]; then
		suc
	else
		warn "$inten"
		# return non-zero exit code on error in repair mode:
		[ -z "$optional_repair_arg" ] || return 1
	fi
}

firewall_run_repair()
{
	set_module_params_firewall

	firewall_run_upgrade repair
}


# vim:ft=sh:

set_module_params_configtroubleshooter()
{
	module="configurations-troubleshooter"
	module_full="Webserver Configurations Troubleshooter"
	module_description="A tool for investigation and correction of possible errors in web servers configurations."
	module_icon=""

	module_version="2.0.0"

	set_module_params

	# Override package name
	packname="plesk-config-troubleshooter"
}


# vim:ft=sh:

uninstall_legacy_configs_zip_extension()
{
	local extension_bin="$PRODUCT_ROOT_D/bin/extension"
	local configs_ext_name="configs"
	local inten="remove legacy '$configs_ext_name' extension if it is installed"

	echo_try "$inten"
	if [ ! -x "$extension_bin" ]; then
		warn "$inten: $extension_bin is not executable"
		return 0
	fi

	$extension_bin --uninstall "$configs_ext_name" -force >> "$product_log" 2>&1 || :
	suc
}

configtroubleshooter_bootstrapper_post_install()
{
	set_module_params_configtroubleshooter

	# No need to attempt removal of old extension on package update, only on clean install
	if [ -f "/tmp/plesk-$module-clean-install.flag" ]; then
		uninstall_legacy_configs_zip_extension
		rm -f "/tmp/plesk-$module-clean-install.flag"
	fi
}


phpmyadmin_bootstrapper_prep_install()
{
	# Take care about upgrades from versions < 11.5.30,
	# because in this versions phpmyadmin was included in plesk-core
	# Need to drop it after 11.5.9 EOL.
	phpmyadmin_config_backup
}


phpmyadmin_bootstrapper_post_install()
{
	install_phpmyadmin
	upgrade_phpmyadmin_to_4_7_plus
}

# vim:ft=sh:


dns_bind_driver_repair()
{
	set_named_params

	pleskrc named stop

	sysconfig_named

	pleskrc named start
}

dnsbinddriver_bootstrapper_post_install()
{
	apparmor_named_install

	# FIXME: account for upgrade case
	set_admin_params
	install_named

	uses_mysql_client

	echo_try "create reverse DNS zone"

	local psa_configured
	psa_configured=`core_read_misc_attr psa_configured`
	[ $? -eq 0 ] || die 'detect that panel is configured'

	if [ "true" = "${psa_configured}" ]; then
		suc
		return
	fi

	local util_name="${PRODUCT_ROOT_D}/admin/sbin/mk_reverse_zone"
	if [ -x "$util_name" ]; then
		if ! "${util_name}" >> "$product_log" 2>&1; then
			warn "create reverse DNS zone - there is error during utility '${util_name}' execution, please fix the problem and run it again."
		else 
			suc
		fi
	else
		die "locate utility '${util_name}'"
	fi
}

# vim:ft=sh:

set_kav_params()
{
	kav_service="kavehost"
	kav_sdk_home="/opt/kav/sdk8l3"
}

kav8_bootstrapper_prep_install()
{
	# On deb-based systems package upgrade with architecture change is treated like replacement (i.e. old
	# package is purged, then new one is installed, *_upgrade_new_pre isn't invoked). Therefore to distinguish
	# clean install and upgrade from SDK 8.0 we need bootstrapper.

	set_kav_params
	[ -f "$kav_sdk_home/bin/appinfo.kli" ] || return 0

	if ! grep -q -F 'Application Id: 1638(666).' "$kav_sdk_home/bin/appinfo.kli" 2>/dev/null; then
		# This is an older SDK version (< 8.5, which actually means 8.0).
		# The flag is removed in bootstrapper post-install.
		touch "/tmp/plesk-kav-upgrade-from-8.0.flag"
	fi
}

kav8_bootstrapper_post_install()
{
	[ -f "/tmp/plesk-kav-upgrade-from-8.0.flag" ] || return 0

	$PRODUCT_ROOT_D/admin/bin/php $PRODUCT_ROOT_D/admin/plib/scripts/kav-key-upgrade.php >>"$product_log" 2>&1 ||
		warn "KAV key upgrade. Kaspersky Anti-Virus service will be stopped until proper key is installed."

	rm -f "/tmp/plesk-kav-upgrade-from-8.0.flag"
}


update_plesk_completion_rules()
{
	local update_plesk_completion_bin="$PRODUCT_ROOT_D/admin/sbin/update_plesk_completion"
	[ -x "$update_plesk_completion_bin" ] || return 0
	# the utility is quite noisy, filter out useless messages
	"$update_plesk_completion_bin" 2>&1 | grep -v '^Usage:' >>$product_log
}

db_bootstrapper_prep_install()
{
	if [ "$do_upgrade" != "1" -a "$do_repair" != "1" ]; then
	# This action doesn't required for clean install
		return
	fi
	pp_echo "===> Plesk database scheme upgrade has been started."
	if $PRODUCT_BOOTSTRAPPER_DIR/db.php migrate; then
		pp_echo "===> Plesk database scheme upgrade has been completed."
	else
		local err_msg="===> Plesk database scheme was not upgraded completely. See installation log for details."
		if [ "$do_repair" = "1" ]; then
			pp_echo "$err_msg"
			return 1
		else 
			simply_die "$err_msg"
		fi
	fi
}

tune_memory_swap()
{
	[ "$do_upgrade" != 1 ] || return 0  # clean install only
	set_file_swap_params
	file_swap_is_required || return 0
	! file_swap_is_switched_off || return 0  # disabled by admin
	! file_swap_status || return 0  # already enabled
	file_swap_enable
}

component_base_prep_install()
{
	selinux_is_active
	if [ "$SELINUX_ENFORCE" = "Enforcing" ]; then
		pp_echo "Set selinux into the Permissive mode"
		touch /var/lock/selinux_disabled.flag && setenforce Permissive
		setenforce Permissive
	fi

	tune_memory_swap
	base_bootstrapper_prep_install
	core_bootstrapper_prep_install
	db_bootstrapper_prep_install
	l10n_bootstrapper_prep_install
	phpmyadmin_bootstrapper_prep_install
	if [ "X$PREP_INSTALL_ROLLBACK_CHECK" = "XYes" ]; then
		simply_die "We die to check bootstrapper rollback"
	fi
}

component_base_post_install()
{
	base_bootstrapper_post_install
	libpamplesk_bootstrapper_post_install
	core_bootstrapper_post_install
	managementnode_bootstrapper_post_install
	l10n_bootstrapper_post_install
	phpmyadmin_bootstrapper_post_install

	update_plesk_completion_rules

	if [ -f "/var/lock/selinux_disabled.flag" ]; then
		pp_echo "Set selinux into the Enforcing mode"
		setenforce Enforcing
		rm -f /var/lock/selinux_disabled.flag
	fi
}


set_optional_component_params()
{
	set_component_pkgname_list

	optional_components="
		horde
		roundcube
		mail-pc-driver
		mail-qc-driver
		mail-mc-driver
		psa-vhost
		dns-bind-driver
		spammng
		courier-imap
		dovecot
		drweb-configurator
		kav8
		fail2ban-configurator
		config-troubleshooter
	"

	# hack to forbid sh_optimizer to wipe bootstrapper instructions out
	true horde_bootstrapper_prep_install horde_bootstrapper_post_install
	true roundcube_bootstrapper_prep_install roundcube_bootstrapper_post_install
	true mailpcdriver_bootstrapper_post_install
	true mailqcdriver_bootstrapper_post_install
	true mailmcdriver_bootstrapper_post_install
	true psavhost_bootstrapper_post_install
	true dnsbinddriver_bootstrapper_post_install
	true spammng_bootstrapper_post_install
	true courierimap_bootstrapper_post_install
	true dovecot_bootstrapper_post_install
	true drwebconfigurator_bootstrapper_prep_install drwebconfigurator_bootstrapper_post_install
	true kav8_bootstrapper_prep_install kav8_bootstrapper_post_install
	true fail2banconfigurator_bootstrapper_prep_install fail2banconfigurator_bootstrapper_post_install
	true configtroubleshooter_bootstrapper_post_install
}

is_package_installed_apt()
{
	[ "`dpkg-query -W --showformat='${Status}' $1 2>/dev/null`" = "install ok installed" ]
}

is_component_installed()
{
	local component=$1
	local package=$(eval "echo \$$(echo COMPONENT_PKGNAME_`echo $component | tr -d '-'`)")

	[ -n "$package" ] || return 1

	is_package_installed_apt $package
}

is_component_installed_old()
{
	# assume that packages in components weren't renamed
	is_component_installed "$@"
}

component_optional_action()
{
	local action=$1
	local component=$2

	if [ "X${component}" = "XBASE" ] && is_function "component_base_${action}"; then 
		"component_base_${action}"
	elif is_function "`echo $component | tr -d '-'`_bootstrapper_${action}"; then
		"`echo $component | tr -d '-'`_bootstrapper_${action}"
	else
		p_echo "Warning: there are no bootstrapper ${action} actions for component '$component'"
	fi
}

component_optional_prep_install()
{
	component_optional_action prep_install $1

	if [ "X$PREP_INSTALL_ROLLBACK_CHECK" = "XYes" ]; then
		simply_die "We die to check bootstrapper rollback"
	fi
}

component_optional_post_install()
{
	component_optional_action post_install $1
}

configure_local_components()
{
	deferred_webserver_apache_configure
	deferred_ftp_proftpd_configure

	deferred_awstats_configure
}

prepare_repair_actions()
{
	[ -s "$PRODUCT_ROOT_D/version" -o -s "$PRODUCT_ROOT_D/core.version" ] && get_prev_version "$PRODUCT_ROOT_D"

	set_local_params
	uses_mysql_client

	check_db
	safe_prep_db_backup_all
}

fix_services_registration()
{
	local pleskrc_bin="$PRODUCT_ROOT_D/admin/sbin/pleskrc"
	local services_list="sw-cp-server sw-engine qmail courier-imapd courier-imaps courier-pop3d courier-pop3s dovecot kavehost drwebd spamassassin pc-remote nginx psa"
	for service in $services_list; do
		if "$pleskrc_bin" "$service" exists; then
			register_service "$service" || true
			"$pleskrc_bin" "$service" status >/dev/null 2>&1 || 				"$pleskrc_bin" "$service" start || true
		fi
	done
}

reset_lock_manager()
{
	local cache_dir="$PRODUCT_ROOT_D/var/cache"

	set_admin_params

	[ ! -d "$cache_dir" ] && mkdir -p "$cache_dir"
	set_ac $admin_user root 0700 "$cache_dir"

	find "$cache_dir" -type f -name "SharedLockManagerStorage*" -print0 | xargs -0 rm -f

	add_user_to_group $admin_user lock-manager
}

run_repair_actions()
{
	local ret=0
	local failed_actions=

	pre_accumulate_error ret failed_actions

	pp_echo "Started bootstrapper repair procedure${PLESK_INSTALLER_FAST_REPAIR:+ (fast mode)}. This may take a while."
	pp_echo "Certain actions may be skipped if not applicable."
	pp_echo

	prepare_repair_actions
	set_optional_component_params

	pp_echo " Finishing up upgrade procedures and rerunning previously failed upgrade actions..."

	reset_lock_manager || accumulate_error "reset lock manager"

	apsc_run_repair || accumulate_error "fix apsc"
	reset_sitebuilder_passwd  no_force || accumulate_error "fix credentials for WPB database"
	reset_siteeditor_passwd no_force || accumulate_error "fix credentials for WPB database"
	reset_phpmyadmin_passwd no_force || accumulate_error "fix credentials for PhpMyAdmin database"
	reset_atmail_passwd no_force || accumulate_error "fix credentials for Atmail database"
	core_run_upgrade_prep_stage repair || accumulate_error "cumulative Plesk upgrade and repair revertible stage"
	db_bootstrapper_prep_install || accumulate_error "upgrade Plesk database scheme"
	run_bl_upgrade --repair || accumulate_error "upgrade Plesk business logic"
	core_run_upgrade_post_stage repair || accumulate_error "cumulative Plesk upgrade and repair final stage"

	! is_component_installed 'spammng' || spamassassin_run_repair || accumulate_error "fix SpamAssassin configuration"
	! is_component_installed 'courier-imap' || courier_imap_run_repair || accumulate_error "fix Courier IMAP configuration"
	! is_component_installed 'roundcube' || roundcube_run_repair || accumulate_error "fix RoundCube configuration"
	! is_component_installed 'horde' || horde_run_repair || accumulate_error "fix Horde configuration"
	! is_component_installed 'drweb-configurator' || drweb_run_repair || accumulate_error "fix DrWeb configuration"
	! is_component_installed 'fail2ban-configurator' || fail2ban_run_repair || accumulate error "fix Fail2Ban configuration"
	! is_component_installed 'firewall' || firewall_run_repair || accumulate_error "fix Firewall configuration"
	! is_component_installed 'dns-bind-driver' || dns_bind_driver_repair || accumulate_error "fix DNS Bind configuration"

	if [ -z "${PLESK_INSTALLER_SKIP_MCHK}" ]; then
		pp_echo " Reconfiguring mail subsystem..."
		mail_restore_all ${PLESK_INSTALLER_FAST_REPAIR:-full} || accumulate_error "restore mail"
	fi

	pp_echo " Reconfiguring Apache web server..."
	run_now_webserver_apache_configure || accumulate_error "fix Apache configuration"

	pp_echo " Reconfiguring ProFTPD FTP server..."
	run_now_ftp_proftpd_configure || accumulate_error "fix ProFTPD configuration"

	pp_echo " Reconfiguring AWStats web statistics..."
	run_now_awstats_configure || accumulate_error "fix AWStats configuration"

	pp_echo " Reconfiguring WatchDog..."
	wd_adapt || accumulate_error "fix WatchDog configuration"

	pp_echo " Restoring SELinux contexts..."
	if [ -n "$PLESK_INSTALLER_FAST_REPAIR" ]; then
		relabel_plesk_directories --verbose || accumulate_error "fix SELinux contexts"
	else
		relabel_plesk_directories --verbose "$HTTPD_VHOSTS_D" "$PLESK_MAILNAMES_D" || accumulate_error "fix SELinux contexts"
	fi

	fix_services_registration || accumulate_error "fix services registration"

	pp_echo " Reconfiguring SSL ciphers and protocols..."
	configure_ssl_ciphers_protocols || accumulate_error "configure SSL ciphers and protocols"

	if [ -z "$PLESK_INSTALLER_FAST_REPAIR" ]; then
		pp_echo " Regenerating web servers' configuration files..."
		httpdmng_reconfigure all || accumulate_error "regenerate web servers configuration files"
	fi

	pp_echo " Cleaning active Panel sessions..."
	clean_panel_sessions || accumulate_error "clean Plesk panel sessions"

	pp_echo
	pp_echo "Bootstrapper repair finished."
	if [ -n "$failed_actions" ]; then
		pp_echo "Errors occurred while performing the following actions: $failed_actions."
		pp_echo "Check '$product_log' and '$product_problems_log' for details."
	else
		pp_echo "If problems persist, please check installer logs ('$product_log' and '$product_problems_log') for errors."
	fi
	pp_echo "If you can't resolve the issue on your own, please address Parallels support."

	return $ret
}

check_bootstrapper_version()
{
	check_product_root >/dev/null 2>&1
	if [ $do_upgrade -eq 0 ]; then
		return 0
	fi

	local short_product_version=`form_undotted_version ${product_version}`
	if [ $short_product_version -lt $prev_short_version ]; then
		p_echo "It seems that there is a newer version (${prev_version}) of ${PRODUCT_NAME} installed"
		p_echo "so that a new bootstrapper script should be used instead."
		exit 1
	fi

	return 0
}

usage()
{
	local prog="`basename $0`"
	echo ""
	echo "Bootstrap setup actions for ${PRODUCT_FULL_NAME} ${product_version}"
	echo ""
	echo "Usage: $prog <prep-install|post-install|rerun|perform-deferred-actions> [component|BASE]"
	echo "       $prog <repair|perform-deferred-actions>"
	echo ""
	echo "Running '$prog repair' with PLESK_INSTALLER_FAST_REPAIR=1 environment"
	echo "variable set will skip some of the most time-consuming restore actions."
	echo ""
}

if [ "X${PLESK_INSTALLER_DEBUG}" != "X" ]; then
    set -x
fi

reexec_with_clean_env "$@"

export PLESK_INTERNAL_PHP_EXEC=1

product_build="debian8.0.build1705181214.14"
product_default_conf
initial_conf
check_bootstrapper_version

script_action="$1"
script_target="$2"
case $script_action in 
	prep-install)
		repair_action="$script_action"
		bootstrapper_transaction_begin
		set_optional_component_params
		if [ -z "$script_target" ]; then
			component_base_prep_install
			# We don't know here whether a certain optional component is going to be installed.
			# So do not execute prep-install actions for it here.
			# However, we may detect upgrade. In this case prep-install is executed.
			for comp in $optional_components; do
				is_component_installed_old $comp && component_optional_prep_install $comp;
			done
		else
			component_optional_prep_install $script_target
		fi
		bootstrapper_transaction_end
	;;
	post-install)
		repair_action="repair"
		bootstrapper_transaction_begin
		set_optional_component_params
		if [ -z "$script_target" ]; then
			component_base_post_install
			for comp in $optional_components; do
				is_component_installed $comp && component_optional_post_install $comp;
			done
			configure_local_components
			perform_deferred_actions
		else
			component_optional_post_install $script_target
		fi
		bootstrapper_transaction_end
	;;
	rerun)
		repair_action="rerun"
		do_repair=1
		set_log_action_name "installation_rerun"
		bootstrapper_transaction_begin
		transaction_add_commit_action "reset_upgrade_failure_flag"
		transaction_add_rollback_action "reset_maintenance_mode_flag"
		set_maintenance_mode_flag
		set_optional_component_params
		if [ -z "$script_target" ]; then
			component_base_prep_install
			for comp in $optional_components; do 
				is_component_installed $comp && component_optional_prep_install $comp;
			done
			for comp in $optional_components; do
				is_component_installed $comp && component_optional_post_install $comp;
			done
			component_base_post_install
			configure_local_components
			perform_deferred_actions
		else
			component_optional_prep_install $script_target
			component_optional_post_install $script_target
		fi
		bootstrapper_transaction_end
	;;
	repair)
		ret=0
		repair_action="repair"
		do_repair=1
		set_log_action_name "repair"
		bootstrapper_transaction_begin
		transaction_add_commit_action "reset_upgrade_failure_flag"
		transaction_add_rollback_action "reset_maintenance_mode_flag"
		set_maintenance_mode_flag
		run_repair_actions || ret=1
		perform_deferred_actions || ret=1
		bootstrapper_transaction_end
		exit $ret
	;;
	perform-deferred-actions)
		bootstrapper_transaction_begin
		perform_deferred_actions
		bootstrapper_transaction_end
	;;
	dump-incomplete-deferred-actions)
		dump_incomplete_deferred_actions
	;;
	*)
		usage
		exit 1
	;;
esac
